/*
 *  WCOL: conversion check procedures
 *      by k-chinen@is.aist-nara.ac.jp, 1994, 1995
 *
 *  $Id: conv.c,v 1.5 1995/10/10 14:40:03 k-chinen Exp k-chinen $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <signal.h>
#include <fcntl.h>
#include <limits.h>
#include <string.h>

#include "wcol.h"

#include "conv.h"


conv_elem conv_tbl[CONV_TBL_NUM+1];
static int conv_tbl_len = 0;


int AppendConvertEntry(char *atype, char *aext, char *aout, char *aprog)
{
    if(atype==NULL) return -1;
    strcpy(conv_tbl[conv_tbl_len].target,       atype);

    if(aext==NULL)  return -1;
    strcpy(conv_tbl[conv_tbl_len].target_ext,   aext);

    if(aout==NULL)  return -1;
    strcpy(conv_tbl[conv_tbl_len].output,       aout);

    if(aprog==NULL) return -1;
    strcpy(conv_tbl[conv_tbl_len].program,      aprog);

    conv_tbl_len++;

    return 0;
}


int Read_Convtable(char *name)
{
#define GET_TMP \
    q = tmp;\
    while(*p && (*p=='\t'||*p==' ')) p++; if(*p!='"') continue; p++;\
    while(*p && *p!='"') { if(*p=='\\') p++; *q++ = *p++;} p++; *q='\0'

    FILE *fp;
    char buf[STRING_SIZE];
    char tmp[STRING_SIZE];
    char *p,*q;
    int n;

    conv_tbl_len = 0;

    if(name==NULL || name[0]=='\0') {
        Error("Read_Convtable: no convert table file");
        conv_tbl[0].target[0] = '.';
        return 1;
    }

    if((fp=fopen(name,"r"))==NULL) {
        Error("Read_Convtable: Cannot open %s",name);
        conv_tbl[0].target[0] = '.';
        return 1;
    }

    n = 0;
    while(n<CONV_TBL_NUM && fgets(buf,STRING_SIZE,fp)!=NULL) {
        if(buf[0]=='#')
            continue;

        p = buf;

        GET_TMP; strcpy(conv_tbl[n].target, tmp);
        GET_TMP; strcpy(conv_tbl[n].target_ext, tmp);
        GET_TMP; strcpy(conv_tbl[n].output, tmp);
        GET_TMP; strcpy(conv_tbl[n].program, tmp);

        Trace("conv_tbl[%d] \"%s\" or \"%s\" --[%s]-> \"%s\"\n",
            n, conv_tbl[n].target, conv_tbl[n].target_ext,
            conv_tbl[n].program, conv_tbl[n].output);

        n++;
    }
    conv_tbl[n].target[0] = '.';

    fclose(fp);

    conv_tbl_len = n;

    return 0;

#undef GET_TMP
}


int ConvProgramName(char *dest, char *mask, char *in, char *out)
{
    char *p;
    char *q;
    char *r;
    int ret=0x0;

    Trace("ConvProgramName: [%s]\n\tin [%s], out [%s]\n",mask,in,out);

    p = mask;
    q = dest;
    while(*p) {
        if(*p=='\\') {
            p++;
            *q++ = *p++;
        }
        else
        if(*p=='@') {
            if(*(p+1)=='i') {
                r = in;
                while(*r)
                    *q++ = *r++;
                ret = ret | CONV_HAS_INPUT;
            }
            if(*(p+1)=='o') {
                r = out;
                while(*r)
                    *q++ = *r++;
                ret = ret | CONV_HAS_OUTPUT;
            }
            p+=2;
        }
        else {
            *q++ = *p++;
        }
    }
    *q = '\0';

    Trace("ConvProgramName: after [%s]\n",dest);

    return ret;
}




/*
 * Check hook or not
 */
int check_hook(char *name, char *type)
{
    int i;

    if(conv_tbl_len<=0) {
        return -1;
    }

    for(i=0;i<conv_tbl_len;i++) {
        if(conv_tbl[i].target[0] &&
          strncmp(type,conv_tbl[i].target,strlen(conv_tbl[i].target))==0) {
            Trace("check_hook: Found convertor stream %s, table %s (%d)\n",
                type,conv_tbl[i].target,i);
            return i;
        }
        i++;
    }

    for(i=0;i<conv_tbl_len;i++) {
        if(conv_tbl[i].target_ext[0] &&
            ExtCmp(name,conv_tbl[i].target_ext)==0) {
            Trace("check_hook: Found convertor %s ext=%s (%d)\n",
                type,conv_tbl[i].target_ext,i);
            return i;
        }
        i++;
    }

    return -1;
}

/*
 *
 */
int IsConvEntry()
{
    return conv_tbl_len;
}


