/*
 *  WCOL : External QuickSort
 *      original is G.H.Gonnet ?
 *      modify by k-chinen@is.aist-nara.ac.jp, 1995
 *
 * $Id: eqsort.c,v 1.2 1995/12/04 17:15:05 k-chinen Exp k-chinen $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <sys/types.h>
#include <sys/file.h>
#include <sys/stat.h>
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif


#define _EQSORT_TEST

#define _DEBUG_EQSORT
#define _TRACE_EQSORT


#ifdef DEBUG_EQSORT
extern show_recs(int fd, int start, int end);
#endif


/*
 * Disk access procedures
 */
int read_record(int fd, int pos, void *rec, size_t esize)
{
    int c;

    c = lseek(fd, (off_t)(pos*esize), SEEK_SET);
    if(c<0) {
        fprintf(stderr, "read_record: lseek error : pos %d, c %d\n", pos, c);
    }
    c = read(fd, rec, esize);
    if(c<=0) {
        fprintf(stderr, "read_record: read error : pos %d, c %d\n", pos, c);
    }

    return c;
}

int write_record(int fd, int pos, void *rec, size_t esize)
{
    int c;

    c = lseek(fd, (off_t)(pos*esize), SEEK_SET);
    if(c<0) {
        fprintf(stderr, "write_record: lseek error : pos %d, c %d\n", pos, c);
    }
    c = write(fd, rec, esize);
    if(c<=0) {
        fprintf(stderr, "write_record: write error : pos %d, c %d\n", pos, c);
    }

    return c;
}

#if 0
#define read_record(int fd, int pos, void *rec, size_t esize) \
    {
    if((chk = lseek(fd, (off_t)(pos*esize), SEEK_SET))<0) {
        fprintf(stderr, "read_record: lseek error : pos %d, c %d\n", pos, chk);
    }
    else
    if((chk = read(fd, rec, esize))<=0) {
        fprintf(stderr, "read_record: read error : pos %d, c %d\n", pos, chk);
    }
    }

#endif




/*
 * Internal sort and copy operations
 *      for fast speed, We use procedures with macros.
 */
#if 1

int int_sort(void *Buff, int start, int end, size_t esize,
        int (*f_comp)(const void*, const void*))
{
#ifdef TRACE_EQSORT
    fprintf(stderr, "int_sort: start=%d,end=%d\n", start, end);
#endif
    qsort((void*)((char*)Buff+(start*esize)), end-start+1, esize, f_comp);
}

let_array(void *array, int pos, void *elem, size_t esize)
{
    memcpy(((char*)array)+esize*pos, elem, esize);
}

let_rec(void *dst, void *src, size_t esize)
{
    memcpy(dst, src, esize);
}

#else

#define int_sort(Buff, start, end, esize, f_cmp) \
    qsort((void*)((char*)Buff+(start*esize)), end-start+1, esize, f_cmp)

#define let_array(array, pos, elem, esize) \
    memcpy(((char*)array)+esize*pos, elem, esize)

#define let_rec(dst, src, esize) \
    memcpy(dst, src, esize)

#endif




/*
 * main body of external quick sort
 */
int _ext_qsort(int fd, int start, int end, int blen, size_t esize, void *Buff,
        void *MinRec, void *MaxRec, int (*f_comp)(const void*, const void*))
{
#define POS(x)  (void*)(((char*)Buff+(x)*esize))
    int i, j, rlow, rupp, wlow, wupp, InBuff;
    void *LastRead, *MaxLower, *MinUpper;

    LastRead = (void*)malloc(esize);
    MaxLower = (void*)malloc(esize);
    MinUpper = (void*)malloc(esize);

#ifdef DEBUG_EQSORT
    fprintf(stderr, "_ext_qsort: fd %d, range start,end=[%6d,%6d], blen %d\n",
        fd, start, end, blen);
#endif
    while(start<end) {
        InBuff = 0;
        rupp = wupp = end;
        rlow = wlow = start;
        let_rec(MaxLower, MinRec, esize);
        let_rec(MinUpper, MaxRec, esize);
        i = start - 1;
        j = end + 1;
        while (rupp >= rlow) {
            if(rlow-wlow < wupp - rupp)
                read_record(fd, rlow++, LastRead, esize);
            else
                read_record(fd, rupp--, LastRead, esize);
            if(InBuff < blen) {
                let_array(Buff, InBuff, LastRead, esize);
                InBuff++;
                int_sort(Buff, 0, InBuff-1, esize, f_comp);
            }
            else {
                if(f_comp(LastRead, POS(blen-1)) == 1) {
                    if(f_comp(LastRead, MinUpper) == 1) {
                        j = wupp;
                    }
                    else {
                        let_rec(MinUpper, LastRead, esize);
                    }
                    write_record(fd, wupp--, LastRead, esize);
                }
                else
                if(f_comp(LastRead, POS(0)) == -1) {
                    if(f_comp(LastRead, MaxLower)) {
                        i= wlow;
                    }
                    else {
                        let_rec(MaxLower, LastRead, esize);
                    }
                    write_record(fd, wlow++, LastRead, esize);
                }
                else
                if(wlow-start < end-wupp) {
                    write_record(fd, wlow++, POS(0), esize);
                    let_rec(MaxLower, POS(0), esize);
                    let_array(Buff, 0, LastRead, esize);
                    int_sort(Buff, 0, blen-1, esize, f_comp);
                }
                else {
                    write_record(fd, wupp--, POS(blen-1), esize);
                    let_rec(MinUpper, POS(blen-1), esize);
                    let_array(Buff, blen-1, LastRead, esize);
                    int_sort(Buff, 0, blen-1, esize, f_comp);
                }
            }
        }
        while(InBuff>0) {
            write_record(fd, wupp--, POS(--InBuff), esize);
        }
        if(i-start < end-j) {
            _ext_qsort(fd, start, i, blen, esize, Buff, MinRec, MaxRec, f_comp);
            start = j;
        }
        else {
            _ext_qsort(fd, j, end, blen, esize, Buff, MinRec, MaxRec, f_comp);
            end = i;
        }
    }
    free(LastRead);
    free(MaxLower);
    free(MinUpper);

    return 1;
#undef POS
}





/*
    eqsort - external quick sort

    void eqsort(target-filename, range-start, range-end, element-size,
                buffer-length, min-sample-record, max-sample-record,
                compare-function)

    NOTES
        - generate internal buffer on heap
        - use qsort() for internal sort
        - if 'end' is -1, set end of file to 'end'.
        - if 'end' is over end of file, set end of file to 'end'
*/

int eqsort(char *fn, int start, int end, size_t esize, int blen,
            void *minrec, void *maxrec, int (*comp)(const void*, const void*))
{
    int     fd;
    void    *buf;
    time_t  before, after;
    struct stat sbuf;
    int     len;

    if((fd=open(fn, O_RDWR))<0) {
        fprintf(stderr, "Cannot open file-'%s'\n", fn);
        return -1;
    }
    if((buf=malloc(esize*(blen+1)))==NULL) {
        fprintf(stderr, "Not enough memory\n");
        close(fd);
        return -1;
    }

    /* scan number of records */
    if(fstat(fd, &sbuf)) {
        fprintf(stderr, "fstat() error\n");
        return -1;
    }
    len = (long)sbuf.st_size / esize;

#if 0
    fprintf(stderr, "; file size=%d, record size=%d\n",
        (long) sbuf.st_size, esize);
    fprintf(stderr, "; found %d record(s) in file\n", len);
#endif

    /* decriment, because zero origin */
    if(end==-1 || end>len) {
        if(len>0) {
            end = len-1;
        }
        /* no record */
        else {
            end = 0;
        }
    }

#ifdef DEBUG_EQSORT
    /*
    fprintf(stderr, "BEFORE\n");
    show_recs(fd, start, end);
    */
#endif

    time(&before);
    _ext_qsort(fd, start, end, blen, esize, buf, minrec, maxrec, comp);
    time(&after);

#if 0
    fprintf(stderr, "; eqsort: take %d secs\n", after-before);
    fflush(stderr);
#endif

#ifdef DEBUG_EQSORT
    /*
    fprintf(stderr, "AFTER\n");
    show_recs(fd, start, end);
    */
#endif

    close(fd);
    free(buf);

    return end;
}



#ifdef EQSORT_TEST

typedef struct {
    int  num;
    char name[512];
    int  len;
    int  val;
} sample_record_t;


int sample_rec_comp(const void *xx, const void *yy)
{
    sample_record_t *x=(sample_record_t*)xx, *y=(sample_record_t*)yy;
    if(x->val > y->val)         return  1;
    else if(x->val < y->val)    return -1;
        else                    return  0;
}


show_recs(int fd, int start, int end)
{
    sample_record_t rec;
    int i, last, ck;

    ck = 0;
    last = -1;
    for(i=start;i<end;i++) {
        read_record(fd, i, &rec, sizeof(sample_record_t));
#if 0
        fprintf(stderr, "%5d: %5d %5d %5d %s\n",
            i, rec.num, rec.val, rec.len, rec.name);
#endif
        fprintf(stderr, "%5d: %5d %5d\n", i, rec.num, rec.val);
        if(rec.val<last)
            ck = 1;
        last = rec.val;
    }
    if(ck) {
        fprintf(stderr, "; *not* sorted\n");
    }
    else {
        fprintf(stderr, "; *sorted*\n");
    }
}

main()
{
    sample_record_t min, max;

    min.val = INT_MIN;
    max.val = INT_MAX;
    eqsort("/var/tmp/wcol_gc.work", 0, 1000, sizeof(sample_record_t), 100,
            &min, &max, sample_rec_comp);
}


#endif
