/*
 *
 *      by k-chinen@is.aist-nara.ac.jp, 1998
 *
 * $Id$
 */


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <setjmp.h>
#include <time.h>
#ifdef TIME_WITH_SYS_TIME
#include <sys/time.h>
#endif
#include <signal.h>
#include <errno.h>


#include "wcol.h"
#include "sigs.h"
#include "socket.h"
#include "icp.h"
#include "msg.h"

#include "accept.h"



/*
 *  return:
 *      -1  fail to connect
 *      0   success
 *      1   no reply
 *
 */
int ICP_Request(char *hostname, int portno, char *url)
{
    int netd;
    char buf[BUF_SIZE];
    int nread;
    static int count=0;

    if((netd = UDP_connect(hostname, portno))<0) {
        Error("ICP_Request: cannot setup UDP socket");
        return -1;
    }

    ICP_SendQuery(netd, ++count, (unsigned char*)url);
    fflush(stderr);
    fflush(stdout);

#if 0
    if((nread = read(netd, buf, BUF_SIZE))<0) {
        Trace("ICP_Request: No reply\n");
        return 1;
    }

    if(trace_flag) {
        ICP_DumpMessage(stdout, (unsigned char*)buf);
        fflush(stdout);
    }
#endif
	close(netd);

    return 0;
}


typedef struct {
    char url[INFONAMESIZE];
} flood_req_t;

flood_req_t *req_list;
int req_count;


int make_req_list(int length)
{
    if(( req_list = (flood_req_t*) malloc(sizeof(flood_req_t)*length))==NULL) {
        fprintf(stderr, "not enought memory\n");
        req_count = -1;
        return -1;
    }
    req_count = 0;

    return length;
}


int
show_list()
{
    int i;

    fprintf(stderr, "# %d\n", req_count);
    for(i=0;i<req_count;i++) {
        fprintf(stderr, " %5d: %s\n", i, req_list[i].url);
    }

    return 0;
}

int
append_list(char *target)
{
#if 0
    fprintf(stderr, "target '%s'\n", target);
#endif
    strcpy(req_list[req_count++].url, target);

    return 0;
}

int
append_list_from_file(char *fname)
{
    FILE *fp;
    char line[INFONAMESIZE];

    fprintf(stderr, "filename '%s'\n", fname);

    if((fp=fopen(fname, "r"))==NULL) {
        fprintf(stderr, "cannot open file '%s'\n", fname);
        return -1;
    }

    while(fgets(line, INFONAMESIZE, fp)!=NULL) {
        Chop_CRLF(line);

        if(line[0]=='#') {
            continue;
        }

        append_list(line);
    }

    close(fp);

    fprintf(stderr, "read %d target(s)\n", req_count);

    return 0;
}


int
query_usage()
{
    fprintf(stderr, "icp's query mode, please check arguments\n\n");
    fprintf(stderr, "usage: icp -q host port# url\n\n");
    return 0;
}


sigjmp_buf icp_query_jumpbuf;

static RETSIGTYPE giveup_icp_query()
{
    SIGLONGJMP(icp_query_jumpbuf, 1);
}


int
main_loop(char *host, char *aport, int connection_width, int sleep_time)
{
    int port;
    int n, c;
    struct timeval currenttime;
    int chk;

    port = atoi(aport);
    fprintf(stderr, "host '%s', port '%s'(%d)\n",
        host, aport, port);
    fprintf(stderr, "connection width %d, sleeping time %d\n",
        connection_width, sleep_time);

    connection_width = 50;
    sleep_time = 1;

    n = 0;
    while(n<req_count) {
        gettimeofday(&currenttime, NULL);

        for(c=0;n<req_count && c<connection_width;c++) {
            fprintf(stderr,
                "%d.%06d: %3d: %s\n",
                currenttime.tv_sec, currenttime.tv_usec,
                n, req_list[n].url);
            chk = ICP_Request(host, port, req_list[n].url);
            n++;
        }

        fflush(stderr);
        fflush(stdout);
        sleep(sleep_time);
    }

    return 0;
}



void
usage()
{
    fprintf(stderr,
"flood - ICP Query program\n\
                by k-chinen@is.aist-nara.ac.jp, 1998\n\
\n\
usage:\n\
    flood [option] host port file [files] \n\
\n\
options:\n\
    -h          help\n\
    -t          trace mode\n\
    -L num      request list length\n\
    -w num      width (the number of connection in same time)\n\
    -s num      sleep time\n\
\n\
example:\n\
    %% flood foo.bar.com 3130 target-list\n\
    %% flood -s 1 foo.bar.com 3130 ping-list\n\
\n");

}



main(int argc, char **argv)
{
    int query_mode = 0, listen_mode = 0;
    int listen_port = ICP_PORTNO;
    int list_length=10000;

    int flag;
    int i;
    int cwidth, stime;
    extern char *optarg;
    extern int  optind;

    /* for query mode */
    char qhost[STRING_SIZE], qport[STRING_SIZE];
    int try_limit;

    /* set defaults */
    try_limit = 10;
    trace_flag = 1;
    qhost[0] = '\0';
    qport[0] = '\0';
    cwidth = 10;
    stime  = 5;

    /* parse args */
    while((flag = getopt(argc, argv, "htL:w:s:")) != -1) {
        switch(flag) {
        case 'h':
            usage();
            exit(1);
            break;
        case 't':
            trace_flag = 1 - trace_flag;
            break;
        case 'L':
            list_length = atoi(optarg);
            break;
        case 'w':
            cwidth = atoi(optarg);
            break;
        case 's':
            stime = atoi(optarg);
            break;
        default:
            usage();
            exit(4);
        }
    }

    make_req_list(list_length);

    if(optind<argc) {
        strcpy(qhost, argv[optind]);
        optind++;
    }
    else {
        usage();
        exit(2);
    }

    if(optind<argc) {
        strcpy(qport, argv[optind]);
        optind++;
    }
    else {
        usage();
        exit(2);
    }



    for(i=optind; i<argc; i++) {
        fprintf(stderr, " %3d: %s\n", i, argv[i]);
        append_list_from_file(argv[i]);
    }

    show_list();

    main_loop(qhost, qport, cwidth, stime);

    exit(0);
}

