/*
 * ftp.c - ftp routines
 *      original by T.Yoeshida, 1996
 *      modify by k-chinen@is.aist-nara.ac.jp, 1997
 *
 * $Id$
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <ctype.h>
#include <time.h>
#include <errno.h>
#include <netdb.h>
#include <signal.h>
#include <setjmp.h>

#include <sys/types.h>
#include <sys/uio.h>
/*
#include <sys/filio.h>
*/

#include <sys/socket.h>
/*
#include <netinet/in_systm.h>
#include <netinet/in.h>
*/
#include <netinet/in.h>
#include <arpa/inet.h>


#include <errno.h>


#include "wcol.h"
#include "sigs.h"
#include "url.h"
#include "hname.h"
#include "socket.h"

#include "fdio.h"
#include "ftp.h"


#ifndef STRING_SIZE
#define STRING_SIZE 2048
#endif

#ifndef BUF_SIZE
#define BUF_SIZE 10240
#endif



#ifdef __STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif


extern int      pipe_broken;
extern int      errno;


int ftp_accept_timeout=10;





#ifdef __STDC__
int             putline(const int fd, const char *fmt,...)
#else
int             putline(va_alist) va_dcl
#endif
{
    va_list         args;

#ifndef __STDC__
    int             fd;
    char           *fmt;
#endif
    char            buffer[BUF_SIZE];
    int             nwrite;


#ifdef __STDC__
    va_start(args, fmt);
#else
    va_start(args);
    fd = va_arg(args, int);
    fmt = va_arg(args, char *);
#endif

    vsprintf(buffer, fmt, args);
    Trace("<<      :%s", buffer);
try:
    nwrite = write(fd, buffer, strlen(buffer));
    if (pipe_broken) {
        pipe_broken = 0;
    }
    if (nwrite < 0) {
        if (errno == EINTR) {
            goto try;
        }
        else {
            Trace("putline: fail write() (%d)\n", errno);
        }
    }

    va_end(args);

    return nwrite;
}



char *
getline(int fd, char *buf, int n, int *rcode)
{
    char           *ret;
    char            rbuf[256];
    register char  *p, *q;
    int             c;

    memset(buf, '\0', n);

#if 0
    ret = x_getline(fd, buf, n);
#endif
    ret = fdgets(fd, buf, n);
    if (ret == NULL) {
        *rcode = -1;
    }
    else {
        p = buf;
        q = rbuf;
        c = 0;
        if(isdigit(*p)) {
            while (*p && c < 256 && isdigit(*p)) {
                *q++ = *p++;
                c++;
            }
            *q = '\0';
            *rcode = atoi(rbuf);
            if (*p == '-') {
                *rcode = -*rcode;
            }
        }
        else {
            *rcode = -1;
        }
    }

    Trace(">> %4d :%s", *rcode, buf);

    return ret;
}








/*
 *
 * return
 *      success 0
 *      error   1
 */
int
FTP_User(int sd, char *uname)
{
    int             rcode;
    char            buf[BUF_SIZE];

    putline(sd, "USER %s\n", uname);

    getline(sd, buf, BUF_SIZE, &rcode);
    if (rcode == 331) {
        return 0;
    }

    return 1;

}

/*
 *
 * return
 *      success 0
 *      error   1
 */
int
FTP_Pass(int sd, char *passwd)
{
    int             rcode;
    char            buf[BUF_SIZE];

    putline(sd, "PASS %s\n", passwd);

loop:
    getline(sd, buf, BUF_SIZE, &rcode);
    if (rcode<0) {
        goto loop;
    }
    if (rcode == 230) {
        Trace("Login !\n");
        return 0;
    }
    else if (rcode == 230) {
        Trace("Login Fail !, bad password\n");
        return 1;
    }
    else {
        Trace("Login Fail !, error happen\n");
        return 1;
    }

}

/*
 *
 * return
 *      success 0
 *      error   1
 */
int
FTP_Type(int sd, char mode)
{
    int             rcode;
    char            buf[BUF_SIZE];

    putline(sd, "TYPE %c\n", mode);

    getline(sd, buf, BUF_SIZE, &rcode);
    if (rcode == 200) {
        return 0;
    }
    else {
        return 1;
    }
}

/*
 * return
 *      success !NULL
 *      error   NULL
 */
char
*FTP_Pwd(int sd)
{
    int             rcode;
    static char     buf[BUF_SIZE];

    putline(sd, "PWD\n");

    getline(sd, buf, BUF_SIZE, &rcode);
    if (rcode == 257) {
        Trace("got current directory !\n");
        return buf;
    }
    else {
        Trace("could not get current directory !\n");
        return NULL;
    }
}

/*
 * return
 *      success 0
 *      error   1
 */
int
FTP_Cd(int sd, char *dir)
{
    int             rcode;
    static char     buf[BUF_SIZE];

    putline(sd, "CWD %s\n", dir);

loop:
    getline(sd, buf, BUF_SIZE, &rcode);
    if (rcode<0) {
        goto loop;
    }
    if (rcode == 250) {
        return 0;
    }

    return 1;
}

/*
 * return
 *      success 0
 *      error   1
 */
int
FTP_Quit(sd)
    int             sd;
{
    int             rcode;
    static char     buf[BUF_SIZE];

    putline(sd, "QUIT\n");

    getline(sd, buf, BUF_SIZE, &rcode);

    if (rcode == 221) {
        Trace("complete\n");
        return 0;
    }

    return 1;
}


static sigjmp_buf accept_timeout_jmp;

static RETSIGTYPE
TimeoutAccept()
{
    SIGLONGJMP(accept_timeout_jmp, 1);
}


/*
 * Establish Data Connection via PORT
 */
int DataInitconnActive(sd)
{
    int             rcode;
    char            buf[BUF_SIZE];

    struct sockaddr_in listen_addr;
    register char  *p, *a;

    int             result, len;
    int             on = 1;
    int             listen_fd;
    struct sockaddr_in from;
    int             data, fromlen = sizeof(from), tos;


#if 1
    len = sizeof(listen_addr);
    if (getsockname(sd, (struct sockaddr *) & listen_addr, &len) < 0) {
        perror("ftp: getsockname");
        return (-1);
    }

    listen_fd = socket(AF_INET, SOCK_STREAM, 0);
    if (listen_fd < 0) {
        perror("ftp: socket");
        close(listen_fd);
        return (-1);
    }
    printf("DATA SOCK fd# %d\n", listen_fd);

    if (setsockopt(listen_fd, SOL_SOCKET, SO_REUSEADDR,
            (char *) &on, sizeof(on)) < 0) {
        perror("ftp: setsockopt (reuse address)");
        close(listen_fd);
        return (-1);
    }

    listen_addr.sin_port = 0;
    if (bind(listen_fd,
            (struct sockaddr *) &listen_addr, sizeof(listen_addr)) < 0) {
        perror("ftp: bind");
        close(listen_fd);
        return (-1);
    }
/*
    if (options & SO_DEBUG &&
        setsockopt(listen_fd, SOL_SOCKET, SO_DEBUG,
            (char *)&on, sizeof (on)) < 0) {
        perror("ftp: setsockopt (ignored)");
    }
*/

    if (listen(listen_fd, 1) < 0)
        perror("ftp: listen");

#else

    if((listen_fd = TCP_bind_port(0, 256))<=0) {
        Error("fail bind port");
        return -1;
    }
#endif


    len = sizeof(listen_addr);
    if (getsockname(listen_fd, (struct sockaddr *) & listen_addr, &len) < 0) {
        perror("ftp: getsockname");
        close(listen_fd);
        return (-1);
    }


    a = (char *) &listen_addr.sin_addr;
    p = (char *) &listen_addr.sin_port;

#define UC(b)   (((int)b)&0xff)


    putline(sd, "PORT %d,%d,%d,%d,%d,%d\n",
        UC(a[0]), UC(a[1]), UC(a[2]), UC(a[3]),
        UC(p[0]), UC(p[1]));
    getline(sd, buf, BUF_SIZE, &rcode);


#undef UC

    if(SIGSETJMP(accept_timeout_jmp, 1)) {
        SetSignalHandler(SIGALRM, SIG_IGN);

        Trace("Timeout, give up to wait server's active connection\n");

        close(listen_fd);
        return -1;
    }

    SetSignalHandler(SIGALRM, TimeoutAccept);
    alarm(ftp_accept_timeout);

    data = accept(listen_fd, (struct sockaddr *) & from, &fromlen);
    if (data < 0) {
        perror("ftp: accept");
        (void) close(listen_fd), listen_fd = -1;
        return (-1);
    }
    (void) close(listen_fd);

    SetSignalHandler(SIGALRM, SIG_IGN);

    return data;
}

int OldDataInitconnActive(sd)
{
    int             rcode;
    char            buf[BUF_SIZE];

    struct sockaddr_in listen_addr;
    register char  *p, *a;
    int             result, len;
    int             on = 1;
    int             listen_fd;
    struct sockaddr_in from;
    int             data, fromlen = sizeof(from), tos;


    listen_fd = socket(AF_INET, SOCK_STREAM, 0);
    if (listen_fd < 0) {
        perror("ftp: socket");
        return (-1);
    }
    printf("DATA SOCK fd# %d\n", listen_fd);

    len = sizeof(listen_addr);
    if (getsockname(sd, (struct sockaddr *) & listen_addr, &len) < 0) {
        perror("ftp: getsockname");
        close(listen_fd);
        return (-1);
    }

    if (setsockopt(listen_fd, SOL_SOCKET, SO_REUSEADDR,
            (char *) &on, sizeof(on)) < 0) {
        perror("ftp: setsockopt (reuse address)");
        close(listen_fd);
        return (-1);
    }

    listen_addr.sin_port = 0;
    if (bind(listen_fd,
            (struct sockaddr *) &listen_addr, sizeof(listen_addr)) < 0) {
        perror("ftp: bind");
        close(listen_fd);
        return (-1);
    }
/*
    if (options & SO_DEBUG &&
        setsockopt(listen_fd, SOL_SOCKET, SO_DEBUG,
            (char *)&on, sizeof (on)) < 0) {
        perror("ftp: setsockopt (ignored)");
    }
*/

    if (listen(listen_fd, 1) < 0)
        perror("ftp: listen");

    len = sizeof(listen_addr);
    if (getsockname(listen_fd, (struct sockaddr *) & listen_addr, &len) < 0) {
        perror("ftp: getsockname");
        close(listen_fd);
        return (-1);
    }


    a = (char *) &listen_addr.sin_addr;
    p = (char *) &listen_addr.sin_port;

#define UC(b)   (((int)b)&0xff)


    putline(sd, "PORT %d,%d,%d,%d,%d,%d\n",
        UC(a[0]), UC(a[1]), UC(a[2]), UC(a[3]),
        UC(p[0]), UC(p[1]));
    getline(sd, buf, BUF_SIZE, &rcode);


#undef UC

    if(SIGSETJMP(accept_timeout_jmp, 1)) {
        SetSignalHandler(SIGALRM, SIG_IGN);

        Trace("Timeout, give up to wait server's active connection\n");

        close(listen_fd);
        return -1;
    }

    SetSignalHandler(SIGALRM, TimeoutAccept);
    alarm(10);

    data = accept(listen_fd, (struct sockaddr *) & from, &fromlen);
    if (data < 0) {
        perror("ftp: accept");
        (void) close(listen_fd), listen_fd = -1;
        return (-1);
    }
    (void) close(listen_fd);

    SetSignalHandler(SIGALRM, SIG_IGN);

    return data;
}


/*
 * Establish Data Connection via PASV
 */
int DataInitconnPassive(sd)
    int             sd;
{
    int             rcode;
    char            buf[BUF_SIZE];

    int             data;
    struct sockaddr_in data_addr;
    short           port;
    char            addr_str[256];

    putline(sd, "PASV\n");
    getline(sd, buf, BUF_SIZE, &rcode);

    if(rcode==227) {
        /* for example, server return such thing */
        /* 227 Entering Passive Mode (133,1,119,4,18,90) */
    }
    else {
        return -1;
    }

    /*
     * parse address and port
     */
    {
        register char  *p, *q;
        int             c;
        int             hi, low;

        p = buf;

        while(*p && isdigit(*p)) {
            p++;
        }
        if(*p!=' '){
            goto fail;
        }
        p++;

/*
        Trace("message [%s]\n", p);
*/

        while(*p && *p!='(') {
            p++;
        }
        if(*p!='(') {
            goto fail;
        }
        p++;

/*
        Trace("specify addr and port [%s]\n", p);
*/

        memset(addr_str, 0, 256);
        q = addr_str;
        c = 0;
        while(*p && (isdigit(*p)||*p==',')) {
            if(*p==',') {
                c++;
                if(c==4)
                    break;
                *q++ = '.';
                *p++;
            }
            else {
                *q++ = *p++;
            }
/*
            Trace("\taddr %-20s %s", addr_str, p);
*/
        }
        *q = '\0';
        if(*p!=',') {
            goto fail;
        }
        p++;
/*
        Trace("address [%s]\n", addr_str);
*/

        hi = 0;
        while(*p && isdigit(*p)) {
            hi = hi*10 + (*p - '0');
/*
            Trace("\tport hi  %4d %s", hi, p);
*/
            p++;
        }
        if(*p!=',') {
            goto fail;
        }
        p++;

        low = 0;
        while(*p && isdigit(*p)) {
            low = low*10 + (*p - '0');
/*
            Trace("\tport low %4d %s", low, p);
*/
            p++;
        }
        if(*p!=')') {
            goto fail;
        }
        p++;

        port = hi * 256 + low;
/*
        Trace("port [%d] (%d,%d)\n", port, hi, low);
*/


        goto success;
    }
fail:
    Trace("strange address or port\n");

    return -1;

success:
    Trace("address [%s] port [%d]\n", addr_str, port);

    fflush(stdout);
    fflush(stderr);


    data = socket(AF_INET, SOCK_STREAM, 0);
    if (data < 0) {
        perror("ftp: socket");
        return (-1);
    }
    printf("DATA SOCK fd# %d\n", data);


    memset(&data_addr, 0, sizeof(data_addr));
    data_addr.sin_family = AF_INET;
    data_addr.sin_port   = htons(port);
    data_addr.sin_addr.s_addr = inet_addr(addr_str);


conn:
    if(connect(data, (struct sockaddr*) &data_addr,
            sizeof(struct sockaddr_in)) < 0) {
        if(errno == EINTR)
            goto conn;

        Trace("fail connect() (%d)\n", errno);

        close(data);
        return -1;
    }


    return data;
}


int DataInitconn(sd, ctype)
    int sd;
    int ctype;
{
    if(ctype==CONN_TYPE_PASV) {
        return DataInitconnPassive(sd);
    }
    else
    if(ctype==CONN_TYPE_PORT) {
        return DataInitconnActive(sd);
    }
    else {
        Trace("DataInitconn: strange connectino type %d\n", ctype);

        return -1;
    }
}

/*************************/
int
DataSend(sd, data, lpath, local, mode)
    int             sd;
    int             data;
    char           *lpath, *local;
    char            mode;
{
    int             rcode;
    char            buf[BUF_SIZE];

    register int    c, d;
    int             bytes = 0;
    char           *bufp;
    char            fname[256], remote[256];
    char            dbuf[BUFSIZ];
    FILE           *fin;
    FILE           *dout;

    struct sockaddr_in from;
    int             s, fromlen = sizeof(from), tos;

    strcpy(remote, local);

    putline(sd, "STOR %s\n", remote);
    getline(sd, buf, BUF_SIZE, &rcode);
    if (rcode == 500) {
        return 0;
    }

#if 0
    memset(smes, ' ', sizeof(smes));
    sprintf(smes, "STOR %s\n", remote);
    write(sd, smes, strlen(smes));

    memset(rmes, ' ', sizeof(rmes));
    len = read(sd, rmes, sizeof(rmes));
    if (len < 0)
        return (len);
#endif


#if 0
    printf(">>>%30s\n", rmes);
#endif

    s = accept(data, (struct sockaddr *) & from, &fromlen);
    if (s < 0) {
        perror("ftp: accept");
        (void) close(data), data = -1;
        return (-1);
    }
    (void) close(data);
    data = s;

    printf("Data-conn accept OK\n");

    sprintf(fname, "%s/%s", lpath, local);
    fin = fopen(fname, "r");
    if (fin == NULL)
        return (-1);

    dout = fdopen(data, "w");

    switch (mode) {
        case TYPE_I:
            errno = d = 0;
            while ((c = read(fileno(fin), dbuf, sizeof(dbuf))) > 0) {
                bytes += c;
                for (bufp = dbuf; c > 0; c -= d, bufp += d)
                    if ((d = write(data, bufp, c)) <= 0)
                        break;
            }
            if (c < 0)
                fprintf(stderr, "local: %s: %s\n", local,
                        strerror(errno));
            if (d <= 0) {
                if (d == 0)
                    fprintf(stderr, "netout: write returned 0?\n");
                else if (errno != EPIPE)
                    perror("netout");
                bytes = -1;
            }
            break;

        case TYPE_A:
            while ((c = getc(fin)) != EOF) {
                if (c == '\n') {
                    if (ferror(dout))
                        break;
                    (void) putc('\r', dout);
                    bytes++;
                }
                (void) putc(c, dout);
                bytes++;
            }

            if (ferror(fin))
                fprintf(stderr, "local: %s: %s\n", local,
                        strerror(errno));
            if (ferror(dout)) {
                if (errno != EPIPE)
                    perror("netout");
                bytes = -1;
            }
            break;
    }
/*
        (void) gettimeofday(&stop, (struct timezone *)0);
*/
    (void) fclose(fin);
    (void) fclose(dout);

    if (bytes > 0)
/*
                ptransfer("sent", bytes, &start, &stop);
*/
        printf("%ld Bytes Sent\n", bytes);


    return 0;
}

/*************************/
int
DataRecvActive(int sd, int data, char *rpath, char *lpath, char mode)
{
    int             rcode;
    char            buf[BUF_SIZE];

    int             bytes = 0, c, d;
    int             bare_lfs;
    char           *bufp;
    char            dbuf[BUFSIZ];
    char            fname[256], remote[256];
    FILE           *fout;
    FILE           *din;

    Trace(
      "DataRecvActive: sd %d, data %d, mode %d\n\trpath '%s'\n\tlpath '%s'\n",
      sd, data, mode, rpath, lpath);

    putline(sd, "RETR %s\n", rpath);
    getline(sd, buf, BUF_SIZE, &rcode);
    if (rcode == 500) {
        Trace("error or down");
        return -1;
    }
    if (rcode == 550) {
        Trace("error or down");
        return -1;
    }
    if (rcode == 425) {
        Trace("error or down");
        return -1;
    }


    sprintf(fname, "%s", lpath);
    fout = fopen(fname, "w");
    if (fout == NULL)
        return -1;

    din = fdopen(data, "r");

    switch (mode) {
        case TYPE_I:
            errno = d = 0;
            while ((c = read(fileno(din), dbuf, sizeof(dbuf))) > 0) {
                if ((d = write(fileno(fout), dbuf, c)) != c)
                    break;
                bytes += c;
            }
            if (c < 0) {
                if (errno != EPIPE)
                    perror("netin");
                bytes = -1;
            }
            if (d < c) {
                if (d < 0)
                    fprintf(stderr, "file: %s: %s\n", rpath,
                            strerror(errno));
                else
                    fprintf(stderr, "%s: short write\n", rpath);
            }
            break;

        case TYPE_A:

            while ((c = getc(din)) != EOF) {
                if (c == '\n')
                    bare_lfs++;
                while (c == '\r') {
#if 0
                                while (hash && (bytes >= hashbytes)) {
                                        (void) putchar('#');
                                        (void) fflush(stdout);
                                        hashbytes += HASHBYTES;
                                }
#endif
                    bytes++;
                    if ((c = getc(din)) != '\n') {
                        if (ferror(fout))
                            goto break2;
                        (void) putc('\r', fout);
                        if (c == '\0') {
                            bytes++;
                            goto contin2;
                        }
                        if (c == EOF)
                            goto contin2;
                    }
                }
                (void) putc(c, fout);
                bytes++;
        contin2:;
    } break2:
            if (bare_lfs) {
                printf("WARNING! %d bare linefeeds received in ASCII mode\n", bare_lfs);
                printf("File may not have transferred correctly.\n");
            }

            if (ferror(din)) {
                if (errno != EPIPE)
                    perror("netin");
                bytes = -1;
            }
            if (ferror(fout))
                fprintf(stderr, "file: %s: %s\n", lpath,
                        strerror(errno));
            break;
    }
    fclose(fout);
    close(data);

    return 0;
}


#if 0
int
DataRecvPassive(int sd, int data, char *rpath, char *lpath, char mode)
{
    int             rcode;
    char            buf[BUF_SIZE];

    int             bytes = 0, c, d;
    int             bare_lfs;
    char           *bufp;
    char            dbuf[BUFSIZ];
    char            fname[256], remote[256];
    FILE           *fout;
    FILE           *din;

#if 0
    struct sockaddr_in from;
    int             s, fromlen = sizeof(from), tos;
#endif


    putline(sd, "RETR %s\n", rpath);
    getline(sd, buf, BUF_SIZE, &rcode);
    if (rcode == 500) {
        return 0;
    }
    if (rcode == 425) {
        return 0;
    }


    sprintf(fname, "%s", lpath);
    fout = fopen(fname, "w");
    if (fout == NULL)
        return (-1);

    din = fdopen(data, "r");

    switch (mode) {
        case TYPE_I:
            errno = d = 0;
            while ((c = read(fileno(din), dbuf, sizeof(dbuf))) > 0) {
                if ((d = write(fileno(fout), dbuf, c)) != c)
                    break;
                bytes += c;
            }
            if (c < 0) {
                if (errno != EPIPE)
                    perror("netin");
                bytes = -1;
            }
            if (d < c) {
                if (d < 0)
                    fprintf(stderr, "file: %s: %s\n", rpath,
                            strerror(errno));
                else
                    fprintf(stderr, "%s: short write\n", rpath);
            }
            break;

        case TYPE_A:

            while ((c = getc(din)) != EOF) {
                if (c == '\n')
                    bare_lfs++;
                while (c == '\r') {
#if 0
                                while (hash && (bytes >= hashbytes)) {
                                        (void) putchar('#');
                                        (void) fflush(stdout);
                                        hashbytes += HASHBYTES;
                                }
#endif
                    bytes++;
                    if ((c = getc(din)) != '\n') {
                        if (ferror(fout))
                            goto break2;
                        (void) putc('\r', fout);
                        if (c == '\0') {
                            bytes++;
                            goto contin2;
                        }
                        if (c == EOF)
                            goto contin2;
                    }
                }
                (void) putc(c, fout);
                bytes++;
        contin2:;
    } break2:
            if (bare_lfs) {
                printf("WARNING! %d bare linefeeds received in ASCII mode\n", bare_lfs);
                printf("File may not have transferred correctly.\n");
            }

            if (ferror(din)) {
                if (errno != EPIPE)
                    perror("netin");
                bytes = -1;
            }
            if (ferror(fout))
                fprintf(stderr, "local: %s: %s\n", lpath,
                        strerror(errno));
            break;
    }
    fclose(fout);
    close(data);

    return 0;
}


#endif



/*
 *
 */
int
FTP_Put(int sd, int ctype, char *path, char *fname, char mode)
{
    int             data;
    int             rcode;
    char            buf[BUF_SIZE];


    data = DataInitconn(sd, ctype);
    if(data<0) {
        return -1;
    }
    DataSend(sd, data, path, fname, mode);

    getline(sd, buf, BUF_SIZE, &rcode);


    return 0;
}




int
FTP_GetPassive(int sd, char *rpath, char *lpath, char mode)
{
    int  rcode;
    char buf[BUF_SIZE];
    int  data;
    int chk;

    Trace("FTP_GetPassive:\n\tremote '%s'\n\tlocal  '%s'\n", rpath, lpath);

    if(rpath==NULL || rpath[0]=='\0') {
        Error("empty remote path\n");
        return -1;
    }
    if(lpath==NULL || lpath[0]=='\0') {
        Error("empty local path\n");
        return -1;
    }

    data = DataInitconnPassive(sd);
    if(data<0) {
        return -1;
    }
    if(chk=DataRecvActive(sd, data, rpath, lpath, mode)) {
        Error("FTP_GetPassive: Fail retrive data <%d>\n", chk);
        return -1;
    }

    getline(sd, buf, BUF_SIZE, &rcode);


    return 0;
}

int
FTP_GetActive(int sd, char *path, char *fname, char mode)
{
    int  rcode;
    char buf[BUF_SIZE];
    int  data;

    Trace("FTP_GetActive:\n\tremote '%s'\n\tlocal  '%s'\n", path, fname);

    if(path==NULL || path[0]=='\0') {
        Error("empty remote path\n");
        return -1;
    }
    if(fname==NULL || fname[0]=='\0') {
        Error("empty local path\n");
        return -1;
    }

    data = DataInitconnActive(sd);
    if(data<0) {
        return -1;
    }
    if(DataRecvActive(sd, data, path, fname, mode)==0) {
        Error("FTP_GetActive: Fail retrive data\n");
        return -1;
    }

    getline(sd, buf, BUF_SIZE, &rcode);


    return 0;
}


/*
 *
 */

int FTP_Get(int sd, int ctype, char *path, char *fname, char mode)
{
    if(ctype==CONN_TYPE_PASV) {
        return FTP_GetPassive(sd, path, fname, mode);
    }
    else {
        return FTP_GetActive(sd, path, fname, mode);
    }
}


#ifdef FTP_TEST

/*
 *
 * return:
 *      success  0
 *      error   -1
 */

int
FTP_Retrieve(char *user, char *pass, char *xurl, char *tmp, int mode)
{
    url_t *url;
    int  port;
#if 0
    char fn[STRING_SIZE], tmp[STRING_SIZE];
#endif
    char buffer[BUF_SIZE];

    int cfd;
    int rcode;

    url = NULL;
    url = URL_NewVal(xurl);
    if(url==NULL) {
        Error("no mmemory");
    }
    URL_Normalize(url);

    Trace("==\n");
    Trace("user     '%s'\n", user);
    Trace("pass     '%s'\n", pass);
    Trace("--\n");
    Trace("URL      '%s'\n", url->full_loc);
    Trace("protocol '%s'\n", url->proto);
    Trace("server   '%s'\n", url->host);
    port = atoi(url->port);
    Trace("port     '%s', %4d\n", url->port, port);
    Trace("path     '%s'\n", url->path);

#if 0
    AssignFileName(fn, url, BODY_TMP_FILE);
    Trace("local file '%s'\n", fn);
    if(SetupDir(fn)) {
        Error("Cannot setup file '%s'", fn);
        return -1;
    }
#endif

    if((cfd = TCP_connect(url->host, port))<0) {
        Log("; server '%s' port %d is down", url->host, port);
        URL_Delete(url);
        return -1;
    }

c1:
    getline(cfd, buffer, BUF_SIZE, &rcode);
    if(rcode<=0) {
        goto c1;
    }

    if(FTP_User(cfd, user)) {
        URL_Delete(url);
        close(cfd);
        return -1;
    }
    if(FTP_Pass(cfd, pass)) {
        URL_Delete(url);
        close(cfd);
        return -1;
    }
    if(FTP_Type(cfd, TYPE_I)) {
        URL_Delete(url);
        close(cfd);
        return -1;
    }
/*
    if(FTP_Pwd(cfd, dir)) {
        URL_Delete(url);
        close(cfd);
        return -1;
    }
*/
    Trace("local file '%s'\n", tmp);

    /* automatic */
    if(mode==-1) {
        if(FTP_GetPassive(cfd, url->path, tmp, TYPE_I)) {
            URL_Delete(url);
            close(cfd);
            if(FTP_GetActive(cfd, url->path, tmp, TYPE_I)) {
                URL_Delete(url);
                close(cfd);
            }
            return -1;
        }
    }
    else
    /* active */
    if(mode==1) {
        if(FTP_GetActive(cfd, url->path, tmp, TYPE_I)) {
            URL_Delete(url);
            close(cfd);
            return -1;
        }
    }
    else
    /* passive */
    if(mode==2) {
        if(FTP_GetPassive(cfd, url->path, tmp, TYPE_I)) {
            URL_Delete(url);
            close(cfd);
            return -1;
        }
    }


    URL_Delete(url);

    return 0;
}






int
usage()
{
    fprintf(stdout, "\n");
    fprintf(stdout, "usage: ftp [options] URL\n");
    fprintf(stdout, "\n");
    fprintf(stdout, "option: -h         help\n");
    fprintf(stdout, "        -A         active connection\n");
    fprintf(stdout, "        -P         passive connection\n");
    fprintf(stdout, "        -u string  user\n");
    fprintf(stdout, "        -p string  password\n");
    fprintf(stdout, "        -X file    store file\n");

    return 0;
}




main(int argc, char **argv)
{
    int flg;
    int cmode = -1;
    char user[STRING_SIZE], pass[STRING_SIZE];
    char path[STRING_SIZE];
    extern char *optarg;
    extern int optind;

    trace_flag = 1;
    strcpy(user, "anonymous");
    strcpy(pass, "dummy@hyper.net");
    strcpy(path, "/var/tmp/x.ftp");

    while((flg=getopt(argc, argv, "vhAPu:p:X:"))!=-1) {
        switch(flg) {
        case 'h':
            usage();
            exit(0);
            break;
        case 'A':
            cmode = 1;
            break;
        case 'P':
            cmode = 2;
            break;
        case 'u':
            strcpy(user, optarg);
            break;
        case 'p':
            strcpy(pass, optarg);
            break;
        case 'X':
            strcpy(path, optarg);
            break;
        }
    }

    Trace("connection direction mode: ");
    if(cmode==-1) {
        Trace("AUTOMATIC (passive -> active)\n");
    }
    if(cmode==1) {
        Trace("ACTIVE\n");
    }
    if(cmode==2) {
        Trace("PASSIVE\n");
    }

    FTP_Retrieve(user, pass, argv[optind], path , cmode);

}

#endif

