/*
 * ftphand.c - FTP handlers
 *      modify by k-chinen@is.aist-nara.ac.jp, 1997
 *
 * $Id: ftphand.c,v 1.1 1997/09/15 02:00:00 k-chinen Exp k-chinen $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <ctype.h>
#include <time.h>
#include <errno.h>
#include <netdb.h>
#include <signal.h>
#include <setjmp.h>
#include <time.h>
#ifdef TIME_WITH_SYS_TIME
#include <sys/time.h>
#endif

#include <sys/types.h>
#include <sys/uio.h>
/*
#include <sys/filio.h>
*/

#include <sys/socket.h>
/*
#include <netinet/in_systm.h>
#include <netinet/in.h>
*/
#include <netinet/in.h>
#include <arpa/inet.h>


#include <errno.h>


#include "wcol.h"
#include "msg.h"
#include "url.h"
#include "request.h"
#include "sigs.h"
#include "hname.h"
#include "socket.h"
#include "misc.h"

#include "fdio.h"
#include "ftp.h"
#include "ftphand.h"

#include "global.h"


static sigjmp_buf retrieve_jmp_point;

static RETSIGTYPE giveup_retrieve()
{
    SIGLONGJMP(retrieve_jmp_point, 1);
}



/*
 *
 * arguments:
 *      sd      control stream of FTP
 *      data    data stream of FTP
 *      cfd     client stream of HTTP
 *      dfd     cache file descriptor
 *      rpath   remote path
 *      lpath   local path
 *
 *
 *
 *      +------+ sd   +------+     +------+
 *      |FTP   |----->|FTP   |     |HTTP  |
 *      |server|----->|hander|---->|client|
 *      +------+ data +------+ cfd +------+
 *                       \dfd
 *                      +-----+
 *                      |cache|
 *                      +-----+
 *
 */
/*
 * return
 *      error   !0
 *      success 0
 */
int
ForwardStore_Passive(int sd, int data, int cfd, int dfd, char *rpath,
        off_t *sum_read, off_t *sum_write_client, off_t *sum_write_cache
        )
{
    int  rcode;
    char line[STRING_SIZE];
    char buf[BUF_SIZE];
/*
    off_t sum_read;
    off_t sum_write_client, sum_write_cache;
*/
    off_t nread, nwrite;
    int  ret;


    Trace("ForwardStore_Passive: sd %d, data %d, cfd %d, dfd %d\n",
        sd, data, cfd, dfd);
    Trace("  rpath '%s'\n", rpath);


    ret = 0;

    putline(sd, "RETR %s\n", rpath);
    getline(sd, line, STRING_SIZE, &rcode);
    if (rcode == 500) {
        return 1;
    }
    if (rcode == 550) {
        return 1;
    }
    if (rcode == 425) {
        return 1;
    }


    if(SIGSETJMP(retrieve_jmp_point, 1)) {
        Error("ForwardStore_Passive: long waitting for retrieve, give up.");
        SetSignalHandler(SIGALRM, SIG_IGN);

        return -1;
    }
    SetSignalHandler(SIGALRM, giveup_retrieve);
    alarm(7200);

    *sum_read  = (off_t)0;
    *sum_write_client = (off_t)0;
    *sum_write_cache = (off_t)0;


    pipe_broken = 0;
    nwrite =
        fdprintf(cfd, "HTTP/1.0 200\r\nVia: FTPHandler of Wcol-E\r\n\r\n");
    if(pipe_broken) {
        Error("ForwardStore: fail write HTTP header");

        pipe_broken = 0;
    }

    *sum_write_client += nwrite;


    while(1) {

r1:
        pipe_broken = 0;
        nread = read(data, buf, BUF_SIZE);
        if(nread<0) {
            if(errno==EINTR) {
                goto r1;
            }
            Error("ForwardStore: fail read() %ld/%ld/%ld (%d)",
                (long)sum_read,
                (long)sum_write_client, (long)sum_write_cache, errno);

            ret = -2;
            break;
        }
        if(nread==0) {
            /* connection is over */
            break;
        }

        *sum_read += (off_t) nread;

w1:
        if(cfd>=0) {
            pipe_broken = 0;
            nwrite = write(cfd, buf, nread);

            if(pipe_broken) {
                close(cfd);
                cfd = -1;

                pipe_broken = 0;
            }
            if(nwrite<0) {
                if(errno==EINTR) {
                    goto w1;
                }

            Error("ForwardStore: fail write() client %d %ld/%ld/%ld (%d)",
                cfd, (long)sum_read,
                (long)sum_write_client, (long)sum_write_cache, errno);

                close(cfd);
                cfd = -1;

                ret = -3;
#if 0
                break;
#endif
            }
            *sum_write_client += (off_t) nwrite;
        }

w2:
        if(dfd>=0) {
            pipe_broken = 0;
            nwrite = write(dfd, buf, nread);

            if(nwrite<0) {
                if(errno==EINTR) {
                    goto w2;
                }

            Error("ForwardStore: fail write() cache %d %ld/%ld/%ld (%d)",
                dfd, (long)sum_read,
                (long)sum_write_client, (long)sum_write_cache, errno);

                ret = -4;
                break;
            }
            *sum_write_cache += (off_t) nwrite;
        }

    }

    Trace("ForwardStore_Passive: ret %d\n", ret);
    Trace("  read %ld\n", (long)*sum_read);
    Trace("  write client %ld / cache %ld \n",
        (long)*sum_write_client, (long)*sum_write_cache);

    return ret;
}

int
List_ForwardStore_Passive(int sd, int data, int cfd, int dfd, char *rpath,
        off_t *sum_read, off_t *sum_write_client, off_t *sum_write_cache,
        char *xurl, int *http_ret
        )
{
    int  rcode;
    char line[STRING_SIZE];
    char buf[BUF_SIZE];

/*
    off_t sum_read;
    off_t sum_write_client, sum_write_cache;
*/
    off_t nread, nwrite;
    int  ret;

    char raw_list[BUF_SIZE];
    off_t rlen;
    char cooked_list[BUF_SIZE];
    off_t clen;
    char *cpos;
    off_t sum_list;
    int c;

    time_t now;

    char tline[STRING_SIZE];


    Trace("List_ForwardStore_Passive: sd %d, data %d, cfd %d, dfd %d\n",
        sd, data, cfd, dfd);
    Trace("  rpath '%s'\n", rpath);


    ret = 0;

    *sum_read  = (off_t)0;
    *sum_write_client = (off_t)0;
    *sum_write_cache = (off_t)0;


    putline(sd, "CWD %s\n", rpath);
    getline(sd, line, STRING_SIZE, &rcode);
    if (rcode == 500) {
        return 1;
    }
    if (rcode == 550) {
        return 1;
    }
    if (rcode == 425) {
        return 1;
    }


    if(*rpath && rpath[strlen(rpath)-1]!='/') {
        Log("; it is not include slash in tail (%s)", rpath);

        nwrite = fdprintf(cfd, "HTTP/1.0 302\r\nLocation: %s/\r\n\r\n",
            xurl);

        *sum_read = (off_t)0;
        *sum_write_client = nwrite;
        *sum_write_cache = (off_t)0;

        *http_ret = 302;

        return 0;
    }


#if 0
    putline(sd, "NLST %s\n", rpath);
#endif
    putline(sd, "LIST %s\n", rpath);
    getline(sd, line, STRING_SIZE, &rcode);
    if (rcode == 500) {
        return 1;
    }
    if (rcode == 550) {
        return 1;
    }
    if (rcode == 425) {
        return 1;
    }

    time(&now);

    if(SIGSETJMP(retrieve_jmp_point, 1)) {
        Error("List_ForwardStore_Passive: long waitting for retrieve, give up.");
        SetSignalHandler(SIGALRM, SIG_IGN);

        return -1;
    }
    SetSignalHandler(SIGALRM, giveup_retrieve);
    alarm(7200);


    pipe_broken = 0;
    nwrite =
        fdprintf(cfd, "HTTP/1.0 200\r\nVia: FTPHandler of Wcol-E\r\n\r\n");
    if(pipe_broken) {
        Error("ForwardStore: fail write HTTP header");

        pipe_broken = 0;
    }
    *sum_write_client += nwrite;

    *http_ret = 200;

    while(1) {

r1:
        pipe_broken = 0;
        nread = read(data, buf, BUF_SIZE);
        if(nread<0) {
            if(errno==EINTR) {
                goto r1;
            }
            Error("ForwardStore: fail read() %ld/%ld/%ld (%d)",
                (long)sum_read,
                (long)sum_write_client, (long)sum_write_cache, errno);

            ret = -2;
            break;
        }
        if(nread==0) {
            /* connection is over */
            break;
        }
        if(*sum_read+nread>BUF_SIZE) {
            Error("ForwardStore: too large message %ld/%ld/%ld",
                (long)sum_read,
                (long)sum_write_client, (long)sum_write_cache);

            ret = -2;
            break;
        }

        memcpy(&raw_list[*sum_read], buf, nread);

        *sum_read += (off_t) nread;

        raw_list[*sum_read] = '\0';
    }

#if 0
    dump_stringchunk("raw_list", raw_list);
#endif



    /***
     *** convert
     ***/
    sum_list = (off_t) 0;
    memset(cooked_list, 0, BUF_SIZE);
    strcat(cooked_list, "<HTML>\n");
    strcat(cooked_list, "<HEAD>\n");
    strcat(cooked_list, "<TITLE>Directory entries of ");
    strcat(cooked_list, xurl);
    strcat(cooked_list, " </TITLE>\n");
    sprintf(line, "<BASE HREF=\"%s\">\n", xurl);
    strcat(cooked_list, line);
    strcat(cooked_list, "</HEAD>\n");

    strcat(cooked_list, "<BODY>\n");
    sprintf(tline, "<H2>%s</H2>\n", xurl);
    strcat(cooked_list, tline);
    sprintf(tline, "<H3>%s</H3>\n", ctime(&now));
    strcat(cooked_list, tline);

    strcat(cooked_list, "<HR>\n");
    strcat(cooked_list, "<PRE>\n");

    sum_list = (off_t) strlen(cooked_list);

    c = -1;
    cpos = raw_list;
    while((cpos = sgets(raw_list, cpos, line, STRING_SIZE))!=NULL) {
#if 0
        dump_string("  line", line);
#endif
        if(strncmp(line, "total ", 6)==0) {
            /* thru */
            memcpy(&cooked_list[sum_list], line, strlen(line));
            sum_list += (off_t) strlen(line);
            c = 0;
        }
        else {
            register char *p, *q;

            Chop_CRLF(line);
            memset(tline, 0, STRING_SIZE);

            p = line+strlen(line);
            while(p>line) {
                if(*p==' ' || *p=='\t') {
                    break;
                }
                --p;
            }

            if(p<=line) {
                /* thru */
                strcpy(tline, line);
            }
            else {
                p++;

                strncat(tline, line, p-line);
                tline[p-line] = '\0';
#if 0
                sprintf(&tline[p-line],
                    "<A HREF=\"%s\">%s</A>\r\n", p, p);
#endif
                if(line[0]=='d') {
                    sprintf(&tline[p-line],
                        "<A HREF=\"%s/\">%s</A>\r\n", p, p);
                }
                else {
                    sprintf(&tline[p-line],
                        "<A HREF=\"%s\">%s</A>\r\n", p, p);
                }

            }

#if 0
            dump_string(" tline", tline);
#endif

            memcpy(&cooked_list[sum_list], tline, strlen(tline));
            sum_list += (off_t) strlen(tline);
            c++;
        }

        cooked_list[sum_list] = '\0';
    }

    strcat(cooked_list, "</PRE>\n");
    strcat(cooked_list, "<HR>\n");

    if(c<=0) {
    sprintf(tline, "<B>no entry in this directory.</B></P>\r\n");
    strcat(cooked_list, tline);
    }
    else {
    sprintf(tline, "<B>%d entri(es) in this directory.</B></P>\r\n", c+1);
    strcat(cooked_list, tline);
    }

    strcat(cooked_list, "Wcol-E\n");
    strcat(cooked_list, "<BODY>\n");
    strcat(cooked_list, "</HTML>\n");
    sum_list += (off_t) strlen(cooked_list);

#if 0
    dump_stringchunk("cooked_list", cooked_list);
#endif

    while(1) {
w1:
        if(cfd>=0) {
            pipe_broken = 0;
            nwrite = write(cfd, cooked_list, sum_list);

            if(pipe_broken) {
                close(cfd);
                cfd = -1;

                pipe_broken = 0;
            }
            if(nwrite<0) {
                if(errno==EINTR) {
                    goto w1;
                }

            Error("ForwardStore: fail write() client %d %ld/%ld/%ld (%d)",
                cfd, (long)sum_read,
                (long)sum_write_client, (long)sum_write_cache, errno);

                close(cfd);
                cfd = -1;

                ret = -3;
#if 0
                break;
#endif
            }
            *sum_write_client += (off_t) nwrite;
        }


w2:
        if(dfd>=0) {
            pipe_broken = 0;
            nwrite = write(dfd, cooked_list, sum_list);

            if(nwrite<0) {
                if(errno==EINTR) {
                    goto w2;
                }

            Error("ForwardStore: fail write() cache %d %ld/%ld/%ld (%d)",
                dfd, (long)sum_read,
                (long)sum_write_client, (long)sum_write_cache, errno);

                ret = -4;
                break;
            }
            *sum_write_cache += (off_t) nwrite;
        }

        break;
    }

    Trace("List_ForwardStore_Passive: ret %d\n", ret);
    Trace("  read %ld\n", (long)*sum_read);
    Trace("  write client %ld / cache %ld \n",
        (long)*sum_write_client, (long)*sum_write_cache);

    return ret;
}



char ftp_user[STRING_SIZE];
char ftp_pass[STRING_SIZE];

/*
 *
 * return:
 *      success  0
 *      error   -1
 */

int
Retrieve_FTPObject(request_t req)
{
    Info info;
    url_t *url;
    int  port;
    char fn[STRING_SIZE], tmp[STRING_SIZE];
    char buffer[BUF_SIZE];
    int sfd;
    int data;
    int dfd;
    int rcode;
    off_t read_size, pass_size, cache_size;
    struct timeval start, end;
    int http_ret;

    gettimeofday(&start, NULL);

    ShowRequest(&req);

    url = NULL;
    url = URL_NewVal(req.url);
    if(url==NULL) {
        Error("no memory");
        goto fail;
    }
    URL_Normalize(url);

    Trace("==\n");
    Trace("user     '%s'\n", ftp_user);
    Trace("pass     '%s'\n", ftp_pass);
    URL_Show(url);
    port = atoi(url->port);

#if 0
    Log("; FTPHandler:  cfd %d, user '%s', pass '%s', url '%s'",
        req.fd, ftp_user, ftp_pass, url->full_loc);
#endif
    Trace("; FTPHandler:  cfd %d, user '%s', pass '%s', url '%s'\n",
        req.fd, ftp_user, ftp_pass, url->full_loc);

    if(SetupFile(tmp, url->full_loc, BODY_TMP_FILE, &dfd)) {
        Error("Cannot setup file '%s'", url->full_loc);
        goto fail;
    }
    Trace("\n\nFile '%s'\n", tmp);

    if((sfd = TCP_connect(url->host, port))<0) {
        Log("; server '%s' port %d is down", url->host, port);

        if(sfd == WCOL_SOCK_SOCKETERROR) {
            http_ret = 500,
            pass_size = HTTP_ErrorMessage_Old(req.fd, http_ret,
                "Socket Error", "Cannot make socket",
                req.ver_major>0);
        }
        else
        if(sfd == WCOL_SOCK_UNKNOWHOST) {
            http_ret = 403,
            pass_size = HTTP_ErrorMessage_Old(req.fd, http_ret,
                "Host Resolution Error",
                "Unknown host. please check it once more",
                req.ver_major>0);
        }
        else
        if(sfd == WCOL_SOCK_CONNECTERROR) {
            http_ret = 403,
            pass_size = HTTP_ErrorMessage_Old(req.fd, http_ret,
                "Connect Error",
                "Cannot connect server. Server is down or busy",
                req.ver_major>0);
        }
        else
        {
            http_ret = 403,
            pass_size = HTTP_ErrorMessage_Old(req.fd, http_ret,
                "Unknown Connect Error",
                "Sorry, I don't know what happend",
                req.ver_major>0);
        }



        goto fail;
    }

c1:
    getline(sfd, buffer, BUF_SIZE, &rcode);
    if(rcode<=0) {
        goto c1;
    }

    if(FTP_User(sfd, ftp_user)) {
        goto fail;
    }
    if(FTP_Pass(sfd, ftp_pass)) {
        goto fail;
    }
    if(FTP_Type(sfd, TYPE_I)) {
        goto fail;
    }
#if 0
    if(FTP_Pwd(sfd, dir)) {
        goto fail;
    }
#endif

    data = DataInitconnPassive(sfd);
    if(data<0) {
        Error("Cannot connect data connection");
        goto fail;
    }

    if(ForwardStore_Passive(sfd, data, req.fd, dfd, url->path,
        &read_size, &pass_size, &cache_size)) {
        Error("fail ForwardStore_Passive() '%s'", url->path);

        data = DataInitconnPassive(sfd);
        if(data<0) {
            Error("Cannot connect data connection");
            goto fail;
        }
        if(List_ForwardStore_Passive(sfd, data, req.fd, dfd, url->path,
            &read_size, &pass_size, &cache_size, url->full_loc, &http_ret)) {
            Error("fail ForwardStore_Passive() '%s'", url->path);


#if 0
            pass_size = fdprintf(req.fd,
                "HTTP/1.0 403\r\n\r\nNo such file or directory (%s)\r\n",
                url->full_loc);
#endif
            http_ret = 403;
            pass_size = HTTP_ErrorMessage_Old(req.fd,
                http_ret,
                "Not found", "No such file or direcotry", req.ver_major>0);


/*
            goto fail;
*/
        }

#if 0
        goto fail;
#endif
    }


success:

    gettimeofday(&end, NULL);

    AssignFileName(fn, url->full_loc, BODY_FILE);
    Trace("\n\nFile '%s'\n", fn);
r1:
    if(rename(tmp, fn)) {
        if(errno==EINTR)
            goto r1;

        Error("Cannot rename old '%s' (%d)", tmp, errno);
        Error("              new '%s' (%d)", fn, errno);
    }


    ClearInfo(&info);
    strcpy(info.attr.name, url->full_loc);
    info.attr.ver_major = req.ver_major;
    info.attr.ver_minor = req.ver_minor;
    info.attr.hsize = 0;
    info.attr.bsize = read_size;
    info.attr.length = read_size;
    info.attr.last   = (time_t)(start.tv_sec);
    info.attr.conn_time = -1;
    info.attr.trans_time = -1;
    info.attr.take_time = WCOL_SUB_TIME(end, start);
    strcpy(info.attr.type, "text/plain");

    info.attr.state = http_ret;
    info.attr.init_state = REQ_CLIENT;
    info.attr.count     = 1;


    What(&info);

    if(SaveInfo(&info, url->full_loc)) {
        Error("Retrieve_FTPObject: cannot save info");
    }
    else {
        Trace("Retrieve_FTPObject: save info\n");
    }

    Log("%d.%06d FETCH %d %s HTTP %d.%d %s %s %s %s %s %d %s %ld %d : %s %s %d %ld %d",
        end.tv_sec, end.tv_usec,
        req.hereport, req.peername,
        req.ver_major, req.ver_minor,
        req.nocache == 1 ? "n" : "-",
        req.ims == -1 ? "-" : "*",
        req.auth[0] == '\0' ? "-" : "X",
        ReqMethodStr(req.method), url->full_loc,
        info.attr.state,
        info.attr.type,
        (long)(pass_size),
        WCOL_SUB_TIME(end, req.issue_time),


        "ORIGIN",
        url->host,
        atoi(url->port),
        (long)(read_size),
        WCOL_SUB_TIME(end, start)

        );


    goto out;

fail:

    gettimeofday(&end, NULL);

    Log("%d.%06d IGNORE %d %s HTTP %d.%d %s %s %s %s %s %d %s %ld %d :",
        end.tv_sec, end.tv_usec,
        req.hereport, req.peername,
        req.ver_major, req.ver_minor,
        req.nocache == 1 ? "n" : "-",
        req.ims == -1 ? "-" : "*",
        req.auth[0] == '\0' ? "-" : "X",
        ReqMethodStr(req.method), url->full_loc,
        http_ret,
        "-",
        (long)(pass_size),
        WCOL_SUB_TIME(end, req.issue_time)

        );

    goto out;

out:
    close(dfd);
    close(req.fd);

    if(url!=NULL) {
        URL_Delete(url);
    }
    close(sfd);

    return 0;

}

int
Send_FTPObject(request_t req, Info *info)
{
    struct timeval start, end;
    off_t  wlen;

    gettimeofday(&start, NULL);
    wlen = Send(&req.fd, req.url, info, req.ver_major, req.ver_minor, 0);
    gettimeofday(&end, NULL);


    Log("%d.%06d SEND %d %s HTTP %d.%d %s %s %s %s %s %d %s %ld %d : %s %s %d %ld %d : %s %d",
        end.tv_sec, end.tv_usec,
        req.hereport, req.peername,
        req.ver_major, req.ver_minor,
        req.nocache == 1 ? "n" : "-",
        req.ims == -1 ? "-" : "*",
    /*
        req.auth[0] == '\0' ? "-" : req.auth,
    */
        req.auth[0] == '\0' ? "-" : "X",
        ReqMethodStr(req.method), req.url,
        info->attr.state,
        info->attr.type,
        (long)wlen,
        WCOL_SUB_TIME(end, req.issue_time),
        "CACHE", "-", -1,
        (long)info->attr.length, info->attr.take_time,
        info->attr.init_state==REQ_CLIENT ?         "CLIENT  " :
            (info->attr.init_state==REQ_PREFETCH ?      "PREFETCH" :
                (info->attr.init_state==REQ_IMPORT ?        "IMPORT  " :
                                                                "UNKNOW  ")),
        info->attr.count+1
      );


    return 0;
}


int
FTPHandler(request_t req)
{
    Info *info;
    int ret;

    if(!req.nocache && (info=Have(req.url))!=NULL) {
        ret = Send_FTPObject(req, info);
        FREE(info);

        return ret;
    }
    else {
        return Retrieve_FTPObject(req);
    }
}



