/*
 * WCOL:
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <time.h>
#ifdef TIME_WITH_SYS_TIME
#include <sys/time.h>
#endif
#include <errno.h>

#include "wcol.h"

#include "fds.h"

#include "request.h"
#include "http.h"
#include "icp.h"
#include "args.h"
#include "accept.h"
#include "conv.h"
#include "manager.h"
#include "handler.h"


/*
 * externals
 */

/* automatic */
extern             char config_file[];
extern              int trace_flag;
extern              int catalyst_flag;
extern              int prefetch_flag;
extern              int mcache_flag;
extern              int remove_cookie_flag;
extern              int remove_referer_flag;
extern              int noreaper_flag;
extern           url_t* fetch_parent;
extern           url_t* prefetch_parent;
extern           url_t* import_sibling;
extern              int http_listen_portno;
extern              int icp_listen_portno;
extern             char state_file[];
extern             char pid_file[];
extern             char log_file[];
extern             char pool_dir[];
extern            off_t cache_size;
extern              int cache_use_rate;
extern           time_t cache_keeptime;
extern           time_t cache_ignorekeeptime;
extern              int gc_interval;
extern             char gc_workfile[];
extern              int gc_bufsize;
extern              int max_handler;
extern              int num_handler;
extern              int max_prompter;
extern              int num_prompter;
extern              int max_speaker;
extern              int num_speaker;
extern              int max_session;
extern              int prefetch_lag_limit;
extern              int num_includedprefetch;
extern              int num_anchorprefetch;
extern             char local_domain[];
extern              int session_timeout;
extern              int session_lifetime;
extern              int query_timeout;
extern              int reception_timeout;
extern             char ftp_user[];
extern             char ftp_pass[];
extern             char dummy_string[];

/* static */
extern pid_t    wcol_pid;
extern char     *progname;

/* end of externals */

char config_file[STRING_SIZE];
char dummy_string[STRING_SIZE];


/*
 * SetGlobalDefaults -
 */
void SetGlobalDefaults()
{
    strcpy(config_file, DEFAULT_CONFIGFILE);
    trace_flag = 0;
    catalyst_flag = 0;
    prefetch_flag = 1;
    mcache_flag = 0;
    remove_cookie_flag = 0;
    remove_referer_flag = 0;
    noreaper_flag = 0;
    fetch_parent = NULL;
    prefetch_parent = NULL;
    import_sibling = NULL;
    http_listen_portno = (int) HTTP_PORTNO;
    icp_listen_portno = (int) ICP_PORTNO;
    strcpy(state_file, DEFAULT_STATEFILE);
    strcpy(pid_file, DEFAULT_PIDFILE);
    strcpy(log_file, DEFAULT_LOGFILE);
    strcpy(pool_dir, DEFAULT_POOLDIR);
    cache_size = (off_t) 100*(1024*1024);
    cache_use_rate = (int) 70;
    cache_keeptime = (time_t) CACHEKEEPTIME;
    cache_ignorekeeptime = (time_t) CACHEIGNOREKEEPTIME;
    gc_interval = (int) 3600;
    strcpy(gc_workfile, DEFAULT_GCWORKFILE);
    gc_bufsize = (int) 1000;
    max_handler = (int) MAX_HANDLER;
    num_handler = (int) NUM_HANDLER;
    max_prompter = (int) MAX_PROMPTER;
    num_prompter = (int) NUM_PROMPTER;
    max_speaker = (int) MAX_SPEAKER;
    num_speaker = (int) NUM_SPEAKER;
    max_session = (int) MAX_SESSION;
    prefetch_lag_limit = (int) 40000;
    num_includedprefetch = (int) 10;
    num_anchorprefetch = (int) 10;
    strcpy(local_domain, "");
    session_timeout = (int) SESSION_TIMEOUT;
    session_lifetime = (int) SESSION_LIFETIME;
    query_timeout = (int) QUERY_TIMEOUT;
    reception_timeout = (int) RECEPTION_TIMEOUT;
    strcpy(ftp_user, "anonymous");
    strcpy(ftp_pass, "");
    strcpy(dummy_string, "");
}
/* end of SetGlobalDefaults() */





#define MAX_AV 10

/*
 * ReadConfigFile -
 */
void ReadConfigFile(char *filename)
{
    arg_t  *args, *pargs;
    int     num;
    int     tint;
    char   *av[MAX_AV];
    int     i, j;


    Trace("ReadConfigFile: filename='%s'\n", filename);
    if((num=ReadArgsFromFile(filename, &args))<=0) {
        Trace("ReadConfigFile: no arguments\n");
        return;
    }

    pargs = args;
    while(pargs) {
        /* project argument value to temporary pointor */
        for(i=0;pargs->vals[i] && i<MAX_AV;i++) {
            av[i] = pargs->vals[i];
        }
        for(j=i;j<MAX_AV;j++) {
            av[j] = NULL;
        }

        if(!pargs->name) {
            /* nothing */
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'C', "ConfigFile")) {
            if(pargs->vals[0]) {
                strcpy(config_file, av[0]);
                goto nextone;
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'T', "Trace")) {
            trace_flag = 1;
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'C', "Catalyst")) {
            catalyst_flag = 1;
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'P', "ProxyServer")) {
            catalyst_flag = 0;
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'O', "Oscillator")) {
            catalyst_flag = 1;
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'P', "Prefetch")) {
            prefetch_flag = 1;
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'C', "CacheOnly")) {
            prefetch_flag = 0;
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'M', "MemoryCache")) {
            mcache_flag = 1;
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'N', "NoMemoryCache")) {
            mcache_flag = 0;
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'N', "NeedCookie")) {
            remove_cookie_flag = 0;
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'R', "RemoveCookie")) {
            remove_cookie_flag = 1;
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'R', "RemoveReferer")) {
            remove_referer_flag = 1;
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'N', "NoReaper")) {
            noreaper_flag = 1;
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'F', "FetchParent")) {
            if(pargs->vals[0]) {
                fetch_parent = URL_NewVal(av[0]);
                goto nextone;
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'P', "PrefetchParent")) {
            if(pargs->vals[0]) {
                prefetch_parent = URL_NewVal(av[0]);
                goto nextone;
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'I', "ImportSibling")) {
            if(pargs->vals[0]) {
                import_sibling = URL_NewVal(av[0]);
                goto nextone;
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'H', "HTTPPort")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    http_listen_portno = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'I', "ICPPort")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    icp_listen_portno = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'S', "StateFile")) {
            if(pargs->vals[0]) {
                strcpy(state_file, av[0]);
                goto nextone;
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'P', "PidFile")) {
            if(pargs->vals[0]) {
                strcpy(pid_file, av[0]);
                goto nextone;
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'L', "LogFile")) {
            if(pargs->vals[0]) {
                strcpy(log_file, av[0]);
                goto nextone;
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'C', "CacheDir")) {
            if(pargs->vals[0]) {
                strcpy(pool_dir, av[0]);
                goto nextone;
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'C', "CacheSize")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    cache_size = (off_t) tint * 1024*1024;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'C', "CacheUseRate")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    cache_use_rate = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'K', "KeepTime")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    cache_keeptime = (time_t) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'N', "NegativeKeepTime")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    cache_ignorekeeptime = (time_t) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'G', "GCInterval")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    gc_interval = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'G', "GCWorkFile")) {
            if(pargs->vals[0]) {
                strcpy(gc_workfile, av[0]);
                goto nextone;
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'G', "GCBufSize")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    gc_bufsize = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'A', "AcceptAddress")) {
            if(pargs->vals[0]) {
                if(AppendAcceptEntry(av[0], av[1])==0) {
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'C', "ConvertType")) {
            if(pargs->vals[0]) {
                if(AppendConvertEntry(av[0], av[1], av[2], av[3])==0) {
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'M', "MaxHandler")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    max_handler = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'N', "NumHandler")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    num_handler = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'M', "MaxPrompter")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    max_prompter = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'N', "NumPrompter")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    num_prompter = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'M', "MaxSpeaker")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    max_speaker = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'N', "NumSpeaker")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    num_speaker = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'M', "MaxSession")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    max_session = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'P', "PrefetchLagLimit")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    prefetch_lag_limit = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'P', "PrefetchIncluded")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    num_includedprefetch = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'P', "PrefetchAnchor")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    num_anchorprefetch = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'L', "LocalDomain")) {
            if(pargs->vals[0]) {
                strcpy(local_domain, av[0]);
                goto nextone;
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'S', "SessionTimeout")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    session_timeout = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'S', "SessionLifetime")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    session_lifetime = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'Q', "QueryTimeout")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    query_timeout = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'R', "ReceptionTimeout")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    reception_timeout = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'S', "Sibling")) {
            if(pargs->vals[0]) {
                if(AppendSiblingEntry(av[0], av[1], av[2], av[3])==0) {
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'F', "FieldRemove")) {
            if(pargs->vals[0]) {
                if(AppendRemovedFieldEntry(av[0], av[1])==0) {
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'F', "FieldChange")) {
            if(pargs->vals[0]) {
                if(AppendChangedFieldEntry(av[0], av[1], av[2])==0) {
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'F', "FTPUser")) {
            if(pargs->vals[0]) {
                strcpy(ftp_user, av[0]);
                goto nextone;
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'F', "FTPPass")) {
            if(pargs->vals[0]) {
                strcpy(ftp_pass, av[0]);
                goto nextone;
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'D', "Dummy")) {
            if(pargs->vals[0]) {
                strcpy(dummy_string, av[0]);
                goto nextone;
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'N', "NeedReferer")) {
            /* nothing */
            goto nextone;
        }
        else {
            Error("Unknow argument '%s'", pargs->name);
            goto nextone;
        }
fail:
        Error("Argment specification ('%s') is wrong", pargs->name);
nextone:
        pargs = pargs->next;
    }

    FreeArgs(args);
}



/*
 * FixGlobals - Fix global vars.
 */
void FixGlobals()
{

    if(fetch_parent)
        URL_Normalize(fetch_parent);
    if(prefetch_parent)
        URL_Normalize(prefetch_parent);
    if(import_sibling)
        URL_Normalize(import_sibling);
    if(max_handler<0) {
        Error("FixGlobals: 'MaxHandler' is too small (min %d)",0);
        max_handler = (int) 0;
    }
    if(max_handler>MAX_HANDLER) {
        Error("FixGlobals: 'MaxHandler' is too large (max %d)",MAX_HANDLER);
        max_handler = (int) MAX_HANDLER;
    }
    if(num_handler<0) {
        Error("FixGlobals: 'NumHandler' is too small (min %d)",0);
        num_handler = (int) 0;
    }
    if(num_handler>max_handler) {
        Error("FixGlobals: 'NumHandler' is too large (max %d)",max_handler);
        num_handler = (int) max_handler;
    }
    if(max_prompter<0) {
        Error("FixGlobals: 'MaxPrompter' is too small (min %d)",0);
        max_prompter = (int) 0;
    }
    if(max_prompter>MAX_PROMPTER) {
        Error("FixGlobals: 'MaxPrompter' is too large (max %d)",MAX_PROMPTER);
        max_prompter = (int) MAX_PROMPTER;
    }
    if(num_prompter<0) {
        Error("FixGlobals: 'NumPrompter' is too small (min %d)",0);
        num_prompter = (int) 0;
    }
    if(num_prompter>max_prompter) {
        Error("FixGlobals: 'NumPrompter' is too large (max %d)",max_prompter);
        num_prompter = (int) max_prompter;
    }
    if(max_speaker<0) {
        Error("FixGlobals: 'MaxSpeaker' is too small (min %d)",0);
        max_speaker = (int) 0;
    }
    if(max_speaker>MAX_SPEAKER) {
        Error("FixGlobals: 'MaxSpeaker' is too large (max %d)",MAX_SPEAKER);
        max_speaker = (int) MAX_SPEAKER;
    }
    if(num_speaker<0) {
        Error("FixGlobals: 'NumSpeaker' is too small (min %d)",0);
        num_speaker = (int) 0;
    }
    if(num_speaker>max_speaker) {
        Error("FixGlobals: 'NumSpeaker' is too large (max %d)",max_speaker);
        num_speaker = (int) max_speaker;
    }
    if(max_session<1) {
        Error("FixGlobals: 'MaxSession' is too small (min %d)",1);
        max_session = (int) 1;
    }
    if(max_session>MAX_SESSION) {
        Error("FixGlobals: 'MaxSession' is too large (max %d)",MAX_SESSION);
        max_session = (int) MAX_SESSION;
    }
    if(prefetch_lag_limit<3600) {
        Error("FixGlobals: 'PrefetchLagLimit' is too small (min %d)",3600);
        prefetch_lag_limit = (int) 3600;
    }
    if(session_timeout<1) {
        Error("FixGlobals: 'SessionTimeout' is too small (min %d)",1);
        session_timeout = (int) 1;
    }
    if(session_timeout>360000) {
        Error("FixGlobals: 'SessionTimeout' is too large (max %d)",360000);
        session_timeout = (int) 360000;
    }
    if(session_lifetime<1) {
        Error("FixGlobals: 'SessionLifetime' is too small (min %d)",1);
        session_lifetime = (int) 1;
    }
    if(session_lifetime>3600000) {
        Error("FixGlobals: 'SessionLifetime' is too large (max %d)",3600000);
        session_lifetime = (int) 3600000;
    }
    if(query_timeout<1) {
        Error("FixGlobals: 'QueryTimeout' is too small (min %d)",1);
        query_timeout = (int) 1;
    }
    if(query_timeout>10000) {
        Error("FixGlobals: 'QueryTimeout' is too large (max %d)",10000);
        query_timeout = (int) 10000;
    }
    if(reception_timeout<0) {
        Error("FixGlobals: 'ReceptionTimeout' is too small (min %d)",0);
        reception_timeout = (int) 0;
    }
    if(reception_timeout>120) {
        Error("FixGlobals: 'ReceptionTimeout' is too large (max %d)",120);
        reception_timeout = (int) 120;
    }
}





ShowGlobals(FILE *s)
{
    fprintf(s,
"======================================================================\n\
Flag Name                CurrentValue\n\
----------------------------------------------------------------------\n");

    fprintf(s, "  f  ConfigFile          '%s'\n",
        config_file);
    fprintf(s, "  t  Trace                %s\n",
        trace_flag==1 ? "*ON*" : "*OFF*" );
    fprintf(s, "  c  Catalyst             %s\n",
        catalyst_flag==1 ? "*ON*" : "*OFF*" );
    fprintf(s, "  -  ProxyServer          %s\n",
        catalyst_flag==0 ? "*ON*" : "*OFF*" );
    fprintf(s, "  -  Oscillator           %s\n",
        catalyst_flag==1 ? "*ON*" : "*OFF*" );
    fprintf(s, "  p  Prefetch             %s\n",
        prefetch_flag==1 ? "*ON*" : "*OFF*" );
    fprintf(s, "  -  CacheOnly            %s\n",
        prefetch_flag==0 ? "*ON*" : "*OFF*" );
    fprintf(s, "  -  MemoryCache          %s\n",
        mcache_flag==1 ? "*ON*" : "*OFF*" );
    fprintf(s, "  -  NoMemoryCache        %s\n",
        mcache_flag==0 ? "*ON*" : "*OFF*" );
    fprintf(s, "  -  NeedCookie           %s\n",
        remove_cookie_flag==0 ? "*ON*" : "*OFF*" );
    fprintf(s, "  k  RemoveCookie         %s\n",
        remove_cookie_flag==1 ? "*ON*" : "*OFF*" );
    fprintf(s, "  e  RemoveReferer        %s\n",
        remove_referer_flag==1 ? "*ON*" : "*OFF*" );
    fprintf(s, "  r  NoReaper             %s\n",
        noreaper_flag==1 ? "*ON*" : "*OFF*" );
    fprintf(s, "  n  FetchParent          %s\n",
        fetch_parent? fetch_parent->full_loc : "*NONE*" );
    fprintf(s, "  j  PrefetchParent       %s\n",
        prefetch_parent? prefetch_parent->full_loc : "*NONE*" );
    fprintf(s, "  g  ImportSibling        %s\n",
        import_sibling? import_sibling->full_loc : "*NONE*" );
    fprintf(s, "  P  HTTPPort            %8d \n",
        http_listen_portno);
    fprintf(s, "  O  ICPPort             %8d \n",
        icp_listen_portno);
    fprintf(s, "  s  StateFile           '%s'\n",
        state_file);
    fprintf(s, "  i  PidFile             '%s'\n",
        pid_file);
    fprintf(s, "  l  LogFile             '%s'\n",
        log_file);
    fprintf(s, "  d  CacheDir            '%s'\n",
        pool_dir);
    fprintf(s, "  C  CacheSize           %8d Mbytes (x 1024*1024)\n",
        cache_size/(1024*1024));
    fprintf(s, "  R  CacheUseRate        %8d \n",
        cache_use_rate);
    fprintf(s, "  K  KeepTime            %8d sec [%s]\n",
        cache_keeptime, display_sec(cache_keeptime));
    fprintf(s, "  N  NegativeKeepTime    %8d sec [%s]\n",
        cache_ignorekeeptime, display_sec(cache_ignorekeeptime));
    fprintf(s, "  G  GCInterval          %8d \n",
        gc_interval);
    fprintf(s, "  -  GCWorkFile          '%s'\n",
        gc_workfile);
    fprintf(s, "  -  GCBufSize           %8d \n",
        gc_bufsize);
    fprintf(s, "  M  MaxHandler          %8d \n",
        max_handler);
    fprintf(s, "  H  NumHandler          %8d \n",
        num_handler);
    fprintf(s, "  -  MaxPrompter         %8d \n",
        max_prompter);
    fprintf(s, "  -  NumPrompter         %8d \n",
        num_prompter);
    fprintf(s, "  -  MaxSpeaker          %8d \n",
        max_speaker);
    fprintf(s, "  -  NumSpeaker          %8d \n",
        num_speaker);
    fprintf(s, "  -  MaxSession          %8d \n",
        max_session);
    fprintf(s, "  L  PrefetchLagLimit    %8d msec\n",
        prefetch_lag_limit);
    fprintf(s, "  A  PrefetchIncluded    %8d \n",
        num_includedprefetch);
    fprintf(s, "  I  PrefetchAnchor      %8d \n",
        num_anchorprefetch);
    fprintf(s, "  -  LocalDomain         '%s'\n",
        local_domain);
    fprintf(s, "  -  SessionTimeout      %8d msec\n",
        session_timeout);
    fprintf(s, "  -  SessionLifetime     %8d msec\n",
        session_lifetime);
    fprintf(s, "  -  QueryTimeout        %8d msec\n",
        query_timeout);
    fprintf(s, "  -  ReceptionTimeout    %8d sec\n",
        reception_timeout);
    fprintf(s, "  -  FTPUser             '%s'\n",
        ftp_user);
    fprintf(s, "  -  FTPPass             '%s'\n",
        ftp_pass);
    fprintf(s, "  z  Dummy               '%s'\n",
        dummy_string);
}


ShowProgramStatus(FILE *outport)
{
    fprintf(outport, "Wcol-E (pre-release) program-version %s, data-id %s\n",
        WCOL_VER, WCOL_INFO_ID);

    fflush(outport);
}

ShowUsage()
{
    FixGlobals();   /* check in temporary */
    ShowProgramStatus(stdout);
    ShowGlobals(stdout);
}

int SetProcessComment(char *comment)
{
    char line[STRING_SIZE];

    sprintf(line, "wcol : %s", comment);
    return SetProcessState(line);
}


/*
 * ReadCommandlineOption
 */
void ReadCommandlineOptions(int argc, char **argv)
{
    int flag;

    extern char *optarg;
    extern int  optind;

    while((flag = getopt(argc, argv,
        "hvf:tcpkern:j:g:P:O:s:i:l:d:C:R:K:N:G:M:H:L:A:I:z:")) != -1) {

        switch(flag) {
        case 'h':
            ShowUsage();
            exit(WCOL_EXIT_NORMAL);
            break;
        case 'v':
            ShowProgramStatus(stdout);
            exit(WCOL_EXIT_NORMAL);
            break;
        case 'f':
            strcpy(config_file, optarg);
            ReadConfigFile(config_file);
            break;
        case 't':
            trace_flag = 1 - trace_flag;
            break;
        case 'c':
            catalyst_flag = 1 - catalyst_flag;
            break;
        case 'p':
            prefetch_flag = 1 - prefetch_flag;
            break;
        case 'k':
            remove_cookie_flag = 1 - remove_cookie_flag;
            break;
        case 'e':
            remove_referer_flag = 1 - remove_referer_flag;
            break;
        case 'r':
            noreaper_flag = 1 - noreaper_flag;
            break;
        case 'n':
            fetch_parent = URL_NewVal(optarg);
            break;
        case 'j':
            prefetch_parent = URL_NewVal(optarg);
            break;
        case 'g':
            import_sibling = URL_NewVal(optarg);
            break;
        case 'P':
            http_listen_portno = (int) atoi(optarg);
            break;
        case 'O':
            icp_listen_portno = (int) atoi(optarg);
            break;
        case 's':
            strcpy(state_file, optarg);
            break;
        case 'i':
            strcpy(pid_file, optarg);
            break;
        case 'l':
            strcpy(log_file, optarg);
            break;
        case 'd':
            strcpy(pool_dir, optarg);
            break;
        case 'C':
            cache_size = (off_t) atoi(optarg) * 1024*1024;
            break;
        case 'R':
            cache_use_rate = (int) atoi(optarg);
            break;
        case 'K':
            cache_keeptime = (time_t) atoi(optarg);
            break;
        case 'N':
            cache_ignorekeeptime = (time_t) atoi(optarg);
            break;
        case 'G':
            gc_interval = (int) atoi(optarg);
            break;
        case 'M':
            max_handler = (int) atoi(optarg);
            break;
        case 'H':
            num_handler = (int) atoi(optarg);
            break;
        case 'L':
            prefetch_lag_limit = (int) atoi(optarg);
            break;
        case 'A':
            num_includedprefetch = (int) atoi(optarg);
            break;
        case 'I':
            num_anchorprefetch = (int) atoi(optarg);
            break;
        case 'z':
            strcpy(dummy_string, optarg);
            break;
        }
    }
}



