/*
 *  WCOL: Handler state procedures
 *      by k-chinen@is.aist-nara.ac.jp, 1996
 *
 *  $Id: hstate.c,v 1.1 1997/02/12 17:21:21 k-chinen Exp k-chinen $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <signal.h>
#ifdef HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif
#include <sys/uio.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <limits.h>

#include <sys/ipc.h>
#include <sys/shm.h>


#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif

#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif


#include <signal.h>
#include <errno.h>

#include "wcol.h"
#include "msg.h"
#include "request.h"
#include "ipc.h"
#include "sigs.h"
#include "fds.h"

#include "session.h"
#include "manager.h"
#include "handler.h"
#include "prompter.h"
#include "reaper.h"
/* */
#include "global.h"
/* */


#if 1
char state_file[STRING_SIZE];
#endif

/*
 * templeat
 */
#if 0
char *count_format_PLAIN=
    "wake %4d ( keep %4d / busy %4d / idle %4d ), max %d\n\n";
char *count_format_HTML=
"<table>\n\
  <tr><td>wake <td>%4d\n\
  <tr><td>max  <td>%4d\n\
  <tr><td>keep <td>%4d\n\
  <tr><td>busy <td>%4d\n\
  <tr><td>idle <td>%4d\n\
</table>\n";
#else
char *count_format_PLAIN=
    "\nwake handler %4d / %4d (max %4d), session %4d\n";
char *count_format_HTML=
"<table>\n\
  <tr><th>handler<td>%4d<td> %4d (max %4d)\n\
  <tr><th>session<td>%4d\n\
</table>\n";
#endif

char *head_format_PLAIN=
    ";  #   B   PID   SECS   SID#    ID#   URL\n";
char *head_format_HTML=
    "<table>\n\
  <tr><th>#<th>B<th>PID<th>SECS<th>SID#<th>ID#<th>URL\n";

char *line_format_PLAIN=" %4d %3d %5d %5d %7d/%7d %s\n";
char *line_format_HTML=
"  <tr><td align=\"right\">%4d\
<td align=\"right\">%3d<td align=\"right\">%5d\
<td align=\"right\">%5d\
<td align=\"right\">%5d\
<td align=\"right\">%5d<td>%s\n";

char *tail_format_PLAIN="\n";
char *tail_format_HTML="</table>\n";




static int _list_handlers(FILE *out,
        modinfo_t *handlers, int len, time_t now, int outtype)
{
    int i, c, nk, nb, ni;
    char *count_format, *head_format, *line_format, *tail_format;


    if(handlers==NULL) {
        Log("; handlers is NULL");
        return 0;
    }

    if(outtype==0) {
        count_format = count_format_PLAIN;
        head_format  = head_format_PLAIN;
        line_format  = line_format_PLAIN;
        tail_format  = tail_format_PLAIN;
    }
    else {
        count_format = count_format_HTML;
        head_format  = head_format_HTML;
        line_format  = line_format_HTML;
        tail_format  = tail_format_HTML;
    }

    fflush(out);

#if 0
    c = nk = nb = ni = 0;
    for(i=0;i<len;i++) {
        if(handlers[i].wake==0)
            continue;
        if(handlers[i].busy==2)
            nk++;
        else
        if(handlers[i].busy==1)
            nb++;
        else
        if(handlers[i].busy==0)
            ni++;

        c++;
    }
    fprintf(out, count_format, c, nk, nb, ni, MAX_HANDLER);
#else
    c = nk = nb = ni = 0;
    for(i=0;i<len;i++) {
        if(handlers[i].wake==0)
            continue;

        nb += handlers[i].busy;

        c++;
    }
    fprintf(out, count_format, c, len, MAX_HANDLER, nb);
#endif



    fprintf(out, head_format);
    for(i=0;i<len;i++) {
        if(handlers[i].wake==0)
            continue;
        fprintf(out, line_format,
            i,
            handlers[i].busy,
            handlers[i].pid,
            -1,
#if 0
            handlers[i].req.sid, handlers[i].req.id,
#endif
            handlers[i].z_req,  handlers[i].z_sess,
#if 0
            handlers[i].url
#else
            " "
#endif
        );

    }
    fprintf(out, tail_format);
    fflush(out);

    return 0;
}

int list_handlers(modinfo_t *handlers, int len, time_t now)
{
    return _list_handlers(stdout, handlers, len, now, 0);
}
int list_handlers_PLAIN(modinfo_t *handlers, int len, time_t now)
{
    return _list_handlers(stdout, handlers, len, now, 0);
}
int list_handlers_with_HTML(modinfo_t *handlers, int len, time_t now)
{
    return _list_handlers(stdout, handlers, len, now, 1);
}





static int handler_state_fd = -1;


int open_handlers_state(time_t now)
{

    if((handler_state_fd = open(state_file, O_WRONLY|O_CREAT, 0644))<0) {
        Error("open_handlers_state: cannot open file '%s'", state_file);
        return -1;
    }

    lseek(handler_state_fd, (off_t)0L, SEEK_SET);
    if(write(handler_state_fd, &now, sizeof(time_t))<0) {
        return 1;
    }

    return 0;
}

int close_handlers_state()
{
    close(handler_state_fd);
    handler_state_fd = -1;

    return 0;
}


int write_handlers_state(modinfo_t *handlers, int len, time_t now)
{
    int i,tmp;

    if(handler_state_fd == -1) {
        if(open_handlers_state(now)) {
            return 0;
        }
    }

    /*
     * record
     *  - DATE
     *  - MAXHANDLER
     *  - number of handlers
     *  - handler #1's state
     *      ...
     *  - handler #n's state
     *
     */

    lseek(handler_state_fd, (off_t)0L, SEEK_SET);
    if(write(handler_state_fd, &now, sizeof(time_t))<0) {
        return 1;
    }
    tmp = MAX_HANDLER;
    if(write(handler_state_fd, &tmp, sizeof(tmp))<0) {
        return 1;
    }
    if(write(handler_state_fd, &len, sizeof(len))<0) {
        return 1;
    }
    for(i = 0; i < len; i++) {
        if(write(handler_state_fd,  &handlers[i], sizeof(modinfo_t))<0) {
            return 1;
        }
    }

    return 0;
}









#ifdef HSTATE_TEST



char *display_sec(time_t);

/*
 *
 * return:
 *      0   success
 *     -1   error (state file is not found)
 *      1   error
 */

key_t hstate_shmid;

int read_shm_display_handlers_state(char *handlers_state_file, int use_HTML,
        int (*listing_func)(modinfo_t*, int, time_t))
{
    time_t last_time;
    int i, len, max;
    time_t now;
    char   ctime_str[STRING_SIZE], ltime_str[STRING_SIZE];

    static modinfo_t *handlers = (modinfo_t*) NULL;
    static int last_max=-1;

    static int _count=0;

    _count++;
    time(&now);

    hstate_shmid = assign_pool(WCOL_SHMKEY_HSTATE);
    if(hstate_shmid == -1 ) {
        Error("init_handlers: cannot alloc shm");
        return -1;
    }

#if 1
    show_pool_stat(hstate_shmid);
#endif

    if((handlers = (modinfo_t*)shmat(hstate_shmid, (void*)NULL, 0))==NULL) {
/*
        exit(WCOL_EXIT_NOMEMORY);
*/
        Fatal("init_handlers: fail to shmat()  0 [%d](%d)", _count, errno);
        return 1;
    }
    if(((int)handlers) == -1 ) {
        Fatal("init_handlers: fail to shmat() -1 [%d](%d)", _count, errno);
        return 1;
    }

    len = max = MAX_HANDLER;

    strcpy(ctime_str, ctime(&now));
    strcpy(ltime_str, ctime(&last_time));

    if(use_HTML) {
        fprintf(stdout,
"<h2> Wcol's handlers states</h2>\n\
\n\
<table>\n\
  <tr><th align=\"left\">Current Time<td align=\"right\">%10d <td> %s\
</table>\n\n", now, ctime_str);

        list_handlers_with_HTML(handlers, len, now);
    }
    else {
        fprintf(stdout, "=== last handlers states\n");
        fprintf(stdout, "Current Time %10d        : %s", now, ctime_str);

        list_handlers(handlers, len, now);
    }
    fprintf(stdout, "\n");

    shmdt(handlers);

    handler_state_fd = -1;

    return 0;
}

int read_file_display_handlers_state(char *handlers_state_file, int use_HTML,
        int (*listing_func)(modinfo_t*, int, time_t))
{
    time_t last_time;
    int i, len, max;
    time_t now;
    char   ctime_str[STRING_SIZE], ltime_str[STRING_SIZE];

    static modinfo_t *handlers = (modinfo_t*) NULL;
    static int last_max=-1;

    time(&now);

    if((handler_state_fd =
            open(handlers_state_file, O_RDONLY|O_CREAT, 0644))<0) {
        Error("read_display_handlers_state: cannot open file '%s'",
            state_file);
        return -1;
    }

    /*
     * read
     *  - DATE
     *  - MAXHANDLER
     *  - number of handlers
     *  - handler #1's state
     *      ...
     *  - handler #n's state
     */

    lseek(handler_state_fd, (off_t)0L, SEEK_SET);
    if(read(handler_state_fd, &last_time, sizeof(time_t))!=sizeof(time_t)) {
        Error("read_display_handlers_state: no time stamp");
        return 1;
    }
    if(read(handler_state_fd, &max, sizeof(max))!=sizeof(max)) {
        Error("read_display_handlers_state: no max of handlers");
        return 1;
    }
    if(read(handler_state_fd, &len, sizeof(len))!=sizeof(len)) {
        Error("read_display_handlers_state: no number of handlers");
        return 1;
    }

    /* max is changed */
    if(max != last_max) {
        free(handlers);
        handlers = NULL;
    }
    if(handlers==NULL) {
        if((handlers = (modinfo_t*)MALLOC(sizeof(modinfo_t)*max))==NULL) {
            Fatal("read_display_handlers_state: not enought memory.");
            return 1;
        }
        else {
            last_max = max;
        }
    }

    for(i = 0; i < len; i++) {
        if(read(handler_state_fd,  &handlers[i], sizeof(modinfo_t))
                !=sizeof(modinfo_t)) {
            Error(
    "read_display_handlers_state: read error, when read %d th handler.", i);
            return 1;
        }
    }

    strcpy(ctime_str, ctime(&now));
    strcpy(ltime_str, ctime(&last_time));

    if(use_HTML) {
#if 0
        Trace("<h2> Wcol's handlers states</h2>\n");
        Trace("<table>\n");
        Trace("<tr><td  align=\"left\">Current Time <td align=\"right\">%10d <td> %s", now, ctime(&now));
        Trace("<tr><td>Time-stamp   <td align=\"left\">%10d <td align=\"right\"> %s", last_time, ctime(&last_time));
        Trace("<tr><td>             <td align=\"left\">%10d sec(s) <td align=\"right\"> %s\n",
                        now-last_time, display_sec(now-last_time));
        Trace("</table>\n");
#endif

        fprintf(stdout,
"<h2> Wcol's handlers states</h2>\n\
\n\
<table>\n\
  <tr><th align=\"left\">Current Time<td align=\"right\">%10d <td> %s\
  <tr><th align=\"left\">Time-stamp  <td align=\"right\">%10d <td> %s\
  <tr><th align=\"left\">            <td align=\"right\">%10d <td> (%s)\n\
</table>\n\n", now, ctime_str, last_time, ltime_str,
            now-last_time, display_sec(now-last_time));

        list_handlers_with_HTML(handlers, len, now);
    }
    else {
        fprintf(stdout, "=== last handlers states\n");
        fprintf(stdout, "Current Time %10d        : %s", now, ctime_str);
        fprintf(stdout, "Time-stamp   %10d        : %s", last_time, ltime_str);
        fprintf(stdout, "             %10d sec(s) :   %s\n\n",
                        now-last_time, display_sec(now-last_time));

        list_handlers(handlers, len, now);
    }
    fprintf(stdout, "\n");

    close(handler_state_fd);
    handler_state_fd = -1;

    return 0;
}



char state_file[STRING_SIZE];




static int usage()
{
    fprintf(stdout,
"hstate - print Wcol's handlers states\n\
    by k-chinen@is.aist-nara.ac.jp, 1996\n\
\n\
usage: hstate [options]\n\
\n\
option:\n\
    -h          help, print this messages.\n\
    -f file     specify state file. (default '%s')\n\
    -l          loop\n\
    -w #sec     wait interval\n\
    -c          for cgi-bin, use HTML\n\
\n\
", state_file);

    return 0;
}


int loop_flag=0;
int wait_interval=5;
int use_HTML=0;
int (*func)(modinfo_t*, int, time_t);

void Init(int argc, char **argv)
{
    int flag;

    extern char *optarg;
    extern int  optind;

    /* defaults */
    trace_flag = 1;
    strcpy(state_file, DEFAULT_STATEFILE);

    while((flag = getopt(argc, argv, "hf:lw:c"))!= -1) {
        switch(flag) {
        case 'h':
            usage();
            exit(0);
            break;
        case 'f':
            strcpy(state_file, optarg);
            break;
        case 'l':
            loop_flag = 1 - loop_flag;
            break;
        case 'w':
            wait_interval = atoi(optarg);
            break;
        case 'c':
            use_HTML = 1 - use_HTML;
        }
    }

    if(use_HTML) {
        func=list_handlers;
    }
    else {
        func=list_handlers_with_HTML;
    }
}



main(int argc, char **argv)
{
    Init(argc, argv);
#ifdef MEMORY_CHECK
    show_memoryusage();
#endif

    if(!loop_flag) {
        read_shm_display_handlers_state(state_file, use_HTML, func);
#ifdef MEMORY_CHECK
        show_memoryusage();
#endif
    }
    else {
        do {
            read_shm_display_handlers_state(state_file, use_HTML, func);
#ifdef MEMORY_CHECK
            show_memoryusage();
#endif
            sleep(wait_interval);
        } while(1);
    }
}



#endif /* HSTATE_TEST */



