/*
 *  WCOL:   HTTP procedures
 *      by k-chinen@is.aist-nara.ac.jp, 1994, 1995, 1996
 *
 * $Id: http.c,v 1.3 1997/09/15 01:58:09 k-chinen Exp k-chinen $
 */

/*
 * NOTE:
 *      see RFC1945 (HTTP/1.0) and internet drafts around WWW.
 */


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif


/*
 * If you set DEBUG_HTTP, you will get details of rountines in this file.
 */
#define _DEBUG_HTTP
#define _DEBUG_REQUEST          /* debug request */
#define _DEBUG_HPARSE           /* debug header parsing */
#define _TRACE_RECV_REQUEST
#define _TRACE_RECV_REQUEST_DETAIL
#define _TRACE_SEND_REQUEST
#define _TRACE_SEND_REQUEST_DETAIL



#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <fcntl.h>
#include <limits.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <signal.h>
#include <setjmp.h>
#include <errno.h>
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif

#ifdef __STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif


#include "wcol.h"
#include "sigs.h"
#include "info.h"
#include "misc.h"
#include "socket.h"
#include "request.h"
#include "http.h"
#include "handler.h"

#include "ctree.h"
#include "field.h"

#ifdef WCOL_SUB_TIME
#undef WCOL_SUB_TIME
#endif

#define WCOL_SUB_TIME(x,y) \
        ( (x.tv_sec  - y.tv_sec )*1000 + (x.tv_usec - y.tv_usec)/1000 )


#if 1
url_t *fetch_parent;
url_t *prefetch_parent;
url_t *import_sibling;
#endif

extern int remove_cookie_flag;
extern int remove_referer_flag;
extern int mcache_flag;



#if 1
static char *get_request_base_format = "GET %s HTTP/1.0\r\n\
Accept: */*\r\nUser-Agent: WWW-Collector-E/%s\r\n";
static char *get_request_regular_format = "GET %s HTTP/1.0\r\n\
Accept: */*\r\nUser-Agent: WWW-Collector-E/%s\r\n\r\n";
static char *get_request_authed_format = "GET %s HTTP/1.0\r\n\
Accept: */*\r\nUser-Agent: WWW-Collector-E/%s\r\n%s\r\n\r\n";
#else
static char *get_request_base_format = "GET %s HTTP/1.0\r\n\
Accept: */*\r\nHost: %s\r\nUser-Agent: WWW-Collector-E/%s\r\n";
static char *get_request_regular_format = "GET %s HTTP/1.0\r\n\
Accept: */*\r\nHost: %s\r\nUser-Agent: WWW-Collector-E/%s\r\n\r\n";
static char *get_request_authed_format = "GET %s HTTP/1.0\r\n\
Accept: */*\r\nHost: %s\r\nUser-Agent: WWW-Collector-E/%s\r\n%s\r\n\r\n";
#endif

static char *error_http1_format =
  "HTTP/1.0 %d %s\r\nMIME-version: 1.0\r\nContent-type: text/html\r\n\r\n";
static char *error_msg_format =
  "<HEAD><TITLE>Wcol Warning Message</TITLE></HEAD>\r\n<BODY>\r\n\
<H1>%d %s</H1>\r\n%s\r\n<HR>\r\n<em>Wcol</em>\r\n</BODY>\r\n";






int HTTP_ErrorMessage_Old(int fd,
    int err_no, char *smessage, char *lmessage, int http1)
{
    char msg[STRING_SIZE];
    int ret;

    ret = -1;

    if(http1) {
        if(smessage!=NULL && smessage[0]!='\0') {
            sprintf(msg, error_http1_format, err_no, smessage);
            ret = write(fd, msg, strlen(msg));
        }
    }

    if(smessage!=NULL && smessage[0]!='\0'
        && lmessage!=NULL && lmessage[0]!='\0') {
        sprintf(msg, error_msg_format, err_no, smessage, lmessage);
        ret = write(fd, msg, strlen(msg));
    }

    return ret;
}



int HTTP_ErrorMessage(int fd, int major_ver, int minor_ver,
            int rcode, char *smessage, char *fmt, ...)
{
    va_list args;
    char tmp[STRING_SIZE];
    char line[STRING_SIZE];
    int ret;

    va_start(args, fmt);
    vsprintf(line, fmt, args);
    va_end(args);

    if(major_ver>0) {
        if(smessage!=NULL && smessage[0]!='\0') {
            sprintf(tmp, error_http1_format, rcode, smessage);
            ret = write(fd, tmp, strlen(tmp));
        }
    }

    if(smessage!=NULL && smessage[0]!='\0'
        && line[0]!='\0') {
        sprintf(tmp, error_msg_format, rcode, smessage, line);
        ret = write(fd, tmp, strlen(tmp));
    }

    return ret;
}



static
int verify_strcat(char *dst, int len, char *src)
{
	int dst_len;
	int src_len;

	dst_len = strlen(dst);
	src_len = strlen(src);
	if(dst_len+src_len < len) {
		strcat(dst, src);
		return 0;
	}
	else {
		return -1;
	}
}

static
int verify_linecat(char *dst, int len, char *src)
{
	int dst_len;
	int src_len;

	dst_len = strlen(dst);
	src_len = strlen(src)+2;
	if(dst_len+src_len < len) {
		strcat(dst, src);
		strcat(dst, "\r\n");
		return 0;
	}
	else {
		return -1;
	}
}

static
int verify_fieldcat(char *dst, int len, char *fname, char *fvalue)
{
	int dst_len;
	int name_len;
	int value_len;

	dst_len 	= strlen(dst);
	name_len 	= strlen(fname)+2;
	value_len 	= strlen(fvalue)+2;
	if(dst_len+name_len+value_len <len) {
		strcat(dst, fname);
		strcat(dst, ": ");
		strcat(dst, fvalue);
		strcat(dst, "\r\n");
		return 0;
	}
	else {
		return -1;
	}
}



/*
 * return:
 *      error   -1
 *      success length of writen with bytes
 *
 */
int HTTP_NewRequest(int fd, char *host, char *path,
        char *auth, char *cookie, int nocache, int noprefetch)
{
    char buf[HTTP_HEADER_SIZE+1];
    char tmp[STRING_SIZE];
    off_t nwrite;

#ifdef DEBUG_HTTP
    Trace("HTTP_NewRequest:* URL='%s'\n", path);
#endif
	memset(buf, 0, sizeof(char)*(HTTP_HEADER_SIZE+1));

	if(verify_strcat(buf, HTTP_HEADER_SIZE-3, "GET ")) 	goto overrun;
	if(verify_strcat(buf, HTTP_HEADER_SIZE-3, path)) 	goto overrun;
	if(verify_strcat(buf, HTTP_HEADER_SIZE-3, " HTTP/1.0\r\n"))
		goto overrun;

    if(host && *host) {
		if(verify_fieldcat(buf, HTTP_HEADER_SIZE-3, "Host", host))
			goto overrun;
    }

	if(verify_fieldcat(buf, HTTP_HEADER_SIZE-3, "Accept","*/*"))
		goto overrun;

	if(verify_strcat(buf, HTTP_HEADER_SIZE-3, "User-Agent: " ))
		goto overrun;
  	if(verify_strcat(buf, HTTP_HEADER_SIZE-3, "WWW-Collector-E/"))
		goto overrun;
	if(verify_linecat(buf, HTTP_HEADER_SIZE-3, WCOL_VER))
		goto overrun;

    if(auth && *auth) {
		if(verify_linecat(buf, HTTP_HEADER_SIZE-3, auth))
			goto overrun;
    }

    if(!remove_cookie_flag && cookie && *cookie) {
		if(verify_linecat(buf, HTTP_HEADER_SIZE-3, cookie))
			goto overrun;
    }

    if(nocache) {
		if(verify_fieldcat(buf, HTTP_HEADER_SIZE-3, 
				"Pragma", "no-cache"))
			goto overrun;
    }
    if(noprefetch){
		if(verify_fieldcat(buf, HTTP_HEADER_SIZE-3, 
				"Pragma", "no-prefetch"))
			goto overrun;
    }

	if(verify_strcat(buf, HTTP_HEADER_SIZE-1, "\r\n"))
		goto overrun;


#ifdef TRACE_SEND_REQUEST
	dump_stringchunk("Send*", buf);
    Trace("@@@length of this request %d\n", strlen(buf));
#endif

    pipe_broken = 0;
    if((nwrite =
		 (off_t)write(fd, buf, sizeof(char)*strlen(buf)))<=(off_t)0) {
        Error("HTTP_NewRequest: Cannot write request to server, '%s'",
			path);
        return -1;
    }
    if(pipe_broken) {
        pipe_broken = 0;
        Error("HTTP_NewRequest: PIPEBROKEN: path '%s'", path);
        return -1;
    }

    return nwrite;

overrun:
 	Error(
 	  "HTTP_NewRequest: request buffer overrun [current length %d]\n",
		HTTP_HEADER_SIZE-1);

#ifdef TRACE_HTTP
	   	dump_stringchunk("Cur", buf);
#endif

	return -1;
}



/*
 * return:
 *      error   -1
 *      success length of writen with bytes
 *
 */
int HTTP_CookedRequest(int fd, int nocache,
			char *host, char *path, char *rbuf)
{
    char buf[HTTP_HEADER_SIZE+1];
    char tmp[STRING_SIZE];
    char *pos;
    char fname[STRING_SIZE];
    register char *p, *q;
    void *chk;
    int have_host=0;
    off_t nwrite;
	int  c=0;

#ifdef DEBUG_HTTP
    Trace("HTTP_CookedRequest: URL='%s'\n", path);
#endif

	memset(buf, 0, sizeof(char)*(HTTP_HEADER_SIZE+1));

#ifdef TRACE_SEND_REQUEST
    dump_stringchunk("Raw ^", rbuf);
#endif

	if(verify_strcat(buf, HTTP_HEADER_SIZE-3, "GET "))  goto overrun;
    if(verify_strcat(buf, HTTP_HEADER_SIZE-3, path))    goto overrun;
    if(verify_strcat(buf, HTTP_HEADER_SIZE-3, " HTTP/1.0\r\n"))
        goto overrun;

    pos = rbuf;
    pos = sgets(rbuf, pos, tmp, STRING_SIZE);
    if(pos==NULL) {
        Error("HTTP_CookedRequest: empty request");
        return -1;
    }

#ifdef TRACE_SEND_REQUEST
    Trace("skip simple request (URL) in original reqeust\n  '%s'\n", tmp);
#endif

	c = 1;
    while((pos = sgets(rbuf, pos, tmp, STRING_SIZE))!=NULL) {
		/* empty field */
        if(*tmp=='\0' || *tmp=='\r' || *tmp=='\n') {
            break;
        }

        if(!have_host && strncasecmp_first_all(tmp, 5, 'H', "Host:")) {
            have_host = 1;
        }

        if(strncasecmp_first_all(tmp, 6, 'P', "Proxy-")) {
            goto next_field;
        }

        if(nocache &&
          strncasecmp_first_all(tmp, 18, 'I', "If-Modified-Since:")) {
            goto next_field;
        }


        if(remove_cookie_flag &&
          strncasecmp_first_all(tmp, 7, 'C', "Cookie:")) {
            goto next_field;
        }
        if(remove_referer_flag &&
          strncasecmp_first_all(tmp, 8, 'R', "Referer:")) {
            goto next_field;
        }

        p = tmp;
        q = fname;
        while(*p && *p!=':') {
            *q++ = *p++;
        }
        *q = '\0';
#ifdef TRACE_SEND_REQUEST_DETAIL
        Trace("field [%s]\n", fname);
#endif

		/* remove specified field */
        if(RSReq && (chk = ctree_search_case(RSReq, fname)) !=NULL) {
            goto next_field;
        }
		/* change specified field */
        if(CSReq && (chk = ctree_search_case(CSReq, fname)) != NULL) {
			if(verify_fieldcat(buf, HTTP_HEADER_SIZE-3,
					fname, (char*)chk))
				goto overrun;
        }
		else {
			if(verify_strcat(buf, HTTP_HEADER_SIZE-3, tmp))
				goto overrun;
		}


next_field:
		/* rest of header is empty */
        if(*pos=='\0' || *pos=='\r' || *pos=='\n') {
            break;
        }

		c++;
    }


    if(!have_host && host && *host) {
		if(verify_fieldcat(buf, HTTP_HEADER_SIZE-3, "Host", host))
				goto overrun;
    }

#if 0
    sprintf(tmp, "Via: 1.0 :%s \r\n", here_hostname);
    strcat(buf, tmp);
#endif


    if(verify_strcat(buf, HTTP_HEADER_SIZE-3, "X-Proxy-Agent: " ))
        goto overrun;
    if(verify_strcat(buf, HTTP_HEADER_SIZE-3, "WWW-Collector-E/"))
        goto overrun;
    if(verify_linecat(buf, HTTP_HEADER_SIZE-3, WCOL_VER))
        goto overrun;

    if(verify_strcat(buf, HTTP_HEADER_SIZE-1, "\r\n"))
        goto overrun;


#ifdef TRACE_SEND_REQUEST
    dump_stringchunk("Send^", buf);
    Trace("@@@length of this request %d\n", strlen(buf));
#endif

    pipe_broken = 0;
    if((nwrite = (off_t)write(fd, buf, sizeof(char)*strlen(buf)))<=(off_t)0) {
        Error("HTTP_CookedRequest: Cannot write request to server, '%s' (%d)",
            path, errno);
        return -1;
    }
    if(pipe_broken) {
        pipe_broken = 0;
        Error("HTTP_CookedRequest: PIPEBROKEN: path '%s'", path);
        return -1;
    }

    return nwrite;

overrun:
 	Error(
 	  "HTTP_CookedRequest: request buffer overrun [current length %d]\n",
		HTTP_HEADER_SIZE-1);

#ifdef TRACE_HTTP
	   	dump_stringchunk("Cur", buf);
#endif
	return -1;
}



/*
 *
 */
int HTTP_Request(int fd, char *url, char *auth)
{
    char protocol[STRING_SIZE];
    char servername[STRING_SIZE], portname[STRING_SIZE];
    char path[STRING_SIZE];

    char buf[STRING_SIZE];

#ifdef DEBUG_HTTP
    Trace("HTTP_Request: URL='%s'\n", url);
#endif

    /* for proxy access */
    if(fetch_parent!=NULL) {
#ifdef USE_SHORT_URL

        /* remove ':80' in URL */
        ParseURL(url, protocol, servername, portname, path);
        if(strcmp_first_all(protocol, 'h', "http")==0 &&
            (strcmp(portname, "80")==0 || strcmp(portname, "http")==0)) {
                sprintf(path, "%s://%s%s", protocol, servername, path);

                /*
                Trace("HTTP_Request: remove ':80' URL='%s'\n", path);
                */
        }
        else {
            strcpy(path, url);

            /*
            Trace("HTTP_Request: direct URL='%s'\n", path);
            */
        }
#else
        strcpy(path, url);
#endif
    }
    else {
        ParseURL(url, protocol, servername, portname, path);
    }


    /* if auth is NULL or Empty, don't talk about Authorization */
    if(auth && *auth) {
        sprintf(buf, get_request_authed_format, path, WCOL_VER, auth);
    }
    else {
        sprintf(buf, get_request_regular_format, path, WCOL_VER);
    }

#ifdef DEBUG_HTTP
    Trace("\tRequest is '%s'\n", buf);
#endif

    pipe_broken = 0;
    if(write(fd, buf, sizeof(char)*strlen(buf))<=0) {
        Error("HTTP_Request: Cannot write request to server, URL='%s'", url);
        return -1;
    }
    if(pipe_broken) {
        pipe_broken = 0;
        Error("HTTP_Request: PIPEBROKEN: URL='%s'", url);
        return -1;
    }

    return 0;
}


/*
 * Parse Simple Request
 *
 *  return:
 *      1   error
 *      0   success (HTTP/0.9 or HTTP/1.0)
 */
static HTTP_ParseSimpleRequest(char *line, int *method, char *url,
        char *ver_major, char *ver_minor)
{
    char tmp[INFONAMESIZE+1];
    register char *p, *q;
    register int c;

#if 0
    Trace("HTTP_ParseSimpleRequest: '%s'\n", line);
#endif

    /* init */
    *method = METHOD_UNKNOW;
    *url    = '\0';
    *ver_major = 0;
    *ver_minor = 9;

    p = line;

    /* parse method */
    q = tmp;
    c = 0;
    while(*p && *p!=' ' && *p!='\t' && *p!='\r' && *p!='\n'
        && c<INFOVERSIZE) {
        *q++ = *p++;
        c++;
    }
    *q = '\0';
    if(!*p) {
        return 1;                               /* METHOD ... error */
    }
    else {
        if(strcmp_first_all(tmp,'G',"GET"))     *method = METHOD_GET;
        if(strcmp_first_all(tmp,'H',"HEAD"))    *method = METHOD_HEAD;
        if(strcmp_first_all(tmp,'P',"POST"))    *method = METHOD_POST;
    }

    /* parse url */
    p++;
#ifdef DEBUG_REQUEST
    Trace("Pa '%s'\n", p);
#endif
    q = tmp;
    c = 0;
    while(*p && *p!=' ' && *p!='\t' && *p!='\r' && *p!='\n'
        && c<INFONAMESIZE) {
        *q++ = *p++;
        c++;
    }
    *q = '\0';
    if(c>=INFONAMESIZE) {
        Error("HTTP_ParseSimpleRequest: too long URL '%s'", tmp);
        return 1;           /* error, long url */
    }
    strcpy(url, tmp);

    if(!*p) {
        return 0;           /* METHOD URL ... HTTP/0.9 success */
    }


    p++;
#ifdef DEBUG_REQUEST
    Trace("Pa '%s'\n", p);
#endif

    {
        int mj, mi;

        sscanf(p, "HTTP/%d.%d", &mj, &mi);
        *ver_major = mj;
        *ver_minor = mi;
#ifdef DEBUG_REQUEST
        Trace("Pa Ver   '%s' -> %d.%d -> %d.%d\n",
            p, mj, mi, *ver_major, *ver_minor);
#endif
    }


    return 0;               /* METHOD URL VER ... HTTP/1.0 or later success */
}




/* extern */ int CheckSupportedRequest(request_t req);



static int local_wcol_request_count = 0;

/*
 * HTTP_ParseRequest - parse HTTP request
 *
 *  return  0   success
 *          1   error
 *          -1  empty
 *          -2  internal error (no disk, no file descriptor)
 */
int HTTP_ParseRequest(int sock_fd, request_t *req, char *req_cont)
{
    char rbuf[HTTP_HEADER_SIZE+1];       /* header and sentinel */
    int  len, sum;
    char line[STRING_SIZE];
    char *head_top=NULL, *body_top=NULL;
    register char *p,*q;
    register int c;
    int  header_end;
    char head_fn[STRING_SIZE];
    int  fd;
    int  slen;
    struct sockaddr_in tmp_buf;

    /*
     * Init
     */
    memset(req, 0, sizeof(request_t));
    req->mode       = REQ_CLIENT;
    req->level      = 0;
    req->sid        = - (int)wcol_pid;
    req->id         = ++local_wcol_request_count;
    req->auth[0]    = '\0';
    req->cookie[0]   = '\0';
    req->ims        = -1;
    req->fd         = sock_fd;
    req->state      = REQ_NOP;
    gettimeofday(&req->issue_time, NULL);
    req->keep_alive = 0;


    slen = sizeof(tmp_buf);
    memset(&tmp_buf, 0, slen);

    if(getpeername(sock_fd, (struct sockaddr*)&tmp_buf, &slen)) {
        Error("HTTP_ParseRequest: fail getpeername() (%d)", errno);

        strcpy(req->peername, "-");
    }
    else {
        p = (char*)inet_ntoa( tmp_buf.sin_addr );
        if(p!=NULL) {
            strcpy(req->peername, p);

#ifdef CHECK_PEERADDRRESS
            /***
             *** check peername
             ***/
            {
                register int c_p,c_w;
                register char *p;

                p = req->peername;
                c_p = c_w = 0;
                while(*p) {
                    if(*p=='.') {
                        c_p++;
                    }
                    if(*p==' '||*p=='\t'||*p=='\r'||*p=='\n') {
                        c_w++;
                        *p = '_';   /* fill dummy charactor */
                    }

                    p++;
                }
                if(c_p==0||c_w>0) {
                    Log("; strange peername [%s]", req->peername);
                }
            }
#endif
        }
        else {
            strcpy(req->peername, "-");
        }
    }


    if(getsockname(sock_fd, (struct sockaddr*)&tmp_buf, &slen)) {
        Error("HTTP_ParseRequest: fail getsockname() (%d)", errno);

        req->hereport = -1;
    }
    else {
/*
        req->hereport = tmp_buf.sin_port;
*/
        req->hereport = ntohs(tmp_buf.sin_port);
    }




    header_end   = req->nocache = req->noprefetch = 0;

    req_cont[0] = '\0';


    /*
     * Read clients request
     *      if cannot read it, replace error mssage as dummy request.
     */
try0:
    len = read(sock_fd, rbuf, HTTP_HEADER_SIZE);
    if(len<0) {
        if(errno == EINTR) {
            goto try0;
        }
#ifdef ECONNRESET
        if(errno == ECONNRESET) {
            Error("HTTP_ParseRequest: socket #%d is reseted by peer (%d)",
                sock_fd, errno);
            return -1;
        }
#endif
        Error("HTTP_ParseRequest: Cannot read request from socket #%d (%d)",
            sock_fd, errno);
        return 1;
    }
    if(len==0) {
        Error("HTTP_ParseRequest: got empty request");
        return -1;
    }
    rbuf[len]='\0';
    strcat(req_cont, rbuf);

#ifdef TRACE_RECV_REQUEST_DETAIL
    Trace("\r\nClient said (%d bytes raw):\r\n%s\r\n--- END ---\r\n",len,rbuf);
#endif


    /*
     * Parse simple-request from a client
     */
    /* Cut 1 line for simple request */
    p = rbuf;
    q = line;
    while(*p&&*p!='\n'&&*p!='\r')
        *q++ = *p++;
    *q = '\0';

#ifdef DEBUG_REQUEST
    Trace("Read..  '%s'\n", line);
#endif

    if(HTTP_ParseSimpleRequest(line, &req->method, req->url,
            &req->ver_major, &req->ver_minor)) {
        Error("HTTP_ParseRequest: strange request, cannot parse");
        return 1;
    }
#ifdef DEBUG_REQUEST
    Trace("So...   Method = '%s'\n",    ReqMethodStr(req->method));
    Trace("        URL    = '%s'\n",    req->url);
    Trace("        Ver    = %d.%d\n",   req->ver_major, req->ver_minor);
#endif


#if 0
    /*
     * Check supported request
     */
    if(CheckSupportedRequest(*req)) {
        Error("HTTP_ParseRequest: it is not support '%s'", req->url);
        return 1;
    }
#endif


    /* store top of rest of command to 'head_top' */
    head_top = p;
    if(*head_top=='\r') head_top++;
    if(*head_top=='\n') head_top++;     /* seek to next line */




    /*
     * Parse HTTP/1.0 fields (rest of Header)
     */
    if(req->ver_major>=1) {
#ifdef DEBUG_REQUEST
        Trace("Parse haader. \n");
#endif

        p = head_top;       /* restore rest of command from 'head_top' */
        do {


#ifdef TRACE_RECV_REQUEST_DETAIL
            dump_string("  ", p);
#endif

            if(*p=='\r'||*p=='\n') {
                header_end = 1;
                body_top = p+1;
                break;
            }

#if 0
            if(*p == 'C' && strncmp("Cookie:",p,7)==0)
#else
            if(strncasecmp_first_all(p, 7, 'C', "Cookie:"))
#endif
                                                        {
/*
                Trace("Oh. 'Cookie' found ...'%s'.", p);
*/
                dump_string("  found cookie ", p);
                c = 0;
                q = req->cookie;
                while(*p && *p!='\r' && *p!='\n' && c<STRING_SIZE) {
                    *q++ = *p++;
                    c++;
                }
                *q = '\0';
            }
#if 0
            if(*p == 'A' && strncmp("Authorization:",p,14)==0)
#else
            if(strncasecmp_first_all(p, 14, 'A', "Authorization:"))
#endif
                                                                {
/*
                Trace("Oh. 'Authorization' found ...'%s'.", p);
*/
                dump_string("  found authorization ", p);
                q = req->auth;
                c = 0;
                while(*p && *p!='\r' && *p!='\n' && c<STRING_SIZE) {
                    *q++ = *p++;
                    c++;
                }
                *q = '\0';
            }
            if(*p == 'I' && strncmp("If-Modified-Since:",p,18)==0) {
                p += 18;
                req->ims = parse_date(p);
            }
            if(*p == 'P' && strncmp("Pragma:",p,7)==0) {
                p += 7;
                while(*p==' '||*p=='\t')
                    p++;
                if(*p == 'n' && strncmp("no-cache",p,8)==0) {
                    p += 8;
                    req->nocache = 1;
                }else if(*p == 'n' && strncmp("no-prefetch",p,11)==0){
            p += 11;
            req->noprefetch = 1;
        }
            }
            if(*p == 'P' && strncmp("Proxy-Connection:",p,17)==0) {
                p += 17;
                while(*p==' '||*p=='\t')
                    p++;
#ifdef SUPPORT_CLIENT_KEEPALIVE
                if(*p == 'K' && strncmp("Keep-Alive",p,10)==0) {
                    p += 10;
                    req->keep_alive = 1;
                }
#endif
            }

            /* skip this line */
            while(*p&&*p!='\r'&&*p!='\n')
                p++;
            if(*p=='\r') p++;
            if(*p=='\n') p++;

            if((!header_end) && (!*p)) {
try1:
                if((len = read(sock_fd, rbuf, HTTP_HEADER_SIZE))<=0) {
                    if(errno == EINTR)
                        goto try1;
                    break;
                }
                else {
                    rbuf[len]='\0';
                    strcat(req_cont, rbuf);


#ifdef DEBUG_REQUEST
    Trace("Client said more (%d bytes raw):\r\n%s\r\n--- END ---\r\n",len,rbuf);
#endif
                    p = rbuf;
                }
            }
        } while(!header_end);
    }
    else {
        Trace("No options...\n");
    }



#ifdef TRACE_RECV_REQUEST
    dump_stringchunk("Recv", req_cont);
#endif


#ifdef DEBUG_REQUEST
    Trace("HTTP_ParseRequest:\n");
    ShowRequest(req);
#endif

    return 0;
}


/*
 *
 */
int HTTP_ParseHeaderBuffer(Info *info, char *hbuf)
{
    char line[STRING_SIZE], tmp[STRING_SIZE];
    register char *p, *q;
    register int c;
    char *cpos;


    if(info==NULL)
        return -1;

    cpos = hbuf;
    while((cpos = sgets(hbuf, cpos, line, STRING_SIZE))!=NULL) {
#ifdef DEBUG_HTTP
        dump_string("  field:", line);
#endif

        if(*line=='C' && strncasecmp("Content-type: ",line,14)==0) {
            p = &line[14];
            q = info->attr.type;
            c = 0;
            while(*p && c<INFOTYPESIZE && *p!='\r' && *p!='\n') {
                *q++ = *p++;
                c++;
            }
            *q = '\0';

#ifdef DEBUG_HTTP
            Trace("HTTP_ParseHeaderFile: type '%s'\n", info->attr.type);
#endif
        }
        if(*line=='C' && strncasecmp("Content-Length: ",line,16)==0) {
            p = &line[16];
            q = tmp;
            c = 0;
            while(*p && *p!='\r' && *p!='\n') {
                *q++ = *p++;
                c++;
            }
            *q = '\0';

            info->attr.clength = (off_t)atol(tmp);
#ifdef DEBUG_HTTP
            Trace("HTTP_ParseHeaderFile: length '%s'-> %d\n",
                tmp, info->attr.clength);
#endif
        }
        if(*line=='L' && strncasecmp("last-modified: ",line,15)==0) {
            p = &line[15];
            q = tmp;
            while(*p&&*p!='\r'&&*p!='\n')
               *q++ = *p++;
            *q = '\0';
            info->attr.lmd= parse_date(tmp);

#ifdef DEBUG_HTTP
            Trace("HTTP_ParseHeaderFile: LM '%s' -> %d\n",
                tmp, info->attr.lmd);
#endif
        }
        if(*line=='E' && strncasecmp("expires: ",line,9)==0) {
            p = &line[9];
            q = tmp;
            while(*p&&*p!='\r'&&*p!='\n')
               *q++ = *p++;
            *q = '\0';
            info->attr.expire = parse_date(tmp);

#ifdef DEBUG_HTTP
            Trace("HTTP_ParseHeaderFile: expire '%s' -> %d\n",
                tmp, info->attr.expire);
#endif
        }
    }


#ifdef DEBUG_HTTP
    What(info);
#endif

    return 0;
}


/*
 *
 */
static int HTTP_ParseHeaderFile(char *name,
    char *type, off_t *len,time_t *lm, time_t *ex)
{
    char fname[STRING_SIZE];
    FILE *fp;
    char line[STRING_SIZE], tmp[STRING_SIZE];
    register char *p, *q;
    register int c;


    *type = '\0';
    *ex   = -1;
    *len  = -1;

    if(AssignFileName(fname, name, HEAD_FILE)) {
        return 1;
    }
    if((fp=fopen(fname,"r"))==NULL) {
        Error("HTTP_ParseHeaderFile: Cannot open HEAD file '%s'.",fname);

        return 1;
    }

    while(fgets(line, STRING_SIZE, fp)!=NULL) {
#ifdef DEBUG_HTTP
        Trace("\t> %s", line);
#endif

        if(*line=='C' && strncasecmp("Content-type: ",line,14)==0) {
            p = &line[14];
            q = type;
            c = 0;
            while(*p && c<INFOTYPESIZE && *p!='\r' && *p!='\n') {
                *q++ = *p++;
                c++;
            }
            *q = '\0';

#ifdef DEBUG_HTTP
            Trace("HTTP_ParseHeaderFile: type '%s'\n", type);
#endif
        }
        if(*line=='C' && strncasecmp("Content-Length: ",line,16)==0) {
            p = &line[16];
            q = tmp;
            c = 0;
            while(*p && *p!='\r' && *p!='\n') {
                *q++ = *p++;
                c++;
            }
            *q = '\0';

            *len = (off_t)atol(tmp);
#ifdef DEBUG_HTTP
            /*
            Trace("HTTP_ParseHeaderFile: length '%s'-> %d\n", tmp, atoi(tmp));
            */
            Trace("HTTP_ParseHeaderFile: length '%s'-> %d\n", tmp, *len);
#endif
        }
        if(*line=='L' && strncasecmp("last-modified: ",line,15)==0) {
            p = &line[15];
            q = tmp;
            while(*p&&*p!='\r'&&*p!='\n')
               *q++ = *p++;
            *q = '\0';
            *lm = parse_date(tmp);

#ifdef DEBUG_HTTP
            Trace("HTTP_ParseHeaderFile: LM '%s' -> %d\n", tmp, *lm);
#endif
        }
        if(*line=='E' && strncasecmp("expires: ",line,9)==0) {
            p = &line[9];
            q = tmp;
            while(*p&&*p!='\r'&&*p!='\n')
               *q++ = *p++;
            *q = '\0';
            *ex = parse_date(tmp);

#ifdef DEBUG_HTTP
            Trace("HTTP_ParseHeaderFile: expire '%s' -> %d\n", tmp, *ex);
#endif
        }
    }

    fclose(fp);

#ifdef DEBUG_HTTP
    Trace("HTTP_ParseHeaderFile: name '%s'\n\ttype '%s', lmd=%d, ex=%d\n",
        name, type, *lm, *ex);
#endif

    return 0;
}









static sigjmp_buf stop_trap_env;

static RETSIGTYPE IntrTarp()
{
    Log("; IntrTarp: SIGINT is arrive");

    SIGLONGJMP(stop_trap_env, 1);
}






#if 0
extern int new_client_fd;
#else
int new_client_fd;
#endif

void FlushCache(int *client_fd, int *head_fd, int *body_fd)
{
    off_t len, pos;
    char buf[BUF_SIZE];

    Trace("FlushCache: client_fd %d, head_fd %d, body_fd %d\n",
            *client_fd, *head_fd, *body_fd);
    Trace("FlushCache: new_client_fd %d\n", new_client_fd);

    *client_fd = new_client_fd;
    new_client_fd = -1;

#if 0
    if(*client_fd==-1)
        return;
#endif


flush_header:
    if(*head_fd==-1) {
        return;
    }
    else {
        pos = len = lseek(*head_fd, 0, SEEK_CUR);
        Trace("FlushCache: header size %d\n", len);
        if(len<0) {
            Error("FlushCache: negative length header");
            return;
        }
        if(len==0) {
            Error("FlushCache: empty header");
            return;
        }

        lseek(*head_fd, 0, SEEK_SET);
        while(len<BUF_SIZE) {
            if(read(*head_fd, buf, BUF_SIZE)<0) {
                Error("FlushCache: header: 1 reed error");
                goto flush_body;
            }
            if(write(*client_fd, buf, BUF_SIZE)<0) {
                Error("FlushCache: header: 1 write error");
                goto flush_body;
            }
            len -= BUF_SIZE;
        }
        if(len>0) {
            if(read(*head_fd, buf, len)<0) {
                Error("FlushCache: header: 2 reed error");
                goto flush_body;
            }
            if(write(*client_fd, buf, len)<0) {
                Error("FlushCache: header: 2 write error");
                goto flush_body;
            }
        }
        len = lseek(*head_fd, pos, SEEK_SET);
    }


flush_body:
    if(*body_fd==-1) {
        return;
    }
    else {
        pos = len = lseek(*body_fd, 0, SEEK_CUR);
        Trace("FlushCache: body size %d\n", len);
        if(len<0) {
            Error("FlushCache: negative length body");
            return;
        }
        if(len==0) {
            Error("FlushCache: empty body");
            return;
        }

        lseek(*body_fd, 0, SEEK_SET);
        while(len<BUF_SIZE) {
            if(read(*body_fd, buf, BUF_SIZE)<0) {
                Error("FlushCache: body: 1 reed error");
                goto out;
            }
            if(write(*client_fd, buf, BUF_SIZE)<0) {
                Error("FlushCache: body: 1 write error");
                goto out;
            }
            len -= BUF_SIZE;
        }
        if(len>0) {
            if(read(*body_fd, buf, len)<0) {
                Error("FlushCache: body: 2 reed error");
                goto out;
            }
            if(write(*client_fd, buf, len)<0) {
                Error("FlushCache: body: 2 write error");
                goto out;
            }
        }
        len = lseek(*body_fd, pos, SEEK_SET);
    }

out:
    return;
}


/*
 * HTTP_Get - GET mothod on HTTP
 *      Get data via HTTP with *raw* location. If you want to access
 *      proxy server, please specify URL to path (e.g, "http://www.ntt.jp/).
 *
 * input:
 *      name:       name (e.g., URL)
 *      keepalive:  keep-alive
 *      nocache:    no-cache
 *      noprefetch:    no-prefetch
 *      istate:     initial state
 *      servername: server name
 *      port:       service port
 *      path:       path in server
 *
 * Note:    There are a lot of data on heep. If you wanna return,
 *          please goto 'clean_up'-label.
 */

int HTTP_Get(int *client_fd, char *name, int keep_alive,
    int nocache, int noprefetch,
    int istate,
    char *servername, int port, char *path, char *auth,  Info **val)
{
    Info *info;
    struct timeval  start_tp,  conn_tp,  end_tp;
    struct timezone start_tpz, conn_tpz, end_tpz;
    long conn, trans, takes;
    char tmp_name[STRING_SIZE];
    char header_fname[STRING_SIZE], body_fname[STRING_SIZE];
    int  head_fd, body_fd, server_fd;
    int  rtime;
    off_t qsum=0;                       /* request to server */
    off_t rcount, pos, hsum, bsum;      /* read from server  */
    off_t wsum=0, nwrite=0;             /* write to client   */
    char buf[BUF_SIZE+2];
    char line[STRING_SIZE+1], header[BUF_SIZE+2];
    int  http0=0;
    int  in_head;
#if 1
    char protocol_ver[INFOVERSIZE];
#endif
    char http_ver_major, http_ver_minor;
    int  server_state = -2;
    char server_ver[STRING_SIZE];
    char tmp[INFOVERSIZE];
    char type[STRING_SIZE];
    off_t clen;
    int  reason;
    int  ret;
    int  bypass_flag;
    int  secret_flag;
    time_t lmd, ex;
    int  chk;
    char *u, *v;
    int  cline;
    char *p, *q;
    int  c;
    int  blen;



    /*
     * init resources (heap and fd)
     */
    head_fd = body_fd = server_fd = -1;
    ret = HREP_UNKNOW;
    bypass_flag = (*client_fd == -1) ? 0 : 1 ;
    if(port==-1) {
        port = HTTP_PORTNO;
    }

    lmd = -1;
    ex  = -1;

    new_client_fd = -1;



    if(auth && *auth) {
        secret_flag = 1;
    }
    else {
        secret_flag = 0;
    }

#if 1
    Trace("HTTP_Get: HTTP %c istate=%d Server=%s Port=%d Path=%s\n",
        secret_flag ? 'S' : '-', istate, servername, port, path);
#endif

    if(SIGSETJMP(stop_trap_env, 1)) {
        Log("; HTTP_Get: INT signal is arrive");
        ret = HREP_INT;

        goto clean_up;
    }
    SetSignalHandler(SIGINT, IntrTarp);


    if(SetupFile(header_fname, name, HEAD_TMP_FILE, &head_fd)) {
        goto inner_err;
    }
    if(SetupFile(body_fname, name, BODY_TMP_FILE, &body_fd)) {
        goto inner_err;
    }
#if 1
    Trace("HTTP_Get: tmpfile '%s'\n", body_fname);
#endif

    gettimeofday(&start_tp, &start_tpz);

    /* we must use sync_sync_connect_byname() */
    if((server_fd = TCP_connect(servername, port))<0) {
        Error("HTTP_Get: Cannot connect to %s ... (%d)", servername, errno);

        reason = server_fd;
        server_fd = -1;
        info = NULL;
        if(reason == WCOL_SOCK_SOCKETERROR)
            goto inner_err;
        else
            goto host_unreach;
    }

    gettimeofday(&conn_tp, &conn_tpz);

    /*
     *  check file descriptors
     */
    Trace("HTTP_Get: FD client/head/body/server=%d/%d/%d/%d: connect...\n",
        *client_fd, head_fd, body_fd, server_fd);
    if(*client_fd==head_fd || head_fd == body_fd || body_fd == server_fd
        || server_fd==*client_fd || *client_fd== body_fd || head_fd==server_fd) {
        Fatal("FD conflict %d/%d/%d/%d", *client_fd, head_fd, body_fd, server_fd);
    }



    /*
     * Send Request to Server
     */
    sprintf(buf, get_request_base_format, path, WCOL_VER);
    if(auth && *auth) {
        strcat(buf, auth);
        strcat(buf, "\r\n");
    }
    if(nocache) {
        strcat(buf, "Pragma: no-cache\r\n");
    }
    if(noprefetch) {
        strcat(buf, "Pragma: no-prefetch\r\n");
    }
    strcat(buf, "\r\n");

#ifdef TRACE_SEND_REQUEST
    Trace("HTTP_Get: Send Request [%s]\n", buf);
#endif

    pipe_broken = 0;
    if((qsum = write(server_fd, buf, sizeof(char)*strlen(buf)))
            < sizeof(char)*strlen(buf) ) {
        Error("HTTP_Get: request send %d/%d", qsum, sizeof(char)*strlen(buf));

        HTTP_ErrorMessage_Old(*client_fd,
            500, "Stream Overflow",
            "Capacity of stream for server was too small", 1);

        info = NULL;
        ret = HREP_INNERERROR;
        goto clean_up;
    }
    if(pipe_broken) {
        pipe_broken = 0;
        Error("HTTP_Get: server stream was broken");

        HTTP_ErrorMessage_Old(*client_fd,
            500, "Stream Broken",
            "Stream for server was broken. Server down or rejected", 1);

        info = NULL;
        ret = HREP_INNERERROR;
        goto clean_up;
    }





    /*
     * Loop for HEAD parsing and writing
     *
     *      - Read Header+Body (HTTP/1 and later)
     *      - Read Body (HTTP/0)
     *
     */
    memset(buf, 0, sizeof(char)*(BUF_SIZE+2)); /* init and set sentinel */

    in_head = 1;
    rtime   = 0;
    hsum    = (off_t)0;
    bsum    = (off_t)0;
    clen    = (off_t)-1;

    /*
     *  Search separator between header and body.
     *  (LF is '\r'. CR is '\n')
     *
     *  2 type response were well known.
     *      - CERN and current NCSA httpd reply was separated CR and LF.
     *      - old NCSA httpd reply was separeted LF.
     *
     *  So, header was separate CR LF CR LF or LF LF.
     *      - \n \n
     *      - \n \r \n
     *   And, skip
     *      - 2chars
     *      - 3chars
     *
     */

    /* XXX - It belive that first block is longer than 5 chars */
try0:
#if 0
    if(new_client_fd != -1) {
        FlushCache(client_fd, &head_fd, &body_fd);
        Trace("HTTP_Get: FD client/head/body/server=%d/%d/%d/%d: connect...\n",
            *client_fd, head_fd, body_fd, server_fd);
    }
#endif
    if((rcount = read(server_fd, buf, BUF_SIZE))<=0) {
        if(errno == EINTR)
            goto try0;
        Trace("HTTP_Get: error in first read() (errno %d)\n", errno);

        goto server_error;
    }

#if 0
try01:
{
    int rc;

    while(rcount<5) {
        if((rc = read(server_fd, &buf[rcount], BUF_SIZE-rcount))<=0) {
            if(errno == EINTR)
                goto try01;
            Trace("HTTP_Get: error in second read() (errno %d)\n", errno);

            goto server_error;
        }
        rcount += rc;
    }
}
#endif



    /*
     * check HTTP/1.0 or not
     *      if found 'HTTP/' in first, it is HTTP/1.0 or later version.
     *      we must read header and parse it.
     */
    if(strncmp_first_all(buf, 5, 'H', "HTTP/")) {

/***
 *** HTTP/1.0
 ***/
        Trace("HTTP_Get: HTTP/1.0 or later version's response\n");

        http0 = 0;
        in_head = 1;
        header[0] = '\0';

        /*
         * parse header
         */
        u = buf;
        while(in_head) {
            v = line;
            cline = 0;

            /* read line */
            do {
#if  0
                Trace("@u = '%c' (%d) addr %p, buf %p \n", *u, *u, u, buf);
#endif
                if(*u=='\0') {
try1:
                    if((rcount=read(server_fd, buf, BUF_SIZE))<=0) {
                        if(errno == EINTR)
                            goto try1;
                        Trace("HTTP_Get: error at read() (errno %d)\n",
                            errno);

                        goto server_error;
                    }
                    buf[rcount] = '\0';

                    u = buf;
                }

                /* append string */
                while(cline<STRING_SIZE && *u && *u!='\n') {
                    *v++ = *u++;
                    cline++;
                }
            } while (*u != '\0' && *u !='\n');

            /* append LF */
            if(cline<STRING_SIZE && *u=='\n') {
                *v++ = *u++;
                cline++;
            }
            *v = '\0';

#if 0
            strcat(header, line);
#endif

#ifdef DEBUG_HPARSE
            Trace("line %p v %p  ", line, v);
            dump_string("    ", line);
#endif

            /* line is terminated LF or CR LF */
            if(line[0]=='\n') {
#ifdef DEBUG_HPARSE
                Trace("    LF termination\n");
#endif
                in_head = 0;
            }
            if(line[0]=='\r' && line[1]=='\n') {
#ifdef DEBUG_HPARSE
                Trace("    CR LF termination\n");
#endif
                in_head = 0;
            }

            if(in_head==0) {
                strcat(header, line);
                break;
            }



            /*
             * check protocol version
             */
            if(rtime==0) {
                p = line;
                q = protocol_ver;
                c = 0;
                while(*p && *p!=' ' && c<INFOVERSIZE) {
                    *q++ = *p++;
                    c++;
                }
                *q = '\0';

                {
                    int mj;
                    int mi;

                    sscanf(protocol_ver, "HTTP/%d.%d", &mj, &mi);
                    http_ver_major = mj;
                    http_ver_minor = mi;
                }
                p = &line[strlen(protocol_ver)];

                server_state = atoi(p);
#ifdef DEBUG_HPARSE
                Trace("HTTP_Get: version %d.%d('%s') state %03d\n",
                    http_ver_major, http_ver_minor, protocol_ver, server_state);
#endif
                if(server_state>=100 && server_state<=999) {
                }
                else {
                Trace("HTTP_Get: version %d.%d('%s') ignore state %03d\n",
                    http_ver_major, http_ver_minor, protocol_ver, server_state);
                Error("HTTP_Get: version %d.%d('%s') ignore state %03d",
                    http_ver_major, http_ver_minor, protocol_ver, server_state);
                }
            }
            else {

                /*
                 * check fields
                 */
                if(*line=='S' && strncasecmp("Server: ", line, 8)==0) {
                    p = &line[8];
                    q = server_ver;
                    c = 0;
                    while(*p && c<STRING_SIZE && *p!='\r' && *p!='\n') {
                        *q++ = *p++;
                        c++;
                    }
                    *q = '\0';

#ifdef DEBUG_HPARSE
                    Trace("HeaderField: server '%s'\n", server_ver);
#endif
                }
                if(*line=='C' && strncasecmp("Content-type: ", line, 14)==0) {
                    p = &line[14];
                    q = type;
                    c = 0;
                    while(*p && c<INFOTYPESIZE && *p!='\r' && *p!='\n') {
                        *q++ = *p++;
                        c++;
                    }
                    *q = '\0';

#ifdef DEBUG_HPARSE
                    Trace("HeaderField: type '%s'\n", type);
#endif
                }
                if(*line=='C' && strncasecmp("Content-Length: ", line, 16)==0) {
                    p = &line[16];
                    q = tmp;
                    c = 0;
                    while(*p && *p!='\r' && *p!='\n') {
                        *q++ = *p++;
                        c++;
                    }
                    *q = '\0';

                    clen = (off_t)atol(tmp);
#ifdef DEBUG_HPARSE
                    Trace("HeaderField: length '%s'-> %d\n", tmp, clen);
#endif
                }
                if(*line=='L' && strncasecmp("last-modified: ", line, 15)==0) {
                    p = &line[15];
                    q = tmp;
                    while(*p&&*p!='\r'&&*p!='\n')
                       *q++ = *p++;
                    *q = '\0';
                    lmd = parse_date(tmp);

#ifdef _DEBUG_HPARSE
                    Trace("HeaderField: LM '%s' -> %d\n", tmp, lmd);
#endif
                }
                if(*line=='E' && strncasecmp("expires: ", line, 9)==0) {
                    p = &line[9];
                    q = tmp;
                    while(*p&&*p!='\r'&&*p!='\n')
                       *q++ = *p++;
                    *q = '\0';
                    ex = parse_date(tmp);

#ifdef DEBUG_HPARSE
                    Trace("HeaderField: expire '%s' -> %d\n", tmp, ex);
#endif
                }
#if 0
                if(remove_cookie_flag && *line=='S' && strncasecmp("Set-Cookie: ", line, 12)==0)
#else
                if(remove_cookie_flag &&
                    strncasecmp_first_all(line, 12, 'S', "Set-Cookie: "))
#endif
                    {

                    goto reject_and_next_field;
#ifdef DEBUG_HPARSE
                    Trace("HeaderField: Cookie '%s'\n", &line[12]);
#endif
                }
            }

correct_and_next_field:
            if(*client_fd != -1) {
#if 1
                dump_string("    ", line);
#endif
                nwrite = write(*client_fd, line, strlen(line));
                if(pipe_broken) {
                    release_socket(client_fd);
                    pipe_broken = 0;
                }
                if(nwrite!=-1)
                    wsum += nwrite;
            }

            strcat(header, line);

reject_and_next_field:

            rtime++;

        }

        /*
         * add optional fields
         */
        line[0] = '\0';
        if(clen!=-1 && keep_alive) {
            Trace("HTTP_GET: Keep-Alive\n");
            strcat(line, "Connection: Keep-Alive\r\n");
        }
        else {
            Trace("HTTP_GET: not Keep\n");
            keep_alive = 0;
        }
        strcat(line, "\r\n");
        if(*client_fd != -1) {
#if 1
            dump_string("    ", line);
#endif
            nwrite = write(*client_fd, line, strlen(line));
            if(pipe_broken) {
                release_socket(client_fd);
                pipe_broken = 0;
            }
            if(nwrite!=-1)
                wsum += nwrite;
        }

#if 0
        Trace("\nheader---\n%s---\n", header);
#endif

        /* save header */
        write(head_fd, header, strlen(header));
        hsum = strlen(header);

        ftruncate(head_fd, hsum);

#if 1
        if(new_client_fd != -1) {
            FlushCache(client_fd, &head_fd, &body_fd);
        Trace("HTTP_Get: FD client/head/body/server=%d/%d/%d/%d: connect...\n",
            *client_fd, head_fd, body_fd, server_fd);
        }
#endif

#if 0
        close(head_fd);
        head_fd = -1;
#endif
    }
    else {


/***
 *** HTTP/0.9
 ***/

        Trace("HTTP_Get: HTTP/0.9 response\n");

        http0 = 1;
        u = buf;
    }

#if 0
    Trace("begin body: buf %p, u %p, rcount %d, u-buf %d, rcount-(u-buf) %d\n",
            buf, u, rcount, (u-buf), rcount-(u-buf));
#endif

    /***
     *** BODY
     ***/


    /*
     * flush rest of buffer as body
     */
    if(rcount-(u-buf)>0) {
        bsum = rcount-(u-buf);
        if(*client_fd != -1) {
            nwrite = write(*client_fd, u, rcount-(u-buf));
            if(pipe_broken) {
                release_socket(client_fd);
                pipe_broken = 0;
            }
            if(nwrite!=-1) {
                wsum += nwrite;
            }
        }
        write(body_fd, u, rcount-(u-buf));
    }



    /* read next block */
try2:
#if 1
    if(new_client_fd != -1) {
        FlushCache(client_fd, &head_fd, &body_fd);
        Trace("HTTP_Get: FD client/head/body/server=%d/%d/%d/%d: connect...\n",
            *client_fd, head_fd, body_fd, server_fd);
    }
#endif

#if 0
    if(clen<0) {
        blen=BUF_SIZE;
    }
    else if(clen-bsum>BUF_SIZE) {
        blen=BUF_SIZE;
    }
    else {
        blen=clen-bsum;
    }
#endif
    blen=BUF_SIZE;

    while((rcount=read(server_fd, buf, blen))>0) {
        bsum += rcount;
        if(*client_fd != -1) {
            nwrite = write(*client_fd, buf, rcount);
            if(pipe_broken) {
                release_socket(client_fd);
                pipe_broken = 0;
            }
            if(nwrite!=-1) {
                wsum += nwrite;
            }
        }
        write(body_fd, buf, rcount);

#if 0
        if(clen<0) {
            blen=BUF_SIZE;
        }
        else if(clen-bsum>BUF_SIZE) {
            blen=BUF_SIZE;
        }
        else {
            blen=clen-bsum;
        }
#endif
    }
    if(rcount <0 &&  errno == EINTR)
        goto try2;


#ifdef DEBUG_HTTP
    Trace("HTTP_Get: * HEAD %6d, BODY %6d, HTTP/0.9? %d, HEAD? %d\n",
        hsum, bsum, http0, in_head);
    Trace("HTTP_Get: * Read [%2d] %6d (%4x) bytes in RAW.\n",
        rtime, rcount, rcount);
#endif

    release_socket(&server_fd);

    ftruncate(body_fd, bsum);
    close(body_fd);
    body_fd = -1;

#if 1
    close(head_fd);
    head_fd = -1;
#endif

    gettimeofday(&end_tp, &end_tpz);


    /*
     *
     */
    ret = HREP_SERVERACCESS;

#if 0
    if(*client_fd != -1) {
        release_socket(client_fd);
    }
#endif

    conn  = WCOL_SUB_TIME(conn_tp, start_tp);
    trans = WCOL_SUB_TIME(end_tp,  conn_tp);
    takes = WCOL_SUB_TIME(end_tp,  start_tp);


    Trace("HTTP_Get: [byte] HEAD  %6d, BODY   %6d, TOTAL %6d (%d)\r\n",
        (long)hsum, (long)bsum, (long)(hsum+bsum), (long)wsum);
    Trace("HTTP_Get: [ms]   CONN. %6d, TRANS. %6d, TOTAL %6d\r\n",
        conn, trans, takes);
    Trace("HTTP_Get: [kbps] ALL/TRANS %8.3f, ALL/TOTAL %8.3f\r\n",
        ((double)wsum)/((double)trans), ((double)wsum)/((double)takes));

#if 0
    if(bsum==0) {
        Error("HTTP_Get: Empty data");

        info = NULL;
        goto clean_up;
    }
#endif




    /*
     * Swap temporary file and formal file.
     */
    AssignFileName(tmp_name, name, HEAD_FILE);
    rename(header_fname, tmp_name);
    AssignFileName(tmp_name, name, BODY_FILE);
    rename(body_fname, tmp_name);



    info = NewInfo();
    if(info!=NULL) {
        time_t now;

        strcpy(info->attr.name, name);
        info->attr.state                = server_state;
        info->attr.ver_major            = http_ver_major;
        info->attr.ver_minor            = http_ver_minor;
        strcpy(info->attr.type, type);
        info->attr.length               = hsum+bsum;
        info->attr.clength              = clen;
        info->attr.lmd                  = lmd;
        info->attr.expire               = ex;
        info->attr.init_state           = istate;
        info->attr.hsize                = hsum;
        info->attr.bsize                = bsum;
        if(bypass_flag) {   /* Now just bypassing, count as 1 */
            info->attr.count            = 1;
        }
        else {
            info->attr.count            = 0;
        }

        info->attr.take_time            = takes;
        info->attr.conn_time            = conn;
        info->attr.trans_time           = trans;

        WCOL_UTIME(&now);
        info->attr.first = info->attr.last = now;

#if 0
        Trace("HTTP_Get: We get new info...\n");
        What(info);
#endif

#if 0
        SaveInfo(info, name);
#else
        SaveAddInfo(info, name);
#endif


#if 1
        if(clen != -1 && clen != bsum) {
            Error("HTTP_Get: diff %d, server said %d, read %d, %d+%d, write %d",
                (long)(clen-bsum),
                (long)clen,
                (long)bsum,
                (long)hsum, (long)bsum,
                (long)wsum);
        }
#endif

#if 1
        if(trace_flag) {
            /* HTTP/0.9 */
            if(http0) {
                Trace(
"HTTP_Get:         body\n\
        receive %8d\n\
        write   %8d\n\
       ==================\n\
        diff    %8d\n",
                (long)bsum,
                (long)wsum,
                (long)(bsum-wsum));
            }
            /* HTTP/1.0 or later */
            else {
                Trace(
"HTTP_Get:         header     body\n\
        listen         - %8d\n\
        receive %8d %8d = %8d\n\
        write                       %8d\n\
       ======================================\n\
        diff           - %8d   %8d\n",
                (long)clen,
                (long)hsum, (long)bsum, (long)(hsum+bsum),
                (long)wsum,
                (long)(clen-bsum), (long)(hsum+bsum-wsum));
            }
        }
#endif

#ifdef DEBUG_HTTP
        Trace("HTTP_Get: Check 'info' - after save ...\n");
        What(info);
        DumpInfo(info);
#endif

        if(mcache_flag) {
            LoadInfo(info);
        }
    }
    else {
        Error("Cannot built new info area. not enought memory ?");
    }



clean_up:
    SetSignalHandler(SIGINT, SIG_IGN);

    /*
     * Relese resources on heep and return 'info'.
     * But 'info' is return value, don't relese it.
     */

    if(head_fd != -1)           { close(head_fd);   head_fd = -1;   }
    if(body_fd != -1)           { close(body_fd);   body_fd = -1;   }
    if(server_fd != -1)         { release_socket(&server_fd);          }

    /*
     * When server not said length, proxy and client read until
     * connection close. So. we must close stream for client.
     */
    if(!keep_alive || clen == -1) {
        if(*client_fd != -1)    { release_socket(client_fd);        }
    }

    if(info!=NULL) {
        *val = info;

#if 0
        Trace("HTTP_Get: Check 'info' - after cleanup ...\n");
        What(info);
        DumpInfo(info);
        Trace("HTTP_Get: Check 'val' - after cleanup ...\n");
        What(*val);
        DumpInfo(*val);
#endif
    }

    if(bypass_flag && secret_flag) {
        /* */
        Log("; Discard Info (bypass %d, secret %d) '%s'",
            bypass_flag, secret_flag, name);
        /* */
        DiscardInfo(name);
    }

    Trace("HTTP_Get: return %d\n", ret);


    return ret;                         /* return success or not */


inner_err:
    if(*client_fd!=-1) {
        if(reason==WCOL_SOCK_SOCKETERROR) {
            HTTP_ErrorMessage_Old(*client_fd,
                500, "Internal Error",
                "Sorry, Cannot generate sockets", 1);
        }
        else {
            HTTP_ErrorMessage_Old(*client_fd,
                500, "Internal Error",
                "Sorry, Internal errors happen. please contact admin.", 1);
        }
    }
    info = NULL;
    ret = HREP_INNERERROR;
    goto clean_up;


host_unreach:
    {
        char msg[STRING_SIZE];

        if(*client_fd!=-1) {
            if(reason==WCOL_SOCK_UNKNOWHOST) {
                sprintf(msg, "Not found such server('%s')", servername);
                HTTP_ErrorMessage_Old(*client_fd,
                    403, "No Host", msg, 1);
            }
            else
            if(reason==WCOL_SOCK_CONNECTERROR) {
                sprintf(msg, "Server('%s') was down or busy", servername);
                HTTP_ErrorMessage_Old(*client_fd,
                    403, "Connect Error", msg, 1);
            }
            else {
                sprintf(msg, "Server('%s') is unreach", servername);
                HTTP_ErrorMessage_Old(*client_fd,
                    403, "Host Unreach", msg, 1);
            }
        }
    }
    info = NULL;
    ret = HREP_IGNOREREQUEST;
    goto clean_up;


server_error:
    info = NULL;
    ret = HREP_SERVERERROR;
    goto clean_up;


}


