/*
 * icp.h - header for ICP (Internet Cache Protocol)
 *
 * $Id: icp.h,v 1.1 1997/01/23 13:21:09 k-chinen Exp k-chinen $
 */

/*
 * NOTE: this file is modified Squid's file.
 *
 */

#ifndef WCOL_ICP_H
#define WCOL_ICP_H



#ifndef u_num32
#if SIZEOF_INT==4
#define u_num32 unsigned int
#else
#if SIZEOF_LONG==4
#define u_num32 unsigned long
#endif
#endif
#endif


typedef struct {
    unsigned char opcode;       /* opcode */
    unsigned char version;      /* version number */
    unsigned short length;      /* total length (bytes) */
    u_num32 reqnum;             /* req number (req'd for UDP) */
    u_num32 flags;
    u_num32 pad;
    u_num32 shostid;            /* sender host id */
} icp_common_t;

#define ICP_FLAG_HIT_OBJ 0x80000000

#define ICP_COMMON_SZ (sizeof(icp_common_t))
#define ICP_HDR_SZ (sizeof(icp_common_t)+sizeof(u_num32))

typedef enum {
    ICP_OP_INVALID,             /* 00 to insure 0 doesn't get accidently interpreted. */
    ICP_OP_QUERY,               /* 01 query opcode (cl->sv) */
    ICP_OP_HIT,                 /* 02 hit (cl<-sv) */
    ICP_OP_MISS,                /* 03 miss (cl<-sv) */
    ICP_OP_ERR,                 /* 04 error (cl<-sv) */
    ICP_OP_SEND,                /* 05 send object non-auth (cl->sv) */
    ICP_OP_SENDA,               /* 06 send object authoritative (cl->sv) */
    ICP_OP_DATABEG,             /* 07 first data, but not last (sv<-cl) */
    ICP_OP_DATA,                /* 08 data middle of stream (sv<-cl) */
    ICP_OP_DATAEND,             /* 09 last data (sv<-cl) */
    ICP_OP_SECHO,               /* 10 echo from source (sv<-os) */
    ICP_OP_DECHO,               /* 11 echo from dumb cache (sv<-dc) */
#ifdef USE_ICP_NOTIFY
    ICP_OP_NOTIFY,              /* 12 notify what is handled (cl->sv) */
#else
    ICP_OP_UNUSED0,             /* 12 */
#endif
    ICP_OP_UNUSED1,             /* 13 */
    ICP_OP_UNUSED2,             /* 14 */
    ICP_OP_UNUSED3,             /* 15 */
    ICP_OP_UNUSED4,             /* 16 */
    ICP_OP_UNUSED5,             /* 17 */
    ICP_OP_UNUSED6,             /* 18 */
    ICP_OP_UNUSED7,             /* 19 */
    ICP_OP_UNUSED8,             /* 20 */
    ICP_OP_RELOADING,           /* 21 access denied while reloading */
    ICP_OP_DENIED,              /* 22 access denied (cl<-sv) */
    ICP_OP_HIT_OBJ,             /* 23 hit with object data (cl<-sv) */
    ICP_OP_END                  /* 24 marks end of opcodes */
} icp_opcode;

#define ICP_OP_HIGHEST (ICP_OP_END - 1)         /* highest valid opcode */

/* Header for QUERY packet */
struct icp_query_s {
    u_num32 q_rhostid;          /* requestor host id */
    char *q_url;                /* variable sized URL data */
};
typedef struct icp_query_s icp_query_t;
#define ICP_QUERY_SZ (sizeof(icp_query_t))

/* Header for HIT packet */
struct icp_hit_s {
    u_num32 h_size;             /* size if known */
    char *h_url;                /* variable sized URL data */
};
typedef struct icp_hit_s icp_hit_t;
#define ICP_HIT_SZ (sizeof(icp_hit_t))

/* Header for MISS packet */
struct icp_miss_s {
    char *m_url;                /* variable sized URL data */
};
typedef struct icp_miss_s icp_miss_t;
#define ICP_MISS_SZ (sizeof(icp_miss_t))

/* Header for ERROR packet */
struct icp_error_s {
    unsigned short e_code;      /* error code */
    char *e_msg;                /* variable sized string message */
};
typedef struct icp_error_s icp_error_t;
#define ICP_ERROR_SZ (sizeof(icp_error_t))

#define ICP_ERROR_MSGLEN        256     /* max size for string, incl '\0' */

/* Error Codes - These can come back in the response packet */
typedef enum {
    ICP_ERROR_INVALID,          /* invalid (not used) */
    ICP_ERROR_BADVERS,          /* version error */
    ICP_ERROR_BADURL,           /* bad URL */
    ICP_ERROR_BADFLAGS,         /* bad flags */
    ICP_ERROR_TIMEDOUT,         /* couldn't get data */
    ICP_ERROR_ACCESS,           /* authorization problem */
    ICP_ERROR_INTERNAL          /* cache server internal err */
} icp_error_code;

/* Header for SEND packet */
struct icp_send_s {
    u_num32 s_rhostid;          /* requestor host id */
    char *s_url;                /* variable sized url */
};
typedef struct icp_send_s icp_send_t;
#define ICP_SEND_SZ (sizeof(icp_send_t))

/* Header for SENDA packet */
struct icp_senda_s {
    u_num32 sa_rhostid;         /* requestor host id */
    char *sa_url;               /* variable sized url */
};
typedef struct icp_senda_s icp_senda_t;
#define ICP_SENDA_SZ (sizeof(icp_senda_t))


/* Header for DATABEGIN packet */
struct icp_datab_s {
    u_num32 db_ttl;             /* time to live */
    u_num32 db_ts;              /* timestamp when gotten from owner */
    u_num32 db_size;            /* size of object if known */
    char *db_data;              /* variable sized data */
};
typedef struct icp_datab_s icp_datab_t;
#define ICP_DATAB_SZ (sizeof(icp_datab_t))


/* Header for DATA packet */
typedef struct icp_data_s {
    u_num32 d_offset;           /* offset into object for d_data  */
    char *d_data;               /* variable sized data */
} icp_data_t;
#define ICP_DATA_SZ (sizeof(icp_databe_t))


/* ICP message type. */
typedef struct icp_message_s {
    icp_common_t header;
    union {
        icp_query_t query;
        icp_hit_t hit;
        icp_miss_t miss;
        icp_error_t error;
        icp_send_t send;
        icp_senda_t senda;
        icp_datab_t data_begin;
        icp_data_t data;        /* Shared between DATA and DATAEND. */
    } op;
} icp_message_t;
#define ICP_MESSAGE_SZ (sizeof(icp_message_t))

/* Version */
#define ICP_VERSION_1           1
#define ICP_VERSION_2           2
#define ICP_VERSION_3           3
#define ICP_VERSION_CURRENT     ICP_VERSION_2



/***
 ***
 ***/

#define ICP_PORTNO              3130


extern char *_icp_opcode_str[];

#define ICPOperationStr(v) ((v>=ICP_OP_INVALID&&v<=ICP_OP_END) ? \
            _icp_opcode_str[v] : _icp_opcode_str[ICP_OP_INVALID])

#endif /* WCOL_ICP_H */

