/*
 *  WCOL: Information control procedures
 *      by k-chinen@is.aist-nara.ac.jp, 1994, 1995, 1996, 1997
 *
 * $Id: info.c,v 1.15 1996/11/26 01:03:54 k-chinen Exp k-chinen $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <signal.h>
#include <fcntl.h>
#include <limits.h>
#include <string.h>
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif
#include <math.h>
#include <errno.h>

#include "wcol.h"

#include "request.h"


#define _DEBUG_INFO


time_t cache_keeptime       = CACHEKEEPTIME;
time_t cache_ignorekeeptime = CACHEIGNOREKEEPTIME;
#if 1
int mcache_flag;
#endif






void ClearInfo(Info *info)
{
    memset(info,0,sizeof(Info));

    /* set default values */
    strcpy(info->attr.id,WCOL_INFO_ID);
    info->attr.name[0]      = '\0';
    info->attr.state        = -1;
    info->attr.ver_major    = 0;
    info->attr.ver_minor    = 9;
    info->attr.type[0]      = '\0';
    info->attr.length       = -1;
    info->attr.clength      = -1;
    info->attr.lmd          = -1;
    info->attr.expire       = -1;
    info->attr.init_state   = -1;
    info->attr.hsize        = -1;
    info->attr.bsize        = -1;
    info->attr.count        = -1;
    info->attr.first        = -1;
    info->attr.last         = -1;
    info->attr.take_time    = -1;
    info->attr.conn_time    = -1;
    info->attr.trans_time   = -1;
    info->attr.phead        = NULL;
    info->attr.pbody        = NULL;

}

/* constractor */
Info *NewInfo()
{
    Info *info;

    info = (Info*)MALLOC(sizeof(Info));
    if(info==NULL) {
        return NULL;
    }

    ClearInfo(info);

    return info;
}


/* */
Info *DeleteInfo(Info *v)
{
    if(v==NULL)
        return NULL;

    if(v->attr.phead)
        FREE(v->attr.phead);
    if(v->attr.pbody)
        FREE(v->attr.pbody);

    return NULL;
}

/*
 *
 */
int What(Info *info)
{
    if(info==NULL) {
        fprintf(stdout,"What: NULL, no data !!!\n");
        return 1;
    }

    fprintf(stdout,"--- Info ---\n");
    fprintf(stdout,"Identifier:     %s\n",info->attr.id);
    fprintf(stdout,"Name:           %s\n",info->attr.name);
    fprintf(stdout,"State:          %d\n",info->attr.state);
    fprintf(stdout,"Version:        %d.%d\n",
        info->attr.ver_major, info->attr.ver_minor);
    fprintf(stdout,"Type:           %s\n",info->attr.type);
    fprintf(stdout,"Length:         %9d\n",info->attr.length);
    fprintf(stdout,"CLength:        %9d\n",info->attr.clength);
    fprintf(stdout,"Last Modified:  %9d ; %s",
        info->attr.lmd, info->attr.lmd<0 ? "\n":ctime(&info->attr.lmd));
    fprintf(stdout,"Expires:        %9d ; %s",
        info->attr.expire,info->attr.expire<0 ? "\n":ctime(&info->attr.expire));
    fprintf(stdout,"Count:          %9d\n",info->attr.count);
    fprintf(stdout,"Header Size:    %9d\n",info->attr.hsize);
    fprintf(stdout,"Body Size:      %9d\n",info->attr.bsize);
    fprintf(stdout,"Connect Time:   %9d\n",info->attr.conn_time);
    fprintf(stdout,"Translate Time: %9d\n",info->attr.trans_time);
    fprintf(stdout,"Take Time:      %9d\n",info->attr.take_time);
    fprintf(stdout,"First Date:     %9d ; %s",
        info->attr.first, info->attr.first<0 ? "\n":ctime(&info->attr.first));
    fprintf(stdout,"Last Date:      %9d ; %s",
        info->attr.last, info->attr.last<0 ? "\n":ctime(&info->attr.last));
    fprintf(stdout,"Init State:     %d\n", info->attr.init_state);

#if 0
    switch(info->attr.init_state) {
    case 1:
        fprintf(stdout,"Init State:     Requested\n");
        break;
    case 2:
        fprintf(stdout,"Init State:     Prefetched\n");
        break;
    default:
        fprintf(stdout,"Init State:     Unknow [%d]\n",info->attr.init_state);
        break;
    }
#endif

    fprintf(stdout,"Header Address: %p\n", info->attr.phead);
    fprintf(stdout,"Body Address:   %p\n", info->attr.pbody);
    fflush(stdout);

    return 0;
}

static char *_hex_chars="0123456789ABCDEF";

int DumpInfo(Info *info)
{
    int i,x;
    char *p;
    char line[STRING_SIZE];
    int  chksum;

    if(info==NULL) {
        fprintf(stdout,"DumpInfo: NULL, no data !!!\n");
        return 1;
    }


    p = (char*)info;
    memset(line, ' ', STRING_SIZE);
    line[78] = '\n';
    line[79] = '\0';

    chksum = 0;
    for(i=0;i<sizeof(Info);i++, p++) {
        chksum += *p;
    }
    p = (char*)info;

    fprintf(stdout, "--- Info (dump) --- base address %p, checksum %d\n",
        p, chksum);
    fflush(stdout);

    for(i=0;i<sizeof(Info);i++, p++) {
        x = (i%16)*3;
        line[x]   = _hex_chars[*p/16];
        line[x+1] = _hex_chars[*p%16];

        x = 50+i%16;
        if(*p>=' ' && *p<0x7f) {
            line[x] = *p;
        }
        else {
            line[x] = '.';
        }

        if(i%16==15) {
            fputs(line, stdout);
            fflush(stdout);
        }
    }
    fflush(stdout);

    return 0;
}



void LoadInfo(Info *info)
{
    int fd;
    int c;
    char *hcache, *bcache;
    char path[STRING_SIZE];

loadhead:
    info->attr.phead = (void*)NULL;

    /* less than MCACHE_THRESHOLD (10k ?) */
    if(info->attr.hsize>0 && info->attr.hsize<MCACHE_THRESHOLD) {
        if(AssignFileName(path, info->attr.name, HEAD_FILE)) {
            goto loadhead_fail;
        }
        if((fd=open(path, O_RDONLY))<0) {
            Error("LoadInfo: head, Cannot open file '%s' (%d).", path, errno);
            goto loadhead_fail;
        }

        hcache = (char*)MALLOC(info->attr.hsize);
        c = read(fd, hcache, info->attr.hsize);
    }

    info->attr.phead = (void*)hcache;
    goto loadbody;

loadhead_fail:
    FREE(hcache);
    goto loadbody;


loadbody:
    info->attr.pbody = (void*)NULL;

    /* less than 1000k */
    if(info->attr.bsize>0 && info->attr.bsize<1000*1024) {
        if(AssignFileName(path, info->attr.name, BODY_FILE)) {
            goto loadbody_fail;
        }
        if((fd=open(path, O_RDONLY))<0) {
            Error("LoadInfo: body, Cannot open file '%s' (%d) .", path, errno);
            goto loadbody_fail;
        }

        bcache = (char*)MALLOC(info->attr.bsize);
        c = read(fd, bcache, info->attr.bsize);
    }

    info->attr.pbody = (void*)bcache;
    goto out;

loadbody_fail:
    FREE(bcache);
    goto out;


out:
    Trace("LoadInfo: phead (%d) %#x, pbody (%d) %#x\n",
        info->attr.hsize, info->attr.phead, info->attr.bsize, info->attr.pbody);
}



/*
 * ReadInfoRaw - read Info directry.
 */
int ReadInfoRaw(char *path, Info *info)
{
    int  fd;
    int  c;

#ifdef DEBUG_INFO
    /*
    Trace("ReadInfoRaw: path='%s'\n", path);
    */
#endif


    if((fd=open(path,O_RDONLY))<0) {
#ifdef DEBUG_INFO
        Error("ReadInfoRaw: cannot open '%s' (%d)", path, errno);
#endif
        return -1;
    }

    if((c = read(fd, info, sizeof(Info)))<=0) {
        Error("ReadInfoRaw: empty? read error '%s' (%d)", path, errno);
        close(fd);
        return -1;
    }
    close(fd);

    if(c!=sizeof(Info)) {
        Error("ReadInfoRaw: strange size '%s'", path);
        return -1;
    }

    return 0;
}


/*
 * CheckInfoAttr - Check attributes
 *
 * Expire information
 *      1) these are specifieded expire date.
 *      2) these are older than keep time.
 *      3) these are older than current access.
 *
 *
 *               <----- dist ---->
 *      lmd     last            now
 *       *-------*- - - -x- - - -*-
 *       <- gap ->     last
 *                      +gap
 *
 *              when dist > gap , information is will be fresh.
 *
 * inputs:
 *      info:       target
 *      ref_time:   reference time (it is mean 'current time')
 *
 * return:
 *      0:  safe
 *      1:  ignore
 *     -1:  gray
 */
CheckInfoAttr(Info info, time_t ref_time)
{
    time_t gap;
    time_t dist;

    /* ignore ID */
    if(strncmp(info.attr.id, WCOL_INFO_ID, strlen(WCOL_INFO_ID)) != 0) {
        return 1;
    }
    /* strage state */
    if(info.attr.state <= 0) {
        return 1;
    }

    /* old negative cache */
    if((info.attr.state<100 || info.attr.state>=400) &&
        ref_time - info.attr.last < cache_ignorekeeptime) {

#ifdef DEBUG_INFO
        Trace("; old negative cache (state %03d, diff %7d, last %d, current=%d) '%s'\n",
            info.attr.state,
            ref_time - info.attr.last, info.attr.last, ref_time,
            info.attr.name);
#endif

        return 1;
    }
    /* expired  */
    if(info.attr.expire != -1 && info.attr.expire < ref_time) {
#ifdef DEBUG_INFO
        Trace("; expired cache (last %d, current %d) '%s'\n",
                info.attr.last, ref_time, info.attr.name);
#endif
        return 1;
    }
    /* very old */
    if(ref_time - info.attr.last > cache_keeptime) {
#ifdef DEBUG_INFO
        Trace("; very old cache (last %d, current %d) '%s'\n",
                info.attr.last, ref_time, info.attr.name);
#endif
        return 1;
    }
    /* old meybe */
    if(info.attr.lmd>0) {
        gap  = info.attr.last - info.attr.lmd;
        dist = ref_time - info.attr.last;

        if(gap>0 && dist>0 && dist > gap) {
#ifdef DEBUG_INFO
            Trace("; aging %5.2f (gap %9d, dist %9d) '%s'\n",
                ((double)dist)/gap, gap, dist, info.attr.name);
#endif

            return -1;
        }
    }

    return 0;
}



/*
 * Have - check specified information is exist and is it fine.
 *
 * return:
 *      NULL    error (not found or wrong)
 *      !NULL   have it
 */
Info *Have(char *name)
{
    Info tmp,*ret;
    char path[STRING_SIZE];
    time_t now;
    int success=0;
    struct   stat sbuf;
    char store_name[STRING_SIZE];
    int ereason;
    int correct;

#ifdef DEBUG_INFO
    Trace("Have: check '%s'\n", name);
#endif

#ifdef USE_NOTE
  {
    int chk;

    if(chk=IsURL_NOTEorBODY(name, store_name)) {
        name = store_name;
        Trace("\ttruncate name to '%s'\n", name);
    }
  }
#endif



    ret = (Info*)NULL;

#ifdef USE_LOCK
    if(CheckLock(name)) {
#ifdef DEBUG_INFO
        Trace("Have: locked '%s'\n", name);
#endif
        return ret;
    }
#endif


    /*
     * check 'have body ?'
     */
s1:
    AssignFileName(path, name, BODY_FILE);
    if(stat(path, &sbuf)) {
        ereason = errno;
        if(ereason==EINTR)
            goto s1;
#ifdef DEBUG_INFO
        if(ereason==ENOENT) {
            Trace("Have: not found body-file '%s'.\n",path);
        }
        else {
            Error("Have: fail stat() (%d) '%s'", ereason, path);
            /*
            Trace("Have: unknow error '%s'\n",path);
            */
        }
#endif

        goto fail;
    }

    /* empty */
    if(sbuf.st_size==(off_t) 0) {
#ifdef DEBUG_INFO
        Trace("Have: found body-file '%s'. but it is empty\n", path);
#endif

        goto fail;
    }

    if(S_ISDIR(sbuf.st_mode)) {
#ifdef DEBUG_INFO
        Trace("Have: found body-file '%s'. but it is directory\n", path);
#endif

        ClearDir(path);

        goto fail;
    }
    else
    if(!S_ISREG(sbuf.st_mode)) {
#ifdef DEBUG_INFO
        Trace("Have: found body-file '%s'. but it is not regular file\n", path);
#endif
        goto fail;
    }


    /*
     * check 'have info ?'
     */
    AssignFileName(path, name, INFO_FILE);
    if(ReadInfoRaw(path, &tmp)) {
#ifdef DEBUG_INFO
        Trace("Have: not found or ignore info-file '%s'.\n", path);
#endif
        goto fail;
    }

    /* clear memory cache attibutes */
    tmp.attr.phead = NULL;
    tmp.attr.pbody = NULL;


    time(&now);
    if((correct = CheckInfoAttr(tmp, now))==0) {
        ret = (Info*)MALLOC(sizeof(Info));
        memcpy((void*)ret,(void*)&tmp,sizeof(Info));

#ifdef DEBUG_INFO
        What(ret);
#endif
    }
    else {
#ifdef DEBUG_INFO
        Trace("Have: info is exist. check attributes -> %d\n", correct);

        ret = (Info*)MALLOC(sizeof(Info));
        memcpy((void*)ret,(void*)&tmp,sizeof(Info));

        What(ret);
        FREE(ret);
        ret = (void*)NULL;
#endif
        goto fail;
    }


safe:
    /* for memory cache */
    if(mcache_flag) {
        LoadInfo(ret);
    }


#ifdef DEBUG_INFO
    Trace("Have: exist '%s' ? %#x\n", name, ret);
#endif

    return ret;

fail:

#ifdef DEBUG_INFO
    Trace("Have: not exist '%s'\n", name);
#endif

    return (Info*)NULL;
}





/*
 * SaveInfo - save info.  if it is exist already, overwrite.
 *
 * return:
 *      1   error
 *      0   success
 */
int SaveInfo(Info *info, char *name)
{
    int  info_fd, count;
    char info_path[STRING_SIZE];


    if(AssignFileName(info_path, name, INFO_FILE)) {
        return 1;
    }
    if(SetupDir(info_path)) {
        Error("SaveInfo: Fail setup INFO \"%s\".",info_path);
        FREE(info_path);
        return 1;
    }

    if((info_fd=open(info_path,O_WRONLY|O_CREAT,0644))<0) {
        Error("SaveInfo: Cannot open INFO \"%s\" (%d).",info_path, errno);

        return 1;
    }

    if((count=write(info_fd,info,sizeof(Info)))<sizeof(Info)) {
        Error("SaveInfo: Cannot write INFO.(%d/%d) (%d)",
            count, sizeof(Info), errno);

        return 1;
    }

    ftruncate(info_fd, sizeof(Info));
    close(info_fd);

    return 0;
}



/*
 * SaveAddInfo - save info. if it is exist already, read old one and
 *      set generate date to old date.
 *
 * return:
 *      1   error
 *      0   success
 */
int SaveAddInfo(Info *info, char *name)
{
    int  info_fd, count;
    char info_path[STRING_SIZE];
    Info old_info;
    time_t generate_date;


    if(AssignFileName(info_path, name, INFO_FILE)) {
        return 1;
    }
    if(SetupDir(info_path)) {
        Error("SaveAddInfo: Fail setup INFO \"%s\".",info_path);
        FREE(info_path);
        return 1;
    }

    if(ReadInfoRaw(info_path, &old_info)) {
        generate_date = -1;
    }
    else {
        generate_date = old_info.attr.first;
    }

    if((info_fd=open(info_path,O_WRONLY|O_CREAT,0644))<0) {
        Error("SaveAddInfo: Cannot open INFO \"%s\" (%d).", info_path, errno);

        return 1;
    }

    if(generate_date>0) {
        info->attr.first = generate_date;
    }

    if((count=write(info_fd,info,sizeof(Info)))<sizeof(Info)) {
        Error("SaveAddInfo: Cannot write INFO.(%d/%d) (%d)",
            count, sizeof(Info), errno);

        return 1;
    }

    ftruncate(info_fd, sizeof(Info));
    close(info_fd);

    return 0;
}




/*
 *
 * return   1: done (success)
 *          0: error
 */
int UpdateInfo(char *name)
{
    Info tmp;
    char path[STRING_SIZE];
    time_t now;
    int fd;
    int c;
    int done=0;

#ifdef USE_LOCK
    Lock(name);
#endif

    if(AssignFileName(path, name,INFO_FILE)) {
        return 0;
    }
    if((fd=open(path,O_RDWR))<0) {
        Error("UpdateInfo: Cannot open file '%s' (%d).", path, errno);
    }
    else {
#if 0
        time(&now);
#endif
        WCOL_UTIME(&now);
        if((c=read(fd, &tmp, sizeof(Info)))) {
            if(strncmp(tmp.attr.id,WCOL_INFO_ID,strlen(WCOL_INFO_ID))!=0) {
                Error("UpdateInfo: Ignore ID URL='%s' (%d).", path, errno);
            }
            else {
                tmp.attr.count ++;

#if 0
                if(tmp.attr.init_state == REQ_PREFETCH) {
                    tmp.attr.init_state = REQ_CLIENT;
                }
#endif

/*
                Trace("* UpdateInfo: name='%s' count=%d.\n",
                    tmp.attr.name,tmp.attr.count);
*/
                tmp.attr.last = now;

                lseek(fd, 0L, SEEK_SET);
                if((c = write(fd, &tmp, sizeof(Info)))==sizeof(Info)) {
#ifdef DEBUG_INFO
                    Trace("UpdateInfo: inc. to %d, '%s' .\n",
                        tmp.attr.count, name);
#endif
                    done = 1;
                }
                else {
                    Error("UpdateInfo: Cannot write to '%s' (size=%d) (%d).",
                        path, c, errno);
                }
            }
        }
        else {
            Error("UpdateInfo: Empty URL='%s'.",path);
        }
        close(fd);
    }

#ifdef USE_LOCK
    UnLock(name);
#endif

    return done;
}


/*
 *
 */
int DiscardInfo(char *name)
{
    char path[STRING_SIZE];

#ifdef USE_LOCK
    Lock(name);
#endif

    if(!AssignFileName(path, name, INFO_FILE))      unlink(path);
#if 1
    if(!AssignFileName(path, name, HEAD_FILE))      unlink(path);
    if(!AssignFileName(path, name, BODY_FILE))      unlink(path);
    if(!AssignFileName(path, name, REQUEST_FILE))   unlink(path);
    if(!AssignFileName(path, name, HIST_FILE))      unlink(path);
#endif


#ifdef USE_LOCK
    UnLock(name);
#endif


#ifdef DEBUG_INFO
    Trace("DiscardInfo: name=%s.\n", name);
#endif

    return 0;
}



/*
 * information evaluation function for cache
 *
 * input:
 *      info;       target information
 *      ref_time:   reference time
 *
 * return:
 *      evaulation value
 */

int EvalVal(Info *info, time_t ref_time)
{
    int  val;
    time_t diff_time;

    if((diff_time = ref_time - info->attr.last)<0) {
        diff_time = 0;
    }

    val = EVAL_OFFSET + info->attr.count * 10
        - (int)log((double)(info->attr.hsize+info->attr.bsize)+sizeof(Info)+1.0)
        - (int)log((double)diff_time+1.0);

#ifdef REAPER_DEBUG
    Trace("@[ %5d + %3d * 10 - %2d - %2d -> %5d ]",
        EVAL_OFFSET,
        info->attr.count,
        (int)log((double)(info->attr.hsize+info->attr.bsize)+sizeof(Info)+1.0),
        (int)log((double)diff_time+1.0),
        val);
#endif

    return val;
}

int CheckEvalVal(Info *info, time_t ref_time)
{
    int  val;
    time_t diff_time;

    if((diff_time = ref_time - info->attr.last)<0) {
        diff_time = 0;
    }

    fprintf(stderr, "\noffset %10d\n",  EVAL_OFFSET);
    fprintf(stderr, "count  %10d\n",    info->attr.count);
    fprintf(stderr, "size   %10.0f: %10d + %10d + %10d + 1.0\n",
        (double)(info->attr.hsize+info->attr.bsize)+sizeof(Info)+1.0,
            (int)info->attr.hsize, (int)info->attr.bsize, sizeof(Info));
    fprintf(stderr, "time   %10f: [%10d - %10d] %10d + 1.0\n",
        (double)diff_time+1.0,
            (int)ref_time, (int)info->attr.last, (int)diff_time);

    val = EvalVal(info, ref_time);

    Trace("@[ %5d + %3d * 10 - %2d - %2d -> %5d ]",
        EVAL_OFFSET,
        info->attr.count,
        (int)log((double)(info->attr.hsize+info->attr.bsize)+sizeof(Info)+1.0),
        (int)log((double)diff_time+1.0),
        val);

    return val;
}



