/*
 *  WCOL: Inter Process Communication by sendmsg() of BSD
 *      by k-chinen@is.aist-nara.ac.jp, 1995, 1996, 1997
 *
 *  $Id: ipc.c,v 1.3 1996/07/26 18:40:16 k-chinen Exp k-chinen $
 */
/*
 * NOTE:    If you use buggy sendmsg() and your platform have STREAM
 *      interfaces, set BUGGY_SENDMSG to avoid troubles in buggy sendmsg().
 *      When BUGGY_SENDMSG is setted, use putmsg() for file descriptor
 *      (clients port) and record (information about job or reports)
 *      passing.
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

/*
 * OSF/1 on AXP have 4.3BSD like functions and 4.4BSD like functions.
 * We need 4.4BSD like functions. Then we select that functions.
 */
#if defined(__osf__) && defined(__alpha)
#ifndef _SOCKADDR_LEN
#define _SOCKADDR_LEN
#endif
#ifndef HAVE_MSG_CONTROL
#define HAVE_MSG_CONTROL
#endif
#endif /* __osf__ && __alpha */

/*
 * IRIX 6.4 definded both 4.3BSD like functions and 4.4BSD like functions.
 * But 4.4BSD like functions are not used.
 * Then we select 4.3BSD like functions.
 */
#if defined(__sgi)
#ifdef HAVE_MSG_CONTROL
#undef HAVE_MSG_CONTROL
#endif
#endif /* __sgi */



#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/socket.h>
#include <signal.h>
#include <setjmp.h>
#include <fcntl.h>
#include <limits.h>

#ifdef linux
#include <sys/un.h>
#define CMSG_DATA(x) ((x)->cmsg_data)
#endif

#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif


#include <errno.h>

#include "wcol.h"
#include "msg.h"
#include "sigs.h"


#define _DEBUG_IPC
#define _TRACE_IPC

#if 0
#undef MALLOC
#define MALLOC(x)   alloca(x)
#endif


#ifdef TRACE_IPC


/*
 *
 */
char *data_sign(char *data, int len)
{
    static char buf[32+1];
    int c;
    unsigned long k;

    k = 0L;
    for(c=0;c<len;c++) {
        k += *(data++);
    }
    sprintf(buf, "%010.10lu %05d", k, len);

    return buf;
}

#endif /* TRACE_IPC */


/*
 * There are many buggy implementation of sendmsg().
 * Then this program support fd passing by STREAM of SYSV.
 */
#ifdef BUGGY_SENDMSG



#ifdef HAVE_PUTMSG
#include <stropts.h>
#endif


int SendFRecord(int port, int fd, void *rec, size_t len)
{
    struct strbuf cnt, dat;
    int fdbuf;

    cnt.maxlen  = sizeof(int);
    if(fd<0) {
        cnt.len = 0;
    }
    else {
        cnt.len = sizeof(int);
    }
    cnt.buf     = (char*)&fdbuf;
    fdbuf = fd;

    dat.maxlen  = len;
    dat.len     = len;
    dat.buf     = (char*)rec;

    if( putmsg(port, &cnt, &dat, 0) < 0 ) {
        Error("SendFRecord: putmsg: errno %d", errno);
        return -1;
    }

    if(fd >= 0 ) {
        if(ioctl(port, I_SENDFD, fd) < 0 ) {
            Error("SendFRecord: ioctl: errno %d", errno);
            return -1;
        }
    }

#ifdef TRACE_IPC
    Log("; send putmsg() port# %4d, %d/%db, fd# %4d [%s]",
            port, cnt.len, cnt.maxlen, fd, data_sign(dat.buf, dat.len));
#endif

    return dat.len;
}

int RecvFRecord(int port, int *fd, void *rec, size_t len)
{
    struct strbuf cnt, dat;
    struct strrecvfd fdbuf;
    int flag;

    cnt.maxlen  = sizeof(int);
    cnt.buf     = (char*)&fdbuf;

    dat.maxlen  = len;
    dat.buf     = (char*)rec;

#ifdef DEBUG_IPC
    if(isastream(port)) {
        Trace("RecvFRecord: port(%d) is stream\n", port);
    }
    else {
        Trace("RecvFRecord: port(%d) is NOT stream\n", port);
    }
#endif

try0:
    flag = 0;
    if( getmsg(port, &cnt, &dat, &flag) < 0 ) {
        if(errno == EINTR)
            goto try0;
        Error("RecvFRecord: getmsg: fd# %d, errno %d", port, errno);
        return -1;
    }

    if(cnt.len>0) {
        if(ioctl(port, I_RECVFD, &fdbuf) < 0 ) {
            Error("RecvFRecord: ioctl: errno %d", errno);
            return -1;
        }
        *fd = fdbuf.fd;
#ifdef TRACE_IPC
        Trace("RecvFRecord: get fd %d from remort fd %d\n",
                *fd, *((int*)cnt.buf));
#endif
    }
    else {
        *fd = -1;
    }


#ifdef TRACE_IPC
    Log("; recv getmsg() port# %4d, %d/%db, fd# %4d [%s]",
            port, cnt.len, cnt.maxlen, *fd, data_sign(dat.buf, dat.len));
#endif

    return dat.len;
}



int mkpipe(int spair[2])
{
    int r;

    if(r = pipe(spair)) {
        Error("mkpipe: pipe() (error %d)", errno);
        return -1;
    }

    return r;
}


#else /* BUGGY_SENDMSG */



#ifdef HAVE_MSG_CONTROL
#define CLEN    (sizeof(struct cmsghdr)+sizeof(int))
#endif


/*
 * SendRecord - Send Record (a.k.a. structure) with file descriptor
 *
 *  args:
 *      fd      file descriptor for send
 *      rec     record for send
 *      len     length of 'rec'
 *
 *  return:
 *      >0      success
 *      0       end
 *      -1      error (fail sendmsg(), or no memory)
 */
int SendFRecord(int port, int fd, void *rec, size_t len)
{
    int count;
    struct iovec    ivec[1];
    struct msghdr   msg;
#ifdef HAVE_MSG_CONTROL
    struct cmsghdr  *cmsg;
#endif

#ifdef DEBUG_IPC
    Trace("SendFRecord: sendmsg: port# %4d, fd# %4d, len %6d\n", port, fd, len);
#endif

    ivec[0].iov_base    = (char*) rec;
    ivec[0].iov_len     = len;
    msg.msg_iov         = ivec;
    msg.msg_iovlen      = 1;
    msg.msg_name        = NULL;
    msg.msg_namelen     = 0;
#ifdef HAVE_MSG_FLAGS
    msg.msg_flags       = 0;
#endif

#ifdef HAVE_MSG_CONTROL
    if(fd<0) {
        msg.msg_controllen  = 0;
        msg.msg_control     = NULL;
    }
    else {
        if((cmsg = (struct cmsghdr*)MALLOC(CLEN))
            ==NULL) {
            Error("SendFRecord: sendmsg: not enough memory");
            return -1;
        }
        msg.msg_controllen  = CLEN;
        msg.msg_control     = (char*) cmsg;

        cmsg->cmsg_level    = SOL_SOCKET;
        cmsg->cmsg_type     = SCM_RIGHTS;
        cmsg->cmsg_len      = CLEN;
        *(int*)CMSG_DATA(cmsg) = fd;
    }

#else /* HAVE_MSG_CONTROL */
    if(fd<0) {
        msg.msg_accrightslen= 0;
        msg.msg_accrights   = NULL;
    }
    else {
        msg.msg_accrightslen= sizeof(int);
        msg.msg_accrights   = (char*) &fd;
    }
#endif

#ifdef DEBUG_IPC
  {
    Trace("\t;msg_name      %p, %d\n", msg.msg_name,      msg.msg_namelen);
    Trace("\t;msg_iov       %p, %d\n", msg.msg_iov,       msg.msg_iovlen);
#ifdef HAVE_MSG_CONTROL
    Trace("\t;msg_control   %p, %d\n", msg.msg_control,   msg.msg_controllen);
#else
    Trace("\t;msg_accrights %p, %d\n", msg.msg_accrights, msg.msg_accrightslen);
#endif
  }
#endif

try:
    errno = -1;
    if((count=sendmsg(port, &msg, 0)) < 0) {
        int terr = errno;
        if(terr==EINTR) {
            goto try;
        }

        if(terr==EAGAIN) {
            Log("; SendFRecord: sendmsg: EAGAIN (errno %d)", errno);
        }
        else
        if(terr==EPIPE) {
            Log("; SendFRecord: sendmsg: EPIPE (errno %d)", errno);
        }
        else {
            Fatal("SendFRecord: sendmsg: unknown error (errno %d)", errno);
            return -1;
        }
    }

#ifdef HAVE_MSG_CONTROL
    if(fd<0) {
    }
    else {
        FREE(cmsg);
    }
#endif

#ifdef TRACE_IPC
    Log("; send sendmsg() port# %4d, fd# %4d [%s] ",
        port, fd, data_sign(rec, len));
#endif

    return count;
}



/*
 * ReceiveRecord - Receive Record with file descriptor
 *
 *  return
 *      >0  success
 *      -1  error
 */
int RecvFRecord(int port, int *fd, void *rec, size_t len)
{
    int count;
    struct iovec    ivec[1];
    struct msghdr   msg;
#ifdef HAVE_MSG_CONTROL
    struct cmsghdr  *cmsg;
#endif

#ifdef DEBUG_IPC
    Trace("RecvFRecord: recvmsg: waitting at port=%d ...\n", port);
#endif

    ivec[0].iov_base    = (char*)rec;
    ivec[0].iov_len     = len;
    msg.msg_iov         = ivec;
    msg.msg_iovlen      = 1;
    msg.msg_name        = NULL;
    msg.msg_namelen     = 0;
#ifdef HAVE_MSG_FLAGS
    msg.msg_flags       = 0;
#endif

#ifdef HAVE_MSG_CONTROL
    if((cmsg = (struct cmsghdr*)MALLOC(CLEN))
        ==NULL) {
        Error("RecvFRecord: recvmsg: not enough memory");
        return -1;
    }
#if 1
    *(int*)CMSG_DATA(cmsg) = -1;
#endif
    msg.msg_controllen  = CLEN;
    msg.msg_control     = (char*) cmsg;
#else
    msg.msg_accrightslen= sizeof(int);
    msg.msg_accrights   = (char*) fd;
#endif

try0:
    if((count = recvmsg(port, &msg, 0)) < 0) {
        if(errno == EINTR)
            goto try0;
        /*
        Log("; RecvFRecord: recvmsg: errno=%d", errno);
        */
        Fatal("RecvFRecord: recvmsg: fd# %d, errno=%d", port, errno);
        return -1;
    }


#ifdef HAVE_MSG_CONTROL
    if(msg.msg_controllen==0) {
        *fd = -1;
    }
    else {
        *fd = *(int*)CMSG_DATA(cmsg);
    }
#else
    if(msg.msg_accrightslen==0) {
        *fd = -1;
    }
    else {
        /*
        *fd = msg.msg_accrights
        */
    }
#endif

#ifdef DEBUG_IPC
    Trace("RecvFRecord: recvmsg: port# %4d, fd# %4d, len %6d/%6d\n", port, *fd, count, len);
#endif
#ifdef DEBUG_IPC
  {
    Trace("\t;msg_name      %p, %d\n", msg.msg_name,      msg.msg_namelen);
    Trace("\t;msg_iov       %p, %d\n", msg.msg_iov,       msg.msg_iovlen);
#ifdef HAVE_MSG_CONTROL
    Trace("\t;msg_control   %p, %d\n", msg.msg_control,   msg.msg_controllen);
#else
    Trace("\t;msg_accrights %p, %d (%d) -> fd %d\n",
        msg.msg_accrights, msg.msg_accrightslen,
        *((int*)msg.msg_accrights), *fd);
#endif
  }
#endif


#ifdef HAVE_MSG_CONTROL
    FREE(cmsg);
#endif

#ifdef TRACE_IPC
    Log("; recv recvmsg() port# %4d, fd# %4d [%s]",
        port, *fd, data_sign(rec, len));
#endif

    return count;
}



int mkpipe(int spair[2])
{
    int r;

    if((r = socketpair(AF_UNIX, SOCK_STREAM, 0, spair)) == -1) {
        Error("mkpipe: socketpair() (error %d)", errno);
        return -1;
    }

    return r;
}


#endif /* BUGGY_SENDMSG */





int SendRecord(int port, void *rec, size_t len)
{
    return SendFRecord(port, -1, rec, len);
}

int RecvRecord(int port, void *rec, size_t len)
{
    int dfd;    /* dummy */
    return RecvFRecord(port, &dfd, rec, len);
}



/***
 *** Ntry version
 ***/

sigjmp_buf ipc_jmp_point;
volatile sig_atomic_t ipc_fail_count;

static RETSIGTYPE giveup_ipc()
{
    ipc_fail_count++;
    SIGLONGJMP(ipc_jmp_point, 1);
}




int NSendFRecord(int n, int port, int fd, void *rec, size_t len)
{
    int ret;

    ipc_fail_count = 0;
    if(SIGSETJMP(ipc_jmp_point, 1)) {
        if(ipc_fail_count<n) {
            goto try;
        }

        return -1;
    }

try:
    SetSignalHandler(SIGALRM, giveup_ipc);
    alarm(1);
    ret = SendFRecord(port, -1, rec, len);
    SetSignalHandler(SIGALRM, SIG_IGN);

    return ret;
}



int NRecvFRecord(int n, int port, int *fd, void *rec, size_t len)
{
    int ret;

    ipc_fail_count = 0;
    if(SIGSETJMP(ipc_jmp_point, 1)) {
        if(ipc_fail_count<n) {
            goto try;
        }

        return -1;
    }

try:
    SetSignalHandler(SIGALRM, giveup_ipc);
    alarm(1);
    ret =  RecvFRecord(port, fd, rec, len);
    SetSignalHandler(SIGALRM, SIG_IGN);

    return ret;
}


int NSendRecord(int n, int port, void *rec, size_t len)
{
    return NSendFRecord(n, port, -1, rec, len);
}

int NRecvRecord(int n, int port, void *rec, size_t len)
{
    int dfd;    /* dummy */
    return NRecvFRecord(n, port, &dfd, rec, len);
}



int XSendFRecord(int x, int port, int fd, void *rec, size_t len)
{
    int ret;

    ipc_fail_count = 0;
    if(SIGSETJMP(ipc_jmp_point, 1)) {
#ifdef TRACE_IPC
        Log("; XSendFRecord timeout %d sec  to fd# %3d of pid %5d with fd# %3d",
            x, port, wcol_pid, fd);
#endif

        return -1;
    }

try:
    SetSignalHandler(SIGALRM, giveup_ipc);
    alarm(x);
    ret = SendFRecord(port, fd, rec, len);
    SetSignalHandler(SIGALRM, SIG_IGN);

#ifdef TRACE_IPC
        Log("; XSendFRecord done    %d byte to fd# %3d of pid %5d with fd# %3d",
            ret, port, wcol_pid, fd);
#endif

    return ret;
}

int XSendRecord(int x, int port, void *rec, size_t len)
{
    return XSendFRecord(x, port, -1, rec, len);
}



