/*
 *  WCOL: Launch module (Handler, Prompter and more) procedures
 *      by k-chinen@is.aist-nara.ac.jp, 1994, 1995, 1996, 1997
 *
 *  $Id: launch.c,v 1.11 1996/11/26 12:24:27 k-chinen Exp k-chinen $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <sys/types.h>
#include <sys/time.h>
#include <signal.h>
#ifdef HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif
#include <sys/uio.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <limits.h>

#include <sys/ipc.h>
#include <sys/shm.h>

#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif

#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif

#include <errno.h>


#include "wcol.h"
#include "msg.h"
#include "request.h"
#include "socket.h"
#include "ipc.h"
/*
#include "global.h"
*/

#include "fds.h"
#include "session.h"
#include "manager.h"
#include "hstate.h"
#include "handler.h"
#include "prompter.h"
#include "reaper.h"
#include "speaker.h"


/***
 *** public
 ***/

extern fdset_t manager_fdset;
extern int http_servd;
extern int lastfd;

extern int icp_servd;

extern time_t manager_ref_time;
extern int    catalyst_flag;
extern int    noreaper_flag;
extern int    icp_listen_portno;


#if 1
int max_handler, num_handler;
int max_prompter, num_prompter;
int max_speaker, num_speaker;
#endif


int reaper_fd = -1;
int reaper_pid = -1;
int speaker_fd = -1;
int speaker_pid = -1;
modinfo_t *handlers     = (modinfo_t*) NULL;
modinfo_t *prompters    = (modinfo_t*) NULL;
modinfo_t *speakers     = (modinfo_t*) NULL;



/* private */
int handler_no = -1;
int prompter_no = -1;
int speaker_no = -1;

static int IAmManager = 1;



inline static void report_die(pid_t pid)
{
    int i, h;

    if(pid==reaper_pid) {
        DLog("INFO die %6d reaper - :", pid);
        reaper_pid = -1;
    }
    else {
        if(handlers != NULL) {
            h = -1;
            for(i=0;i<MAX_HANDLER;i++) {
                if(pid == handlers[i].pid) {
                    h = i;
                }
            }
            if(h != -1) {
                DLog("INFO die %6d handler#  %3d : ", pid, h);
                return;
            }
        }

        if(prompters != NULL) {
            h = -1;
            for(i=0;i<MAX_PROMPTER;i++) {
                if(pid == prompters[i].pid) {
                    h = i;
                }
            }
            if(h != -1) {
                DLog("INFO die %6d prompter# %3d : ", pid, h);
                return;
            }
        }

        if(speakers != NULL) {
            h = -1;
            for(i=0;i<MAX_SPEAKER;i++) {
                if(pid == speakers[i].pid) {
                    h = i;
                }
            }
            if(h != -1) {
                DLog("INFO die %6d speaker#  %3d : ", pid, h);
                return;
            }
        }

        DLog("INFO die %6d unknown - :", pid);
    }
}




/*
 * TrapChildDie - Trap child's die. Check child is what.
 *
 */
RETSIGTYPE TrapChildDie()
{
    pid_t pid;

#ifdef HAVE_WAIT3
    /***
     *** use wait3()
     ***/

#ifdef HAVE_SYS_WAIT_H      /* like POSIX */
    int status;
#else
    union wait status;
#endif
    int i, h;

    while((pid = wait3(&status, WNOHANG | WUNTRACED,
        (struct rusage*)NULL)) > 0 ) {

        report_die(pid);
    }

#else /* HAVE_WAIT3 */



#ifdef HAVE_WAITPID
    /***
     *** use waitpid()
     ***/

#ifdef HAVE_SYS_WAIT_H      /* like POSIX */
    while((pid = waitpid((pid_t)-1, (int*) NULL, WNOHANG | WUNTRACED)) > 0 ) {
        report_die(pid);
    }
#else
    while((pid = wait(-1, (union wait*) NULL, WNOHANG | WUNTRACED)) > 0 ) {
        report_die(pid);
    }
#endif

#else   /* HAVE_WAITPID */
    /***
     *** use wait()
     ***/

#ifdef HAVE_SYS_WAIT_H      /* like POSIX */
    while((pid = wait((int*) NULL)) >= 0 ) {
        report_die(pid);
    }
#else
    while((pid = wait((union wait*) NULL)) >= 0 ) {
        report_die(pid);
    }
#endif


#endif  /* HAVE_WAITPID */

#endif /* HAVE_WAIT3 */
}




int find_module_from_pid(pid_t target_pid, int *h)
{
    int ret;
    int i;


    ret = WCOL_MOD_UNKNOW;

    if(target_pid==reaper_pid) {
        ret = WCOL_MOD_REAPER;
    }
    else {
        if(handlers != NULL) {
            *h = -1;
            for(i=0;i<MAX_HANDLER;i++) {
/*
                Trace("    h%d:%d\n", i, handlers[i].pid);
*/
                if(target_pid == handlers[i].pid) {
                    ret = WCOL_MOD_HANDLER;
                    *h = i;
                    goto out;
                }
            }
        }

        if(prompters != NULL) {
            *h = -1;
            for(i=0;i<MAX_PROMPTER;i++) {
/*
                Trace("    p%d:%d\n", i, prompters[i].pid);
*/
                if(target_pid == prompters[i].pid) {
                    ret = WCOL_MOD_PROMPTER;
                    *h = i;
                    goto out;
                }
            }
        }

        if(speakers != NULL) {
            *h = -1;
            for(i=0;i<MAX_SPEAKER;i++) {
/*
                Trace("    s%d:%d\n", i, speakers[i].pid);
*/
                if(target_pid == speakers[i].pid) {
                    ret = WCOL_MOD_SPEAKER;
                    *h = i;
                    goto out;
                }
            }
        }
    }

out:

#if 0
    Trace("find_module_from_pid: target %d -> module (type %d, no %d)\n",
        target_pid, ret, *h);
#else
    Trace("find_module_from_pid: target %d -> module (type %s(%d), no %d)\n",
        target_pid, ModTypeStr(ret), ret, *h);
#endif

    return ret;
}


/*
 *
 */
int KillModuleByPid(pid_t target)
{
    return 0;
}




/*
 * RemoveManagerResources - slaves not need manager's resources.
 *      (e.g., 'handlers' and others)
 */
void RemoveManagerResources()
{
    int i;

    IAmManager = 0;

#if 0
    close_handlers_state();
#endif
    deinit_requestqueue();

    /***
     ***
     ***/
    if(handlers!=NULL) {
        /* close rest of FDs */
        for(i=0;i<MAX_HANDLER;i++) {
            if(handlers[i].wake) {
                close(handlers[i].port);
            }
        }
    }

    if(prompters!=NULL) {
        /* close rest of FDs */
        for(i=0;i<MAX_PROMPTER;i++) {
            if(prompters[i].wake) {
                close(prompters[i].port);
            }
        }
        FREE(prompters);
        prompters = NULL;
    }

    if(speakers!=NULL) {
        /* close rest of FDs */
        for(i=0;i<MAX_PROMPTER;i++) {
            if(speakers[i].wake) {
                close(speakers[i].port);
            }
        }
        FREE(speakers);
        speakers = NULL;
    }

    /***
     ***
     ***/
    if(reaper_fd!=-1) {
        close(reaper_fd);
        reaper_fd = -1;
    }

#ifdef SINGLE_RECEPTION
    if(http_servd!=-1) {
        close(http_servd);
        http_servd = -1;
    }
#endif

#if 0
    if(icp_servd!=-1) {
        close(icp_servd);
        icp_servd = -1;
    }
#endif
}

/*
 * AmIManager - Check this process is manager or not.
 */
int AmIManager()
{
    return IAmManager;
}


/*
 * SwitchMinFD - switch target file descriptor to minmum file descriptor.
 */
int SwitchMinFD(int target)
{
    int newfd;

    /* file descriptor compaction */
    if((newfd = dup(target))<0) {
        Error("SwitchMinFD: fail dup() fd# %4d (%d)", target, errno);
        newfd = target;
    }
    else {
        close(target);
    }

    return newfd;
}


/*
 * FDCompaction
 */
int FDCompaction(int ear)
{
#ifdef USE_FD_COMPACTION
    return SwitchMinFD(ear);
#else
    return ear;
#endif
}

#if 0
int FDCompaction(int ear)
{
    RemoveManagerResources();
    return SwitchMinFD(ear);
}
#endif



key_t hstate_shmid;
key_t session_shmid;
Session* session_pool;



int init_handlers()
{
    int i;


#if 1
    /*
     *
     */
    session_shmid = assign_pool(WCOL_SHMKEY_SESSION);
    if(session_shmid == -1 ) {
        /* nothing */
    }
    else {
        delete_pool(session_shmid);
    }
    session_shmid =
        make_new_pool(WCOL_SHMKEY_SESSION, sizeof(Session)*MAX_HANDLER*MAX_SESSION);
    if(session_shmid == -1) {
        Error("init_handlers: cannot alloc session memory via shmop");

        return -1;
    }

    show_pool_stat(session_shmid);
#if 1
    if((session_pool = (Session*)shmat(session_shmid, (void*)NULL, 0))==NULL) {
        Fatal("init_handlers: cannot allocate memory to sessions array.");
        exit(WCOL_EXIT_NOMEMORY);
    }
    for(i=0;i<MAX_HANDLER*MAX_SESSION;i++) {
        CleanSession(&session_pool[i]);
    }
#endif


#endif


    /*
     *
     */
    if(handlers!=NULL) {
        Error("init_handlers: handler array is not NULL. free it.");
        FREE(handlers);
    }

    hstate_shmid = assign_pool(WCOL_SHMKEY_HSTATE);
    if(hstate_shmid == -1 ) {
        /* nothing */
    }
    else {
        delete_pool(hstate_shmid);
    }

    hstate_shmid =
        make_new_pool(WCOL_SHMKEY_HSTATE, sizeof(modinfo_t)*MAX_HANDLER);
    if(hstate_shmid == -1) {
        Error("init_handlers: cannot alloc handler memory via shmop");

        return -1;
    }

    show_pool_stat(hstate_shmid);

#if 0
    if((handlers = (modinfo_t*)MALLOC(sizeof(modinfo_t)*MAX_HANDLER))==NULL) {
        Fatal("init_handlers: cannot allocate memory to handlers array.");
        exit(WCOL_EXIT_NOMEMORY);
    }
#endif

    if((handlers = (modinfo_t*)shmat(hstate_shmid, (void*)NULL, 0))==NULL) {
        Fatal("init_handlers: cannot allocate memory to handlers array.");
        exit(WCOL_EXIT_NOMEMORY);
    }

    for(i=0;i<MAX_HANDLER;i++) {
        handlers[i].wake  = 0;
        handlers[i].busy  = 0;
        handlers[i].start = -1;
        handlers[i].port  = -1;

        handlers[i].z_req  = 0;
        handlers[i].z_sess = 0;
    }

    return 0;
}



/*
 * find_emptyhandler_slot - find index of new handler in array
 *                          scan handlers[0..max_handler]
 */
static int find_emptyhandler_slot()
{
    int i;

    for(i=0;i<max_handler;i++) {
        if(handlers[i].wake == 0) {
            return i;
        }
    }

    return -1;
}


/*
 *
 */
int LaunchHandler()
{
    int spair[2];
    pid_t cpid;
    int i;
    int newfd;

    if(handlers==NULL) {
        Fatal("LaunchHandler: handler list is NULL");
        return -1;
    }

    handler_no = find_emptyhandler_slot();
    if(handler_no==-1) {
        Trace("LaunchHandler: no slot. give up\n");
        return -1;
    }

    if (mkpipe(spair) == -1) {
        Error("faile mkpipe()");
        return -1;
    }

    switch(cpid = fork()) {
    case -1:
        Error("LaunchHandler: fail fork()");
        return -1;
        /*
        break;
        */

    case 0:
        /* Child (handler) */
        close(spair[0]);        /* close mouth of parent (manager) */

#ifdef DEBUG_HANDLER
        trace_flag = 1;
#endif

        RemoveManagerResources();
        Handler(FDCompaction(spair[1]));

        break;

    default:
        /* Parent (manager) */
        close(spair[1]);        /* close ear of child (handler) */

        /* append handler to 'handlers' */
        handlers[handler_no].pid     = cpid;
        handlers[handler_no].wake    = 1;
        handlers[handler_no].busy    = 0;
        handlers[handler_no].times   = 0;
        handlers[handler_no].port    = spair[0];
#if 0
        handlers[handler_no].url[0]  = '\0';
#endif
        handlers[handler_no].req.sid = 0;
        handlers[handler_no].req.id  = 0;

        if(lastfd<handlers[handler_no].port)
            lastfd = handlers[handler_no].port;

        FD_SET(handlers[handler_no].port, &manager_fdset);
        fd2mod[handlers[handler_no].port].type = WCOL_MOD_HANDLER;
        fd2mod[handlers[handler_no].port].mod  = &handlers[handler_no];

        DLog("INFO launch %6d handler#  %3d : fd# %4d",
            cpid, handler_no, handlers[handler_no].port);

        return handler_no;
        /*
        break;
        */
    }

    /*
    list_handlers(handlers, MAX_HANDLER, manager_ref_time);
    */
    return -1;
}






/*
 * StopHandler - Stop curren request
 *      send INT signal
 */
void StopHandler(int n)
{
    if(handlers[n].wake) {
        Log("; *stop handler  %d, pid %d, wake %d, busy %d, start %d",
            n, handlers[n].pid, handlers[n].wake, handlers[n].busy,
            handlers[n].start);

        kill(handlers[n].pid, SIGINT);
        handlers[n].busy = 0;
    }
}



int LaunchHandlers(int req)
{
    int i;
    int count;

    count = 0;
    for(i=0;i<req;i++) {
        if(LaunchHandler() != -1)
            count++;
    }

    list_handlers(handlers, MAX_HANDLER, manager_ref_time);

    return count;
}




/*
 * KillHandler - kill handler
 *      send TERM signal
 */
void KillHandler(int n)
{
    int f;

    if(handlers[n].wake) {
        Log("; *kill handler  %d, pid %d, wake %d, busy %d, start %d",
            n, handlers[n].pid, handlers[n].wake, handlers[n].busy,
            handlers[n].start);

        if(handlers[n].pid>1) {
            kill(handlers[n].pid, SIGTERM);
        }

        handlers[n].wake = 0;
        handlers[n].busy = 0;
        f = handlers[n].port;

        FD_CLR(f, &manager_fdset);
        close(f);
        fd2mod[f].type = WCOL_MOD_UNKNOW;
        fd2mod[f].mod  = NULL;

    }
}

void KillHandlers()
{
    int i;

    Log("; killhandlers - %p", handlers);

    if(handlers==NULL) {
        Log("; killhandlers - no handler", handlers);
        return;
    }

    for(i=0;i<MAX_HANDLER;i++) {
        KillHandler(i);
    }

    list_handlers(handlers, MAX_HANDLER, manager_ref_time);
}

int KillHandlerByPid(pid_t target)
{
    int n;
    int c;

    Trace("KillHandlerByPid: target(pid#) %d\n", target);
    if((c = find_module_from_pid(target, &n)) == WCOL_MOD_HANDLER) {
        if(n>=0) {
            KillHandler(n);
        }
    }
    return 0;
}



/*
 * ReplaceHandler - Replace current handler to new one
 *      kill handler, and launch new one
 */
void ReplaceHandler(int n)
{
    Log("; replace handler %d (pid %6d)", n, handlers[n].pid);

    KillHandler(n);

    LaunchHandler();
}

int ReplaceHandlerByPid(pid_t target)
{
    int n;
    int c;

    Trace("ReplaceHandlerByPid: target(pid#) %d\n", target);
    if((c = find_module_from_pid(target, &n)) == WCOL_MOD_HANDLER) {
        if(n>=0) {
            ReplaceHandler(n);
        }
    }
    return 0;
}




/***
 ***
 ***/

int init_prompters()
{
    int i;

    if(prompters!=NULL) {
        Error("init_prompters: prompter array is not NULL. free it.");
        FREE(prompters);
    }

    if((prompters = (modinfo_t*)MALLOC(sizeof(modinfo_t)*MAX_PROMPTER))==NULL) {
        Fatal("init_prompters: cannot allocate memory to prompters array.");
        exit(WCOL_EXIT_NOMEMORY);
    }

    for(i=0;i<MAX_PROMPTER;i++) {
        prompters[i].wake  = 0;
        prompters[i].busy  = 0;
        prompters[i].start = -1;
        prompters[i].port  = -1;

        prompters[i].z_sess = 0;
        prompters[i].z_req  = 0;
    }

    return 0;
}

static int find_emptyprompter_slot()
{
    int i;

    for(i=0;i<max_prompter;i++) {
        if(prompters[i].wake == 0) {
            return i;
        }
    }

    return -1;
}


int LaunchPrompter()
{
    int spair[2];
    pid_t cpid;
    int i;

    if(prompters==NULL) {
        Fatal("LaunchPrompter: prompter list is NULL");
        return -1;
    }

    prompter_no = find_emptyprompter_slot();
    if(prompter_no==-1) {
        Trace("LaunchPrompter: no slot. give up\n");
        return -1;
    }

    if (mkpipe(spair) == -1) {
        Error("faile mkpipe()");
    }

    switch(cpid = fork()) {
    case -1:
        Error("LaunchPrompter: fail fork()");
        return -1;
        /*
        break;
        */

    case 0:
        /* close mouth of parent (manager) */
        close(spair[0]);

#ifdef DEBUG_PROMPTER
        trace_flag = 1;
#endif

        RemoveManagerResources();
        Prompter(FDCompaction(spair[1]));

        break;

    default:
        /* Parent (manager) */
        close(spair[1]); /* close ear of child (prompter) */

        /* append handler to 'prompters' */
        prompters[prompter_no].pid     = cpid;
        prompters[prompter_no].wake    = 1;
        prompters[prompter_no].busy    = 0;
        prompters[prompter_no].times   = 0;
        prompters[prompter_no].port    = spair[0];
#if 0
        prompters[prompter_no].url[0]  = '\0';
#endif
        prompters[prompter_no].z_req  = 0;
        prompters[prompter_no].z_sess = 0;

        if(lastfd<prompters[prompter_no].port)
            lastfd = prompters[prompter_no].port;

        FD_SET(prompters[prompter_no].port, &manager_fdset);
        fd2mod[prompters[prompter_no].port].type = WCOL_MOD_PROMPTER;
        fd2mod[prompters[prompter_no].port].mod  = &prompters[prompter_no];

        DLog("INFO launch %6d prompter# %3d : fd# %4d",
            cpid, prompter_no, prompters[prompter_no].port);

    }

    return 0;
}




int LaunchPrompters(int req)
{
    int i;
    int count;

    count = 0;
    for(i=0;i<req;i++) {
        if(LaunchPrompter() != -1)
            count++;
    }

    list_handlers(prompters, MAX_PROMPTER, manager_ref_time);

    return count;
}


void KillPrompter(int n)
{
    int f;

    if(prompters[n].wake) {
        Log("; *kill prompter %d, pid %d, wake %d, busy %d, start %d",
            n, prompters[n].pid, prompters[n].wake, prompters[n].busy,
            prompters[n].start);

        if(prompters[n].pid>1) {
            kill(prompters[n].pid, SIGTERM);
        }

        prompters[n].wake = 0;
        prompters[n].busy = 0;
        f = prompters[n].port;

        FD_CLR(f, &manager_fdset);
        close(f);
        fd2mod[f].type = WCOL_MOD_UNKNOW;
        fd2mod[f].mod  = NULL;
    }
}

void KillPrompters()
{
    int i;

    if(prompters==NULL) {
        return;
    }

    for(i=0;i<MAX_PROMPTER;i++) {
        KillPrompter(i);
    }

    /*
    list_handlers(prompters, MAX_PROMPTER, manager_ref_time);
    */
}



/***
 ***
 ***/



int LaunchReaper()
{
    int spair[2];
    pid_t cpid;
    int i;


    if(noreaper_flag!=0) {
        Trace("LaunchReaper: noreaper mode.\n");
        return 0;
    }

    if (mkpipe(spair) == -1) {
        Error("faile mkpipe()");
    }


    switch(cpid = fork()) {
    case -1:
        Error("LaunchReaper: fail fork()");
        return -1;
        /*
        break;
        */

    case 0:
        /* close mouth of parent (manager) */
        close(spair[0]);

#ifdef DEBUG_REAPER
        trace_flag = 1;
#endif

        RemoveManagerResources();
        Reaper(0, FDCompaction(spair[1]));

        break;

    default:
        reaper_pid = cpid;
        i = 0;

        /* close ear of child (handler) */
        close(spair[1]);


        reaper_fd = spair[0];
#if 1
        /*
         * connect processes without deadlock
         */
#ifdef FNDELAY
        if(fcntl(reaper_fd, F_SETFL, FNDELAY)<0) {
            Error("LaunchReaper: cannot set noblocking mode");
        }
#else
        if(fcntl(reaper_fd, F_SETFL, O_NDELAY)<0) {
            Error("LaunchReaper: cannot set noblocking mode");
        }
#endif

#endif

        FD_SET(reaper_fd, &manager_fdset);
        fd2mod[reaper_fd].type = WCOL_MOD_REAPER;
        fd2mod[reaper_fd].mod  = NULL;

        if(lastfd<reaper_fd)
            lastfd = reaper_fd;

        DLog("INFO launch %6d reaper#   %3d : fd# %4d",
            cpid, i, reaper_fd);


    }

    return 0;
}


int KillReaper()
{
    int f;

    /* no reaper mode */
    if(noreaper_flag!=0) {
        return 0;
    }

    if(reaper_pid>0) {
        Log("; *kill reaper, pid %d", reaper_pid);
        kill(reaper_pid, SIGTERM);
    }
    else {
        Log("; *kill reaper skip, pid %d", reaper_pid);
    }

    f = reaper_fd;

    FD_CLR(f, &manager_fdset);
    close(f);
    fd2mod[f].type = WCOL_MOD_UNKNOW;
    fd2mod[f].mod  = NULL;


    return 0;
}





int init_speakers()
{
    int i;

    if(speakers!=NULL) {
        Error("init_speakers: handler array is not NULL. free it.");
        FREE(speakers);
    }

    if((speakers = (modinfo_t*)MALLOC(sizeof(modinfo_t)*MAX_SPEAKER))==NULL) {
        Fatal("init_speakers: cannot allocate memory to speakers array.");
        exit(WCOL_EXIT_NOMEMORY);
    }

    for(i=0;i<MAX_SPEAKER;i++) {
        speakers[i].wake  = 0;
        speakers[i].busy  = 0;
        speakers[i].start = -1;
        speakers[i].port  = -1;
    }

    return 0;
}


static int find_emptyspeaker_slot()
{
    int i;

    for(i=0;i<max_speaker;i++) {
        if(speakers[i].wake == 0) {
            return i;
        }
    }

    return -1;
}


int LaunchSpeaker()
{
    int spair[2];
    pid_t cpid;
    int i;


    if(speakers==NULL) {
        Error("LaunchSpeaker: speaker list is NULL");
        return -1;
    }

    speaker_no = find_emptyspeaker_slot();
    if(speaker_no==-1) {
        Trace("LaunchSpaeker: no slot. give up\n");
        return -1;
    }

    if (mkpipe(spair) == -1) {
        Error("faile mkpipe()");
    }

    switch(cpid = fork()) {
    case -1:
        Error("LaunchSpeaker: fail fork()");
        return -1;
        /*
        break;
        */

    case 0:
        /* close mouth of parent (manager) */
        close(spair[0]);

#ifdef DEBUG_REAPER
        trace_flag = 1;
#endif

        RemoveManagerResources();
        Speaker(FDCompaction(spair[1]));

        break;

    default:
        /* close ear of child (speaker) */
        close(spair[1]);

        /* append speaker to 'speakers' */
        speakers[speaker_no].pid     = cpid;
        speakers[speaker_no].wake    = 1;
        speakers[speaker_no].busy    = 0;
        speakers[speaker_no].times   = 0;
        speakers[speaker_no].port    = spair[0];
#if 0
        speakers[speaker_no].url[0]  = '\0';
#endif

        if(lastfd<speakers[speaker_no].port)
            lastfd = speakers[speaker_no].port;

        FD_SET(speakers[speaker_no].port, &manager_fdset);
        fd2mod[speakers[speaker_no].port].type = WCOL_MOD_SPEAKER;
        fd2mod[speakers[speaker_no].port].mod  = &speakers[speaker_no];

        DLog("INFO launch %6d speaker#  %3d : fd# %4d",
            cpid, speaker_no, speakers[speaker_no].port);

        return speaker_no;
    }

    return 0;
}

int LaunchSpeakers(int req)
{
    int i;
    int count;

    if((icp_servd = UDP_bind_port(icp_listen_portno, WCOL_LISTEN_LENGTH)) < 0) {
        Error("LaunchSpeakers: cannot bind listen port# %d",
            icp_listen_portno);
        exit(0);
    }
    Trace("LaunchSpeakers: ICP port(%d) is ready as fd# %4d\n",
        icp_listen_portno, icp_servd);

    count = 0;
    for(i=0;i<req;i++) {
        if(LaunchSpeaker() != -1)
            count++;
    }

#if 0
    if(icp_servd >= 0) {
        Trace("LaunchSpeakers: close ICP port fd# %4d\n", icp_servd);
        close(icp_servd);
        icp_servd = -1;
    }
#endif

    list_handlers(speakers, MAX_SPEAKER, manager_ref_time);

    return count;
}


int KillSpeaker(int n)
{
    int f;

    if(speakers[n].wake) {
        Log("; *kill speaker %d, pid %d, wake %d, busy %d, start %d",
            n, speakers[n].pid, speakers[n].wake, speakers[n].busy,
            speakers[n].start);

        if(speakers[n].pid>0) {
            kill(speakers[n].pid, SIGTERM);
        }

        speakers[n].wake = 0;
        speakers[n].busy = 0;
        f = speakers[n].port;

        FD_CLR(f, &manager_fdset);
        close(f);
        fd2mod[f].type = WCOL_MOD_UNKNOW;
        fd2mod[f].mod  = NULL;
    }

    return 0;
}

void KillSpeakers()
{
    int i;

    if(speakers==NULL) {
        return;
    }

    for(i=0;i<MAX_SPEAKER;i++) {
        KillSpeaker(i);
    }

    list_handlers(speakers, MAX_SPEAKER, manager_ref_time);
}



/*
 * ReplaceSpeaker - Replace current speaker to new one
 *      kill speaker, and launch new one
 */
void ReplaceSpeaker(int n)
{
    Log("; replace speaker %d (pid %6d)", n, speakers[n].pid);

    KillSpeaker(n);

    LaunchSpeaker();
}

int ReplaceSpeakerByPid(pid_t target)
{
    int n;
    int c;

    Trace("ReplaceSpeakerByPid: target(pid#) %d\n", target);
    if((c = find_module_from_pid(target, &n)) == WCOL_MOD_SPEAKER) {
        if(n>=0) {
            ReplaceSpeaker(n);
        }
    }
    return 0;
}






/* END of launch.c */

