/*
 *  WCOL: Manager Routines
 *      by k-chinen@is.aist-nara.ac.jp, 1995, 1996, 1997
 *
 *  $Id: manager.c,v 1.11 1996/11/26 12:24:43 k-chinen Exp k-chinen $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <limits.h>
#include <time.h>
#ifdef TIME_WITH_SYS_TIME
#include <sys/time.h>
#endif
#include <sys/resource.h>
#include <signal.h>
#include <errno.h>

#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif

#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif


#include "wcol.h"
#include "msg.h"
#include "socket.h"
#include "request.h"
#include "ipc.h"
#include "sigs.h"
#include "fds.h"

#include "handler.h"
#include "prompter.h"
#include "reaper.h"
#include "session.h"
#include "manager.h"
#include "hstate.h"

/*
#include "global.h"
*/


#ifndef TRACE_INTERVAL
#define TRACE_INTERVAL  (50)
#endif

/* events */
#define WE_TIMEOUT  -1
#define WE_NOP      0       /* no operation */
#define WE_NEW      1       /* new request */
#define WE_REPORT   2       /* regular report from handler */
#define WE_EREPORT  3       /* exception report from handler */
#define WE_PROMPT   4       /* prompter's prompt */
#define WE_PPROMPT  5       /* speaker's sync. prompt */
#define WE_REAPER   6       /* reaper's report */
#define WE_GCTIME   10      /* garbege collection time */



extern int speaker_fd;
extern int speaker_pid;
extern int reaper_fd;
extern int reaper_pid;

extern int max_speaker, num_speaker;
extern int max_prompter, num_prompter;



#if 0
#include "fdset_debug.inc"
#endif


#ifndef TO_PROMPTOR_TIMEOUT
#define TO_PROMPTOR_TIMEOUT (3)
#endif



/* global */
fdset_t manager_fdset;
int http_servd  = -1;
int lastfd = -1;
fd2mod_t fd2mod[WCOL_FD_MAX];

/* global clock */
time_t manager_ref_time = -1;

#if 1
int http_listen_portno;
int icp_listen_portno;
int wcol_request_count=0;
int prefetch_flag;
char pid_file[STRING_SIZE];
#endif

int to_promptor_timeout = TO_PROMPTOR_TIMEOUT;


/* internal */
off_t usearea, freearea;


char *_mod_type_str[WCOL_MOD_SPEAKER+1] = {
        "*unknow*", "manager", "reaper", "handler", "prompter", "speaker",
    };




/*
 * import from schedule.c
 */

extern int n_active, n_busy, n_client, n_prefetch;
extern int n_nop_req, n_busy_req, n_empty_req, n_lock_req;

extern  show_rque () ;
extern  check_rque () ;
extern  int init_requestqueue () ;
extern  int append_request (request_t req ) ;
extern  int remove_request (int sid, int id ) ;
extern  void DispatchJob () ;

/*
 * import from launch.c
 */

extern  void ReplaceHandler (int n ) ;
extern  void StopHandler (int n ) ;



/*
 * GetPrompterPrompt - get prompt to prefetch from prompter (a.k.a. Oracle :-)
 */
int GetPrompterPrompt(modinfo_t *a_prompter)
{
    int c;
    prequest_t prompt;
    int ear;

    ear = a_prompter->port;

try:
    c=RecvRecord(ear, &prompt, sizeof(prequest_t));
    if(c<0) {
        if(errno==EINTR)
            goto try;

        Error("GetPrompterPrompt: RecvRecord() error (%d)", errno);
        return -1;
    }

    if(c==0) {
        /* nothing */
    }
    else
    if(c==sizeof(prequest_t)){
        Trace("GetPrompterPrompt: got prompt about '%s' (%d)\n",
            prompt.req.url, prompt.cont);
#if 1
        if(trace_flag)
            ShowRequest(&prompt.req);
#endif

        prompt.req.fd = -1;

#if 1
        if(trace_flag)
            ShowRequest(&prompt.req);
#endif

        if(prompt.cont==0) {
            a_prompter->busy = 0;
        }
        else {

        /*
        prompt.req.fd   = -1;
        */
#if 1
        prompt.req.sid  = ++wcol_request_count;
#endif
        if(append_request(prompt.req)) {
            Log("; request queue overflow in GetPrompterPrompt()");
            /*
             * prompt was recevied and was not append to queue.
             * so, it is to be garbege.
             */
        }

        }
    }
    else {
        Trace("GetPrompterPrompt: get *ignore* prompt size=%d/%d errno=%d.\n",
            c, sizeof(prequest_t), errno);
    }

    return 0;
}


/*
 * GetProxyPrompt - get prompt to prefetch from speaker.
 *      speaker listen ICP query from other proxies.
 */
int GetProxyPrompt(modinfo_t *a_speaker)
{
    size_t c;
    prequest_t prompt;
    int chk;
    int ear;


    ear = a_speaker->port;


try:
    c=RecvRecord(ear, &prompt, sizeof(prequest_t));
    if(((int)c)<0) {
        if(errno==EINTR)
            goto try;

        Error("GetProxyPrompt: RecvRecord() error (%d)", errno);
        return -1;
    }

    if(c==0) {
        Error("GetProxyPrompt: empty proxy prompt");

        ReplaceSpeakerByPid(a_speaker->pid);

        return -1;
    }
    else
    if(c==sizeof(prequest_t)){
        Trace("GetProxyPrompt: get prompt about '%s'\n", prompt.req.url);
        /*
        prompt.req.mode = REQ_PROXY;
        */
        prompt.req.fd   = -1;
#if 1
        prompt.req.sid   = ++wcol_request_count;
#endif
        /*
        append_request(prompt.req);
        */
        chk = append_uniq_request(prompt.req);
        if(chk<0) {
#if 0
            Log("; request is duplicated in GetProxyPrompt()");
#else
            Trace("GetProxyPrompt: request is duplicated, skip\n");
#endif
            /*
             * prompt was recevied and was not append to queue.
             * don't warry. because it is found in queue.
             * requested resource will be get soon.
             */
        }
        else
        if(chk>0) {
            Log("; request queue overflow in GetProxyPrompt()");
            /*
             * prompt was recevied and was not append to queue.
             * so, it is to be garbege.
             */
        }
    }
    else {
        Error("GetProxyPrompt: get *ignore* prompt size=%d/%d errno=%d",
            c, sizeof(prequest_t), errno);
    }

    return 0;
}


/*
 * SendPrompterRequest - send report to prompter. prompter will reply prompt.
 */
int SendPrompterRequest(request_t req)
{
    int i;
    int target;
    mreport_t report;
    off_t nwrite;

    if(!prefetch_flag || req.noprefetch) {
        return 0;
    }

    report.req = req;


#if 1
    Trace("SendPrompterRequest: sid# %d, id# %d --\n",
        report.req.url, req.sid, req.id);

    report.req.sid   = ++wcol_request_count;
#endif
    Trace("SendPrompterRequest: '%s', sid# %d, id# %d\n",
        report.req.url, req.sid, req.id);

    target = -1;
    for(i=0;i<num_prompter;i++) {
        if(prompters[i].busy==0) {
            target = i;
            break;
        }
        if(manager_ref_time - prompters[target].start > 2 ) {
            prompters[target].busy  = 0;
            target = i;
            break;
        }
    }

    if(target==-1) {
        Trace("SendPrompterRequest: prompters are busy\n");
        return 1;
    }

    Trace("SendPrompterRequest: send to prompter #%d (pid %d) [%d]\n",
        target, prompters[target].pid, target);

    prompters[target].start = manager_ref_time;
    prompters[target].busy  = 1;
#if 0
    strcpy(prompters[target].url, req.url);
#endif

    pipe_broken = 0;
    nwrite = XSendRecord(to_promptor_timeout, prompters[target].port, &report, sizeof(mreport_t));
    if(pipe_broken) {
        pipe_broken = 0;
        Fatal("prompter is stange. Is it Down ?");

        LaunchPrompter();
    }
    Trace("\twrite %d\n", nwrite);


    return 0;
}


request_t *take_request(int rid) ;


/*
 * GetHandlerReport - get handler's report and calc disk space with report.
 *
 * return:
 *      -1  error
 *      0   success
 */
int GetHandlerReport(modinfo_t *a_handler)
{
    hreport_t report;
    int c;
    request_t *rreq;
    /*
    request_t dreq;
    */

    if(a_handler->wake==0) {
        Trace("GetHandlerReport: report from %p. Is it a zombie ?\n", *a_handler);
        return -1;
    }

    pipe_broken = 0;
    if((c=RecvRecord(a_handler->port, &report, sizeof(hreport_t)))<0)
    {
        Error("GetHandlerReport: read error fd %d (%d bytes)",
            a_handler->port, c);
        return -1;
    }
    if(pipe_broken) {
        Fatal("GetHandlerReport: Report of %p (pid=%d) is broken.",
            a_handler, a_handler->pid);

#if 0
        /*
        ReplaceHandler(n);
        */
        kill(a_handler->pid, SIGKILL);
#endif

        ReplaceHandlerByPid(a_handler->pid);

        return -1;
    }
    if(c==0){
        Error("GetHandlerReport: Report of %p (pid=%d) is empty.",
            a_handler, a_handler->pid);

        ReplaceHandlerByPid(a_handler->pid);

        return -1;
    }

#if 0
    Log("; report #%d from %d is arrive", report.req.id, n);
#endif

    /*
     *
     */

    Trace("\nGetHandlerReport: ID #%d of handler %#10x (pid %d)\n\
  %d secs %d bytes\n",
        report.id, a_handler, a_handler->pid,
        manager_ref_time - a_handler->start, c);

#if 1
    if(trace_flag) {
        ShowHReport(&report);
    }
#endif

    if(report.keep_alive==0) {
#if 0
        a_handler->busy--;      /* to be free */
#endif
        a_handler->start = -1;
    }
    else {
                                /* to be keep-alive */
        a_handler->start = -1;
    }


#if 0
    list_handlers(handlers, MAX_HANDLER, manager_ref_time);
#endif

    usearea  += report.diff;
    freearea -= report.diff;

    Trace("; DIFF %10d bytes ( %4d KB )\n",
            (long)report.diff, (long)TOKB(report.diff));
    Trace("; USE  %10d bytes ( %4d MB )\n",
            (long)usearea,  (long)TOMB(usearea));
    Trace("; FREE %10d bytes ( %4d MB )\n",
            (long)freearea, (long)TOMB(freearea));

#if 0
    if(report.keep_alive==0) {
        rreq = take_request(report.id);
        if(rreq!=NULL) {
            remove_request(rreq->sid, rreq->id);
        }
        else {
            Trace("\tnot found such request in queue.\n");
        }
    }
#else
    if(report.id!=0) {
        remove_request(report.sid, report.id);
    }
#endif


    rreq = &report.req;

#if 0
    if(trace_flag) {
        ShowRequest(rreq);
    }
#endif


    if(report.type==HREP_UNKNOW ||
        report.type==HREP_IGNOREREQUEST || report.type==HREP_INNERERROR) {
        Trace("  ERROR     %7d/%d %4d secs\n", rreq->sid, rreq->id,
            manager_ref_time - rreq->issue_time.tv_sec);
    }
    else
    if(report.type==HREP_INT) {
        Trace("  INTERRUPT %7d/%d %4d secs\n", rreq->sid, rreq->id,
            manager_ref_time - rreq->issue_time.tv_sec);
    }
    else
    if(report.type==HREP_TIMEOUT) {
        Trace("  TIMEOUT   %7d/%d %4d secs\n", rreq->sid, rreq->id,
            manager_ref_time - rreq->issue_time.tv_sec);
    }
    else {
        Trace("  DONE      %7d/%d %4d secs\n", rreq->sid, rreq->id,
            manager_ref_time - rreq->issue_time.tv_sec);
    }

    if(report.type==HREP_SERVERACCESS || report.type==HREP_CACHEHIT) {
        if(prefetch_flag && !rreq->noprefetch) {
            if(IsHTML(rreq->url)) {
                SendPrompterRequest(*rreq);
            }
        }
    }

    return 0;
}



GetReaperReport(int ear)
{
    rreport_t rep;
    int c;

    if(ear<0)
        return -1;

    c = RecvRecord(ear, &rep, sizeof(rreport_t));
    Trace("GetReaperReport: time %10d, use %10d, free %10d (%d)\n",
        manager_ref_time, (long)rep.usearea, (long)rep.freearea, c);

    usearea  = rep.usearea;
    freearea = rep.freearea;

    return 0;
}



static time_t next_gctime = -1;

static time_t Calc_next_gctime()
{
    if(next_gctime<0)
        time(&next_gctime);

    next_gctime += 60;

    return next_gctime;
}




void Display_modules_info()
{
    int fn;
    char *str_type;

    fprintf(stdout, "\n\n; Display_modules_info\n");
    fprintf(stdout, "%3s %-8s %-010s %6s %9s %9s\n",
            ";fd#", "type", "address", "pid#", "zreq", "zsess");

    for(fn = 0; fn<WCOL_FD_MAX; fn++) {
        if(fd2mod[fn].type == WCOL_MOD_UNKNOW) {
            continue;
        }

        if(fd2mod[fn].mod == NULL) {
        fprintf(stdout, "%3d: %-8s %-10s\n",
            fn, ModTypeStr(fd2mod[fn].type), "NULL");
        }
        else {
#if 0
        fprintf(stdout, "%3d: %-8s %#010x %6d %s\n",
            fn, ModTypeStr(fd2mod[fn].type), fd2mod[fn].mod,
            fd2mod[fn].mod != NULL ? (fd2mod[fn].mod)->pid : -1,
            fd2mod[fn].mod != NULL ? (fd2mod[fn].mod)->url : "*none*");
#else
        fprintf(stdout, "%3d: %-8s %#010x %6d %9d %9d\n",
            fn, ModTypeStr(fd2mod[fn].type), fd2mod[fn].mod,
            (fd2mod[fn].mod)->pid,
            (fd2mod[fn].mod)->z_req, (fd2mod[fn].mod)->z_sess);
#endif
        }
    }

    if(handlers!=NULL) {
        list_handlers(handlers, MAX_HANDLER, manager_ref_time);
    }
    if(prompters!=NULL) {
        list_handlers(prompters, MAX_PROMPTER, manager_ref_time);
    }
#if 0
    if(speakers!=NULL) {
        list_speakers(speakers, MAX_SPEAKER, manager_ref_time);
    }
#endif
}



static void init_eventhanlder(int port)
{
    int i;

    for(i=0;i<WCOL_FD_MAX;i++) {
        fd2mod[i].type = WCOL_MOD_UNKNOW;
        fd2mod[i].mod  = NULL;
    }


    FD_ZERO(&manager_fdset);

#if 0
    Trace("bind HTTP port(%d)\n", port);
#endif
    if((http_servd=TCP_bind_port(port, WCOL_LISTEN_LENGTH))<=0) {
        Fatal("manager: port %d is not available.", port);
        exit(WCOL_EXIT_BINDERROR);
    }
    else {
        Trace("manager: bind port# %d to fd# %4d\n", port, http_servd);
    }


#ifdef SINGLE_RECEPTION
    FD_SET(http_servd, &manager_fdset);
    fd2mod[http_servd].type = WCOL_MOD_MANAGER;
    fd2mod[http_servd].mod  = NULL;

    if(lastfd<http_servd)
        lastfd = http_servd;
#endif

}


static int get_event(int *etype, modinfo_t **eoption, int waittime)
{
    static int err_count=0;
    int chk;
    fdset_t rdset, wdset, edset;
    struct timeval *ptimeout, timeout;
    int i;

    *etype = WE_NOP;
    *eoption = NULL;
    rdset = wdset = edset = manager_fdset;
    i = 0;


    /* wait several seconds */
    if(waittime>0) {
        timeout.tv_sec = waittime;
        timeout.tv_usec = 0;
        ptimeout = &timeout;
    }
    else {
        ptimeout = NULL;
    }

retry:
#if 0
    time(&manager_ref_time);
    if(manager_ref_time>next_gctime) {
        Calc_next_gctime();
        *etype = WE_GCTIME;
        goto out;
    }
#endif

    errno = -1;
#if 1
    chk = select(lastfd, &rdset, NULL, NULL, ptimeout);
#else
    chk = select(FD_SETSIZE, &rdset, NULL, NULL, ptimeout);
#endif

    time(&manager_ref_time);

    /* error */
    if(chk<0) {
        if(errno == EINTR) {
            goto retry;
        }
        Trace("get_event: errno=%d\n", errno);

    {

        Fatal("get_event : errno %d, count %d", errno, err_count);

        Trace("\n*lastfd=%d\n", lastfd);
        Trace("*mask ");
        for(i=0;i<=lastfd;i++) {
            Trace("%d",i%10);
        }
        Trace("\n");
        Trace("manager_fdset ");
        for(i=0;i<=lastfd;i++) {
            Trace("%d",FD_ISSET(i,&manager_fdset)?1:0);
        }
        Trace("\n");
        Trace("rdset ");
        for(i=0;i<=lastfd;i++) {
            Trace("%d",FD_ISSET(i,&rdset)?1:0);
        }
        Trace("\n");
    }

        err_count++;
        if(err_count>10) {
            Error("get_event: too many error (%d times)", err_count);
            exit(WCOL_EXIT_UNDEF);
        }

        goto out;
    }

#if 1
    /* timeout */
    if(chk==0) {
        *etype = WE_TIMEOUT;
        goto out;
    }
#endif

#if 0
    dump_fdset(&manager_fdset, &rdset, &wdset, &edset);
#endif

#if 0
    for(i=0;i<WCOL_FD_MAX;i++) {
#endif
    for(i=0;i<=lastfd;i++) {
        if(FD_ISSET(i, &rdset)) {
            Trace("get_event: fd# %d/%d -> module type '%s', addr %#10x\n",
                i, lastfd, ModTypeStr(fd2mod[i].type), fd2mod[i].mod);

            if(fd2mod[i].type==WCOL_MOD_UNKNOW) {
                continue;
            }

            switch(fd2mod[i].type) {
            case WCOL_MOD_MANAGER:
                    if(i==http_servd) {
                        *etype = WE_NEW;
                        goto out;
                    }
                    continue;
                break;
            case WCOL_MOD_REAPER:
                    *etype = WE_REAPER;
                    *eoption = NULL;
                    goto out;
                break;
            case WCOL_MOD_HANDLER:
                    *etype = WE_REPORT;
                    *eoption = fd2mod[i].mod;
                    goto out;
                break;
            case WCOL_MOD_PROMPTER:
                    *etype = WE_PROMPT;
                    *eoption = fd2mod[i].mod;
                    goto out;
                break;
            case WCOL_MOD_SPEAKER:
                    *etype = WE_PPROMPT;
                    *eoption = fd2mod[i].mod;
                    goto out;
                break;
            default:
                Trace("get_event: not implemented yet\n");
                Trace("get_event: strange, unknow type module\n");

                FD_CLR(i, &manager_fdset);
                fd2mod[i].type = WCOL_MOD_UNKNOW;
                fd2mod[i].mod  = NULL;
                close(i);

                break;
            }
        }
    }

out:
#if 0
    if(*etype != WE_NOP && *etype != WE_TIMEOUT) {
        Trace("get_event: type=%d, option=%#10x [%d]\n",
            *etype, *eoption, i);
    }
#endif

    return *etype;
}



/*
 * CheckHandlers - remove very long connection
 *
 */
CheckHandlers()
{
    int i;
    int ret;

    ret = 53;

    n_active = 0;
    n_busy = 0;
    n_client = 0;
    n_prefetch = 0;

    for(i=0;i<MAX_HANDLER;i++) {
        ret += handlers[i].pid
                + handlers[i].start*11 + handlers[i].busy * 123
                + handlers[i].times*7;

        if(handlers[i].wake==0 || handlers[i].busy==0) {
            continue;
        }

#if 0
        if(manager_ref_time - handlers[i].start> REQUEST_LIMITTIME) {
            Trace("CheckHandlers: *stop long session %3d, %9d sec (%5d)\n",
                i, manager_ref_time-handlers[i].start, REQUEST_LIMITTIME);
            StopHandler(i);
        }
#endif
    }

    for(i=0;i<MAX_HANDLER;i++) {
        if(handlers[i].wake==0) {
            continue;
        }
        n_active++;
        if(handlers[i].busy==0) {
            continue;
        }
        n_busy++;

        if(handlers[i].req.mode == REQ_CLIENT)
            n_client++;
        if(handlers[i].req.mode == REQ_PREFETCH)
            n_prefetch++;

    }

    return ret;
}



KnockReaper()
{
    Log("; KnockReaper");

    if(reaper_pid>0) {
        /*
        kill(reaper_pid, SIGQUIT);
        */
        kill(reaper_pid, SIGHUP);
    }

    return 0;
}




int report_flag=0;

RETSIGTYPE DoReportNow()
{
    report_flag = 1;
}



static int deinit_done = 0;

void DeInit()
{
    Log("; END ? %s %s %d", WCOL_VER, WCOL_INFO_ID, AmIManager());

    if(AmIManager() && deinit_done == 0) {
        KillSpeakers();
        KillPrompters();
        KillHandlers();
        KillReaper();

        DLog("END %s %s", WCOL_VER, WCOL_INFO_ID);
        Log("; thank you. good bye.");

        deinit_done++;
    }
    else {
        /*
         * not manager, do nothing
         */
    }
}



/*
 * Main body of Manager
 */
int ManagerSetup()
{
    FILE *fp_pid;
    struct rlimit nfile;
    int max_nofile;


    wcol_pid = getpid();
    SetProcessComment("boot");

    time(&manager_ref_time);
#if 0
    open_handlers_state(manager_ref_time);
#endif

    if(!trace_flag) {
        close(STDOUT_FILENO);
    }



#ifdef RLIMIT_NOFILE
    if(getrlimit(RLIMIT_NOFILE, &nfile)<0) {
        Error("getrlimit() fail (%d)", errno);
    }
    else {
        Trace("RLIMIT_NOFILE cur %d, max %d\n",
            nfile.rlim_cur, nfile.rlim_max);
    }

    max_nofile = nfile.rlim_cur = nfile.rlim_max;

    if(setrlimit(RLIMIT_NOFILE, &nfile)<0) {
        Error("settrlimit() fail (%d)", errno);
    }

    if(getrlimit(RLIMIT_NOFILE, &nfile)<0) {
        Error("getrlimit() fail (%d)", errno);
    }
    else {
        Trace("RLIMIT_NOFILE cur %d, max %d\n",
            nfile.rlim_cur, nfile.rlim_max);
    }

#endif /* RLIMIT_NOFILE */

#ifdef RLIMIT_OFILE
    if(getrlimit(RLIMIT_OFILE, &nfile)<0) {
        Error("getrlimit() fail (%d)", errno);
    }
    else {
        Trace("RLIMIT_OFILE cur %d, max %d\n",
            nfile.rlim_cur, nfile.rlim_max);
    }

    max_nofile = nfile.rlim_cur = nfile.rlim_max;

    if(setrlimit(RLIMIT_OFILE, &nfile)<0) {
        Error("settrlimit() fail (%d)", errno);
    }

    if(getrlimit(RLIMIT_OFILE, &nfile)<0) {
        Error("getrlimit() fail (%d)", errno);
    }
    else {
        Trace("RLIMIT_OFILE cur %d, max %d\n",
            nfile.rlim_cur, nfile.rlim_max);
    }
#endif /* RLIMIT_OFILE */




#if 1

    Trace("; constants\n");
    Trace(";\t_GB_CONST_ %12d\n", _GB_CONST_);
    Trace(";\t_MB_CONST_ %12d\n", _MB_CONST_);
    Trace(";\t_KB_CONST_ %12d\n", _KB_CONST_);

    Trace("; size of records\n");
    Trace(";\tInfo          %9d\n", sizeof(Info));
    Trace(";\trequest       %9d\n", sizeof(request_t));
    Trace(";\threport_t     %9d\n", sizeof(hreport_t));
    Trace(";\tmrequest_t    %9d\n", sizeof(mrequest_t));
    Trace(";\tmreport_t     %9d\n", sizeof(mreport_t));
    Trace(";\tprequest_t    %9d\n", sizeof(prequest_t));
    Trace(";\trreport_t     %9d\n", sizeof(rreport_t));
    Trace(";\tSesson        %9d\n", sizeof(Session));
    Trace(";\tfd2sess_t     %9d\n", sizeof(fd2sess_t));

#endif

    init_handlers();
    init_prompters();
    init_speakers();
    init_eventhanlder(http_listen_portno);

    /*
     * store PID file
     */
    if((fp_pid=fopen(pid_file, "w"))==NULL) {
        Error("Cannot open pid file '%s'", pid_file);
    }
    else {
        fprintf(fp_pid, "%d", wcol_pid);
        fclose(fp_pid);
    }


    /*
    init_requestqueue();
    */

#ifdef MEMORY_CHECK
    fprintf(stderr, "*** before slave launching\n");
    show_memoryusage();
#endif

    LaunchReaper();
    LaunchPrompters(num_prompter);
    LaunchSpeakers(num_speaker);
    LaunchHandlers(num_handler);

    return 0;
}


/*
 *
 *  return:
 *      0   success
 *      -1  fail
 */
static int NewConnection(int http_servd, request_t *req)
{
    int     cfd;    /* actual connection (client fd) */

    if((cfd=AcceptRequest(http_servd))<0) {
        Trace("NewConnection: ignore access\n");
        return -1;
    }

    memset(req, 0, sizeof(request_t));

    req->state   = REQ_EMPTY;
    req->mode    = REQ_CLIENT;
    req->sid     = ++wcol_request_count;
    req->id      = 0;
    req->fd      = cfd;
#if 0
    req->issue_time = manager_ref_time;
#endif
    gettimeofday(&req->issue_time, NULL);

    req->method  = METHOD_UNKNOW;
    req->url[0]  = '\0';
    req->ver_major  =   0;
    req->ver_minor  =   0;
    req->keep_alive =   0;

    req->auth[0]    = '\0';
    req->cookie[0]  = '\0';
    req->ims     = -1;
    req->nocache =  0;
    req->noprefetch =  0;

#if 1
    Trace("NewConnection: ID #%d\n", req->sid );
#endif

    return 0;
}




RETSIGTYPE NewExitProgram()
{
    Log("; Catch signal for new exit. goodbye !");

    DeInit();

    exit(WCOL_EXIT_NORMAL);
}







/*
 * Main body of Manager
 */
Manager()
{
    int state, last_state;
    int chk, et;
    modinfo_t *eop;
    request_t req;
    int req_no=0;
    int hidden=1;
    int n_skip=0;


    SetProcessComment("manager");

    init_requestqueue();
    CheckFieldMask();

#ifdef MEMORY_CHECK
    fprintf(stderr, "*** after slave launching\n");
    show_memoryusage();
#endif

#if 0
    SetSignalHandler(SIGINT,    ExitProgram);
#endif
    SetSignalHandler(SIGTERM,   NewExitProgram);
    SetSignalHandler(SIGUSR1,   DoReportNow);

#ifdef HAVE_ATEXIT
    if(atexit(DeInit)) {
        Fatal("main: cannot alloc exit routine via atexit().\n");
        exit(WCOL_EXIT_UNDEF);
    }
#endif

    if(trace_flag)
        ShowSiblings(stdout);

    usearea  = (off_t) -1;
    freearea = (off_t) -1;
#if 0
{
    /*
     * schduler checking
     */
    int i, k;
    request_t treq;

    treq.mode       = REQ_PREFETCH;
    treq.state      = REQ_NOP;
    treq.id         = -1;
    treq.fd         = -1;
    treq.issue_time.tv_sec =  0;
    treq.issue_time.tv_usec =  0;
    strcpy(treq.peername, "-");
    treq.hereport   = -1;
    treq.ver_major  =  1;
    treq.ver_minor  =  0;
    strcpy(treq.auth, "");
    treq.ims        = -1;
    treq.nocache    = 0;
    treq.noprefetch    = 0;

    treq.method     = METHOD_GET;
    strcpy(treq.url, "http://www.gigabit.com/");

    srand((unsigned int)123);

    for(i=1;i<MAX_REQUEST_QUEUE*3;i++) {
        k = (rand() % i)+1;

        treq.sid    = i;

        Trace("### i=%d, k=%d\n", i, k);

        append_request(treq);
        show_rque();

        remove_request(k);
        show_rque();
    }

    exit(0);
}
#endif

#if 1
    if(trace_flag) {
        Display_modules_info();
    }
#endif

    last_state = -1;
    while(1) {
        state = CheckHandlers();
        /*
         * show handlers state
         */
        if(hidden==0) {
            check_rque();
            Log("; *RSTATE* empty %d nop %d busy %d lock %d",
                n_empty_req, n_nop_req, n_busy_req, n_lock_req);
            Log("; *HSTATE* req %d act %d busy %d c %d p %d",
                req_no, n_active, n_busy, n_client, n_prefetch);

#if 1
            if(trace_flag) {
                Display_modules_info();
            }
#endif

            report_flag = 0;
            hidden = 1;
        }

        chk = get_event(&et, &eop, 10);

        if(report_flag==1) {
            hidden = 0;
        }

        if(et==WE_TIMEOUT || et==WE_NOP) {
#if 0
            Trace("event handler timeout, I am ideling... \n");
#endif
        }
        else {
            switch(et) {
            case WE_NEW:
                req_no++;
                if(req_no%TRACE_INTERVAL==0) {
                    hidden = 0;
                }

                chk=NewConnection(http_servd, &req);

                if(append_request(req)) {
                    Error("manager: request queue overflow");
                    HTTP_ErrorMessage_Old(req.fd,
                            503, "Busy", "Request queue was overflow",
                            req.ver_major>0);
                    close(req.fd);
                    Log("; request queue overflow in Manager()");
                }
                break;

            case WE_REPORT:
                if(eop!=NULL)
                    GetHandlerReport(eop);
                break;

#if 0
            case WE_EREPORT:
                if(eop!=NULL)
                    GetHandlerReport(eop);
                break;
#endif

            case WE_PROMPT:
                if(eop!=NULL)
                    GetPrompterPrompt(eop);
                break;


            case WE_PPROMPT:
                GetProxyPrompt(eop);
                break;

            case WE_REAPER:
                GetReaperReport(reaper_fd);
                break;

#if 0
            case WE_GCTIME:
                /*
                Reaper();
                */
                break;
#endif

            default:
                Trace("*unknow* event is happend\n");
                break;
            }
        }

        DispatchJob();


        /*
         * report handlers state
         */
        if(state != last_state) {
#if 0
            write_handlers_state(handlers, MAX_HANDLER, manager_ref_time);
#endif
            n_skip = 0;

            if(trace_flag) {
#if 0
                Display_modules_info();
                list_handlers(handlers, MAX_HANDLER, manager_ref_time);
#endif
            }
        }
        else {
            if(n_skip > 7) {
#if 0
                write_handlers_state(handlers, MAX_HANDLER, manager_ref_time);
#endif
                n_skip = 0;
            }
            else {
                n_skip++;
            }
        }

        last_state = state;

#ifdef MEMORY_CHECK
        show_memoryusage();
#endif
    }

    DeInit();


#if 0
#ifndef HAVE_ATEXIT
    DeInit();
#endif
#endif
}


