/*
 *  WCOL: memory check procedures
 *      by k-chinen@is.aist-nara.ac.jp, 1995
 *
 * $Id: mem.c,v 1.1 1997/03/06 15:58:16 k-chinen Exp k-chinen $
 */

/*
 * These procedures called when Wcol was compile with -DMEMORY_CHECK.
 * Since these procedures generate large log and not release heap areas,
 * you should be enable only debugging.
 *
 * In addition, if you be enable -DTRACE_HEAP, these procedures show
 * more detail about memory chank.
 */


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <sys/types.h>

#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif


#define _TRACE_HEAP

/* overwrap malloc() and free() */

typedef struct {
    void *next;
    void *pos;
    long len;
    int  use;
} cattr;

static void *_chain_root    = NULL;
static long  _chain_count   = 0;
static long  _chain_size    = 0;
static long  _m_count=0;
static long  _f_count=0;
static long  _ig_m_count=0;
static long  _ig_f_count=0;




cattr *find_chain(void *target)
{
    char *p;
    cattr *a;

    p = _chain_root;
    while(p) {
        a = (cattr*)p;
        if(a->pos == target) {
            return a;
        }
        p = a->next;
    }

    return NULL;
}

void show_memoryusage()
{
    char  *p;
    cattr *a;
    long  c, keep, keep_size;
    long  nuse, nfree, nover, nminus;

    fprintf(stderr, "\n;; chain root=%#x count=%d size=%d\n",
        _chain_root, _chain_count, _chain_size);
                     /* 12345 1234567890 1234567890 1234567890 123 12345678 */
    fprintf(stderr, ";;   #      head       next       value    F    size\n");


    c = 0;
    keep = 0;
    keep_size = 0;
    nuse = nfree = nover = nminus = 0;
    p = _chain_root;
    while(p) {
        a = (cattr*)p;
        if(a->use<0 || a->use>1) {
            fprintf(stderr, ";: %5d %10x %10x %10x %3d %8d\n",
                c, p, a->next, a->pos, a->use, a->len);

            if(a->use>1) {
                nover++;
            }
            else {
                nminus++;
            }
        }
        else {
            fprintf(stderr, ";: %5d %10x %10x %10x  %c  %8d\n",
                c, p, a->next, a->pos,
                a->use==1 ? 'U' : (a->use==0 ? '-' : '?' ),
                a->len);
            if(a->use==1) {
                nuse++;
            }
            else {
                nfree++;
            }
        }

        if(a->use>0) {
            keep ++;
            keep_size += a->len;
        }

        c++;
        p = a->next;
    }

#if 0
    fprintf(stderr, ";; total %d keep %d size %d, malloc %d (%d) free %d (%d)\n",
        c, keep, keep_size, _m_count, _ig_m_count, _f_count, _ig_f_count);
    fprintf(stderr, "; use %d free %d over %d minus %d\n",
        nuse, nfree, nover, nminus);
#else
                             /* 1234567890 1234567890 1234567890 | 1234567890 */
    fprintf(stderr, ";@               keep     malloc       free |      total\n");
    fprintf(stderr, ";@ count#  %10d %10d %10d | %10d\n",
                                keep, _m_count, _f_count, c);
    fprintf(stderr, ";@  ignore %10s %10d %10d | %10s\n",
                                "", _ig_m_count, _ig_f_count, "");
    fprintf(stderr, ";@ size    %10d %10s %10s | %10s\n",
                                keep_size, "", "", "");
    fprintf(stderr, ";@ use %d free %d over %d minus %d\n",
        nuse, nfree, nover, nminus);
    fprintf(stderr, ";@\n");
#endif

}

void entry_chain(char *p, size_t x)
{
    cattr *a;

#if 0
    _chain_root = pp;
    *p = NULL;
#endif
    a = (cattr*)p;
    a->next = _chain_root;
    _chain_root = (void*) p;
    _chain_count++;
    _chain_size += x;
}

void *wcol_malloc(size_t x)
{
    char *ret,*get;
    cattr *a;
    int w=0;

    w = x + sizeof(cattr);
    get = malloc(w);
    if(get==NULL) {
        _ig_m_count++;
#ifdef TRACE_HEAP
        fprintf(stderr, "<<malloc:error:%d>>", x);
        fflush(stderr);
#endif
        return NULL;
    }
    ret = get+sizeof(cattr);
    a = (cattr*)get;
    a->pos = ret;
    a->next = NULL;
    a->len = x;
    a->use = 1;
#ifdef TRACE_HEAP
    fprintf(stderr, "<<malloc:ok:%x:%x:%d>>", get, ret, x);
    fflush(stderr);
#endif
    entry_chain(get, x);

    _m_count++;

    return (void*)ret;
}


void wcol_free(void *p)
{
    cattr *c;

    if((c = find_chain(p))==NULL) {
#ifdef TRACE_HEAP
        fprintf(stderr, "<<free:error:%x>>", p);
        fflush(stderr);
#endif
        _ig_f_count++;
        return;
    }

    _f_count++;

    if(c->use>1) {
#ifdef TRACE_HEAP
        fprintf(stderr, "<<free:ignore:%x:%x>>", c, p);
        fflush(stderr);
#endif
    }
    else
    if(c->use==1) {
#ifdef TRACE_HEAP
        fprintf(stderr, "<<free:ok:%x:%x>>", c, p);
        fflush(stderr);
#endif
    }
    else {
#ifdef TRACE_HEAP
        fprintf(stderr, "<<free:dup:%x:%x>>", c, p);
        fflush(stderr);
#endif
    }

    c->use--;
    /*
    free(p);
    */
    c->pos = NULL;
}


void *dummy_malloc(size_t x)
{
    _m_count++;
    return malloc(x);
}

void dummy_free(void *p)
{
    _f_count++;
    free(p);
}


