/*
 *  WCOL: Prompter (parse data, generate prompt and send to manager)
 *      by k-chinen@is.aist-nara.ac.jp, 1994, 1995, 1996, 1997
 *
 *  $Id: prompter.c,v 1.1 1996/11/24 14:46:55 k-chinen Exp k-chinen $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif


#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/uio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <signal.h>
#include <setjmp.h>
#include <fcntl.h>
#include <limits.h>

#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif

#include <errno.h>

#include "wcol.h"
#include "sigs.h"
#include "ipc.h"
#include "request.h"
#include "parse.h"
#include "prompter.h"
/*
#include "global.h"
*/


#if 1
extern int prompter_fd;

int prefetch_lag_limit;
int num_includedprefetch;
int num_anchorprefetch;
#endif



#define _TRACE_PROMPTER



#ifndef PROMPT_TIMEOUT
#define PROMPT_TIMEOUT            (2)         /* 2 sec */
#endif






/*
 * Send prompt to manager
 */
static int SendPrompterPrompt(int port, int s_id, char *url, int cont)
{
    prequest_t prompt;
    int c=-1;
    struct timeval cur;

/*
    Trace("prompt '%s'\n", url);
    Trace("; send '%s'\n", url);
*/

    gettimeofday(&cur, NULL);

    memset(&prompt.req, 0, sizeof(request_t));

    prompt.req.mode             = REQ_PREFETCH;
    prompt.req.sid              = - (int) wcol_pid;
    if(s_id>0) {
    prompt.req.id               = - s_id;
    }
    else {
    Trace("SendPrompterPrompt: ignore s_id %d\n", s_id);
    prompt.req.id               = INT_MAX;
    }
    prompt.req.fd               = -1;
    prompt.req.state            = REQ_NOP;
    prompt.req.issue_time       = cur;

    strcpy(prompt.req.peername, "-");
    prompt.req.hereport         = -1;

    prompt.req.method           = METHOD_GET;
    strcpy(prompt.req.url,      url);

    prompt.req.ver_major        = 1;
    prompt.req.ver_minor        = 0;

    strcpy(prompt.req.auth,     "");
    prompt.req.ims              = -1;
    prompt.req.nocache          = 0;
    prompt.req.noprefetch       = 0;


    prompt.cont                 = cont;

    /*
     * To avoid deadlock with IPC, set timer
     */
    pipe_broken = 0;
    if((c = XSendRecord(PROMPT_TIMEOUT, port, &prompt, sizeof(prompt)))<0) {
        Error("Prompter: Cannot send prompt");
    }
    else
    if(c==0) {
#ifdef TRACE_PROMPTER
        Trace("Prompter: Cannot send prompt, 0 byte\n");
#endif
    }

    return 0;
}




static int Generate_elements(int port, int sid, char *base, int mode,
        int num_prefetch, int conntime_limit)
{
#if 0
    char *path;
#endif
    char path[STRING_SIZE];
    int len;
    int num;
    tags_t *tags;
    char **list;
    char **elem;
    char *nurl;
    Info *target_info;


    num = -1;
    list = NULL;

    Trace("Generate_elements: base URL='%s'\n", base);

    if((target_info=Have(base))==NULL) {
        Trace("Generate_elements: not exist such URL\n");
        goto fail;
    }
    if(target_info->attr.take_time>conntime_limit) {
        Trace("Generate_elements: It is slow, meybe.\n");
        Trace("\tlast connection time is %d[ms], limit is %d[ms]\n",
            target_info->attr.take_time, conntime_limit);

        FREE(target_info);
        goto fail;

    }
    FREE(target_info);

    if((tags = HTML_NewTags())==NULL) {
        Error("Generate_elements: not enough memory");
        goto fail;
    }

    if(AssignFileName(path, base, BODY_FILE)) {
        Trace("Generate_elements: no path about '%s'\n", base);
        goto fail;
    }

    if((len = HTML_ParseTags(tags, path, mode))<=0) {
        Trace("Generate_elements: no tags\n");
        goto notag;
    }

    list = HTML_ConvURLList(tags, &num);
    HTML_DestroyTags_withoutvalue(tags);
    /*
    FREE(tags);
    */

#ifndef TRACE_PROMPTER
    Trace("; %d/%d URL(s) ", num, len);
#else
    Trace("; %d/%d URL(s)\n", num, len);
#endif

    elem = list;
    num = 0;
    while(*elem && num < num_prefetch) {
        if((nurl = URL_JoinStr(base, *elem)) == NULL) {
            goto next;
        }


#ifdef TRACE_PROMPTER
        Trace("; %3d: %s ", num, nurl);
#endif


        /*
         * Ignore or Unsupport URL
         */
        if(Nstrncmp_first_all(nurl, 5, 'h', "http:"))
        {
            /* Nothing */
#ifdef TRACE_PROMPTER
            Trace(" *unsupported*\n");
#endif
        }
        /*
         * Target is exist
         */
        else
        if((target_info=Have(nurl))!=NULL) {
            /* Nothing */
#ifdef TRACE_PROMPTER
            Trace(" *have*\n");
#endif
            FREE(target_info);
        }
        /*
         * New Target
         */
        else {
#ifdef TRACE_PROMPTER
            Trace(" *go* ...\n");
#endif

            SendPrompterPrompt(port, sid, nurl, 1);

            num++;
        }

next:
        FREE(nurl);
        elem++;
    }

end:
#if 0
    FREE(path);
#endif
    HTML_FreeURLList(list);
    FREE(list);

    Trace("; dispatch %d targets\n\n", num);

    return num;

notag:
    HTML_DestroyTags(tags);
    /*
    FREE(tags);
    */

fail:
#if 0
    FREE(path);
#endif
    if(list!=NULL) {
        HTML_FreeURLList(list);
        FREE(list);
    }

    return -1;
}

static void IncludedPrefetching(int port, int sid, char *url,
            int num_prefetch, int conntime_limit)
{
    Generate_elements(port, sid, url, PARSE_INCLUDED,
        num_prefetch, conntime_limit);
}

static void AnchorPrefetching(int port, int sid, char *url,
            int num_prefetch, int conntime_limit)
{
    Generate_elements(port, sid, url, PARSE_REFERED,
        num_prefetch, conntime_limit);
}




/*
 * Prompter
 */
void Prompter(int ear)
{
    mreport_t report;
    int c, nc, ec;

    wcol_pid = getpid();
    SetProcessComment("prompter");
    if(trace_flag) {
        /*
        DLog("WAKE Prompter");
        */
        SwitchLog(DEFAULT_DEBUGLOGDIR, "log of wcol's prompter");
    }


#ifdef MEMORY_CHECK
    show_memoryusage();
#endif
    Trace("Prompter: connection for manager fd# %d\n", ear);

    nc = ec = 0;
    while(1) {
        nc++;
        if((c = RecvRecord(ear, &report, sizeof(mreport_t)))<0)
        {
            ec++;
            Error("Prompter: Cannot read manager's report (%d, %d)", ec, nc);

            if(ec>=100) {
                Log("; too many error (%d, %d). give up.", ec, nc);
                break;
            }
            continue;
        }

        ec = 0; /* reset error counter */
        if(c==0) {
            Trace("Prompter: get empty report\n");
            break;
        }

#ifdef MEMORY_CHECK
        /*
        show_memoryusage();
        */
#endif

#if 1
        Trace("Prompter: get report (%d bytes)\n\tmode=%s\n\turl='%s'\n\tsid# %d, id# %d\n",
            c, ReqModeStr(report.req.mode),
            report.req.url, report.req.sid, report.req.id);
#endif

        if(report.req.mode==REQ_CLIENT || report.req.mode==REQ_IMPORT)
        {
#if 0
            if(IsHTML(report.req.url)) {
                AnchorPrefetching(ear, report.req.sid, report.req.url,
                    num_anchorprefetch, prefetch_lag_limit);
            }
#endif
            AnchorPrefetching(ear, report.req.sid, report.req.url,
                num_anchorprefetch, prefetch_lag_limit);
        }
        else
        if(report.req.mode==REQ_PREFETCH) {
#if 0
            if(IsHTML(report.req.url)) {
                IncludedPrefetching(ear, report.req.sid, report.req.url,
                    num_includedprefetch, prefetch_lag_limit);
            }
            else {
                Trace("\tit is not worth, *thru*.\n\n");
            }
#endif
            IncludedPrefetching(ear, report.req.sid, report.req.url,
                num_includedprefetch, prefetch_lag_limit);
        }
        else {
            Error("Prompter: got ignore report");
            Trace("\n");
        }

        SendPrompterPrompt(ear, report.req.sid, "", 0);

#ifdef MEMORY_CHECK
        show_memoryusage();
#endif
    }
    /*
    Log("DIE Prompter");
    */
    Log("; goodbye !");

    exit(WCOL_EXIT_NORMAL);
}


