/*
 *  WCOL: set process state to hidden memory area for 'ps' or others
 *      by k-chinen@is.aist-nara.ac.jp, 1994, 1995
 *
 *  $Id: pstate.c,v 1.1 1995/12/04 17:14:22 k-chinen Exp k-chinen $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <limits.h>
#include <errno.h>


#ifndef USE_PSTATE

int SetProcessState (char *astr )
{
    return 0;   /* empty */
}

void MarkArgumentsPosition (int arg_num ,char **one_arg )
{
    /* empty */
}


#else /* USE_PSTATE */

static char *_start_cmdline=NULL;
static char *_end_cmdline=NULL;
static int   _len_cmdline=0;
static int   _orig_argc=0;


/*
 * Show command and arguments.
 * If you want to view more information (envrions) for debug,
 * You will it via setting 'rest' to nonzero.
 */
static void show_cmdline(int rest)
{
    char *p;
    int c;

    c = 0;
    p = _start_cmdline;
    while(p<_end_cmdline) {
        if(*p=='\0')
            c++;
        p++;
    }

    printf(" (%2d) ", c);
    putchar('<');
    p = _start_cmdline;
    while(p<_end_cmdline+rest) {
        if(*p=='\0')
            putchar('#');
        else
            putchar(*p);
        p++;
    }
    putchar('>');
    putchar('\n');
}


/*
 * SetProcessState -
 *
 *      Write string in process environ's arguments area (which is
 *      implemented array of string in BSD).
 */
int SetProcessState(char *astr)
{
    if(_start_cmdline==NULL) {
        return 1;
    }
    if(strlen(astr)==0) {
        return 1;
    }
    if((strlen(astr)+1) > (_len_cmdline-_orig_argc)) {
        return 1;
    }

#if 0
    printf("_____ <%s>\n", astr);
    show_cmdline(100);
#endif

    memset(_start_cmdline, ' ', _len_cmdline-1);
    strcpy(_start_cmdline, astr);
    *((char*)_start_cmdline+strlen(astr)) = ' ';
    memset(_end_cmdline-_orig_argc, '\0', _orig_argc-1);

#if 0
    show_cmdline(100);
#endif

    fflush(stdout);
    fflush(stderr);

    return 0;
}


/*
 * MarkArgumentsPosition - mark arguments position in memory
 */
void MarkArgumentsPosition(int arg_num, char **one_arg)
{
    int i;

    _orig_argc = arg_num;

    for(i=0;i<arg_num;i++) {
#if 0
        fprintf(stderr, "%3d: %#x %d %#x [%s]\n",
            i, one_arg[i], strlen(one_arg[i]),
            one_arg[i]+strlen(one_arg[i])+1, one_arg[i]);
#endif

        if(i==0) {
            _start_cmdline = one_arg[0];
        }
    }
    if(_start_cmdline!=NULL) {
        /* last '\0' is sentinel. for safety */
        _end_cmdline = one_arg[arg_num-1]+strlen(one_arg[arg_num-1])+1;
        _len_cmdline = _end_cmdline - _start_cmdline;
    }

#if 0
    fprintf(stderr, "start %#x, end %#x, len %d\n",
        _start_cmdline, _end_cmdline, _len_cmdline);
#endif
}


#endif /* USE_PSTATE */


