/*
 *  WCOL: Reaper Routines
 *      by k-chinen@is.aist-nara.ac.jp, 1995, 1996, 1997
 *
 *  $Id: reaper.c,v 1.10 1996/11/18 20:57:45 k-chinen Exp k-chinen $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/uio.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <limits.h>
#include <time.h>
#ifdef TIME_WITH_SYS_TIME
#include <sys/time.h>
#endif
#include <signal.h>
#include <errno.h>
#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif

#include <math.h>

#include "wcol.h"
#include "msg.h"
#include "misc.h"
#include "ipc.h"
#include "sigs.h"
#include "request.h"
#include "reaper.h"
/*
#include "global.h"
*/


#define _REAPER_TRACE
#define _REAPER_DEBUG


extern char    pool_dir[];

time_t tmpfile_keeptime = 30*60;     /* half hour */


#if 1
off_t   cache_size;
int     cache_use_rate;
int     gc_interval;
char    gc_workfile[STRING_SIZE];
int     gc_bufsize;
#endif


/*
#ifndef REQ_PREFETCH
#define REQ_PREFETCH
#endif
*/


/*
 * calc usage in disk like 'du'
 */
off_t check_du(char *dir, FILE *fp)
{
    DIR    *dp;
    dirent_t *elem;
    off_t    sum;
    struct   stat sbuf;
    char     path[STRING_SIZE];

    Trace("check_du: '%s'\n", dir);

    sum = 0;
    if((dp = opendir(dir))==NULL) {
        Trace("check_du: *empty*\n");
        return (off_t)0;
    }
    while((elem=readdir(dp))!=NULL) {
        sprintf(path, "%s/%s", dir, elem->d_name);
        if(stat(path, &sbuf)) {
            continue;
        }

        /*
         * without . or ..
         */
        if(elem->d_name[0]=='.' &&
             (elem->d_name[1]=='\0'     ||
            ((elem->d_name[1]=='.'  && elem->d_name[2]=='\0')) )) {
            continue;
        }

        sum += sbuf.st_size;

#if 0
        Trace("^ %s %d %d %d\n", path,
            (long)sbuf.st_size, (long)sbuf.st_blksize, (long)sbuf.st_blocks);

        sum += sbuf.st_blksize * sbuf.st_blocks;
#endif


        if(S_ISDIR(sbuf.st_mode)) {
            fprintf(fp, "D %s %d %d\n", path, (long) sbuf.st_size, (long)sum);
            fflush(fp);

            sum += check_du(path, fp);
        }
        else {
            fprintf(fp, "F %s %d %d\n", path, (long) sbuf.st_size, (long)sum);
        }
    }
    closedir(dp);
    fflush(fp);

    Trace("check_du: '%s' sum=%d\n", dir, (long)sum);

    return sum;
}


/*
 *
 */
off_t Calc_DiskUsage(char *base_dir, char *vfname)
{
    off_t chk;
    FILE *fp;

    /* */
    Trace("Calc_DiskUsage: size of struct stat = %d\n", sizeof(struct stat));
    /* */
    Trace("Calc_DiskUsage: base='%s'\n", base_dir);

    if((fp=fopen(vfname,"w"))==NULL) {
        Fatal("Calc_DiskUsage: cannot open file '%s'", vfname);
        return (off_t)-1;
    }

    chk = check_du(base_dir, fp);
    Trace("Calc_DiskUsage: size=%d\n", (long)chk );

    fclose(fp);

    return (off_t) 0;
}
















/*
 * compare memo of informations
 */
int CompMemo(const void *xx, const void *yy)
{
    memo_t *x = (memo_t*) xx, *y =(memo_t*) yy;

    if(x->val < y->val)
        return 1;
    else
    if(x->val > y->val)
        return -1;
    else
        return 0;
}





/*
 * global
 */
static time_t   start_time;
static int      memo_total = 0;
static off_t    memo_totalsize = 0;

extern int CheckEvalVal(Info*, time_t);     /* info.c */


/*
 * record_memo -
 *
 * input:
 *      path    - target filepath
 *      log_fd  - fd for memo
 *
 * return:
 *      >0      success (size of memo)
 *      <=0     error
 */
off_t record_memo(char *path, int log_fd)
{
    Info info;
    int  val;
    int  ck;
    memo_t memo;

#ifdef REAPER_TRACE
#if 0
    if(memo_total%1000==0) {
        Trace("<%d>", memo_total);
    }
#endif
#endif

    if(ReadInfoRaw(path, &info)) {
        Error("fail ReadInfoRaw()");
        return (off_t)-1;
    }
    if(CheckInfoAttr(info, start_time)) {
        return (off_t)-2;
    }

    val = EvalVal(&info, start_time);
    if(val<0) {
        Error("negative val=%d", val);
        if(trace_flag) {
            What(&info);
        }

        CheckEvalVal(&info, start_time);

#ifdef REAPER_DEBUG
    Trace("@# %d + %d * 10 - %d - %d ===> %d",
        EVAL_OFFSET,
        info.attr.last,
        (long)log((double)(info.attr.hsize+info.attr.bsize)+sizeof(Info)+1.0),
        (int)log((double)(start_time - info.attr.last) + 1.0),
        val);
#endif

        return (off_t)-3;
    }

    memo_total++;
    memo.num    = memo_total;
    memo.val    = val;
#if 0
    memo.len    = (off_t)sizeof(Info)+info.attr.hsize+info.attr.bsize;
#else
    memo.len    = NBLK512((off_t)sizeof(Info))
                    + NBLK512(info.attr.hsize) + NBLK512(info.attr.bsize);
#endif
    memset(memo.name, 0, INFONAMESIZE);
    strcpy(memo.name, path);


    if((ck=write(log_fd, &memo, sizeof(memo_t)))==sizeof(memo_t)) {
        memo_totalsize += memo.len;
        return memo.len;
    }
    else {
        Error("record_memo: write error %d bytes errno=%d", ck, errno);
        return (off_t)-4;
    }
}


/*
 * IsOlder -
 *
 * return
 *      1   old;    clear
 *      0   fresh;  not clear
 */
int IsOlder(char *path, struct stat *sbuf)
{
    int ret;
    /*
    Trace("IsOlder: path='%s'\n", path);
    */
    if( start_time - sbuf->st_mtime > tmpfile_keeptime) {
#ifdef REAPER_TRACE
        Trace("OLD %d ", start_time - sbuf->st_mtime);
#endif
        ret = 1;
    }
    else {
#ifdef REAPER_TRACE
        Trace("FRESH %d ", start_time - sbuf->st_mtime);
#endif
        ret = 0;
    }

    return ret;
}


/*
 *  collect_dir - collect memo of information in directory
 *
 *  return  total size (files/directories) in directory
 */
static int collect_dir(char *dir, int log_fd, off_t *file_sum, off_t *dir_sum)
{
    DIR    *dp;
    dirent_t *elem;
    struct   stat sbuf;
    char     path[STRING_SIZE];
    char     ndir[STRING_SIZE];
    int      ftype;
    int      clear_flag;
    int      n_worth;
    int      ecount;
    off_t    c_file_sum, c_dir_sum;
    int      chk;
    off_t    memo_size;
    off_t    this_dir_size;

    char ipath[STRING_SIZE];
    Info tmp_info;

    ecount = 0;
    n_worth = 0;
    *file_sum = (off_t)0;
    *dir_sum = (off_t)0;


    if(stat(dir, &sbuf)) {
        /* error happen, but it is hot file maybe */
        Error("collect_dir: *this* fail stat() (%d) '%s'", errno, dir);
        this_dir_size = 0;

        return -1;
    }
    else {
        this_dir_size = NBLK512(sbuf.st_size);
    }

    if((dp = opendir(dir))==NULL) {
        Trace("collect_dir: *empty* '%s'\n", dir);
        return -1;
    }

    while((elem=readdir(dp))!=NULL) {
        /*
         * empty
         */
        if(elem->d_name[0]=='\0') {
            continue;
        }

        /*
         * without . or ..
         */
        if(elem->d_name[0]=='.' &&
             (elem->d_name[1]=='\0'     ||
            ((elem->d_name[1]=='.'  && elem->d_name[2]=='\0')) )) {
            continue;
        }

        sprintf(path, "%s/%s", dir, elem->d_name);

        if(stat(path, &sbuf)) {
            /* error happen, but it is hot file maybe */
            Error("collect_dir: *child* stat() error (%d) '%s'", errno, path);
            ecount++;
            continue;
        }

        if(S_ISDIR(sbuf.st_mode)) {
            *dir_sum += NBLK512(sbuf.st_size);
            n_worth++;

#ifdef REAPER_TRACE
            Trace("D %s %d - \n", path, (long) sbuf.st_size);
#endif
            chk = collect_dir(path, log_fd, &c_file_sum, &c_dir_sum);
            *file_sum += NBLK512(c_file_sum);
            *dir_sum  += NBLK512(c_dir_sum);
        }
        else
        if(S_ISREG(sbuf.st_mode)) {
            clear_flag = 0;
            ftype = CheckFileType(path);

#ifdef REAPER_TRACE
            if(ftype==BODY_FILE) {
                Trace("F %s      %5d %02x ",
                    path, (long) sbuf.st_size, ftype);
            }
            else {
                Trace("F %s %5d %02x ",
                    path, (long) sbuf.st_size, ftype);
            }
#endif

            switch(ftype) {
            case INFO_FILE:
                if(ReadInfoRaw(path, &tmp_info)) {
#ifdef REAPER_TRACE
                    Trace("IGNORE ");
#endif
                    Error("strange INFO %s", path);
                    clear_flag = 1;
                }
                else {
                    if(tmp_info.attr.expire!=-1
                        && tmp_info.attr.expire < start_time) {
#ifdef REAPER_TRACE
                        Trace("EXPIRE %d ", tmp_info.attr.expire);
#endif
                        clear_flag = 1;
                    }
                    else {
                        if((memo_size = record_memo(path, log_fd))>0) {
#ifdef REAPER_TRACE
                            Trace("MARK %d ",(long)memo_size);
#endif
                        }
                        else {
#ifdef REAPER_TRACE
                            Trace("UNMARK %d ",(long)memo_size);
#endif
                            clear_flag = 1;
                        }
                    }
                }
                break;

            case HIST_FILE:
            case FACE_FILE:
            case NOTE_FILE:
            case HEAD_FILE:
            case BODY_FILE:
                /*
                 * check INFO is exist ?
                 */
                strcpy(ipath, path);
                RemoveExtension(ipath);
                strcat(ipath, INFO_EXT);
                if(ReadInfoRaw(ipath, &tmp_info)) {
#ifdef REAPER_TRACE
                    Trace("IGNORE ");
#endif
                    clear_flag = 1;
                }
                else {
                    if(tmp_info.attr.expire!=-1
                        && tmp_info.attr.expire < start_time) {
#ifdef REAPER_TRACE
                        Trace("EXPIRE %d ", tmp_info.attr.expire);
#endif
                        clear_flag = 1;
                    }
#ifdef REAPER_TRACE
                    Trace("SKIP ");
#endif
                }

                break;

            case LOCK_FILE:
            case REQUEST_FILE:
            case HEAD_TMP_FILE:
            case BODY_TMP_FILE:
#ifdef REAPER_TRACE
                Trace("CHECK ");
#endif
                clear_flag = IsOlder(path, &sbuf);
                break;

            default:
                /* unknown file */
                clear_flag = 1;
                break;
            }


            /*
             *
             */
            if(clear_flag) {
#ifdef REAPER_TRACE
                Trace("*CLEAR*");
#endif
                if(unlink(path)) {
                    Trace("collect_dir: cannot remove file-'%s'\n", path);

                    *file_sum += NBLK512(sbuf.st_size);
                    n_worth++;
                }
            }
            else {
#ifdef REAPER_TRACE
                Trace("*KEEP*");
#endif
                *file_sum += NBLK512(sbuf.st_size);
                n_worth++;
            }

#ifdef REAPER_TRACE
            Trace("\n");
#endif
        }
        else {
            *file_sum += NBLK512(sbuf.st_size);
#ifdef REAPER_TRACE
            Trace("U %s %d %d\n", path, (long) sbuf.st_size,
                (long)(*file_sum+*dir_sum));
#endif
        }
    }
    closedir(dp);


#ifdef REAPER_TRACE
    Trace("S %s %d %d ( %d %d ) ",
            dir, (long)*file_sum, (long)*dir_sum,
            (long)(TOMB(*file_sum)), (long)(TOMB(*dir_sum)));
#endif
    if(n_worth==0 || *file_sum+*dir_sum<=this_dir_size) {
#ifdef REAPER_TRACE
        /*
        Trace("EMPTY %d ", (long)(*file_sum+*dir_sum));
        */
        Trace("EMPTY ");
#endif
        if(rmdir(dir)) {
#ifdef REAPER_TRACE
            Trace("*FAIL* ");
#endif
            *dir_sum -= this_dir_size;
        }
        else {
#ifdef REAPER_TRACE
            Trace("*CLEAR* ");
#endif
        }
    }
    else {
#ifdef REAPER_TRACE
        /*
        Trace("NORMAL %d ", (long)(*file_sum+*dir_sum));
        */
        Trace("NORMAL ");
#endif
    }
#ifdef REAPER_TRACE
    Trace("\n");
#endif

    return 0;
}


/*
 *  collect - collect file/dir size in cache
 *
 */
int collect(char *work_file, off_t *fsum, off_t *dsum)
{
    off_t ret;
    int fd;
    struct stat sbuf;
    int len;
    int chk;
    char cache_dir[STRING_SIZE];

    memo_total = 0;
    memo_totalsize = (off_t)0;

/*
    unlink(work_file);
*/
    if((fd=open(work_file, O_WRONLY|O_CREAT|O_TRUNC, 0644))<0) {
        Error("collect: cannot open working file-'%s' (%d)",
            work_file, errno);
        goto fail;
    }
    strcpy(cache_dir, pool_dir);
#if 0
    strcat(cache_dir, "/http");
#endif
    chk = collect_dir(cache_dir, fd, fsum, dsum);
    ret = *fsum + *dsum;

    Trace("collect: cache size file %d dir %d ret %d ( %d %d %d )\n",
            (long)*fsum, (long)*dsum, (long)ret,
            (long)(TOMB(*fsum)), (long)(TOMB(*dsum)), (long)(TOMB(ret)));
    Trace("collect: record %d mark %d ( %d )\n",
            memo_total, (long) memo_totalsize,
            (long)(TOMB(memo_totalsize)));

    fstat(fd, &sbuf);
    len = (int) sbuf.st_size / sizeof(memo_t);
    Trace("collect: workfile size %d record %d memo %d\n",
        (long)sbuf.st_size, len, memo_total);
    close(fd);

    return len;

fail:
    return -1;
}




/*
 *
 */
int SenseDisk(off_t *usearea, off_t *freearea)
{
    return 0;
}



extern int eqsort(char *fn, int start, int end, size_t esize, int blen,
    void *minrec, void *maxrec, int (*comp)(const void*,const void*));




/*
 * show records for debug
 */
int show_recs(char *fn, int len)
{
    int     fd;
    int     i,c;
    off_t   sum;
    memo_t  mark;
    int     last;
    int     sorted=1;
    int     flow=1;

    if((fd=open(fn, O_RDONLY, 0644))<0) {
        Error("show_recs: cannot open working file-'%s'", fn);
        return 1;
    }

    last = INT_MAX;
    sum = (off_t)0;
    for(i=0;i<len;i++) {
        c = read(fd, &mark, sizeof(memo_t));
#if 0
        if(c<=0 || mark.num!=i+1) {
            Trace("show_recs: read error %dth rec, c %d, errno %d, num %d\n",
                i, c, errno, mark.num);
            break;
        }
#endif
        if(c<=0) {
            Trace("show_recs: read error %dth rec, c %d, errno %d\n",
                i, c, errno);
            break;
        }
        if(flow|| mark.num!=i+1) {
            Trace("show_recs: not squencial\n");
            flow = 0;
        }

        sum += mark.len;
        Trace("%5d:: %5d %5d %8d %10d %s\n",
            i, mark.num, mark.val, (long)mark.len, (long)sum, mark.name);

        if(mark.val>last)
            sorted = 0;
        last = mark.val;
    }

    if(sorted) {
        Trace("; *sorted*\n");
    }
    else {
        Trace("; *not* sorted\n");
    }
    if(flow) {
        Trace("; *sequencial*\n");
    }
    else {
        Trace("; *not* sequencial\n");
    }

    close(fd);

    return 0;
}



int scan_marks(char *fn, int len, off_t *f_sum, off_t *d_sum)
{
    int fd;
    int i, c;
    memo_t mark;

    if((fd=open(fn, O_RDONLY, 0644))<0) {
        Error("scan_marks: cannot open working file-'%s'", fn);
        return -1;
    }

    *f_sum = 0;
    for(i=0;i<len;i++) {
        c = read_record(fd, i, &mark, sizeof(memo_t));
        *f_sum += mark.len;
    }

    Trace("scan_marks: file %d dir %d mark %d\n",
        (long)*f_sum, (long)*d_sum, len);

    close(fd);

    return 0;
}


static int last_val = -1;

#define __REAPER_TRACE

/*
 * sweep - delete data which is less than EvalVal value
 */
off_t sweep(char *fn, int len, off_t limit, off_t *r_unuse)
{
    int     fd;
    int     i,c;
    off_t   sum, dsum, unsum;
    memo_t  mark;
    int     last;
    int     sorted=1;
    int     nval;
    char    path[STRING_SIZE], tmp[STRING_SIZE];
    int     ck;
    Info    info;


    Trace("sweep: records %d, size threshold %d bytes\n", len, (long)limit);

    *r_unuse = -1;

    if((fd=open(fn, O_RDONLY, 0644))<0) {
        Error("sweep: cannot open working file-'%s'", fn);
        return -1;
    }

    last_val = -1;

    last = INT_MAX;
    dsum = unsum = sum = (off_t)0;
    for(i=0;i<len;i++) {
        c = read_record(fd, i, &mark, sizeof(memo_t));
        if(c<=0) {
            Error("read error in %d", i);
            break;
        }

        /*
        Trace("* %7d %7d ;", (long)(sum+(off_t)mark.len-limit), (long)limit);
        */

        /*
         * keep
         */
        if(sum + (off_t)mark.len < limit) {
#ifdef REAPER_TRACE
            Trace("%5d:@ k %6d %4d %8d %10d %s\n",
                i, mark.num, mark.val,
                (long)mark.len, (long)(sum+(off_t)mark.len), mark.name);
#endif
            sum +=  mark.len;
        }
        /*
         * delete (keep, if it is changed)
         */
        else {
#ifdef REAPER_TRACE
            Trace("%5d:@ d %6d %4d %8d %10d %s\n",
                i, mark.num, mark.val,
                (long)mark.len, (long)(sum+(off_t)mark.len), mark.name);
#endif
            if(last_val == -1) {
                last_val = mark.val;
            }

            /* check current evaulation value */
            strcpy(path, mark.name);
            if(ReadInfoRaw(path, &info)) {
                Error(" *fail* to read %s", path);
                sum += mark.len;
            }
            else {
                nval = EvalVal(&info, start_time);
#ifdef REAPER_TRACE
                Trace("\tnew value %d ", nval);
#endif
                if(nval == mark.val) {
                    /* not changed - delete INFO, HEAD and BODY files */
                    RemoveExtension(path);
                    if(path[0]) {
                        strcpy(tmp, path);
                        strcat(tmp, INFO_EXT);
                        if((ck=unlink(tmp))<0)
                            Error("delete(%d) INFO '%s'", errno, tmp);

                        strcpy(tmp, path);
                        strcat(tmp, HEAD_EXT);
                        if((ck=unlink(tmp))<0)
                            Error("delete(%d) HEAD '%s'", errno, tmp);

                        strcpy(tmp, path);
                        if((ck=unlink(tmp))<0)
                            Error("delete(%d) BODY '%s'", errno, tmp);

                        /* count deleted */
                        dsum += mark.len;
                        /* count unused */
                        if(info.attr.init_state==REQ_PREFETCH) {
                            unsum += info.attr.length;
                        }
#ifdef REAPER_TRACE
                        Trace(" *delete*\n");
#endif
                    }
                    else {
#ifdef REAPER_TRACE
                        Trace(" *fail* to delete\n");
#endif
                        sum += mark.len;
                    }
                }
                else {
                    sum += mark.len;
#ifdef REAPER_TRACE
                    Trace(" *keep*\n");
#endif
                }
            }
        }

        if(mark.val>last)
            sorted = 0;
        last = mark.val;
    }

    close(fd);

    Trace("sweep: keep %10d, delete %10d, unuse %10d, threshold %10d ",
        (long)sum, (long)dsum, (long)unsum, last_val);
    if(sorted) {
        Trace(" *sorted*\n");
    }
    else {
        Trace(" *not* sorted\n");
    }

    *r_unuse = unsum;

    return sum;
}

#undef __REAPER_TRACE


static int in_gc=0;



int onmemory_sort(char *fname, int len, int esize,
   int (*f_comp)(const void*, const void*))
{
    void *buf;
    int fd;
    int chk;
    int nread, nwrite;

#if 0
    show_recs(fname, len);
#endif

    if((buf = (void*)MALLOC(esize*len))==NULL) {
        Trace("onmemory_sort: not enought memory (%dbytes)\n", esize*len);
        return -1;
    }

    if((fd=open(fname, O_RDWR, 0644))<0) {
        Trace("onmemory_sort: cannot open '%s' (%d)\n", fname, errno);
        goto error_out;
    }

try_read:
    if((nread=read(fd, buf, len*esize))<0) {
        Trace("onmemory_sort: cannot read %d (%d)\n", nread, errno);
        if(errno==EINTR)
            goto try_read;
        goto error_out;
    }
    else {
        Trace("onmemory_sort: read  %d\n", nread);
    }

    qsort(buf, len, esize, f_comp);

try_seek:
    if(lseek(fd, (off_t)0L, SEEK_SET)<0) {
        Trace("onmemory_sort: cannot seek to head (%d)\n", errno);
        if(errno==EINTR)
            goto try_seek;

        goto error_out;
    }

try_write:
    if((nwrite=write(fd, buf, len*esize))<0) {
        Trace("onmemory_sort: cannot write %d (%d)\n", nwrite, errno);
        if(errno==EINTR)
            goto try_write;
        goto error_out;
    }
    else {
        Trace("onmemory_sort: write %d\n", nwrite);
    }

    close(fd);

    FREE(buf);
    buf = NULL;

#if 0
    show_recs(fname, len);
#endif
    Trace("onmemory_sort: Done (len %d, read %d, write %d)\n",
        len, nread, nwrite);

    return len;

error_out:
    Log("; onmemory_sort: Fail (read %d, write %d)", nread, nwrite);
    if(buf!=NULL) {
        FREE(buf);
        buf=NULL;
    }
    return -1;
}



/*
 * GC - garbage collection via mark-sweep method
 *
 * input:
 *      work_file - work filename
 *
 * retrurn:
 *      disk usage (bytes)
 */
off_t GC(char *work_file)
{
    memo_t  min_sample, max_sample;
    time_t  t_begin, t_end;
    time_t  end_time;
    off_t   s_collect, s_sort, s_sweep, limit, ret;
    off_t   unuse;
    off_t   dsum, fsum;
    int     chk;
    int     len;

    if(in_gc==1) {
        Trace("Oops ! GC working now. Don't run many GC in same time.\n");
        return (off_t)-1;
    }
    in_gc = 1;


    /*
     * start
     */
    WCOL_UTIME(&start_time);
    DLog("GC-START %d %d %s", start_time, gc_bufsize, work_file);

    /*
     * collect
     */
    time(&t_begin);
    len = collect(work_file, &fsum, &dsum);
    time(&t_end);
    if(len<=0) {
        Error("GC: no data");
        in_gc = 0;
        return (off_t)0;
    }
    s_collect = fsum + dsum;
    Trace("GC: collect file %d bytes, dir %d bytes, total %d bytes\n",
        (long)fsum, (long)dsum, (long)s_collect);
    Trace("GC: marks %d , take %d seconds\n", len, t_end-t_begin);


    /*
     * sort
     */
    time(&t_begin);
    if(len>gc_bufsize) {
        min_sample.val = INT_MIN;
        max_sample.val = INT_MAX;
        s_sort = eqsort(work_file, 0, len-1, sizeof(memo_t),
            20, &min_sample, &max_sample, CompMemo);
    }
    else {
        Trace("GC: call onmemory sort routine\n");
        s_sort = onmemory_sort(work_file, len, sizeof(memo_t), CompMemo);
    }
    time(&t_end);
    Trace("GC: sort %d records, %d secs\n", (long)s_sort, t_end-t_begin);


    /*
     * solve threshold
     */
    limit = (off_t)floor(((double)(cache_size-dsum)*
                (double)cache_use_rate)/100.0);
    Trace("GC: cache capacity %d ( %d ), use-rate %d, threshold %d ( %d )\n",
        (long) cache_size, (long)(TOMB(cache_size)), cache_use_rate,
        (long)limit, (long)(TOMB(limit)));

    /*
     * sweep
     */
    time(&t_begin);
    s_sweep = sweep(work_file, len, limit, &unuse);
    time(&t_end);
    Trace("GC: sweep total %d unuse %d bytes , %d secs\n",
            (long)s_sweep, (long)unuse, t_end-t_begin);

    ret = s_sweep + dsum;


    /*
     * report
     */

    WCOL_UTIME(&end_time);

    Log("; GC - end %d, takes %d [%s], files %d",
        end_time, end_time - start_time, display_sec(end_time - start_time),
        (long)len);
    Log("; GC - after %dM, before %dM(%dM+%dM), unuse %dM",
        (long)TOMB(ret),
        (long)TOMB(s_collect),
        (long)TOMB(dsum), (long)TOMB(fsum), (long)TOMB(unuse));


    DLog("GC-END   %d %d %d %d %d %d %d %d",
        end_time, end_time - start_time,
        (long)ret,
        (long)s_collect, (long)dsum, (long)fsum, (long)len,
        (long)unuse);

out:
    in_gc = 0;
    return ret;

fail:
    in_gc = 0;
    return -1;
}



/*
 *
 */
RETSIGTYPE EmargencyGCTrap()
{
    Log("; reaper: EmargencyGC");
}


/*
 *
 */
RETSIGTYPE ConvensionalGCTrap()
{
    Log("; reaper: ConvensionalGC");
}


int SendReaperReport(int ear, off_t usearea, off_t freearea)
{
    rreport_t rep;
    int c;

    Log("; reaper USE %d %d FREE %d %d",
        (long)usearea, (long)TOMB(usearea), (long)freearea, (long)TOMB(freearea));

    rep.usearea  = usearea;
    rep.freearea = freearea;

    c = SendRecord(ear, &rep, sizeof(rreport_t));

    return c;
}





void Reaper(int standalone, int ear)
{
    off_t sum;
    int chk;
    off_t usearea, freearea;

    wcol_pid = getpid();
    if(!standalone && trace_flag) {
        SwitchLog(DEFAULT_DEBUGLOGDIR, "log of wcol's reaper");
    }

/*
    Trace("Reaper: hello, ear=%d\n", ear);
    SenseDisk(&usearea, &freearea);
*/

    sum = -1;

    /*
     * When manager was die, reaper should be die ,
     */
#if 0
    SetSignalHandler(SIGHUP,    ConvensionalGCTrap);
#endif
    SetSignalHandler(SIGQUIT,   ExitProgram);
    SetSignalHandler(SIGTERM,   ExitProgram);

    if(standalone) {
        SetProcessComment("standalone reaper");

        SetSignalHandler(SIGINT,    ExitProgram);

        sum = GC(gc_workfile);
        Trace("Reaper: use space %d bytes\n", (long)sum);
    }
    else {
        SetProcessComment("reaper");

        SetSignalHandler(SIGINT,    ExitProgram);
#if 0
        SetSignalHandler(SIGINT,    EmargencyGCTrap);
#endif

        while(!standalone) {
            sum = GC(gc_workfile);

            /* report cache state */
            if(ear!=-1) {
                chk = SendReaperReport(ear, sum, -1);
                Trace("Reaper: use space %d bytes (%d)\n", (long)sum, chk);
            }

            sleep(gc_interval);
        }
    }

    exit(WCOL_EXIT_NORMAL);
}




#ifdef REAPER_TEST


#if 1
pid_t wcol_pid;
#endif

extern char config_file[];

void Init(int argc, char **argv)
{
    SetGlobalDefaults();
    if(getenv(DEFAULT_CONFIG_ENVNAME)!=NULL) {
        strcpy(config_file, getenv(DEFAULT_CONFIG_ENVNAME));
    }
    ReadConfigFile(config_file);
    ReadCommandlineOptions(argc, argv);

    FixGlobals();

    trace_flag      = 1;

    Trace("Init: pool_dir '%s'\n", pool_dir);

    InitLogger("reaper");   /* generate differenet log */
}


main(int argc, char **argv)
{
    Init(argc, argv);

    Reaper(1, -1);
}

#endif
