/*
 *  WCOL: Accept request and parse procedures.
 *      by k-chinen@is.aist-nara.ac.jp, 1995, 1996
 *
 *  $Id: request.c,v 1.6 1996/12/01 20:35:41 k-chinen Exp k-chinen $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <ctype.h>
#include <time.h>
#include <setjmp.h>
#include <signal.h>

#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif


/*
 * Solaris's accept() is strange
 */

#if defined(__sun)
#ifndef USE_ACCEPT_LOCK
#define USE_ACCEPT_LOCK
#endif
#endif /* __sun */


#include "wcol.h"
#include "sigs.h"
#include "fds.h"
#include "accept.h"
#include "request.h"

#define _DEBUG_REQUEST
#define _DEBUG_REQUEST_DETAIL


char *_req_method_str[4] = {"UNKNOW", "GET", "HEAD", "POST"};
char *_req_state_str[6]  = {"UNKNOW", "EMPTY", "NOP", "LOCK", "BUSY", "DONE"};
char *_req_mode_str[4]   = {"UNKNOW", "CLIENT", "PREFETCH", "IMPORT"};
char _req_mode_char[4]   = "UCPI";


int reception_timeout=RECEPTION_TIMEOUT;


char *mk_sock_in_str(char *buf, struct sockaddr_in *sock)
{
    char tmp[STRING_SIZE];
    char *p_addr;

    buf[0] = '\0';

    p_addr = (char*)&(sock->sin_addr.s_addr);
    sprintf(tmp, "%d.%d.%d.%d", p_addr[0], p_addr[1], p_addr[2], p_addr[3]);

    sprintf(buf, "addr %s, port %d", tmp, sock->sin_port);

    return buf;
}

char *mk_sock_in_host_port(char *buf, int *port, struct sockaddr_in *sock)
{
    char *p_addr;

    p_addr = (char*)&(sock->sin_addr.s_addr);
    sprintf(buf, "%d.%d.%d.%d", p_addr[0], p_addr[1], p_addr[2], p_addr[3]);
    *port = sock->sin_port;

    return buf;
}



/*
 *
 */
void ShowRequest(request_t *req)
{
    char buf[STRING_SIZE];

    fprintf(stdout, "--- Request ---\n");
    fprintf(stdout, "State:     %s\n", ReqStateStr(req->state));
    fprintf(stdout, "Mode:      %s\n", ReqModeStr(req->mode));
    fprintf(stdout, "Level:     %d\n", req->level);
    fprintf(stdout, "Sid:       %d\n", req->sid);
    fprintf(stdout, "Id:        %d\n", req->id);
    fprintf(stdout, "Fd:        %d\n", req->fd);
#if 0
    fprintf(stdout, "Time:      %9d ; %s",
        req->issue_time,    req->issue_time<0 ? "\n":ctime(&req->issue_time));
#endif
    fprintf(stdout, "Time:      %10d.%06d ; %s",
        req->issue_time.tv_sec, req->issue_time.tv_usec,
        req->issue_time.tv_sec<0 ? "\n":ctime(&req->issue_time.tv_sec));

#if 0
    mk_sock_in_str(buf, (struct sockaddr_in*) &req->from);
    fprintf(stdout, "From:      %s\n", buf);
    mk_sock_in_str(buf, (struct sockaddr_in*) &req->here);
    fprintf(stdout, "Here:      %s\n", buf);
#endif
    fprintf(stdout, "Peer:      %s\n", req->peername);
    fprintf(stdout, "Port:      %9d\n", req->hereport);

    fprintf(stdout, "Method:    %s\n", ReqMethodStr(req->method));
    fprintf(stdout, "URL:       %s\n", req->url);
    fprintf(stdout, "Version:   %d.%d\n", req->ver_major, req->ver_minor);
    fprintf(stdout, "KeepAlive: %s\n", req->keep_alive ? "YES":"NO");
    fprintf(stdout, "Auth:      '%s'\n", req->auth);
    fprintf(stdout, "Cookie:    '%s'\n", req->cookie);
    fprintf(stdout, "IMS:       %9d ; %s",
        req->ims,           req->ims<0 ? "\n":ctime(&req->ims));
    fprintf(stdout, "Nocache:   %s\n", req->nocache ? "FORCE":"CHECK");
    fprintf(stdout, "Noprefetch:   %s\n", req->noprefetch ? "FORCE":"CHECK");

    fflush(stdout);
}



/*
 * Request limitor
 */
static sigjmp_buf request_jmp_point;

static RETSIGTYPE giveup_reception()
{
    SIGLONGJMP(request_jmp_point, 1);
}





/*
 * AcceptRequest - accept requst
 *
 * return
 *      >0  success (file descriptor)
 *      -1  error
 *
 */
int AcceptRequest(int reqd)
{
    int fd;
    struct sockaddr_in from;
    int len;
    unsigned long int from_addr;
    unsigned char *target;
    int ac_errno;
    int chk;
    struct timeval now;


#ifdef TRACE_NEWREQUEST
    Trace("AcceptRequest: fd# %d ", reqd);
#endif

ac_start:
#ifdef USE_ACCEPT_LOCK
    if(lock_port(reqd)) {
        return -1;
    }
#endif

    if(SIGSETJMP(request_jmp_point, 1)) {
        Error("AcceptRequest: long waitting of accept() #%d, give up.", fd);
        SetSignalHandler(SIGALRM, SIG_IGN);

#ifdef USE_ACCEPT_LOCK
        unlock_port(reqd);
#endif

        return -1;
        /*
        goto done;
        */
    }

    SetSignalHandler(SIGALRM, giveup_reception);
    alarm(reception_timeout);


    /* not readable or error */
    if((chk=isreadable(reqd, 0, 500000))<=0) {
        SetSignalHandler(SIGALRM, SIG_IGN);

#ifdef USE_ACCEPT_LOCK
        unlock_port(reqd);
#endif
        return -1;
    }


    len = sizeof(from);
acc:
    ac_errno = 0;
    if((fd=accept(reqd, (struct sockaddr*)&from, &len)) < 0 ) {
        ac_errno = errno;
        if(errno == EINTR) {
            goto acc;
        }
    }

done:
    SetSignalHandler(SIGALRM, SIG_IGN);


#ifdef USE_ACCEPT_LOCK
    unlock_port(reqd);
#endif


    if(ac_errno) {
        if(ac_errno == EMFILE) {
            Error("AcceptRequest: too many open file. the process file table is full.");
        }
        else
        if(ac_errno == ENFILE) {
            Error("AcceptRequest: too many open file. the system file table is full.");

        }
        else {
            Error("AcceptRequest: cannot accept request (errno=%d).", ac_errno);
        }

        return -1;
    }

ac_ok:

    /*
        Don't warry. Source address of socket information is expressd
        with network byte-order.
    */
    target = (unsigned char*) &from.sin_addr.s_addr;

    if(!Acceptable(target)) {
        gettimeofday(&now, NULL);

        show_addr("Ignore Access from", target);

        shutdown(fd, 2);
        close(fd);

        Log("%d.%06d IGNORE %d.%d.%d.%d (%02x.%02x.%02x.%02x)",
            now.tv_sec, now.tv_usec,
            target[0], target[1], target[2], target[3],
            target[0], target[1], target[2], target[3]);

        return -1;
    }

    Trace("AcceptRequest: new request fd #%d,", fd);
    show_addr("addr. ", target);

    return fd;
}


int WaitRequest(int reqd)
{
    int ret;

    while((ret=AcceptRequest(reqd))<0) {
    }

    return ret;
}




/*
 *
 */
int CheckSupportedRequest(request_t req)
{
    char protocol[STRING_SIZE];
    char host[STRING_SIZE], port[STRING_SIZE], path[STRING_SIZE];
    int  chk;
    int  ret = 0;

/*
    Trace("CheckSupportedRequest: *** url='%s'\n", req.url);
*/
    /*
     * check URL
     */
    chk = ParseURL(req.url, protocol, host, port, path);

#if 0
    Trace("CheckSupportedRequest: protocol='%s' http? %d, HTTP? %d\n",
        protocol,   strncmp_first_all(protocol, 4, 'h', "http"),
                    strncmp_first_all(protocol, 4, 'H', "HTTP"));
#endif

#if 0
    if(protocol[0]=='\0' ||
            (Nstrncmp_first_all(protocol,4,'h',"http") &&
            Nstrncmp_first_all(protocol,4,'H',"HTTP")))
        ret = 1;
#else
    if(protocol[0]!='\0') {
        if(strncmp_first_all(protocol, 4, 'h', "http") ||
            strncmp_first_all(protocol, 4, 'H', "HTTP")) {
                /* nothing */
        }
        else {
            ret = 1;
        }
    }
    else {
        ret = 1;
    }
#endif


    if(host[0]=='\0')
        ret = 1;
    if(path[0]=='\0')
        ret = 1;

    if(ret) {
        Trace("CheckSupportedRequest: unsupport request\n");
        Trace("\turl='%s'\n", req.url);
        Trace("\tprotocol='%s', host='%s', port='%s'\n",
            protocol, host, port);
        Trace("\tpath='%s'\n", path);
    }
/*
    Trace("CheckSupportedRequest: url='%s' -> %d\n", req.url, ret);
*/

    return ret;
}





/*
 * NewRequest - accept connection, parse request and make actual request.
 */
int NewRequest(int http_servd, request_t *req)
{
    int  fd;
    int  chk;
    char buf[BUF_SIZE];

    if((fd=AcceptRequest(http_servd))<0) {
        Trace("NewRequest: ignore access\n");
        return 1;
    }

    if(SIGSETJMP(request_jmp_point, 1)) {
        Error("NewRequest: long waitting #%d, give up.", fd);
        close(fd);
        return 1;
    }
    SetSignalHandler(SIGALRM, giveup_reception);
    alarm(10);      /* request must be complete within 10 secs */

    chk = HTTP_ParseRequest(fd, req, buf);
    SetSignalHandler(SIGALRM, SIG_IGN);

    /* error or empty */
    if(chk == 1 || chk == -1) {
        Error("NewRequest: ignore request from fd #%d", fd);

        HTTP_ErrorMessage_Old(fd,
            501, "Not implemented", "Unsupport Command", req->ver_major>0);
        close(fd);

        return 1;
    }
    else
    /* internal error */
    if(chk == -2) {
        Error("NewRequest: internal error, check disk space or permissions");

        HTTP_ErrorMessage_Old(fd,
            500, "Internal Error",
            "Check system now. please wait several minites", req->ver_major>0);
        close(fd);

        return -1;
    }

    if(trace_flag) {
        Trace("NewRequest:\n");
        ShowRequest(req);
    }

    return 0;
}




