/*
 *  WCOL: request procedures
 *      by k-chinen@is.aist-nara.ac.jp, 1994, 1995, 1996
 *
 *  $Id: request.h,v 1.1 1995/10/09 10:34:47 k-chinen Exp k-chinen $
 */

#ifndef WCOL_REQUEST_H
#define WCOL_REQUEST_H


/* state */
#define REQ_UNKNOWSTATE (0)
#define REQ_EMPTY       (1)
#define REQ_NOP         (2)
#define REQ_LOCK        (3)
#define REQ_BUSY        (4)
#define REQ_DONE        (5)

/* mode */
#define REQ_UNKNOWMODE  (0)
#define REQ_CLIENT      (1)
#define REQ_PREFETCH    (2)
#define REQ_IMPORT      (3)

/* method */
#define METHOD_UNKNOW   (0)
#define METHOD_GET      (1)
#define METHOD_HEAD     (2)
#define METHOD_POST     (3)


/*
 * type of request
 */
typedef struct {
    int state;                  /* state - wake or sleep */
    int mode;                   /* mode  - prompt handling or not */
    int level;                  /* level - retrieval level */
    int sid;                    /* sid   - session id */
    int id;                     /* id    - reqeust id */
    int fd;                     /* fd    - file descriptor for fd-passing */
    struct timeval issue_time;  /* issue_time - accept or issue time */

#if 0
    struct sockaddr_in from;
    struct sockaddr_in here;
#endif
    char peername[20];
    int  hereport;


    /* HTTP/0 */
    int  method;                /* method - GET, HEAD or POST */
    char url[INFONAMESIZE];
    char ver_major;
    char ver_minor;
    char keep_alive;
    /*
    char refer[INFONAMESIZE];
    */

    /* HTTP/1.0 */
    char auth[STRING_SIZE];
    char cookie[STRING_SIZE];
    time_t ims;
    int  nocache;
    int  noprefetch;
} request_t;




#define ReqMethodStr(v) ((v>=METHOD_GET&&v<=METHOD_POST) ? \
                    _req_method_str[v] : _req_method_str[METHOD_UNKNOW])
#define ReqStateStr(v)  ((v>=REQ_UNKNOWSTATE&&v<=REQ_DONE) ? \
                    _req_state_str[v] : _req_state_str[REQ_UNKNOWSTATE])
#define ReqModeStr(v)   ((v>=REQ_UNKNOWMODE&&v<=REQ_IMPORT) ? \
                    _req_mode_str[v] : _req_mode_str[REQ_UNKNOWMODE])
#define ReqModeChar(v)   ((v>=REQ_UNKNOWMODE&&v<=REQ_IMPORT) ? \
                    _req_mode_char[v] : _req_mode_char[REQ_UNKNOWMODE])



void ShowRequest(request_t *req);

extern char *_req_method_str[];
extern char *_req_state_str[];
extern char *_req_mode_str[];
extern char _req_mode_char[];

#endif


