/*
 * WCOL:
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <time.h>
#ifdef TIME_WITH_SYS_TIME
#include <sys/time.h>
#endif
#include <errno.h>

#include "wcol.h"

#include "fds.h"

#include "request.h"
#include "http.h"
#include "icp.h"
#include "args.h"
#include "accept.h"
#include "conv.h"
#include "manager.h"
#include "handler.h"


/*
 * externals
 */

/* automatic */
extern             char config_file[];
extern              int trace_flag;
extern             char log_file[];
extern             char pool_dir[];
extern            off_t cache_size;
extern              int cache_use_rate;
extern           time_t cache_keeptime;
extern           time_t cache_ignorekeeptime;
extern              int gc_interval;
extern             char gc_workfile[];
extern              int gc_bufsize;
extern             char local_domain[];
extern             char dummy_string[];

/* static */
extern pid_t    wcol_pid;
extern char     *progname;

/* end of externals */

char config_file[STRING_SIZE];
char dummy_string[STRING_SIZE];


/*
 * SetGlobalDefaults -
 */
void SetGlobalDefaults()
{
    strcpy(config_file, DEFAULT_CONFIGFILE);
    trace_flag = 0;
    strcpy(log_file, DEFAULT_LOGFILE);
    strcpy(pool_dir, DEFAULT_POOLDIR);
    cache_size = (off_t) 100*(1024*1024);
    cache_use_rate = (int) 70;
    cache_keeptime = (time_t) CACHEKEEPTIME;
    cache_ignorekeeptime = (time_t) CACHEIGNOREKEEPTIME;
    gc_interval = (int) 3600;
    strcpy(gc_workfile, DEFAULT_GCWORKFILE);
    gc_bufsize = (int) 1000;
    strcpy(local_domain, "");
    strcpy(dummy_string, "");
}
/* end of SetGlobalDefaults() */





#define MAX_AV 10

/*
 * ReadConfigFile -
 */
void ReadConfigFile(char *filename)
{
    arg_t  *args, *pargs;
    int     num;
    int     tint;
    char   *av[MAX_AV];
    int     i, j;


    Trace("ReadConfigFile: filename='%s'\n", filename);
    if((num=ReadArgsFromFile(filename, &args))<=0) {
        Trace("ReadConfigFile: no arguments\n");
        return;
    }

    pargs = args;
    while(pargs) {
        /* project argument value to temporary pointor */
        for(i=0;pargs->vals[i] && i<MAX_AV;i++) {
            av[i] = pargs->vals[i];
        }
        for(j=i;j<MAX_AV;j++) {
            av[j] = NULL;
        }

        if(!pargs->name) {
            /* nothing */
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'C', "ConfigFile")) {
            if(pargs->vals[0]) {
                strcpy(config_file, av[0]);
                goto nextone;
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'T', "Trace")) {
            trace_flag = 1;
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'L', "LogFile")) {
            if(pargs->vals[0]) {
                strcpy(log_file, av[0]);
                goto nextone;
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'C', "CacheDir")) {
            if(pargs->vals[0]) {
                strcpy(pool_dir, av[0]);
                goto nextone;
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'C', "CacheSize")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    cache_size = (off_t) tint * 1024*1024;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'C', "CacheUseRate")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    cache_use_rate = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'K', "KeepTime")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    cache_keeptime = (time_t) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'N', "NegativeKeepTime")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    cache_ignorekeeptime = (time_t) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'G', "GCInterval")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    gc_interval = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'G', "GCWorkFile")) {
            if(pargs->vals[0]) {
                strcpy(gc_workfile, av[0]);
                goto nextone;
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'G', "GCBufSize")) {
            if(pargs->vals[0]) {
                if((tint = atoi(av[0]))>0) {
                    gc_bufsize = (int) tint * 1;
                    goto nextone;
                }
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'L', "LocalDomain")) {
            if(pargs->vals[0]) {
                strcpy(local_domain, av[0]);
                goto nextone;
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'D', "Dummy")) {
            if(pargs->vals[0]) {
                strcpy(dummy_string, av[0]);
                goto nextone;
            }
        }
        else
        if(strcasecmp_first_all(pargs->name, 'P', "ProxyServer")) {
            /* nothing */
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'C', "Catalyst")) {
            /* nothing */
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'O', "Oscillator")) {
            /* nothing */
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'P', "Prefetch")) {
            /* nothing */
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'C', "CacheOnly")) {
            /* nothing */
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'M', "MemoryCache")) {
            /* nothing */
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'N', "NoMemoryCache")) {
            /* nothing */
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'R', "RemoveCookie")) {
            /* nothing */
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'N', "NeedCookie")) {
            /* nothing */
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'R', "RemoveReferer")) {
            /* nothing */
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'N', "NeedReferer")) {
            /* nothing */
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'N', "NoReaper")) {
            /* nothing */
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'F', "FetchParent")) {
            /* nothing */
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'P', "PrefetchParent")) {
            /* nothing */
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'I', "ImportSibling")) {
            /* nothing */
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'S', "StateFile")) {
            /* nothing */
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'P', "PidFile")) {
            /* nothing */
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'F', "FTPUser")) {
            /* nothing */
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'F', "FTPPass")) {
            /* nothing */
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'H', "HTTPPort")) {
            /* nothing */
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'I', "ICPPort")) {
            /* nothing */
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'M', "MaxHandler")) {
            /* nothing */
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'N', "NumHandler")) {
            /* nothing */
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'M', "MaxPrompter")) {
            /* nothing */
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'N', "NumPrompter")) {
            /* nothing */
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'M', "MaxSpeaker")) {
            /* nothing */
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'N', "NumSpeaker")) {
            /* nothing */
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'M', "MaxSession")) {
            /* nothing */
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'P', "PrefetchIncluded")) {
            /* nothing */
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'P', "PrefetchAnchor")) {
            /* nothing */
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'P', "PrefetchLagLimit")) {
            /* nothing */
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'S', "SessionTimeout")) {
            /* nothing */
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'S', "SessionLifetime")) {
            /* nothing */
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'Q', "QueryTimeout")) {
            /* nothing */
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'R', "ReceptionTimeout")) {
            /* nothing */
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'A', "AcceptAddress")) {
            /* nothing */
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'C', "ConvertType")) {
            /* nothing */
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'S', "Sibling")) {
            /* nothing */
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'F', "FieldRemove")) {
            /* nothing */
            goto nextone;
        }
        else
        if(strcasecmp_first_all(pargs->name, 'F', "FieldChange")) {
            /* nothing */
            goto nextone;
        }
        else {
            Error("Unknow argument '%s'", pargs->name);
            goto nextone;
        }
fail:
        Error("Argment specification ('%s') is wrong", pargs->name);
nextone:
        pargs = pargs->next;
    }

    FreeArgs(args);
}



/*
 * FixGlobals - Fix global vars.
 */
void FixGlobals()
{

}





ShowGlobals(FILE *s)
{
    fprintf(s,
"======================================================================\n\
Flag Name                CurrentValue\n\
----------------------------------------------------------------------\n");

    fprintf(s, "  f  ConfigFile          '%s'\n",
        config_file);
    fprintf(s, "  t  Trace                %s\n",
        trace_flag==1 ? "*ON*" : "*OFF*" );
    fprintf(s, "  l  LogFile             '%s'\n",
        log_file);
    fprintf(s, "  d  CacheDir            '%s'\n",
        pool_dir);
    fprintf(s, "  C  CacheSize           %8d Mbytes (x 1024*1024)\n",
        cache_size/(1024*1024));
    fprintf(s, "  R  CacheUseRate        %8d \n",
        cache_use_rate);
    fprintf(s, "  K  KeepTime            %8d sec [%s]\n",
        cache_keeptime, display_sec(cache_keeptime));
    fprintf(s, "  N  NegativeKeepTime    %8d sec [%s]\n",
        cache_ignorekeeptime, display_sec(cache_ignorekeeptime));
    fprintf(s, "  G  GCInterval          %8d \n",
        gc_interval);
    fprintf(s, "  -  GCWorkFile          '%s'\n",
        gc_workfile);
    fprintf(s, "  -  GCBufSize           %8d \n",
        gc_bufsize);
    fprintf(s, "  -  LocalDomain         '%s'\n",
        local_domain);
    fprintf(s, "  z  Dummy               '%s'\n",
        dummy_string);
}


ShowProgramStatus(FILE *outport)
{
    fprintf(outport, "Wcol-E (pre-release) program-version %s, data-id %s\n",
        WCOL_VER, WCOL_INFO_ID);

    fflush(outport);
}

ShowUsage()
{
    FixGlobals();   /* check in temporary */
    ShowProgramStatus(stdout);
    ShowGlobals(stdout);
}

int SetProcessComment(char *comment)
{
    char line[STRING_SIZE];

    sprintf(line, "wcol : %s", comment);
    return SetProcessState(line);
}


/*
 * ReadCommandlineOption
 */
void ReadCommandlineOptions(int argc, char **argv)
{
    int flag;

    extern char *optarg;
    extern int  optind;

    while((flag = getopt(argc, argv,
        "hvf:tl:d:C:R:K:N:G:z:")) != -1) {

        switch(flag) {
        case 'h':
            ShowUsage();
            exit(WCOL_EXIT_NORMAL);
            break;
        case 'v':
            ShowProgramStatus(stdout);
            exit(WCOL_EXIT_NORMAL);
            break;
        case 'f':
            strcpy(config_file, optarg);
            ReadConfigFile(config_file);
            break;
        case 't':
            trace_flag = 1 - trace_flag;
            break;
        case 'l':
            strcpy(log_file, optarg);
            break;
        case 'd':
            strcpy(pool_dir, optarg);
            break;
        case 'C':
            cache_size = (off_t) atoi(optarg) * 1024*1024;
            break;
        case 'R':
            cache_use_rate = (int) atoi(optarg);
            break;
        case 'K':
            cache_keeptime = (time_t) atoi(optarg);
            break;
        case 'N':
            cache_ignorekeeptime = (time_t) atoi(optarg);
            break;
        case 'G':
            gc_interval = (int) atoi(optarg);
            break;
        case 'z':
            strcpy(dummy_string, optarg);
            break;
        }
    }
}



