/*
 *  WCOL: request scheduling routines
 *      by k-chinen@is.aist-nara.ac.jp, 1995, 1996, 1997
 *
 *  $Id: schedule.c,v 1.4 1996/11/24 14:48:53 k-chinen Exp k-chinen $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <limits.h>
#include <time.h>
#ifdef TIME_WITH_SYS_TIME
#include <sys/time.h>
#endif
#include <signal.h>
#include <setjmp.h>
#include <errno.h>

#include "wcol.h"
#include "msg.h"
#include "request.h"
#include "sigs.h"
#include "ipc.h"
#include "fds.h"
/*
#include "global.h"
*/

#include "handler.h"
#include "session.h"
#include "prompter.h"
#include "reaper.h"
#include "manager.h"
#include "hstate.h"


#ifndef DISPATCH_TIMEOUT
#define DISPATCH_TIMEOUT            (5)         /* 5 sec */
#endif

#ifndef PREFETCH_DISPATCH_TIMEOUT
#define PREFETCH_DISPATCH_TIMEOUT   (3*60*1000) /* 3 min */
#endif

#define _TRACE_DISPATCH
#define _TRACE_REMOVE_REQUEST
#define _TRACE_APPEND_REQUEST

#ifndef MAX_REQUEST_QUEUE
#define MAX_REQUEST_QUEUE (64)
#endif



/*
extern int prompter_fd;
extern int prompter_pid;
extern int reaper_fd;
extern int reaper_pid;
*/

extern void ReplaceHandler(int n);
extern void StopHandler(int n);
extern time_t manager_ref_time;


extern int max_speaker,  num_speaker;
extern int max_prompter, num_prompter;
extern int max_handler,  num_handler;



/* global */

int n_active=0, n_busy=0, n_client=0, n_prefetch=0;

int n_nop_req = 0, n_busy_req = 0, n_empty_req = 0, n_lock_req = 0;
int n_client_req = 0, n_prefetch_req = 0, n_import_req = 0;

int rhead=0, rtail=0;
int request_queue_length=MAX_REQUEST_QUEUE;
request_t *rque = (request_t *)NULL;


#if 0
time_t prefetch_dispatch_waittime=60*30;        /* half hour */
#else
time_t prefetch_dispatch_waittime=PREFETCH_DISPATCH_TIMEOUT;
#endif


extern int session_timeout;
extern int prefetch_lag_limit;


void check_rque()
{
    int i,c,t;
#if 0
    time_t now;

    time(&now);
#endif
    int diff;
    struct timeval cur;

    gettimeofday(&cur, NULL);
    n_nop_req = n_busy_req = n_empty_req = n_lock_req = 0;
    n_client_req = n_prefetch_req = n_import_req = 0;

    /* remove old prefetch requests */
    c = 0;

    if(rhead<=rtail)
        t = rtail;
    else
        t = MAX_REQUEST_QUEUE + rtail;

    for(i=rhead;i<t;i++) {
        diff = (int)WCOL_SUB_TIME(cur, rque[i].issue_time);
#if 0
        if(rque[i].state == REQ_NOP &&
            (rque[i].mode == REQ_PREFETCH || rque[i].mode == REQ_IMPORT)) {
            if(WCOL_SUB_TIME(cur, rque[i].issue_time)
                > prefetch_dispatch_waittime)
            {

                remove_request(rque[i].sid, rque[i].id);
                c++;
            }
        }
#else
        /* any very old request */
        if(diff>session_timeout) {
            goto clear_next;
        }

        /* old prefetch request */
        if((rque[i].mode == REQ_PREFETCH || rque[i].mode == REQ_IMPORT)
            && diff>prefetch_dispatch_waittime) {
            goto clear_next;
        }

#if 0
        /* old prefetch request */
        if((rque[i].mode == REQ_PREFETCH || rque[i].mode == REQ_IMPORT)
            && diff>prefetch_lag_limit) {
            goto clear_next;
        }
#endif

        goto next;

clear_next:
        remove_request(rque[i].sid, rque[i].id);
        c++;

next:
        (void)c;    /* dummy */

#endif
    }

    if(c>0) {
#ifdef TRACE_REMOVE_REQUEST
        Log("; remove old request %d", c);
#endif
    }

    /* count request states in queue */
#if 0
    for(i=0;i<MAX_REQUEST_QUEUE;i++)
#endif

    if(rhead<=rtail)
        t = rtail;
    else
        t = MAX_REQUEST_QUEUE + rtail;

    for(i=rhead;i<t;i++) {
        if(rque[i].state == REQ_EMPTY)
            n_empty_req++;
        if(rque[i].state == REQ_NOP)
            n_nop_req++;
        if(rque[i].state == REQ_LOCK)
            n_lock_req++;
        if(rque[i].state == REQ_BUSY)
            n_busy_req++;

        if(rque[i].mode == REQ_CLIENT) {
            n_client_req++;
        }
        if(rque[i].mode == REQ_PREFETCH) {
            n_prefetch_req++;
        }
        if(rque[i].mode == REQ_IMPORT) {
            n_import_req++;
        }
    }
}


void show_rque()
{
    int i,c,t;
#if 1
    struct timeval cur;

    gettimeofday(&cur, NULL);
#endif

    if(rhead<=rtail)
        t = rtail;
    else
        t = MAX_REQUEST_QUEUE + rtail;

#if 0
    Trace("show_rque: head %d, tail %d (rtail %d)\n", rhead, t, rtail);
#endif
    Trace("show_rque: rhead %d, rtail %d -> %d..%d\n", rhead, rtail, rhead, t);

#if 0
    Trace("raw %4s: %5s %3s %s %7s %s\n",
        "#","i","fd","s","method","current-or-last-url");
    for(i=0;i<MAX_REQUEST_QUEUE;i++)
        Trace("raw %4d: %5d %3d %d %7s %s\n",
            i, rque[i].id, rque[i].fd, rque[i].state,
            rque[i].method, rque[i].url);
#endif

#if 1
    /***
     *** laptime
     ***/
    Trace("\tsession_timeout            %9d\n", session_timeout);
    Trace("\tprefetch_dispatch_waittime %9d\n", prefetch_dispatch_waittime);
#if 0
    Trace("\tprefetch_lag_limit         %9d\n", prefetch_lag_limit);
#endif
    /*      1234 12-3 1 1 12345 12345 123456 123456 123... */
    Trace(";  #   %%  M S  SID   ID    TIME   METHOD URL\n");
    c = 0;
    for(i=rhead;i<t;i++) {
        Trace(" %4d %3d %c %c %5d/%5d %6d %-6s %s\n",
            i, i%MAX_REQUEST_QUEUE,
            /*
            rque[i].mode==REQ_PREFETCH ? 'P':'C',
            */
            ReqModeChar(rque[i].mode),
            (rque[i].state==REQ_EMPTY ? '-':
             (rque[i].state==REQ_NOP ? 'N':
              (rque[i].state==REQ_LOCK ? 'L':
               (rque[i].state==REQ_BUSY ? 'B':
                (rque[i].state==REQ_DONE ? 'D':'?'))))),
            rque[i].sid, rque[i].id,
            WCOL_SUB_TIME(cur, rque[i].issue_time),
            ReqMethodStr(rque[i].method),
            rque[i].url);
        c++;
    }
#else
    /***
     *** FD
     ***/
    /*      1234 12-3 1 1 12345 12345 1234 123456 123... */
    Trace(";  #   %%  M S  SID   ID    FD  METHOD URL\n");
    c = 0;
    for(i=rhead;i<t;i++) {
        Trace(" %4d %3d %c %c %5d/%5d %4d %-6s %s\n",
            i, i%MAX_REQUEST_QUEUE,
            /*
            rque[i].mode==REQ_PREFETCH ? 'P':'C',
            */
            ReqModeChar(rque[i].mode),
            (rque[i].state==REQ_EMPTY ? '-':
             (rque[i].state==REQ_NOP ? 'N':
              (rque[i].state==REQ_LOCK ? 'L':
               (rque[i].state==REQ_BUSY ? 'B':
                (rque[i].state==REQ_DONE ? 'D':'?'))))),
            rque[i].sid, rque[i].id,
            rque[i].fd,
            ReqMethodStr(rque[i].method),
            rque[i].url);
        c++;
    }
#endif

    Trace("; %d..%d -> total %d request(s)\n", rhead, rtail, c);

    check_rque();
    Trace("; %10d.%06d empty %d nop %d busy %d lock %d\n",
        cur.tv_sec, cur.tv_usec,
        n_empty_req, n_nop_req, n_busy_req, n_lock_req);
    Trace("; %10d.%06d client %d, prefetch %d, import %d\n",
        cur.tv_sec, cur.tv_usec,
        n_client_req, n_prefetch_req, n_import_req);

}



int init_requestqueue()
{
    int i;

    if(rque == NULL) {
        Trace("init_requestqueue: malloc %d bytes\n",
            sizeof(request_t)*MAX_REQUEST_QUEUE);
        rque = (request_t*)MALLOC(sizeof(request_t)*MAX_REQUEST_QUEUE);
        if(rque == NULL ) {
            Error("init_requestqueue: not enoght memory (%d bytes)",
                sizeof(request_t)*MAX_REQUEST_QUEUE);
            return -1;
        }
    }

    for(i=0;i<MAX_REQUEST_QUEUE;i++) {
        rque[i].state = REQ_EMPTY;
    }

    return 0;
}

int deinit_requestqueue()
{
    if(rque != NULL) {
        FREE(rque);
        rque = NULL;
    }
    return 0;
}


/*
 * append_request
 */
int append_request(request_t req)
{
    int ptail, limit;

    if(rhead>rtail) {
        ptail = MAX_REQUEST_QUEUE+rtail;
    }
    else {
        ptail = rtail;
    }
    limit = MAX_REQUEST_QUEUE+rhead - 1;


#ifdef TRACE_APPEND_REQUEST
    Trace("append_request: request #%d/#%d, fd #%d\n", req.sid, req.id, req.fd);
    Trace("\thead %d, tail %d, max %d -> head %d, tail' %d, limit %d\n",
        rhead, rtail, MAX_REQUEST_QUEUE, rhead, ptail, limit);
    /*
    show_rque();
    */
#endif
    if(ptail<limit) {
        rque[rtail] = req;
        rtail++;
        rtail = rtail%MAX_REQUEST_QUEUE;

#ifdef TRACE_APPEND_REQUEST
        Trace("append_request: add -> head %d, tail %d\n", rhead, rtail);
        /*
        show_rque();
        */
#endif

        return 0;
    }
    else {
#if 0
#ifdef TRACE_APPEND_REQUEST
        Trace("append_request: over flow\n");
#endif
#endif
        Log("; request queue is over flow");
    }

    return 1;
}

int append_uniq_request(request_t req)
{
    int i;
    int limit;

    limit = MAX_REQUEST_QUEUE+rhead-1;

#ifdef TRACE_APPEND_REQUEST
    Trace(
"append_uniq_request: request #%d/#%d, fd #%d, head %d, tail %d, limit %d\n",
        req.sid, req.id, req.fd, rhead, rtail, limit);
#endif

#if 0
    Log("; request #%d/#%d '%s'", req.sid, req.id, req.url);
#endif
    Trace("\t; #%d: '%s'\n", req.id, req.url);

    /* check same request */
    for(i=0;i<MAX_REQUEST_QUEUE;i++) {
        if(rque[i].url[0]==req.url[0] && strcmp(rque[i].url,req.url)==0) {
            Trace("append_uniq_request: found same request in %d\n",
                req.id, rque[i].id);

            return -1;
        }
    }

/*
    show_rque();
*/
    if(rtail<limit) {
        rque[rtail%MAX_REQUEST_QUEUE] = req;
        rtail++;
        rtail = rtail%MAX_REQUEST_QUEUE;

        /*
        show_rque();
        */

        return 0;
    }
#if 1
    Log("; request queue: head %d, tail %d, limit %d\n", rhead, rtail, limit);
#endif

    return 1;
}

request_t *take_request(int rid)
{
    int i;

    for(i=0;i<MAX_REQUEST_QUEUE;i++) {
        if(rque[i].sid==rid) {
            return &rque[i];
        }
    }

    return NULL;
}



/*
 *
 * return
 *       0  success (or no entry)
 *      -1  error
 */

int remove_request(int rsid, int rid)
{
    int i, j;
    int h, t;
    int pos;

#ifdef TRACE_REMOVE_REQUEST
    Trace("remove_request: remove sid#/id# %d/%d\n", rsid, rid);
    Trace("remove_request: before  rhead %4d, rtail %4d\n", rhead, rtail);
    /*
    show_rque();
    */
#endif

    if(rsid==0) {
#ifdef TRACE_REMOVE_REQUEST
        Log("; remove_request: rsid# %d", rsid);
#endif
        return 0;
    }
    if(rid==0) {
#ifdef TRACE_REMOVE_REQUEST
        Log("; remove_request: rid#  %d", rid);
#endif
        return 0;
    }


    if(rhead==rtail) {
        Trace("\tno entry\n");
        return 0;
    }

    pos = -1;
    for(i=0;i<MAX_REQUEST_QUEUE;i++) {
        if(rque[i].sid==rsid && rque[i].id==rid) {
            pos = i;

#ifdef TRACE_REMOVE_REQUEST
            Trace("\tfound in rque[%d], remove\n", pos);
#endif
            rque[pos].state = REQ_EMPTY;
        }
    }
    if(pos<0 || pos>=MAX_REQUEST_QUEUE) {
#ifdef TRACE_REMOVE_REQUEST
        Trace("\tnot found\n");
#endif
        return -1;
    }

#ifdef TRACE_REMOVE_REQUEST
    Trace("remove_request: middle rhead %4d, rtail %4d\n", rhead, rtail);
    /*
    show_rque();
    */
#endif

    /*
     * Compaction
     */
    h = rhead;
    if(rhead<=rtail)
        t = rtail;
    else
        t = MAX_REQUEST_QUEUE + rtail;

    for(i=h;i<t;i++) {
        if(rque[i%MAX_REQUEST_QUEUE].state==REQ_EMPTY) {
            /*
            Trace("\ton %d shift...", i);
            */
            for(j=i;j<t-1;j++) {
                rque[(j)%MAX_REQUEST_QUEUE] = rque[(j+1)%MAX_REQUEST_QUEUE];
            }
            /*
            Trace("\n");
            */
            rque[t%MAX_REQUEST_QUEUE].sid       = 0;
            rque[t%MAX_REQUEST_QUEUE].id        = 0;
            rque[t%MAX_REQUEST_QUEUE].state     = REQ_EMPTY;
            rque[t%MAX_REQUEST_QUEUE].method    = METHOD_UNKNOW;
            rque[t%MAX_REQUEST_QUEUE].url[0]    = '\0';
        }
        /*
        show_rque();
        */
    }

    for(i=h;i<t;i++) {
        if(rque[i%MAX_REQUEST_QUEUE].state==REQ_EMPTY) {
            j = i;
            break;
        }
    }
    rtail = j%MAX_REQUEST_QUEUE;

    t = MAX_REQUEST_QUEUE+rhead-1;
    for(i=rtail;i<t;i++) {
        rque[i%MAX_REQUEST_QUEUE].id    = 0;
        rque[i%MAX_REQUEST_QUEUE].state = REQ_EMPTY;
        rque[i%MAX_REQUEST_QUEUE].url[0]= '\0';
    }

#ifdef TRACE_REMOVE_REQUEST
    Trace("remove_request: after   rhead %4d, rtail %4d\n", rhead, rtail);
    /*
    show_rque();
    */
#endif

    return 0;
}


static int find_freehandler()
{
    int i;

    for(i=0;i<MAX_HANDLER;i++) {
        if(handlers[i].wake!=0 && handlers[i].busy==0)
            return i;
    }

    return -1;
}

#if 1
static int find_lowloadhandler(int limit)
{
    int i;
    int min, min_pos;

    min = limit*2;
    min_pos = -1;

    for(i=0;i<MAX_HANDLER;i++) {
        if(handlers[i].wake!=0 && handlers[i].busy<min) {
            min = handlers[i].busy;
            min_pos = i;
        }
    }
    if(min<limit) {
        Trace("find_lowloadhandler: select handler# %d (load %d)\n",
            min, min_pos);
        return min_pos;
    }

    return -1;
}
#else
static int find_lowloadhandler(int limit)
{
    int i;

    for(i=0;i<MAX_HANDLER;i++) {
        if(handlers[i].wake!=0 && handlers[i].busy<limit)
            return i;
    }

    return -1;
}
#endif

static int find_samehandler(char *url)
{
    int i;

    for(i=0;i<MAX_HANDLER;i++) {
        if(handlers[i].wake!=0 && strcmp(handlers[i].req.url,url)==0)
            return i;
    }

    return -1;
}

static int find_samebusyhandler(char *url)
{
    int i;

    for(i=0;i<MAX_HANDLER;i++) {
        if(handlers[i].wake!=0 && handlers[i].busy!=0
            && strcmp(handlers[i].req.url,url)==0 ) {
            return i;
        }
    }

    return -1;
}

static int find_samefreehandler(char *url)
{
    int i;

    for(i=0;i<MAX_HANDLER;i++) {
        if(handlers[i].wake!=0 && handlers[i].busy==0
            && strcmp(handlers[i].req.url,url)==0 ) {
            return i;
        }
    }

    return -1;
}

static int find_nullfreehandler()
{
    int i;

    for(i=0;i<MAX_HANDLER;i++) {
        if(handlers[i].wake!=0 && handlers[i].busy==0
            && handlers[i].req.url[0]=='\0') {
            return i;
        }
    }

    return -1;
}

static int find_prefetchinghandler()
{
    int i;

    for(i=0;i<MAX_HANDLER;i++) {
        if(handlers[i].wake!=0 && handlers[i].busy!=0
            && handlers[i].req.mode==REQ_PREFETCH ) {
            Trace("find_prefetchinghandler: *stop %d, pid %d\n",
                i, handlers[i].pid);
            StopHandler(i);

            return i;
        }
    }

    return -1;
}



static int find_sameprefetchinghandler(char *url)
{
    int i;

    for(i=0;i<MAX_HANDLER;i++) {
        if(handlers[i].wake!=0 && handlers[i].busy!=0
            && strcmp(handlers[i].req.url,url)==0
            && handlers[i].req.mode==REQ_PREFETCH ) {
            Trace("find_sameprefetchinghandler: *add %d, pid %d\n",
                i, handlers[i].pid);

            return i;
        }
    }

    return -1;
}





/*
 * Dispatch request to handler
 */
static int dispatch_request(int p, int n)
{
    time_t now;
    mrequest_t request;
    int prefetching_handler = 0;


    time(&now);
    Trace("; %10d.%06d dispatch\n", now, 0);

    request = rque[p];

    Trace("Dispatch request #%d/#%d to handler# %d (pid %d) fd# %d ",
        request.sid, request.id, n, handlers[n].pid, request.fd);

    if(handlers[n].busy>0 && handlers[n].req.mode == REQ_PREFETCH)  {
        prefetching_handler = 1;
        Trace("(prefetching handler) ");
    }
    else {
        Trace("(client request handler) ");
    }



    /*
     * To avoid deadlock with IPC, set timer
     */
    pipe_broken = 0;
#ifdef SINGLE_RECEPTION
    XSendFRecord(DISPATCH_TIMEOUT, handlers[n].port,
        request.fd, &request, sizeof(mrequest_t));
#else
    request.fd = -1;
    XSendRecord(DISPATCH_TIMEOUT, handlers[n].port, &request, sizeof(mrequest_t));
#endif /* SINGLE_RECEPTION */

    if(pipe_broken) {
        Fatal("stream to handler# %d (pid %d) is broken", n, handlers[n].pid);
        KillHandler(n);
    }
#if 0
    if(prefetching_handler) {
        Trace("send SIGUSR2 to %d\n", handlers[n].pid);
        kill(handlers[n].pid, SIGUSR2);
    }
#endif


    handlers[n].times ++;
#if 0
    handlers[n].busy ++;
#endif
    strcpy(handlers[n].req.url, request.url);
    handlers[n].req = request;
    handlers[n].start = now;


    if(rque[p].fd != -1) {
        /*
        Trace("dispatch_request: client's fd %d. try to close\n", rque[p].fd);
        */
        if(close(rque[p].fd)) {
            Error("fail to close file(%d)", errno);
        }
        rque[p].fd = -1;
    }
    rque[p].state = REQ_BUSY;

done:
    Trace("done\n");

    if(trace_flag) {
        list_handlers(handlers, MAX_HANDLER, manager_ref_time);
    }


    return 0;

fail:
    Trace("fail\n");

    return 1;
}



/*
 * Locked client request processing is FIFO
 */
static void dispatch_locked_client_request()
{
    int limit;
    int i,p;
    int n;


    if(rtail<rhead)
        limit = MAX_REQUEST_QUEUE+rtail;
    else
        limit = rtail;

    for(i=rhead;i<limit;i++) {
        p = i % MAX_REQUEST_QUEUE;
        if(rque[p].state == REQ_LOCK && rque[p].mode == REQ_CLIENT) {
            Trace("dispatch_locked_client_request: i %d p %d id #%d...",
                i, p, rque[p].id);

#if 0
            /* lock while exist busy hander which handle same URL */
            if((n = find_samebusyhandler(rque[p].url))>=0) {
                Trace("busy, heandler %d getting now\n", n);
                continue;
            }

            if((n = find_samefreehandler(rque[p].url))>=0) {
                Trace("Dispatch %d, it handled this URL and free\n", n);
                goto found_ch;
            }
            if((n = find_nullfreehandler())>=0) {
                Trace("Dispatch %d, it is free, and idle\n", n);
                goto found_ch;
            }
#endif
            if((n = find_freehandler())>=0) {
                Trace("Dispatch %d, it is free\n", n);
                goto found_ch;
            }
#if 1
            if((n = find_sameprefetchinghandler(rque[p].url))>=0) {
                Trace("Dispatch %d, it process same object prefetching\n", n);
                goto found_ch;
            }
#endif
#if 0
            /* stop prefetching handler and use for client request */
            if((n = find_prefetchinghandler())>=0) {
                Trace("Dispatch %d, it process other object prefetching\n", n);
                goto found_ch;
            }
#endif
#if 1
            if((n = find_lowloadhandler(max_session))>=0) {
                Trace("Dispatch %d, because it is low load\n", n);
                goto found_ch;
            }
#endif

            if(n==-1) {
                /*
                 * try lauch more handler
                 */
                n = LaunchHandler();
                if(n==-1) {
                    Trace("Sorry, Cannot find handler...\n");
                    break;
                }
            }
found_ch:
            dispatch_request(p, n);
        }
    }
}


/*
 * Client request processing is FIFO
 */
static void dispatch_client_request()
{
    int limit;
    int i,p;
    int n;


    if(rtail<rhead)
        limit = MAX_REQUEST_QUEUE+rtail;
    else
        limit = rtail;

    for(i=rhead;i<limit;i++) {
        p = i % MAX_REQUEST_QUEUE;
        if((rque[p].state == REQ_NOP || rque[p].state == REQ_EMPTY )
                && (rque[p].mode == REQ_CLIENT
                        || rque[p].mode == REQ_IMPORT)) {
            Trace("dispatch_client_request: i %d p %d id #%d...\n\t",
                i, p, rque[p].id);

#if 0
            if((n = find_samebusyhandler(rque[p].url))>=0) {
                Trace("lock, heandler %d getting now\n", n);
                rque[p].state = REQ_LOCK;
                handlers[n].req.mode = REQ_CLIENT;  /* handle as client */
                continue;
            }
#endif

            if((n = find_samefreehandler(rque[p].url))>=0) {
#if 1
                if(rque[p].fd==-1 || rque[p].sid<0) {
                    Trace("Cancel, it is proxy (fd %d, sid %d)\n",
                        rque[p].fd, rque[p].sid);
                    remove_request(rque[p].sid, rque[p].id);
                    continue;
                }
#endif
                Trace("Dispatch client   %d, it handled this URL and free\n", n);
                goto found_ch;
            }
            if((n = find_nullfreehandler())>=0) {
                Trace("Dispatch client   %d, it is free, and idle\n", n);
                goto found_ch;
            }
            if((n = find_freehandler())>=0) {
                Trace("Dispatch client   %d, it is free\n", n);
                goto found_ch;
            }
#if 1
            if((n = find_sameprefetchinghandler(rque[p].url))>=0) {
                Trace("Dispatch client   %d, it process same object prefetching\n", n);
                goto found_ch;
            }
#endif
#if 0
            /* stop prefetching handler and use for client request */
            if((n = find_prefetchinghandler())>=0) {
                Trace("Dispatch client    %d, it process prefetching\n", n);
                goto found_ch;
            }
#endif
#if 1
            if((n = find_lowloadhandler(max_session))>=0) {
                Trace("Dispatch client     %d, because it is low load\n", n);
                goto found_ch;
            }
#endif

            if(n==-1) {
                if(n_active>=max_handler) {
                    Trace("all handlers are up\n");
                    break;
                }

                /*
                 * try lauch more handler
                 */
                n = LaunchHandler();
                if(n==-1) {
                    Trace("Sorry, Cannot find handler...\n");
                    break;
                }
            }
found_ch:
            dispatch_request(p, n);
        }
    }
}



/*
 * Prefetch request processing is LIFO
 */
static void dispatch_prefetch_request()
{
    int limit;
    int i,p;
    int n;
    int max_prefetch;

#if 0
    max_prefetch = ((max_handler - n_client - 1) * 2) / 3 ;
#else
    max_prefetch = ((max_handler*max_session - n_client - 1) * 2) / 3 ;
#endif

#if 1
    Trace("dispatch_prefetch_request:\n\tmax - handler %d, session %d\n",
        max_handler, max_session);
    Trace("\tsession - client %d prefetch %d -> max_prefetch %d\n",
        n_client, n_prefetch, max_prefetch);
#endif

    if(max_prefetch < 1) {
        return;
    }

    if(rtail<rhead)
        limit = MAX_REQUEST_QUEUE+rtail;
    else
        limit = rtail;

    for(i=limit;i>=rhead;i--) {
        p = i % MAX_REQUEST_QUEUE;
        if(rque[p].state == REQ_NOP && rque[p].mode == REQ_PREFETCH) {
            Trace("dispatch_prefetch_request: i %d p %d id #%d...\n\t",
                i, p, rque[p].id);
            if(n_prefetch >= max_prefetch) {
                Trace("no slot to prefetch\n");
                break;
            }

            if((n = find_freehandler())>=0) {
                Trace("Dispatch prefetch %d, it is free\n", n);
                goto found_ph;
            }

            if((n = find_lowloadhandler(max_session))>=0) {
                Trace("Dispatch prefetch %d, because it is low load\n", n);
                goto found_ph;
            }

            if(n==-1) {
                if(n_active>=max_handler) {
                    Trace("all handlers are up\n");
                    break;
                }

                /*
                 * try lauch more handler
                 */
                n = LaunchHandler();
                if(n==-1) {
                    Trace("Sorry, Cannot find handler...\n");
                    break;
                }
            }
found_ph:
            dispatch_request(p, n);
            n_prefetch++;
        }
    }
}


/*
 *
 */
void DispatchJob()
{
#if 0
    CheckHandlers();
#endif

    check_rque();

    dispatch_locked_client_request();
    dispatch_client_request();
    dispatch_prefetch_request();

#ifdef TRACE_DISPATCH
    if(trace_flag) {
        show_rque();
    }
#endif

#if 0
    if(trace_flag) {
        list_handlers(handlers, MAX_HANDLER, manager_ref_time)
    }
#endif
}

