/*
 *  WCOL: shared memory controll
 *      by k-chinen@is.aist-nara.ac.jp, 1997
 *
 *  $Id$
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <signal.h>
#ifdef HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif
#include <sys/uio.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <limits.h>

#include <sys/ipc.h>
#include <sys/shm.h>


#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif

#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif

#ifndef ESC
#define ESC '\x1b'
#endif


#include <signal.h>
#include <errno.h>

#include "wcol.h"
#include "msg.h"
#include "request.h"
#include "ipc.h"
#include "sigs.h"
#include "fds.h"

#include "session.h"
#include "manager.h"
#include "handler.h"
#include "prompter.h"
#include "reaper.h"
/* */
#include "global.h"
/* */


#define SHOW    0
#define DELETE  1


static int usage()
{
    fprintf(stdout,
"scntl - shared memory controllor\n\
    by k-chinen@is.aist-nara.ac.jp, 1997\n\
\n\
usage: scntl [options] [command]\n\
\n\
command:\n\
    -h          help, print this messages\n\
    -s          show (default)\n\
    -d          delete\n\
\n\
option:\n\
    -S          without session\n\
    -H          without handler\n\
\n\
    \n");

    return 0;
}

Session *session_pool;
modinfo_t *handlers;

int flag_handler = 1;
int flag_session = 1;
int command = SHOW;

key_t session_shmid;
key_t hstate_shmid;



void Init(int argc, char **argv)
{
    int flag;
    int i;

    extern char *optarg;
    extern int  optind;

    /* defaults */
    trace_flag = 1;

    while((flag = getopt(argc, argv, "vhsdSH"))!= -1) {
        switch(flag) {
        case 'v':
            Trace("parameters:\n");
            Trace("\tMAX_HANDLER %d, MAX_SESSION %d\n",
                MAX_HANDLER, MAX_SESSION);
            break;
        case 'h':
            usage();
            exit(0);
            break;
        case 'S':
            flag_session = 0;
            break;
        case 'H':
            flag_handler = 0;
            break;
        case 's':
            command = SHOW;
            break;
        case 'd':
            command = DELETE;
            break;
        }
    }





}



main(int argc, char **argv)
{
    int c;
    extern Session *session_pool;
    extern int handler_no;

    Init(argc, argv);
#ifdef MEMORY_CHECK
    show_memoryusage();
#endif


    /*
     *
     */
    if(flag_session) {
    session_shmid = assign_pool(WCOL_SHMKEY_SESSION);
    if(session_shmid == -1 ) {
        Error("not found shared memory area for sessions");
        exit(WCOL_EXIT_NOMEMORY);
    }

    if((session_pool = (Session*)shmat(session_shmid, (void*)NULL, 0))==NULL) {
        Error("cannot allocate shared memory area for sessions (%d)", errno);
        exit(WCOL_EXIT_NOMEMORY);
    }
/*
    Trace("Init: sstate %8x\n", session_pool);
*/
    }

    /*
     *
     */
    if(flag_handler) {
    hstate_shmid = assign_pool(WCOL_SHMKEY_HSTATE);
    if(hstate_shmid == -1 ) {
        Error("not found shared memory area for handlers");
        exit(WCOL_EXIT_NOMEMORY);
    }

    if((handlers = (modinfo_t*)shmat(hstate_shmid, (void*)NULL, 0))==NULL) {
        Error("cannot allocate shared memory area for handlers (%d)", errno);
        exit(WCOL_EXIT_NOMEMORY);
    }
/*
    Trace("Init: hstate %8x\n", handlers);
*/
    }

    switch (command) {
    case SHOW:
        if(flag_session) {
            show_pool_stat(session_shmid);
        }
        if(flag_handler) {
            show_pool_stat(hstate_shmid);
        }

        break;

    case DELETE:
        if(flag_session) {
            show_pool_stat(session_shmid);
            delete_pool(session_shmid);
        }
        if(flag_handler) {
            show_pool_stat(hstate_shmid);
            delete_pool(hstate_shmid);
        }

        break;

    }

}






