/*
 *  WCOL: send (HTTP) procedures
 *      by k-chinen@is.aist-nara.ac.jp, 1994, 1995, 1996, 1997
 *
 *  $Id: send.c,v 1.3 1997/09/15 02:00:33 k-chinen Exp k-chinen $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <signal.h>
#include <fcntl.h>
#include <limits.h>
#include <string.h>
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif
#include <errno.h>

#include "wcol.h"
#include "sigs.h"
#include "conv.h"
#include "jfilter.h"

#include "misc.h"


#define _DEBUG_SEND
#define _TRACE_HEADER


#if 1
int remove_cookie_flag;
int remove_referer_flag;
#endif


/*
 * SendHeadDirect -
 *  return:
 *      >=0 success (header size)
 *      -1  error
 */
static int SendHeadDirect(int *sfd, char *name,
    char ver_major, char ver_minor, int keep_alive, int discard)
{
    char hpath[STRING_SIZE];
    char buf[BUF_SIZE];
    int  head_fd;
    int  hsum;
    int  nread, nwrite;

    /*
    Trace("Send: HEAD direct name=%s\n", name);
    */

    if(AssignFileName(hpath, name, HEAD_FILE)) {
        return -1;
    }
    if((head_fd=open(hpath, O_RDONLY))<0) {
        Error("SendHeadDirect: cannot open '%s'", hpath);
        return -1;
    }

    hsum = 0;
    while((nread=read(head_fd, buf, BUF_SIZE))>0) {
try:
        pipe_broken = 0;
        nwrite = write(*sfd, buf, nread);

        if(pipe_broken) {
            pipe_broken = 0;
            close(head_fd);

            return -1;
        }

        if(nwrite<0) {
            if(errno==EINTR) {
                goto try;
            }

            Error("SendHeadDirect: write() fail (%d)", errno);

            close(head_fd);
            return -1;
        }
        if(nwrite!=nread) {
            Error("SendHeadDirect: write() fail nread %d, nwrite %d",
                nread, nwrite);

            close(head_fd);
            return -1;
        }

        hsum += nwrite;
    }
    close(head_fd);


    Trace("SendHeadDirect: HEAD direct %d bytes.\n", hsum);

    return hsum;
}





/*
 *
 */
int SendHeadBuf(int *sfd, char *orig_header, char *name, char *ntype,
    char ver_major, char ver_minor, int keep_alive, int discard)
{
    char cooked_header[BUF_SIZE];
    char field[STRING_SIZE], tmpfield[STRING_SIZE];
    int  sum,c;
    char *cpos;

    if(ntype==NULL) {
    Trace("SendHeadBuf: URL='%s'\n\tremove cookie %d, type *unknown*\n",
        name, remove_cookie_flag);
    }
    else {
    Trace("SendHeadBuf: URL='%s'\n\tremove cookie %d,type %s\n",
        name, remove_cookie_flag, ntype);
    }

    Trace("SendHeadBuf: HEAD original %5d bytes.\n", strlen(orig_header));

    cooked_header[0] = '\0';
    pipe_broken=0;
    sum = 0;

#ifdef TRACE_HEADER
    /* test */
    Trace("----\n");
    cpos = orig_header;
    while((cpos = sgets(orig_header, cpos, field, BUF_SIZE))!=NULL) {
        dump_string("  0 ", field);
    }
#endif

#ifdef TRACE_HEADER
    Trace("----\n");
#endif

    cpos = orig_header;
    while((cpos = sgets(orig_header, cpos, field, BUF_SIZE))!=NULL) {
        /*
        Chop_CRLF(field);
        */
        if(field[0]=='\0' || field[0]=='\r' || field[0]=='\n' ) {
            break;
        }
#ifdef TRACE_HEADER
        dump_string("  1 ", field);
#endif

        /*
         * hook output content-type
         */
        if(ntype!=NULL && ntype[0]!='\0' &&
                strncasecmp_first_all(field, 14, 'C', "Content-type: ")) {
            sprintf(tmpfield, "X-%s", field);
            strcat(cooked_header, tmpfield);

            sprintf(tmpfield,"Content-Type: %s\r\n", ntype);
            strcat(cooked_header, tmpfield);

            Trace("SendHeadBuf: Change C-Type to '%s'\n", ntype);
        }
        else
        /*
         * remove cookie
         */
        if (remove_cookie_flag &&
                strncasecmp_first_all(field, 12, 'S', "Set-Cookie: ")) {
            /* nothing */
            Trace("SendHeadBuf: Skip (Remove) Cookie\n");
        }
        /*
         * otherwise, thru
        */
        else {
            strcat(cooked_header, field);
        }
    }

#if 1
    if(ver_major>0) {
        sprintf(field, "X-ProxyAgent: WWW-Collector-E/%s\r\n", WCOL_VER);
        strcat(cooked_header, field);
        sprintf(field, "X-Magick: %d/%d\r\n", wcol_pid, 0);
        strcat(cooked_header, field);

        sprintf(field, "X-URL: %s\r\n", name);
        strcat(cooked_header, field);
    }

    if(keep_alive) {
        Trace("SendHeadBuf: Keep-Alive\n");

        sprintf(field, "Connection: Keep-Alive\r\n");
        strcat(cooked_header, field);
    }
#endif


#ifdef USE_NOTE
#ifdef NOTE_SAMEWINDOW
    sprintf(field, "Window-target: Wcol-BODY\n");
    strcat(cooked_header, field);
#endif
#endif

    sprintf(field, "\r\n");
    strcat(cooked_header, field);



    Trace("SendHeadBuf: HEAD cooked   %5d bytes.\n", strlen(cooked_header));
#ifdef TRACE_HEADER
    /* test */
    Trace("----\n");
    cpos = cooked_header;
    while((cpos = sgets(cooked_header, cpos, field, BUF_SIZE))!=NULL) {
        dump_string("  2 ", field);
    }
#endif


w1:
    c = write(*sfd, cooked_header, strlen(cooked_header));
    if(c<0) {
        if(errno==EINTR) {
            goto w1;
        }
        Error("SendHeadBuf: write error fd# %d, ret %d, at label-'w1' (%d)",
            *sfd, c, errno);

        if(pipe_broken) {   /* *sfd is broken ? */
            pipe_broken=0;
            Trace("SendHeadBuf: PIPEBROKEN: HEAD cooked %d bytes.\n", sum);
        }
    }

    return strlen(cooked_header);
}




/*
 * SendHead -
 *
 *  HTTP/0 is not need header. So, if you need header, you use HTTP/1.
 *  if other version is appear, you should support vesion.
 *
 *  return:
 *      >=0 success (header size)
 *      -1  error
 */


int SendHead(int *sfd, char *name, char *ntype,
    char ver_major, char ver_minor, int keep_alive, int discard)
{
    char hpath[STRING_SIZE];
    FILE *head_fp;
    int  head_fd;
    char orig_header[BUF_SIZE];
    int  c;


    Trace("SendHead: URL='%s'\n", name);

    if(AssignFileName(hpath, name, HEAD_FILE)) {
        return -1;
    }


    if((head_fd=open(hpath, O_RDONLY))<0) {
        Error("SendHead: cannot open header file '%s'\n", hpath);
        return -1;
    }

r1:
    c = read(head_fd, orig_header, BUF_SIZE);
    if(c<0) {
        if(errno==EINTR) {
            goto r1;
        }
        else {
            Error("SendHead: cannot read header file '%s'\n", hpath);
            return -1;
        }
    }
    close(head_fd);

    /* Call actual procedure */
    return SendHeadBuf(sfd, orig_header, name, ntype, ver_major, ver_minor,
                keep_alive, discard);
}


/*
int SendHeadBuf(int *sfd, char *orig_header, char *name, char *ntype,
    char ver_major, char ver_minor, int keep_alive, int discard)
*/







/*
 * SendDirect
 *      *sfd:       output socket
 *      name:       logical name
 *      ver_major:  version major
 *      ver_minor:  version minor
 */
static int SendDirect(int *sfd, char *name, Info *info,
    char ver_major, char ver_minor, int keep_alive,
    int discard)
{
    char buf[BUF_SIZE];
    int  c,hsum, bsum;
    char path[STRING_SIZE];
    int  fd;
    off_t nwrite;

/*
    Trace("SendDirect: socket=#%d, name='%s', ver='%s', discard=%d.\n",
        *sfd, name, ver, discard);
*/
    Trace("SendDirect: socket=#%d, name='%s', ver=%d.%d, discard=%d, bsize=%d\n",
        *sfd, name, ver_major, ver_minor, discard, (long)info->attr.bsize);

    /*
     * Head Transfer
     */
#if 1
    if(strncmp(name, "http:", 5)==0 && ver_major>0) {
  /*
        hsum = SendHead(sfd, name, info->attr.type,
                    ver_major, ver_minor, keep_alive, discard);
   */
        hsum = SendHead(sfd, name, NULL,
                    ver_major, ver_minor, keep_alive, discard);
    }
    else {
        hsum = 0;
    }
#else
    if(strncmp(name, "http:", 5)==0 && ver_major>0) {
        hsum = SendHeadDirect(sfd, name,
                    ver_major, ver_minor, keep_alive, discard);
    }
    else {
        hsum = 0;
    }
#endif

    if(AssignFileName(path, name, BODY_FILE)) {
        return -1;
    }
    if((fd=open(path, O_RDONLY))<0) {
        Error("SendDirect: Cannot open file %s.", path);

        return -1;
    }

    bsum = 0;
    while((c=read(fd, buf, sizeof(buf)))>0) {
try:
        pipe_broken=0;
        nwrite = write(*sfd, buf, c);
        if(pipe_broken) {   /* *sfd is broken ? */
            pipe_broken=0;
            close(fd);

            return -1;
        }
        if(nwrite<0) {
            if(errno==EINTR) {
                goto try;
            }
            else {
                Error("SendDirect: write() fail (%d)", errno);

                close(fd);
                return -1;
            }
        }
        if(nwrite!=c) {
            Error("SendDirect: write() fail : read %d bytes, write %d bytes",
                c, nwrite);

            close(fd);
            return -1;
        }

        bsum += nwrite;
    }
    close(fd);

    Trace("SendDirect: content-length %d / BODY %d, HEAD %d\n",
        (long)info->attr.clength, bsum, hsum);

#if 0
    if(shutdown(*sfd, 2)) {
        int _errno_ = errno;
        Error("SendDirect: (%d) fd %d is cannot shutdown", _errno_, *sfd);
        *sfd = -1;
    }
    if(close(*sfd)) {
        int _errno_ = errno;
        Fatal("SendDirect: (%d) fd %d is cannot close", _errno_, *sfd);
        *sfd = -1;
    }
#endif
    if(!keep_alive || info->attr.clength==-1) {
        if(info->attr.clength==-1) {
          Trace("SendDirect: content-length is empty. close client stream.\n");
        }
        if(!keep_alive) {
            Trace("SendDirect: no keep-alive, close client stream.\n");
        }

        if(*sfd!=-1) {
            if(shutdown(*sfd, 2)) {
                int _errno_ = errno;
                Error("SendDirect: fd# %d fail shutdown() (%d)", *sfd, _errno_);
            }
            if(close(*sfd)) {
                int _errno_ = errno;
                Fatal("SendDirect: fd# %d fail close() (%d)", *sfd, _errno_);
            }
            *sfd = -1;
        }
    }

    if(discard) {
        DiscardInfo(name);
    }
    else {
        UpdateInfo(name);
    }

    return hsum + bsum;
}






/*
 * Send -
 *
 *  return:
 *      -1  error
 */
int SendConv(int *sfd, char *name, Info *info,
    char ver_major, char ver_minor, int keep_alive,
    int discard)
{
    char buf[BUF_SIZE];
    int  c,sum;
    int  fd,trans_type;
    char path[STRING_SIZE];
    char *tmp_path;
    int  pos;
    char pgn[STRING_SIZE];
    int  chpid;
    int  codetype;


    Trace("SendConv: socket=#%d, name='%s', ver=%d.%d, discard=%d.\n",
        *sfd, name, ver_major, ver_minor, discard);

    if(AssignFileName(path, name, BODY_FILE)) {
        return -1;
    }
    tmp_path = tmpnam((char*)NULL);
    pos = check_hook(name,info->attr.type);

    /* Use inner-function */
    if(conv_tbl[pos].program[0]=='!') {
        if(strcmp(conv_tbl[pos].program, "!tosjis")!=0 &&
            strcmp(conv_tbl[pos].program, "!toeuc")!=0 &&
            strcmp(conv_tbl[pos].program, "!tojis")!=0) {

            Error("SendConv: Not support function. What is '%s' ?",
                conv_tbl[pos].program);
            return SendDirect(sfd, name, info,
                        ver_major, ver_minor, keep_alive, discard);
        }

        /* check code via 100 lines parsing */
        codetype = CheckFileKanjiCode(path, 100);
        if(codetype == -1) {
            return SendDirect(sfd, name, info,
                        ver_major, ver_minor, keep_alive, discard);
        }

        if(strncmp(name, "http:", 5)==0 && ver_major>0) {
            SendHead(sfd, name, conv_tbl[pos].output,
                ver_major, ver_minor, keep_alive, discard);
        }

        if((fd=open(path, O_RDONLY))<0) {
            Error("SendConv: cannot open %s .",tmp_path);
            return -1;
        }

        if(strcmp(conv_tbl[pos].program, "!tosjis")==0) {
            Trace("SendConv: tosjis() start ... ");
            ToSJIS(codetype, fd, *sfd);
        }
        else
        if(strcmp(conv_tbl[pos].program, "!toeuc")==0) {
            Trace("SendConv: toeuc() start ... ");
            ToEUC(codetype, fd, *sfd);
        }
        else
        if(strcmp(conv_tbl[pos].program, "!tojis")==0) {
            Trace("SendConv: toeuc() start ... ");
            ToJIS(codetype, fd, *sfd);
        }

        close(fd);
        Trace("done.\n");
    }

    /* Call external program */
    else {
        trans_type =
            ConvProgramName(pgn, conv_tbl[pos].program, path, tmp_path);

        Trace("SendConv: exec `%s'\n", pgn);

        if((chpid=vfork())== -1) {
            Error("SendConv: Cannot vfork()!!!");
        }
        if(chpid == 0) {
            if(strncmp(name, "http:", 5)==0 && ver_major>0) {
                SendHead(sfd, name, conv_tbl[pos].output,
                    ver_major, ver_minor, keep_alive, discard);
            }

            if(trans_type & CONV_HAS_OUTPUT) {
                /*
                 *  Have output-file
                 *      1. convert input to tmp-file as output
                 *      2. send tmp-file to socket.
                 */

                system(pgn);

                if((fd=open(tmp_path, O_RDONLY))<0) {
                    Error("SendConv: cannot open %s .",tmp_path);
                }
                pipe_broken = 0;
                sum = 0;
                while((c=read(fd, buf, BUF_SIZE))>0) {
                    write(*sfd, buf, c);
                    sum += c;

                    if(pipe_broken) {   /* fd *sfd is broken ? */
                        pipe_broken=0;
                        close(fd);
                        unlink(tmp_path);
                        exit(WCOL_EXIT_UNDEF);
                    }
                }

                close(fd);
                unlink(tmp_path);
            }
            else {
                /*
                 *  Not have output-file
                 *      1. duplicate socket for output (stdout)
                 *      2. call convert program
                 */

                if(dup2(*sfd, 1)<0) {
                    Error("Fail dup2()");
                }

                system(pgn);
            }


            if(shutdown(*sfd, 2)) {
                int _errno_ = errno;
                Error("SendConv: (%d) fd %d is cannot shutdown", _errno_, *sfd);
                *sfd = -1;
            }
            if(close(*sfd)) {
                int _errno_ = errno;
                Fatal("SendConv: (%d) fd %d is cannot close", _errno_, *sfd);
                *sfd = -1;
            }

            exit(WCOL_EXIT_NORMAL);
        }
    }

/*
    shutdown(*sfd, 0);
*/
    shutdown(*sfd, 2);
    close(*sfd);
    *sfd = -1;


    if(discard) {
        DiscardInfo(name);
    }
    else {
        UpdateInfo(name);
    }


    return sum;
}


int SendDirectFromMemory(int *sfd, char *name, Info *info,
    char ver_major, char ver_minor, int keep_alive,
    int discard)
{
    int  c, hsum, bsum;

/*
    Trace("SendDirectFromMemory: socket=#%d, name='%s', ver='%s', discard=%d.\n",
        *sfd, name, ver, discard);
*/
    Trace("SendDirectFromMemory: socket=#%d, name='%s', ver=%d.%d, discard=%d, bsize=%d\n",
        *sfd, name, ver_major, ver_minor, discard, info->attr.bsize);

    /*
     * Head Transfer
     */
    if(strncmp(name, "http:", 5)==0 && ver_major>0) {
        if(info->attr.phead==NULL) {
            hsum = SendHeadDirect(sfd, name,
                        ver_major, ver_minor, keep_alive, discard);
        }
        else {
            /*
             */
            c = write(*sfd, info->attr.phead, info->attr.hsize);
            hsum = c;
        }
    }
    else {
        hsum = 0;
    }

    Trace("SendDirectFromMemory: head %d/%d\n", hsum, info->attr.hsize);

    pipe_broken=0;
    bsum = 0;
    while(bsum<info->attr.bsize
        && (c = write(*sfd, info->attr.pbody, info->attr.bsize))!=0) {
        Trace("\tcurrent %d, sum %d\n", c, bsum);
        if(c<=0) {
            break;
        }
        bsum += c;

        if(pipe_broken) {   /* *sfd is broken ? */
            pipe_broken=0;
            break;
        }
    }


/*
    shutdown(*sfd, 0);
*/
    shutdown(*sfd, 2);
    close(*sfd);
    *sfd = -1;

out:
    Trace("SendDirectFromMemory: body %d/%d\n", bsum, info->attr.bsize);

    if(discard) {
        DiscardInfo(name);
    }
    else {
        UpdateInfo(name);
    }

    return hsum + bsum;
}


int SendNote(int *sfd, char *name, Info *info,
    char ver_major, char ver_minor, int keep_alive,
    int discard)
{
    char buf[BUF_SIZE];
    char tmp[STRING_SIZE];
    int size;
    char *unknow_msg = "*unknow*\n";

    Trace("SendNote: it is Note for '%s'\n", name);

    if(ver_major>=1) {
#ifdef NOTE_SAMEWINDOW
        sprintf(buf,"HTTP/1.0 200\n\
Content-Type: text/html\n\
Window-target: Wcol-NOTE\n\
\n");
#else
        sprintf(buf,"HTTP/1.0 200\n\
Content-Type: text/html\n\
Window-target: Wcol-NOTE-Win\n\
\n");
#endif /* NOTE_SAMEWINDOW */
        size = write(*sfd, buf, strlen(buf));
    }
    else {
        size = 0;
    }

#ifdef NOTE_SAMEWINDOW
    sprintf(buf,"<HTML>\n\
<pre>\n\
<b>Length:</b> server said %d, receive %d (head %d, total %d)\n\
<b>Connection Time:</b> %d (connect %d, trans %d) <b>Count:</b> %d\n\
<b>Last Modified:</b> %s\
<b>Expires:</b> %s\
<b>Last Touch:</b> %s\
<b>Type:</b> %s\n\
<b>URL:</b> %s\n\
</pre>\n\
</HTML>\n",
    info->attr.clength, info->attr.bsize, info->attr.hsize, info->attr.length,
    info->attr.take_time, info->attr.conn_time, info->attr.trans_time,
    info->attr.count,
    (info->attr.lmd == -1)      ? "*unknow*\n" : ctime(&info->attr.lmd),
    (info->attr.expire == -1)   ? "*unknow*\n" : ctime(&info->attr.expire),
    (info->attr.last == -1)     ? "*unknow*\n" : ctime(&info->attr.last),
    info->attr.type,
    info->attr.name
    );

    size += write(*sfd, buf, strlen(buf));

#else
    sprintf(buf,"<HTML>\n\
<pre>\n\
<b>URL:</b> %s\n\
<b>Type:</b> %s\n\
<b>Length:</b> server said %d, receive %d (head %d, total %d)\n\
<b>Connection Time:</b> %d (connect %d, trans %d)\n\
<b>Count:</b> %d\n",
        info->attr.name,
        info->attr.type,
        info->attr.clength, info->attr.bsize,
            info->attr.hsize, info->attr.length,
        info->attr.take_time, info->attr.conn_time, info->attr.trans_time,
        info->attr.count
    );

    if(info->attr.lmd<0) {
        sprintf(tmp, "<b>Last Modified:</b> <%d> %s",
            info->attr.lmd, unknow_msg);
    }
    else {
        sprintf(tmp, "<b>Last Modified:</b> <%d> %s",
            info->attr.lmd, ctime(&info->attr.lmd));
    }
    strcat(buf, tmp);

    if(info->attr.expire<0) {
        sprintf(tmp, "<b>Expires:</b>    <%d> %s",
            info->attr.expire, unknow_msg);
    }
    else {
        sprintf(tmp, "<b>Expires:</b>    <%d> %s",
            info->attr.expire, ctime(&info->attr.expire));
    }
    strcat(buf, tmp);

    if(info->attr.last<0) {
        sprintf(tmp, "<b>Last Touch:</b> <%d> %s",
            info->attr.last, unknow_msg);
    }
    else {
        sprintf(tmp, "<b>Last Touch:</b> <%d> %s",
            info->attr.last, ctime(&info->attr.last));
    }
    strcat(buf, tmp);

    strcat(buf, "</pre>\n</HTML>\n");

    size += write(*sfd, buf, strlen(buf));
#endif


    return size;
}



int SendFace(int *sfd, char *name, Info *info,
    char ver_major, char ver_minor, int keep_alive,
    int discard)
{
    char buf[BUF_SIZE];
    int size;
    char path[STRING_SIZE];
    char *p, *q;

    Trace("SendFace: it is Face for '%s'\n", name);

    if(ver_major>=1) {
        sprintf(buf,"HTTP/1.0 200\n\
Content-Type: text/html\n\
\n");
/*
Window-target: _top\n\
*/
        size = write(*sfd, buf, strlen(buf));
    }
    else {
        size = 0;
    }

    strcpy(path, info->attr.name);
    p = path;
    q = NULL;
    while(*p) {
        q = p;
        p++;
    }
    if(q && *q=='/') {
        strcat(path, PADDING_NAME);
    }


#ifdef NOTE_SAMEWINDOW
    Trace("Send: it is FACE for '%s'\n", name);
    sprintf(buf, "<html>\n\
<frameset rows=\"10%%,*\" name=\"Wcol-FACE\">\n\
    <frame src=\"%s**NOTE\" name=\"Wcol-NOTE\">\n\
    <frame src=\"%s**BODY\" name=\"Wcol-BODY\">\n\
</frameset>\n\
</html>\n", path, path);
#else
    Trace("Send: it is FACE for '%s'\n", name);
    sprintf(buf, "<html>\n\
<frameset rows=\"0,*\">\n\
    <frame src=\"%s**NOTE\" marginwidth=0 marginheight=0>\n\
    <frame src=\"%s**BODY\" marginwidth=0 marginheight=0>\n\
</frameset>\n\
</html>\n", path, path);
#endif /* NOTE_SAMEWINDOW */

    size += write(*sfd, buf, strlen(buf));

#if 0
    close(*sfd);
    *sfd = -1;
#endif

    return size;
}


/*
 * Send -
 *
 *  return:
 *      -1  error
 *    ! -1  success (length of resource)
 */
int Send(int *sfd, char *name, Info *info,
    char ver_major, char ver_minor, int keep_alive,
    int discard)
{
    int size;
    int k;

#if 1
    Trace("Send: socket=#%d, name='%s', ver=%d.%d, discard=%d.\n",
        *sfd, name, ver_major, ver_minor, discard);
#endif
    if(*sfd==-1) {
        return 0;
    }

#if 1
    Trace("\t->%d.\n", check_hook(name,info->attr.type));
    Trace("\tphead %#x,pbody=%#x\n", info->attr.phead, info->attr.pbody);
#endif


#ifdef USE_NOTE
    if(strcmp(info->attr.type,"text/html")==0) {
        char tmp_buf[STRING_SIZE];

        k=IsURL_NOTEorBODY(name, tmp_buf);
        Trace("Send: check NOTE or BODY -> %d\n", k);

        if(k==1) {
            size = SendNote(sfd, name, info,
                    ver_major, ver_minor, keep_alive, discard);

            return size;
        }
        else
        if(k==2) {
            strcpy(name, tmp_buf);

            Trace("Send: it is BODY for '%s'\n", name);
        }
        else {
            size = SendFace(sfd, name, info,
                    ver_major, ver_minor, keep_alive, discard);

            return size;
        }
    }
#endif

    if(IsConvEntry() && check_hook(name,info->attr.type)>=0) {
        size = SendConv(sfd, name, info,
                    ver_major, ver_minor, keep_alive, discard);
    }
    else {
        if(info->attr.pbody!=NULL) {
            size = SendDirectFromMemory(sfd, name, info,
                        ver_major, ver_minor, keep_alive, discard);
        }
        else {
            size = SendDirect(sfd, name, info,
                        ver_major, ver_minor, keep_alive, discard);
        }
    }

#if 0
#endif

    /*
    Log("SEND %s %d", info->attr.name, size);
    */


    return size;
}


