/*
 * WCOL: session
 *          by k-chinen@is.aist-nara.ac.jp, 1995, 1997
 *
 * $Id: session.h,v 1.4 1995/10/09 07:10:46 k-chinen Exp k-chinen $
 */

#ifndef WCOL_SESSION_H
#define WCOL_SESSION_H


#ifndef MAX_SESSION
#define MAX_SESSION     14
#endif

#ifndef HAVE_FD_SET_T
#ifndef WCOL_FD_LIMIT
#ifdef NOFILE
#define WCOL_FD_LIMIT NOFILE
#else
#define WCOL_FD_LIMIT (64)
#endif
#endif
#define HAVE_FDSET_T
#endif



/*
 * It is hurisutic parameter :-)
 */

#ifndef SESSION_TIMEOUT
#define SESSION_TIMEOUT     (300000)
#endif

#ifndef SESSION_LIFETIME
#define SESSION_LIFETIME    (3600000)
#endif

#ifndef QUERY_TIMEOUT
#define QUERY_TIMEOUT       (2000)
#endif





#ifndef HEADER_BUFLEN
#define HEADER_BUFLEN       KB(5)
#endif
#ifndef BODY_BUFLEN
#define BODY_BUFLEN         KB(40)
#endif



typedef enum {
    CONN_NOP=0,
    CONN_COMMAND,
    CONN_HTTPLISTEN, CONN_HTTP, CONN_HTTPREQUEST, CONN_HTTPTIMEOUT,
    CONN_ICPCONNECT, CONN_ICPQUERY, CONN_ICPREPLY, CONN_ICPTIMEOUT,
    CONN_ERROR
} Connection_type;


typedef enum {
    SESS_NOP=0,
    SESS_CLIENT, SESS_PREFETCH, SESS_IMPORT
} Session_type;


typedef enum {
    PHASE_UNKNOW=0, PHASE_IDLE, PHASE_START,
    PHASE_REQUEST, PHASE_THINK,
    PHASE_WAIT, PHASE_SEND,
    PHASE_QUERY, PHASE_CONNECTSIBLING,
    PHASE_NAMERESOLV,
    PHASE_CONNECTSERVER, PHASE_HEADER, PHASE_BODY, PHASE_END,
    PHASE_ERROR
} Session_phase;

typedef struct {
    Session_type    type;
    Session_phase   phase;

    request_t       req;

    url_t           *url;
    char            _url[URL_LENGTH];

    int             use_full_location;

    /* server */
    int             server_type;
    char            server_name[STRING_SIZE];
    int             server_port;


    struct timeval  start_tp, conn_tp, end_tp;

    struct timeval  begin;
    struct timeval  hot;


    int             command_fd;
    int             server_fd, client_fd;
    int             header_fd, body_fd;

    char            header_fn[STRING_SIZE], body_fn[STRING_SIZE];

    /*
     * buffers (contents and sentinel)
     */
    char            request_buf[HEADER_BUFLEN+1];
    char            header_buf[HEADER_BUFLEN+1];
    char            body_buf[BODY_BUFLEN+1];
    char            linebuf[BUF_SIZE+1];
    char            rcv_mode;

    /*
     * length
     *      len     - length
     *      clen    - content length (server said)
     *      wlen    - write length
     *      hsum    - header length
     *      bsum    - body length
     */
    off_t           len, clen, wlen;
    off_t           hsum, bsum;
    off_t           rsum;

    /*
     * count of query to sibling
     */
    int             query_sibling;

    Info            info;
} Session;



typedef struct {
    Connection_type type;
    struct timeval  timeout;
    Session         *sess;
    int             sibling_id;
} fd2sess_t;





extern Session      *session_list;
extern int          last_session;

extern fdset_t      session_fdset;
extern fd2sess_t    fd2sess[];
extern int          last_fd;
extern int          servd;


extern int          max_session;


/*
 * func. for session
 */

char SessTypeStr(Session_type pt);
char SessPhaseChar(Session_phase ph);
char* SessPhaseStr(Session_phase ph);


Session *ScanActiveSession (int servd, int ear, fdset_t fdset,
    time_t interval, int *stype, int *state, int *active);
Session *ScanWaitSession ();

Session* NewSession();
void     DeleteSession(Session *sess);

int      StartSession(Session *sess, int servd, int ear, request_t *orig_req);
int      JoinSessionList(Session *sess);
void     LeaveSessionList(Session *sess);
int      LeaveWaitSessionList(Session *sess );

int      ConnectSession(Session *sess);
int      AbortSession (Session *sess );
int      CloseSession(Session *sess );


#if 0
int      ReachSession(Session *sess );
#endif
int      ReadSession(Session *sess );


/*
 * func. for sessions
 */

void     InitSessions();
int      CheckSessions();
void     DisplaySessions(int all);

#endif /* WCOL_SESSION_H */
