
/*
 * sheard memory procedures
 *      by k-chinen@is.aist-nara.ac.jp, 1996, 1997
 *
 * $Id$
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <errno.h>



int make_new_pool(key_t shmkey, size_t pool_size)
{
    int chk;
    int flg;

    flg = 0666;
    flg = flg | IPC_CREAT;

    chk = shmget(shmkey, (int)pool_size, flg);
    if(chk==-1) {
        fprintf(stderr, "make_new_pool: fail shmget() (%d)\n", errno);
        return chk;
    }
    return chk;
}

int delete_pool(int shmid)
{
    int chk;

    chk = shmctl(shmid, IPC_RMID, NULL);

    return chk;
}

int assign_pool(key_t shmkey)
{
    int chk;
    int flg;

    flg = 0666;
    /*
    flg = flg | IPC_CREAT;
    */

    chk = shmget(shmkey, (int)0, flg);
    if(chk==-1) {
        fprintf(stderr, "assign_pool: fail shmget() (%d)\n", errno);
        return chk;
    }
    return chk;
}



int show_pool_stat(int shmid)
{
    struct shmid_ds act_buf, *buf;
    int chk;

    buf = &act_buf;

    fprintf(stdout, "=====\n");
    fprintf(stdout, "ID %d\n", shmid);

    chk = shmctl(shmid, IPC_STAT, buf);

    if(chk==-1) {
        fprintf(stderr, "fail shmctl() (%d)\n", errno);
        return -1;
    }

    fprintf(stdout, "---\n");
    fprintf(stdout, " key  = %d %#x\n",
            buf->shm_perm.key, buf->shm_perm.key);
    fprintf(stdout, " seq  = %d\n", buf->shm_perm.seq);
    fprintf(stdout, " mode = %#o\n", buf->shm_perm.mode);
    fprintf(stdout, " current uid/gid = %d/%d\n",
            buf->shm_perm.uid, buf->shm_perm.gid);
    fprintf(stdout, " creator uid/gid = %d/%d\n",
            buf->shm_perm.cuid, buf->shm_perm.cgid);
    fprintf(stdout, "---\n");
#if 0
    fprintf(stdout, " segment size = %d %#x ( %d KB, %d MB)\n",
        buf->shm_segsz, buf->shm_segsz,
        buf->shm_segsz/1024, buf->shm_segsz/1048576);
#else
    fprintf(stdout, " segment size = %d %#x (%s)\n",
        buf->shm_segsz, buf->shm_segsz,
        conv_byte((off_t)buf->shm_segsz));
#endif

#if 0
    fprintf(stdout, " region = %p\n", buf->shm_reg);
#endif
    fprintf(stdout, " last    pid = %d\n", buf->shm_lpid);
    fprintf(stdout, " creator pid = %d\n", buf->shm_cpid);
    fprintf(stdout, " n  attached process = %d\n", buf->shm_nattch);
#if 0
    fprintf(stdout, " cn attrched process = %d\n", buf->shm_cnattch);
#endif
    fprintf(stdout, " time shmat/shmdt/change = %d/%d/%d\n",
            buf->shm_atime, buf->shm_dtime, buf->shm_ctime);

    fflush(stdout);

    return 0;
}




#ifdef SHM_TEST


/*
    0x7fffafe0
*/

#define WCOL_SHMKEY     0x0123
#define WCOL_SHMSIZE    0x100000

key_t wcol_shmkey;

int   wcol_shmflag;
int   wcol_shmid;
int   wcol_shmsize;



main(int argc, char **argv)
{
    pid_t cpid;
    int  i;
    int  flag;
    void *pos;
    int *k;

    extern char *optarg;
    extern int  optind;


    wcol_shmkey  = (key_t) WCOL_SHMKEY;
    wcol_shmsize = WCOL_SHMSIZE;


    while((flag =  getopt(argc, argv, "hmd")) != -1) {
        switch(flag) {
        case 'm':
            fprintf(stdout, "MAKE NEW POOL\n");

#if 1
            if((wcol_shmid = make_new_pool(wcol_shmkey, WCOL_SHMSIZE))==-1) {
                printf("error when make new pool\n");
                exit(3);
            }
            else {
                printf("success make new pool id=%d\n", wcol_shmid);
            }
            show_pool_stat(wcol_shmid);
#else
            /* try */
            for(wcol_shmkey=1 ; wcol_shmkey<10000 ; wcol_shmkey += 256 ) {
                fprintf(stdout, "key=%d\n", wcol_shmkey);
                if((wcol_shmid = make_new_pool(wcol_shmkey, WCOL_SHMSIZE))==-1) {
                    continue;
                }
                printf("success make new pool: key=%d, id=%d\n",
                    wcol_shmkey, wcol_shmid);
                show_pool_stat(wcol_shmid);

                break;
            }
            if(wcol_shmid == -1) {
                fprintf(stderr, "error when make new pool\n");
                exit(3);
            }

#endif
            break;

        case 'd':
            fprintf(stdout, "DELETE POOL\n");

            wcol_shmid = assign_pool(wcol_shmkey);
            delete_pool(wcol_shmid);

            exit(0);

            break;
        }
    }


#if 0
    cpid = fork();

    /* child */
    if(cpid==0) {
        wcol_shmid = assign_pool(wcol_shmkey);
        if(wcol_shmid==-1) {
            fprintf(stderr, "child exit\n");
            exit(2);
        }
    }
    else {
        wcol_shmid = assign_pool(wcol_shmkey);
        if(wcol_shmid==-1) {
            fprintf(stderr, "parent exit\n");
            exit(2);
        }
    }

    if(cpid==0)
        sleep(1);
#endif


    wcol_shmid = assign_pool(wcol_shmkey);
    /*
    pos = shmat(wcol_shmid, (void*)NULL, SHM_SHATTR);
    */
    pos = shmat(wcol_shmid, (void*)NULL, 0);
    if(pos==NULL || ((int)pos) == -1) {
        fprintf(stderr, "fail shmat() pos=%p (errno=%d)\n", pos, errno);
        exit(2);
    }

    for(i=0;i<3;i++) {
        k = (int*)pos;

        printf("\nprocess %d:\n", getpid());
        show_pool_stat(wcol_shmid);
        fprintf(stdout, "pos   %p\n", pos);
        fprintf(stdout, "value %d\n", *k);

        *k = *k+1;

        sleep(2);
    }

}

#endif /* SHM_TEST */
