/*
 * WCOL: procedures for sibling
 *      by k-chinen@is.aist-nara.ac.jp, 1997
 *
 * $Id$
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>


#include "wcol.h"
#include "msg.h"
#include "socket.h"
#include "hname.h"
#include "sibling.h"


static int sirial_id=0;

int num_sibling=0;
sibling_info_t *sibling_list[MAX_SIBLING];


void DumpSiblings()
{
    int i;
    sibling_info_t *s;

    if(num_sibling<=0)  {
        Trace("no siblings\n");
        return;
    }

    Trace("Siblings: %d\n", num_sibling);
    for(i=0;i<num_sibling;i++) {
        s = sibling_list[i];

        Trace("%2d: %-40s %2d HTTP %5d, ICP %5d\n",
            i, s->name, s->type, s->http_port, s->icp_port);
    }
}


/*
 * A2SiblingType - convert ASCII to sibling type (bit-masked value)
 */
int A2SiblingType(char *astr)
{
    char word[STRING_SIZE];
    register char *p, *q;
    register int c;
    int ret;

#if 0
    Trace("A2SiblingType: type specify string '%s'\n", astr);
#endif

    ret = 0;
    c = 0;
    p = astr;
    q = &word[0];
    while(*p) {
        if(isalpha(*p)) {
            *q++ = *p++;

            *q = '\0';
            c=0;
        }
        else {
            p++;
            c=1;
        }

        if(c || !*p) {
            *q++ = '\0';
            q = &word[0];

            if(strcasecmp_first_all(word, 'Q', "Query")) {
                ret |= SIBLING_TYPE_QUERY;
            }
            else
            if(strcasecmp_first_all(word, 'N', "Notify")) {
                ret |= SIBLING_TYPE_NOTIFY;
            }
        }
    }

#if 0
    Trace("A2SiblingType: return %#x\n", ret);
#endif

    return ret;
}


int SiblingType2A(int atype, char *buf)
{
    buf[0] = '\0';

    if(atype & SIBLING_TYPE_QUERY) {
        if(buf[0])
            strcat(buf, ",query");
        else
            strcat(buf, "query");
    }
    if(atype & SIBLING_TYPE_NOTIFY) {
        if(buf[0])
            strcat(buf, ",notify");
        else
            strcat(buf, "notify");
    }

    return 0;
}


/*
 *
 */
void ShowSiblings(FILE *outfp)
{
    int i;
    sibling_info_t *s;
    char tmp[STRING_SIZE];

    if(num_sibling<=0)  {
        fprintf(outfp, "Siblings: no siblings\n");
        fflush(outfp);

        return;
    }

    fprintf(outfp, "Siblings: %d sibling(s)\n", num_sibling);

    fprintf(outfp, " #: ID  %-36s %16s %5s %5s\n",
        "Hostname", "Type", "HTTP", "ICP");

    for(i=0;i<num_sibling;i++) {
        s = sibling_list[i];

        SiblingType2A(s->type, tmp);

        fprintf(outfp, "%2d: %03d %-36s %16s %5d %5d\n",
            i, s->id, s->name, tmp, s->http_port, s->icp_port);
    }
    fflush(outfp);
}




/*
 *
 */
sibling_info_t *find_sibling(int q_id)
{
    int i;

    for(i=0;i<MAX_SIBLING;i++) {
        if(sibling_list[i]==NULL)
            continue;

        if(sibling_list[i]->id==q_id)
            return sibling_list[i];
    }

    return NULL;
}







int AppendSibling(char *a_name, int a_type, int a_http_port, int a_icp_port)
{
    sibling_info_t sinfo, *s;

    if(num_sibling>=MAX_SIBLING) {
        return -1;
    }

    sirial_id++;

    Trace("AppendSibling: %2d/%2d: ID %d, name '%s', type %d, http/icp %d/%d\n",
        sirial_id, num_sibling, MAX_SIBLING,
        a_name, a_type, a_http_port, a_icp_port);

    sinfo.id   = sirial_id;
    sinfo.type = a_type;
    strcpy(sinfo.name, a_name);
    if(find_addr(a_name, a_icp_port, &(sinfo.addr))) {
        Error("AppendSibling: Cannot found host '%s'", a_name);
        return -1;
    }

    sinfo.http_port = a_http_port;
    sinfo.icp_port = a_icp_port;

    sinfo.usecount = 0;

    if((s=(sibling_info_t*) MALLOC(sizeof(sibling_info_t)))==NULL) {
        Error("AppendSibling: not enough memory");
        return -1;
    }
    sibling_list[num_sibling] = s;
    *(sibling_list[num_sibling]) = sinfo;

    num_sibling++;

    /*
    ShowSiblings(stderr);
    */

    return 0;
}


int AppendSiblingEntry(char *a_name, char *a_type,
        char *a_http_port, char *a_icp_port)
{
#if 0
    int c;

    c = trace_flag;
    trace_flag = 1;

    Trace("AppendSiblingEntry: %s %s %s %s\n",
                a_name, a_type, a_http_port, a_icp_port);
    Trace("\ttype '%s' -> %#x\n", a_type, A2SiblingType(a_type));

    trace_flag = c;
#endif
    if(a_name && *a_name &&
        a_type && *a_type &&
          a_http_port && *a_http_port &&
            a_icp_port && *a_icp_port) {

        return AppendSibling(a_name, A2SiblingType(a_type),
                atoi(a_http_port), atoi(a_icp_port));

    }
    else {
        Log("; 'Sibling' configuration is wrong, please check syntax");

        return -1;
    }
}

