/*
 * speaker.c - speak have object or not to other proxies via ICP
 *      by k-chinen@is.aist-nara.ac.jp, 1996
 *
 * $Id: speaker.c,v 1.3 1997/08/31 18:02:42 k-chinen Exp k-chinen $
 */

/*
 * NOTE:
 *
 */


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <signal.h>
#include <setjmp.h>
#include <errno.h>

#include "wcol.h"
#include "socket.h"
#include "ipc.h"
#include "icp.h"
#include "sigs.h"
/*
#include "global.h"
*/

#include "request.h"
#include "prompter.h"
#include "accept.h"


extern int catalyst_flag;
extern int icp_listen_portno;


static int local_wcol_request_count=0;



#ifndef PROMPT_TIMEOUT
#define PROMPT_TIMEOUT            (2)         /* 2 sec */
#endif



static int SendProxyPrompt(int port,
        char *a_peername, int a_hereport, char *url)
{
    static long count=0;
    prequest_t prompt;
    int c;
#if 0
    time_t now;
#endif
    struct timeval cur;

/*
    Trace("prompt '%s'\n", url);
*/
    Trace(";\t'%s'\n", url);

#if 0
    time(&now);
#endif

    if(catalyst_flag) {
        prompt.req.mode         = REQ_IMPORT;
        prompt.req.sid          = -1;
        prompt.req.id           = ++local_wcol_request_count;
    }
    else {
        prompt.req.mode         = REQ_PREFETCH;
        prompt.req.sid          = -1;
        prompt.req.id           = ++local_wcol_request_count;
    }

    prompt.req.state            = REQ_NOP;
    prompt.req.method           = METHOD_GET;
#if 0
    prompt.req.issue_time       = now;
#endif
    /*
    gettimeofday(&cur);
    */
    gettimeofday(&cur, NULL);
    prompt.req.issue_time       = cur;

    strcpy(prompt.req.peername, a_peername);
    prompt.req.hereport = a_hereport;

    strcpy(prompt.req.url,      url);
    prompt.req.ver_major        = 1;
    prompt.req.ver_minor        = 0;
    strcpy(prompt.req.auth,     "");
    prompt.req.ims              = -1;
    prompt.req.nocache          = 0;
    prompt.req.noprefetch          = 0;

    /*
     * To avoid deadlock with IPC, set timer
     */
    pipe_broken = 0;
    if((c = XSendRecord(PROMPT_TIMEOUT, port, &prompt, sizeof(prompt)))<0) {
        Error("Speaker: Cannot send prompt");
    }
    else
    if(c==0) {
#ifdef TRACE_PROMPTER
        Trace("Speaker: Cannot send prompt, 0 byte\n");
#endif
    }

    /*
    Trace("Speaker: send prompt %d byte(s)\n", c);
    */

fail:

    return 0;
}


int icp_servd = -1;


void Speaker(int mouth)
{
    struct sockaddr_in from;
    int from_len;
    int nread;
    unsigned char buf[KB(10)];
    icp_common_t *header;
    unsigned char *target;
    unsigned char *peer_addr;
    Info *info_ptr;
    int req_count=0;
    int nc;
    int chk;
#if 0
    char peer_name[STRING_SIZE];
#endif
    char *peer_name;
    int here_port;
    struct timeval cur;
    int nwrite;


    wcol_pid = getpid();
    SetProcessComment("speaker");
    if(trace_flag) {
        /*
        DLog("WAKE Speaker");
        */
        SwitchLog(DEFAULT_DEBUGLOGDIR, "log of wcol's speaker");
    }

    here_port = icp_listen_portno;


    SetSignalHandler(SIGTERM,   ExitProgram);

    if(icp_servd<0) {
        if((icp_servd = UDP_bind_port(icp_listen_portno, WCOL_LISTEN_LENGTH)) < 0) {
            Error("Speaker: cannot bind listen ICP port (%d)",
                icp_listen_portno);
            exit(0);
        }
        Trace("Speaker: new port(%d) is ready as fd #%d\n",
            icp_listen_portno, icp_servd);
    }
    else {
        Trace("Speaker: port(%d) is setted as fd #%d, already\n",
            icp_listen_portno, icp_servd);
    }

    req_count = 0;
    while(1) {

#if 0
        /*
         * wait ICP message from other system.
         */
        if((chk=isreadable(icp_servd, 30, 0))>0) {
            continue;
        }
        if(chk<0) {
            exit(WCOL_EXIT_UNDEF);
        }
#endif

#ifdef USE_ACCEPT_LOCK
        if(lock_port(icp_servd)) {
            continue;
        }
#endif

        /*
         * Is it sure ? socket is readable ?
         */
        /* not readable */
        if((chk=isreadable(icp_servd, 10, 0))==0) {
#ifdef USE_ACCEPT_LOCK
            unlock_port(icp_servd);
#endif
            continue;
        }
        /* error */
        if(chk<0) {
            exit(WCOL_EXIT_UNDEF);
        }

        /* readable ... */

        nc = 0;
        from_len = sizeof(struct sockaddr);
wait_query:
        if((nread = recvfrom(icp_servd, (char*)buf, sizeof(buf), 0,
                    (struct sockaddr*)&from, &from_len))<0) {
            if(errno == EINTR) {
                goto wait_query;
            }
            Log("; speaker error happen (%d)", errno);
            if(nc++ > 5) {
                Log("; strange speaker exit");
                exit(1);
            }

            goto wait_query;
        }
        if(nread == 0) {
            Trace("\nSpeaker: Empty query\n");

            goto wait_query;
        }

#ifdef USE_ACCEPT_LOCK
        unlock_port(icp_servd);
#endif


        gettimeofday(&cur, NULL);
        req_count++;
        peer_name = (char*)inet_ntoa( from.sin_addr );

        Trace("\nSpeaker: %d th, addr %s, read %d bytes ",
            req_count, peer_name, nread);

#if 1
        ICP_DumpMessage(stdout, buf);
#endif
        header = (icp_common_t*) buf;
        if(header->opcode == ICP_OP_QUERY) {
            target = &buf[ICP_HDR_SZ];
            Trace("\tQuery URL (%d) = '%s'\n", strlen(target), target);


            if(strlen(target)>URL_LENGTH) {
                Trace("\tToo long URL (%d chars), give up\n", strlen(target));

                Log("%d.%06d QUERY %d %s ICP %d - - - %s %s %s %d :",
                    cur.tv_sec, cur.tv_usec, here_port, peer_name,
                    header->version, _icp_opcode_str[header->opcode],
                    target, "LOST", 0
                    );
            }
            else {
                if((info_ptr=Have((char*)target))!=NULL) {
                    nwrite = ICP_SendHit(icp_servd, &from, from_len,
                                    header, target);
                    FREE(info_ptr);

                    Log("%d.%06d QUERY %d %s ICP %d - - - %s %s %s - %d - :",
                        cur.tv_sec, cur.tv_usec, here_port, peer_name,
                        header->version, _icp_opcode_str[header->opcode],
                        target, "HIT", nwrite
                    );
                }
                else {
                    nwrite = ICP_SendMiss(icp_servd, &from, from_len,
                                    header, target);

                    Log("%d.%06d QUERY %d %s ICP %d - - - %s %s %s - %d - :",
                        cur.tv_sec, cur.tv_usec, here_port, peer_name,
                        header->version, _icp_opcode_str[header->opcode],
                        target, "MISS", nwrite
                    );
                }

                if(IsHTML((char*)target)) {
                    SendProxyPrompt(mouth, peer_name, here_port, (char*)target);
                }
            }
        }
        else
        if(header->opcode == ICP_OP_NOTIFY) {

            target = &buf[ICP_HDR_SZ];
            Trace("\tQuery URL (%d) = '%s'\n", strlen(target), target);

            if(strlen(target)>URL_LENGTH) {
                Trace("\tToo long URL (%d chars), give up\n", strlen(target));

                Log("%d.%06d NOTIFY %d %s ICP %d - - - %s %s %s - %d - :",
                    cur.tv_sec, cur.tv_usec, here_port, peer_name,
                    header->version, _icp_opcode_str[header->opcode],
                    target, "LOST", 0
                );
            }
            else {
                Log("%d.%06d NOTIFY %d %s ICP %d - - - %s %s %s - %d - :",
                    cur.tv_sec, cur.tv_usec, here_port, peer_name,
                    header->version, _icp_opcode_str[header->opcode],
                    target, "-", 0
                );

                if(IsHTML((char*)target)) {
                    SendProxyPrompt(mouth, peer_name, here_port, (char*)target);
                }
            }
        }
        else {
            Trace("Unsupport ICP message\n");
#if 1
            ICP_DumpMessage(stdout, buf);
#endif
        }
    }

    Fatal("Speaker: complete.");


    close(icp_servd);
    icp_servd = -1;

    exit(0);
}



