/*
 *  WCOL: show sessions state
 *      by k-chinen@is.aist-nara.ac.jp, 1997
 *
 *  $Id: sstate.c,v 1.5 1997/09/15 02:00:25 k-chinen Exp k-chinen $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <signal.h>
#ifdef HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif
#include <sys/uio.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <limits.h>

#include <sys/ipc.h>
#include <sys/shm.h>


#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif

#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif

#ifndef ESC
#define ESC '\x1b'
#endif


#include <signal.h>
#include <errno.h>

#include "wcol.h"
#include "msg.h"
#include "request.h"
#include "ipc.h"
#include "sigs.h"
#include "fds.h"

#include "session.h"
#include "manager.h"
#include "handler.h"
#include "prompter.h"
#include "reaper.h"
/* */
#include "global.h"
/* */



static int usage()
{
    fprintf(stdout,
"sstate - print Wcol's sessions states\n\
    by k-chinen@is.aist-nara.ac.jp, 1997\n\
\n\
usage: sstate [options]\n\
\n\
option:\n\
    -h          help, print this messages\n\
    -l          loop\n\
    -i #sec     wait interval\n\
    -d          decoration\n\
    -s          short\n\
    -w          wide\n\
\n\
Enviroment:\n\
    max sessions = %d  (handler %d, max of sessions per handler %d)\n\
\n\
Output:\n\
    num - number of session\n\
    pos - position in pool\n\
    h   - handler\n\
    s   - session\n\
    T   -   type (Client,Prefetch,Import)\n\
    P   -   phase (Start,Lock,Query,Connect,Header,Body,End)\n\
    sid -   id\n\
    id  - request id\n",
                        MAX_HANDLER*MAX_SESSION, MAX_HANDLER, MAX_SESSION);

    return 0;
}

Session *session_pool;
modinfo_t *handlers;




void ShowSessions_Short(int decoration, int label)
{
    Session *s;
    int i,j,p;
    int c;
    time_t now;
    int nlen;
    char now_str[STRING_SIZE];
    int n_wake;
    int Ntype[256], Nphase[256];


    static int last_z_req=0, last_z_sess=0;
    int cur_z_req, cur_z_sess;


    time(&now);
#if 0
    strcpy(now_str, ctime(&now));
#endif
    sprintf(now_str, "%d", now);

    nlen = 0;
    {
        char *p;

        p = now_str;
        while(*p) {
            if(*p=='\r'||*p=='\n') {
                *p = '\0';
                break;
            }
            p++;
            nlen++;
        }
    }

    n_wake = 0;
    cur_z_req = 0;
    cur_z_sess = 0;
    for(j=0;j<MAX_HANDLER;j++) {
        if(handlers[j].wake) {
            n_wake++;

            if(handlers[j].z_req>0) {
                cur_z_req  += handlers[j].z_req;
            }
            if(handlers[j].z_sess>0) {
                cur_z_sess += handlers[j].z_sess;
            }
        }
    }

    for(i=0;i<256;i++) {
        Ntype[i] = 0;
        Nphase[i] = 0;
    }

    for(j=0;j<MAX_HANDLER;j++) {
        if(handlers[j].wake) {

            for(i=0;i<MAX_SESSION;i++) {
                p = j*MAX_SESSION + i;
                s = &session_pool[p];

                Ntype[((int)s->type)%256]++;
                Nphase[((int)s->phase)%256]++;
            }
        }
    }

    if(label%24==0) {
        char  fline1[STRING_SIZE], fline2[STRING_SIZE];
        char  tmp[STRING_SIZE];
        int   pos;
        char  *p, *q;

        sprintf(fline1, "%-*s %-4s", nlen, "#", "max");
        sprintf(fline2, "%-*s %4d",  nlen, "#", MAX_HANDLER);

        sprintf(tmp, " %-19s", "total");
        strcat(fline1, tmp);
        sprintf(tmp, " %9d %9d", cur_z_req, cur_z_sess);
        strcat(fline2, tmp);

        sprintf(tmp, " %-4s %-4s", "total", "max");
        strcat(fline1, tmp);
        sprintf(tmp, " %4d/%4d", n_wake*MAX_SESSION, MAX_HANDLER*MAX_SESSION);
        strcat(fline2, tmp);

        if(label==0) {
            printf("#\n");
            puts(fline1);
            puts(fline2);
        }
        else {
            puts(fline2);
            puts(fline1);
        }


        /*
         *
         */
        printf("#\n");

        sprintf(fline1, "%-*s hand", nlen, "#");
#if 0
        sprintf(fline2, "%-*s %4d",  nlen, "#", MAX_HANDLER);
#endif
        sprintf(fline2, "%-*s %4s",  nlen, "#", "");

        sprintf(tmp, " %-19s", "change");
        strcat(fline1, tmp);
        sprintf(tmp, " %9s %9s", "req", "sess");
        strcat(fline2, tmp);

        pos = strlen(fline1);
        for(i=(int)SESS_NOP;i<=(int)SESS_IMPORT;i++){
            sprintf(tmp, " %4c", ' ');
            strcat(fline1, tmp);
        }
        q = &fline1[pos];
        p = " mode";
        while(*p) {
            *q++ = *p++;
        }
        for(i=(int)SESS_NOP;i<=(int)SESS_IMPORT;i++){
            sprintf(tmp, " %4c", SessTypeStr(i));
            strcat(fline2, tmp);
        }


        pos = strlen(fline1);
        for(i=(int)PHASE_UNKNOW;i<=(int)PHASE_ERROR;i++){
            sprintf(tmp, " %4c", ' ');
            strcat(fline1, tmp);
        }
        q = &fline1[pos];
        p = " phase";
        while(*p) {
            *q++ = *p++;
        }
        for(i=(int)PHASE_UNKNOW;i<=(int)PHASE_ERROR;i++){
            sprintf(tmp, " %4c", SessPhaseChar(i));
            strcat(fline2, tmp);
        }

        puts(fline1);
        puts(fline2);
    }

    printf("%s %4d", now_str, n_wake);
    printf(" %9d %9d", cur_z_req-last_z_req, cur_z_sess-last_z_sess);

    for(i=(int)SESS_NOP;i<=(int)SESS_IMPORT;i++){
        printf(" %4d", Ntype[i]);
    }
    for(i=(int)PHASE_UNKNOW;i<=(int)PHASE_ERROR;i++){
        printf(" %4d", Nphase[i]);
    }

    printf(" |\n");

    fflush(stdout);
    fflush(stderr);

    last_z_req  = cur_z_req;
    last_z_sess = cur_z_sess;

    return;
}

void ShowSessions(int decoration, int all)
{
    Session *s;
    int i,j,p;
    int c;
    time_t now;
    char now_str[STRING_SIZE];
    int n_wake;
    char hst;


    time(&now);
    strcpy(now_str, ctime(&now));

    n_wake = 0;
    for(j=0;j<MAX_HANDLER;j++) {
        if(handlers[j].wake) {
            n_wake++;
        }
    }

    if(decoration) {

        fprintf(stdout,"%c[1;1H%c[J",ESC,ESC);
        fprintf(stdout,"Wake %d handler(s) // ", n_wake);
        fprintf(stdout,"Time: %d %s", now, now_str);
        fprintf(stdout,"%c[7m",ESC);
    }
    else {
        fprintf(stdout,"\n=== Wake %d // Time %d %s", n_wake, now, now_str);
    }

        /* 1234 123 1 123 1 1 123456 123456 1234.. */
    fprintf(stdout,"#num  #h U  #s T P -sid-- --id-- URL..\n");

    if(decoration) {
        fprintf(stdout,"%c[m", ESC);
    }

    c = 0;
    for(j=0;j<MAX_HANDLER;j++) {
        hst = 'D';
        if(all || handlers[j].wake) {
            if(handlers[j].wake) {
                hst = 'U';
            }

            for(i=0;i<MAX_SESSION;i++) {
                p = j*MAX_SESSION + i;
                s = &session_pool[p];

                if(all || s->phase != PHASE_IDLE) {
                    fprintf(stdout,"%4d %03d %c %03d %c %c %6d %6d %s\n",
                        c, j, hst, i,
                        SessTypeStr(s->type),
                        SessPhaseChar(s->phase),
                        s->req.sid, s->req.id,
                        s->_url);

                    c++;    /* count */
                }

            }
        }
    }

    fflush(stdout);
    fflush(stderr);

    return;
}

void ShowSessions_Wide(int decoration, int all)
{
    Session *s;
    int i,j,p;
    int c;
    time_t now;
    char now_str[STRING_SIZE];
    int n_wake;
    char hst;


    time(&now);
    strcpy(now_str, ctime(&now));

    n_wake = 0;
    for(j=0;j<MAX_HANDLER;j++) {
        if(handlers[j].wake) {
            n_wake++;
        }
    }

    if(decoration) {

        fprintf(stdout,"%c[1;1H%c[J",ESC,ESC);
        fprintf(stdout,"Wake %d handler(s) // ", n_wake);
        fprintf(stdout,"Time: %d %s", now, now_str);
        fprintf(stdout,"%c[7m",ESC);
    }
    else {
        fprintf(stdout,"\n=== Wake %d // Time %d %s", n_wake, now, now_str);
    }

        /* 1234 123 1 123 1 1 1234567 12345 1234 1234 1234 1234 1234 1234567 1234.. */
    fprintf(stdout,"#num  #h U  #s T P --lap-- -hot- -c#- -s#- -h#- -b#- head --body- URL..\n");

    if(decoration) {
        fprintf(stdout,"%c[m", ESC);
    }

    c = 0;
    for(j=0;j<MAX_HANDLER;j++) {
        hst = 'D';
        if(all || handlers[j].wake) {
            hst = 'U';

            for(i=0;i<MAX_SESSION;i++) {
                p = j*MAX_SESSION + i;
                s = &session_pool[p];

                if(all || s->phase != PHASE_IDLE) {
                    fprintf(stdout,"%4d %03d %c %03d %c %c %7d %5d %4d %4d %4d %4d %4ld %7ld %s\n",
                        c, j, hst, i,
                        SessTypeStr(s->type),
                        SessPhaseChar(s->phase),
                        now - s->begin.tv_sec,
                        now - s->hot.tv_sec,
                        s->client_fd, s->server_fd, s->header_fd, s->body_fd,
                        (long)s->hsum, (long)s->bsum,
                        s->_url);

                    c++;    /* count */
                }

            }
        }
    }

    fflush(stdout);
    fflush(stderr);

    return;
}


int loop_flag=0;
int wait_interval=5;
int deco_flag=0;
int all_flag=0;

int short_flag=0;
int wide_flag=0;

key_t session_shmid;
key_t hstate_shmid;



void Init(int argc, char **argv)
{
    int flag;
    int i;

    extern char *optarg;
    extern int  optind;

    /* defaults */
    trace_flag = 1;

    while((flag = getopt(argc, argv, "vhaf:li:dsw"))!= -1) {
        switch(flag) {
        case 'v':
            Trace("parameters:\n");
            Trace("\tMAX_HANDLER %d, MAX_SESSION %d\n",
                MAX_HANDLER, MAX_SESSION);
            break;
        case 'h':
            usage();
            exit(0);
            break;
        case 'a':
            all_flag = 1 - all_flag;
            break;
        case 'l':
            loop_flag = 1 - loop_flag;
            break;
        case 'i':
            wait_interval = atoi(optarg);
            break;
        case 'd':
            deco_flag = 1 - deco_flag;
            break;
        case 's':
            short_flag = 1 - short_flag;
            break;
        case 'w':
            wide_flag = 1 - wide_flag;
            break;
        }
    }



    /*
     *
     */
    session_shmid = assign_pool(WCOL_SHMKEY_SESSION);
    if(session_shmid == -1 ) {
        Error("not found shared memory area for sessions");
        exit(WCOL_EXIT_NOMEMORY);
    }

    if((session_pool = (Session*)shmat(session_shmid, (void*)NULL, 0))==NULL) {
        Error("cannot allocate shared memory area for sessions (%d)", errno);
        exit(WCOL_EXIT_NOMEMORY);
    }
    /*
    show_pool_stat(session_shmid);
    Trace("Init: sstate %8x\n", session_pool);
    */

    /*
     *
     */
    hstate_shmid = assign_pool(WCOL_SHMKEY_HSTATE);
    if(hstate_shmid == -1 ) {
        Error("not found shared memory area for handlers");
        exit(WCOL_EXIT_NOMEMORY);
    }

    if((handlers = (modinfo_t*)shmat(hstate_shmid, (void*)NULL, 0))==NULL) {
        Error("cannot allocate shared memory area for handlers (%d)", errno);
        exit(WCOL_EXIT_NOMEMORY);
    }
    /*
    show_pool_stat(hstate_shmid);
    Trace("Init: hstate %8x\n", handlers);
    */

}



main(int argc, char **argv)
{
    int c;
    extern Session *session_pool;
    extern int handler_no;

    Init(argc, argv);
#ifdef MEMORY_CHECK
    show_memoryusage();
#endif

    c = 0;
    do {
        if(wide_flag) {
            ShowSessions_Wide(deco_flag, all_flag);
        }
        else
        if(short_flag) {
            ShowSessions_Short(deco_flag, c);
        }
        else {
            ShowSessions(deco_flag, all_flag);
        }
        if(loop_flag) {
            sleep(wait_interval);
        }

        c++;
        if(c>=1000)
            c=0;

    } while(loop_flag);
}






