/*
 *  WCOL: The prefetching proxy server
 *
 *  wcol.h - global header
 *      by k-chinen@is.aist-nara.ac.jp, 1994 - 1997
 *
 *  $Id: wcol.h,v 1.9 1996/11/18 20:58:21 k-chinen Exp k-chinen $
 */

#ifndef WCOL_H
#define WCOL_H  /* block for forever(or nest) loop */


/*
 * Constants
 */

#define WCOL_VER      "0.10981"

#define WCOL_SHMKEY_HSTATE  (0x0123)
#define WCOL_SHMKEY_SESSION (0x0125)
#define WCOL_SHMKEY_CACHE   (0x0127)


#define SERVER_UNKNOW       (0)
#define SERVER_ORIGIN       (1)
#define SERVER_SIBLING      (2)
#define SERVER_PARENT       (3)



/* EXIT CODE */
#define WCOL_EXIT_NORMAL        0
#define WCOL_EXIT_UNDEF         1
#define WCOL_EXIT_BINDERROR     2
#define WCOL_EXIT_NOMEMORY      3
#define WCOL_EXIT_PIPEBROKEN    4


/*
 * Basic header files
 */
#include "range.h"
#include "msg.h"
#include "url.h"
#include "info.h"
#include "base.h"   /* path.h */

extern pid_t wcol_pid;
extern char config_file[];


/*
 * memory
 */
#ifndef MEMORY_CHECK

#define MALLOC(x)   malloc(x)
#define FREE(x)     free(x)

#else /* MEMORY_CHECK */

#define MALLOC(x)   wcol_malloc(x)
#define FREE(x)     wcol_free(x)

#endif /* MEMORY_CHECK */


/*
 * misc macros
 */
#define strcmp_first_all(x, y0, y)      (x[0]==y0 && strcmp(x,y)==0)
#define Nstrcmp_first_all(x, y0, y)     (x[0]!=y0 || strcmp(x,y)!=0)
#define strcmp_first3(x, y0, y1, y2)    (x[0]==y0 && x[1]==y1 && x[2]==y2)
#define Nstrcmp_first3(x, y0, y1, y2)   (x[0]!=y0 || x[1]!=y1 || x[2]!=y2)
#define strncmp_first_all(x, n, y0, y)  (x[0]==y0 && strncmp(x,y,n)==0)
#define Nstrncmp_first_all(x, n, y0, y) (x[0]!=y0 || strncmp(x,y,n)!=0)
#define strcasecmp_first_all(x, y0, y) \
            ((x[0]==y0 || tolower(x[0])==tolower(y0)) && strcasecmp(x,y)==0)
#define Nstrcasecmp_first_all(x, y0, y) \
            ((x[0]!=y0 && tolower(x[0])!=tolower(y0)) || strcasecmp(x,y)!=0)
#define strncasecmp_first_all(x, n, y0, y) \
            ((x[0]==y0 || tolower(x[0])==tolower(y0)) && strncasecmp(x,y,n)==0)
#define Nstrncasecmp_first_all(x, n, y0, y) \
            ((x[0]!=y0 && tolower(x[0])!=tolower(y0)) || strncasecmp(x,y,n)!=0)



/***
 ***
 ***/

#define NBLK(radix, x)  ((x)/(radix)+(((x)%(radix))!=0))
#define NBLK512(x)      (NBLK(512,(x))*512)

#define BCEIL(x, blen)  (((x)/(blen)+1)*(blen))

#define alloc_str(x)    ((char*)MALLOC(sizeof(char)*BCEIL((x)+1, STRING_BLEN)))


#define STRDUP(d,s) \
    ((d=alloc_str(strlen(s)))==NULL ? NULL : strcpy(d, s), d)


/***
 ***
 ***/

#define WCOL_LTIME(x)       (time(x))
#define WCOL_UTIME(x)       (time(x), *(x)=mktime(gmtime(x)))
#define WCOL_TO_UTIME(x)    (mktime(gmtime(x)))
#define WCOL_TO_LTIME(x)    (mktime(localtime(x)))

        /* x(sec,usec) - y(sec,usec) */
#define WCOL_SUB_TIME(x,y) \
        ( (x.tv_sec  - y.tv_sec )*1000 + (x.tv_usec - y.tv_usec)/1000 )

#if 0

#define WCOL_ADD_TIME(x,y) {\
        x.tv_usec += y.tv_usec%1000000; \
        x.tv_sec  += y.tv_sec + y.tv_usec/1000000; \
        }

#else

        /* x(sec,usec) += z(msec) */
#define WCOL_ADD_MSEC(x,y) {\
        x.tv_usec += (y%1000)*1000; \
        x.tv_sec  += y/1000+(x.tv_usec/1000000); \
        x.tv_usec = x.tv_usec%1000000; \
        }
        /* x(sec,usec) += z(usec) */
#define WCOL_ADD_USEC(x,y) {\
        x.tv_sec  += y/1000000; \
        x.tv_usec += y%1000000+(x.tv_usec/1000000); \
        x.tv_usec = x.tv_usec%1000000; \
        }


#endif




/***
 ***
 ***/

#ifdef NOTE_SAMEWINDOW

#ifndef USE_NOTE
#define USE_NOTE
#endif

#else

#ifdef NOTE_WINDOW

#ifndef USE_NOTE
#define USE_NOTE
#endif

#else

#ifdef USE_NOTE

#ifndef NOTE_WINDOW
#define NOTE_WINDOW
#endif

#endif /* USE_NOTE */

#endif /* NOTE_WINDOW */

#endif /* NOTE_SAMEWINDOW */



#ifndef DEFAULT_CONFIG_ENVNAME
#define DEFAULT_CONFIG_ENVNAME "WCOL"
#endif


#ifndef DEFAULT_CONFIGFILE
#define DEFAULT_CONFIGFILE      "/etc/wcol.conf"
#endif

#ifndef DEFAULT_POOLDIR
#define DEFAULT_POOLDIR         "/var/tmp/www-pool"
#endif
#ifndef DEFAULT_LOGFILE
#define DEFAULT_LOGFILE         "/var/tmp/wcol.log"
#endif
#ifndef DEFAULT_PIDFILE
#define DEFAULT_PIDFILE         "/var/tmp/wcol.pid"
#endif
#ifndef DEFAULT_STATEFILE
#define DEFAULT_STATEFILE       "/var/tmp/wcol.state"
#endif
#ifndef DEFAULT_GCWORKFILE
#define DEFAULT_GCWORKFILE      "/var/tmp/wcol.gc_work"
#endif
#ifndef DEFAULT_DEBUGLOGDIR
#define DEFAULT_DEBUGLOGDIR     "/var/tmp"
#endif



#endif /* WCOL_H */
